@@ -36,34 +36,6 @@
 #include "shared.h"
 #include "benchmark_shared.h"
 
-static int run(void)
-{
-	int i, ret = 0;
-
-	ret = ft_init_fabric();
-	if (ret)
-		return ret;
-
-	if (!(opts.options & FT_OPT_SIZE)) {
-		for (i = 0; i < TEST_CNT; i++) {
-			if (!ft_use_size(i, opts.sizes_enabled))
-				continue;
-			opts.transfer_size = test_size[i].size;
-			init_test(&opts, test_name, sizeof(test_name));
-			ret = pingpong();
-			if (ret)
-				return ret;
-		}
-	} else {
-		init_test(&opts, test_name, sizeof(test_name));
-		ret = pingpong();
-		if (ret)
-			return ret;
-	}
-
-	return ft_finalize();
-}
-
 int main(int argc, char **argv)
 {
 	int op, ret;
@@ -101,13 +73,12 @@ int main(int argc, char **argv)
 
 	hints->ep_attr->type = FI_EP_RDM;
 	hints->caps = FI_MSG;
-	hints->mode |= FI_CONTEXT;
+	hints->mode |= FI_CONTEXT | FI_CONTEXT2;
 	hints->domain_attr->mr_mode = opts.mr_mode;
-	hints->domain_attr->threading = FI_THREAD_DOMAIN;
 	hints->tx_attr->tclass = FI_TC_LOW_LATENCY;
 	hints->addr_format = opts.address_format;
 
-	ret = run();
+	ret = run_pingpong();
 
 	ft_free_res();
 	return ft_exit_code(ret);
