@@ -38,38 +38,38 @@
  *
  *	RDMA write host memory --> device memory:
  *
- *	    ./fi-rdmabw-xe -m device -d 0 &
+ *	    ./fi_xe_rdmabw -m device -d 0 &
  *	    sleep 1 &&
- *	    ./fi-rdmabw-xe -m host -t write localhost
+ *	    ./fi_xe_rdmabw -m host -t write localhost
  *
  *	RDMA read host memory <-- device memory:
  *
- *	    ./fi-rdmabw-xe -m device -d 0 &
+ *	    ./fi_xe_rdmabw -m device -d 0 &
  *	    sleep 1 &&
- *	    ./fi-rdmabw-xe -m host -t read localhost
+ *	    ./fi_xe_rdmabw -m host -t read localhost
  *
  *	RDMA write between memory on the same device:
  *
- *	    ./fi-rdmabw-xe -m device -d 0 &
+ *	    ./fi_xe_rdmabw -m device -d 0 &
  *	    sleep 1 &&
- *	    ./fi-rdmabw-xe -m device -d 0 -t write localhost
+ *	    ./fi_xe_rdmabw -m device -d 0 -t write localhost
  *
  *	RDMA write test between memory on different devices:
  *
- *	    ./fi-rdmabw-xe -m device -d 0 &
+ *	    ./fi_xe_rdmabw -m device -d 0 &
  *	    sleep 1 &&
- *	    ./fi-rdmabw-xe -m device -d 1 -t write localhost
+ *	    ./fi_xe_rdmabw -m device -d 1 -t write localhost
  *
  *	RDMA write test between memory on different devices, use specified
  *	network interface (as OFI domain name):
  *
- *	    ./fi-rdmabw-xe -m device -d 0 -D mlx5_0 &
+ *	    ./fi_xe_rdmabw -m device -d 0 -D mlx5_0 &
  *	    sleep 1 &&
- *	    ./fi-rdmabw-xe -m device -d 1 -D mlx5_1 -t write localhost
+ *	    ./fi_xe_rdmabw -m device -d 1 -D mlx5_1 -t write localhost
  *
  *	For more options:
  *
- *	    ./fi-rdmabw-xe -h
+ *	    ./fi_xe_rdmabw -h
  */
 
 #include <stdio.h>
@@ -98,7 +98,8 @@
 #define MIN_PROXY_BLOCK	(131072)
 #define TX_DEPTH	(128)
 #define RX_DEPTH	(1)
-#define MAX_NICS	(4)
+#define MAX_NICS	(32)
+#define MAX_RAW_KEY_SIZE (256)
 
 enum test_type {
 	READ,
@@ -107,21 +108,31 @@ enum test_type {
 	RECV, /* internal use only */
 };
 
+struct raw_key {
+	uint64_t	size;
+	uint8_t		key[MAX_RAW_KEY_SIZE];
+};
+
 static struct business_card {
 	int num_nics;
 	int num_gpus;
 	struct {
-		struct {
-			uint64_t	one;
-			uint64_t	two;
-			uint64_t	three;
-			uint64_t	four;
+		union {
+			struct {
+				uint64_t	one;
+				uint64_t	two;
+				uint64_t	three;
+				uint64_t	four;
+			};
+			uint8_t bytes[1024];
 		} ep_name;
 	} nics[MAX_NICS];
 	struct {
 		uint64_t	addr;
 		uint64_t	rkeys[MAX_NICS];
+		struct raw_key 	raw_keys[MAX_NICS];
 	} bufs[MAX_GPUS];
+	int use_raw_key;
 } me, peer;
 
 static char			*server_name;
@@ -129,6 +140,7 @@ static char			*prov_name;
 static char			*domain_names;
 static int			client;
 static int			ep_type = FI_EP_RDM;
+static int			use_raw_key;
 
 struct nic {
 	struct fi_info		*fi, *fi_pep;
@@ -155,6 +167,7 @@ struct buf {
 static int			num_gpus;
 static struct buf		bufs[MAX_GPUS];
 static struct buf		proxy_buf;
+static struct buf		sync_buf;
 static int			buf_location = MALLOC;
 
 static int			use_proxy;
@@ -162,6 +175,8 @@ static int			proxy_block = MAX_SIZE;
 static int			use_sync_ofi;
 static int			verify;
 static int			prepost;
+static int			batch = 1;
+static size_t			max_size = MAX_SIZE;
 
 static void init_buf(size_t buf_size, char c)
 {
@@ -187,6 +202,11 @@ static void init_buf(size_t buf_size, char c)
 			exit(-1);
 		}
 	}
+
+	if (!xe_alloc_buf(page_size, page_size, MALLOC, 0, &sync_buf.xe_buf)) {
+		fprintf(stderr, "Couldn't allocate sync buf.\n");
+		exit(-1);
+	}
 }
 
 static void check_buf(size_t size, char c, int gpu)
@@ -228,6 +248,8 @@ static void free_buf(void)
 
 	if (use_proxy)
 		xe_free_buf(proxy_buf.xe_buf.buf, proxy_buf.xe_buf.location);
+
+	xe_free_buf(sync_buf.xe_buf.buf, sync_buf.xe_buf.location);
 }
 
 /*
@@ -323,11 +345,10 @@ static int init_nic(int nic, char *domain_name, char *server_name, int port,
 	if (buf_location != MALLOC)
 		hints->caps |= FI_HMEM;
 	hints->mode = FI_CONTEXT;
-	hints->domain_attr->control_progress = FI_PROGRESS_MANUAL;
 	hints->domain_attr->data_progress = FI_PROGRESS_MANUAL;
 	hints->domain_attr->mr_mode = FI_MR_ALLOCATED | FI_MR_PROV_KEY |
 				      FI_MR_VIRT_ADDR | FI_MR_LOCAL |
-				      FI_MR_HMEM | FI_MR_ENDPOINT;
+				      FI_MR_HMEM | FI_MR_ENDPOINT | FI_MR_RAW;
 	if (domain_name)
 		hints->domain_attr->name = strdup(domain_name);
 
@@ -392,12 +413,16 @@ static int init_nic(int nic, char *domain_name, char *server_name, int port,
 		goto done;
 	}
 
+	if (fi->domain_attr->mr_mode & FI_MR_RAW)
+		use_raw_key = 1;
+
 	for (i = 0; i < num_gpus; i++) {
 		iov.iov_base = bufs[i].xe_buf.buf;
-		iov.iov_len = MAX_SIZE;
+		iov.iov_len = max_size * batch;
 		mr_attr.mr_iov = &iov;
 		mr_attr.iov_count = 1;
-		mr_attr.access = FI_REMOTE_READ | FI_REMOTE_WRITE;
+		mr_attr.access = FI_REMOTE_READ | FI_REMOTE_WRITE |
+				 FI_READ | FI_WRITE | FI_SEND | FI_RECV;
 		mr_attr.requested_key = i + 1;
 		mr_attr.iface = bufs[i].xe_buf.location == MALLOC ?
 					FI_HMEM_SYSTEM : FI_HMEM_ZE;
@@ -414,10 +439,11 @@ static int init_nic(int nic, char *domain_name, char *server_name, int port,
 
 	if (buf_location == DEVICE && use_proxy) {
 		iov.iov_base = proxy_buf.xe_buf.buf;
-		iov.iov_len = MAX_SIZE;
+		iov.iov_len = max_size * batch;
 		mr_attr.mr_iov = &iov;
 		mr_attr.iov_count = 1;
-		mr_attr.access = FI_REMOTE_READ | FI_REMOTE_WRITE;
+		mr_attr.access = FI_REMOTE_READ | FI_REMOTE_WRITE |
+				 FI_READ | FI_WRITE | FI_SEND | FI_RECV;
 		mr_attr.requested_key = i + 1;
 		mr_attr.iface = FI_HMEM_ZE;
 		mr_attr.device.ze = xe_get_dev_num(i);
@@ -431,6 +457,23 @@ static int init_nic(int nic, char *domain_name, char *server_name, int port,
 		proxy_buf.mrs[nic] = mr;
 	}
 
+	iov.iov_base = sync_buf.xe_buf.buf;
+	iov.iov_len = 4;
+	mr_attr.mr_iov = &iov;
+	mr_attr.iov_count = 1;
+	mr_attr.access = FI_SEND | FI_RECV;
+	mr_attr.requested_key = i + 2;
+	mr_attr.iface = FI_HMEM_SYSTEM;
+	mr_attr.device.ze = 0;
+	EXIT_ON_ERROR(fi_mr_regattr(domain, &mr_attr, 0, &mr));
+
+	if (fi->domain_attr->mr_mode & FI_MR_ENDPOINT) {
+		EXIT_ON_ERROR(fi_mr_bind(mr, (fid_t)ep, 0));
+		EXIT_ON_ERROR(fi_mr_enable(mr));
+	}
+
+	sync_buf.mrs[nic] = mr;
+
 done:
 	nics[nic].fi = fi;
 	nics[nic].fi_pep = fi_pep;
@@ -448,8 +491,8 @@ static void show_business_card(struct business_card *bc, char *name)
 {
 	int i, j;
 
-	printf("%s:\tnum_nics %d num_gpus %d ", name, bc->num_nics,
-		bc->num_gpus);
+	printf("%s:\tnum_nics %d num_gpus %d use_raw_key %d", name,
+		bc->num_nics, bc->num_gpus, bc->use_raw_key);
 	for (i = 0; i < bc->num_nics; i++) {
 		printf("[NIC %d] %lx:%lx:%lx:%lx ", i,
 			bc->nics[i].ep_name.one, bc->nics[i].ep_name.two,
@@ -471,19 +514,20 @@ static void init_ofi(int sockfd, char *server_name, int port, int test_type)
 	int err;
 	size_t len;
 	char *domain_name;
+	char *saveptr;
 
 	EXIT_ON_NULL((context_pool = init_context_pool(TX_DEPTH + 1)));
 
 	num_nics = 0;
 	if (domain_names) {
-		domain_name = strtok(domain_names, ",");
+		domain_name = strtok_r(domain_names, ",", &saveptr);
 		while (domain_name && num_nics < MAX_NICS) {
 			err = init_nic(num_nics, domain_name, server_name,
 				       port, test_type);
 			if (err)
 				return;
 			num_nics++;
-			domain_name = strtok(NULL, ",");
+			domain_name = strtok_r(NULL, ",", &saveptr);
 		}
 	} else {
 		err = init_nic(num_nics, NULL, server_name, port, test_type);
@@ -500,10 +544,18 @@ static void init_ofi(int sockfd, char *server_name, int port, int test_type)
 						 &me.nics[j].ep_name, &len));
 			me.bufs[i].rkeys[j]= bufs[i].mrs[j] ?
 						fi_mr_key(bufs[i].mrs[j]) : 0;
+			if (use_raw_key && bufs[i].mrs[j]) {
+				me.bufs[i].raw_keys[j].size = MAX_RAW_KEY_SIZE;
+				fi_mr_raw_attr(bufs[i].mrs[j],
+					       (void *)(uintptr_t)me.bufs[i].addr,
+					       me.bufs[i].raw_keys[j].key,
+					       &me.bufs[i].raw_keys[j].size, 0);
+			}
 		}
 	}
 	me.num_nics = num_nics;
 	me.num_gpus = num_gpus;
+	me.use_raw_key = use_raw_key;
 
 	show_business_card(&me, "Me");
 
@@ -521,6 +573,26 @@ static void init_ofi(int sockfd, char *server_name, int port, int test_type)
 		exit(-1);
 	}
 
+	if (me.use_raw_key != peer.use_raw_key) {
+		printf("The use of raw key doesn't match. Exiting\n");
+		exit(-1);
+	}
+
+	if (use_raw_key) {
+		for (i = 0; i < peer.num_gpus; i++) {
+			for (j = 0; j < peer.num_nics; j++) {
+				if (!peer.bufs[i].rkeys[j])
+					continue;
+				EXIT_ON_ERROR(fi_mr_map_raw(nics[j].domain,
+							    peer.bufs[i].addr,
+							    peer.bufs[i].raw_keys[j].key,
+							    peer.bufs[i].raw_keys[j].size,
+							    &peer.bufs[i].rkeys[j],
+							    0));
+			}
+		}
+	}
+
 	if (ep_type == FI_EP_MSG)
 		return;
 
@@ -536,7 +608,19 @@ static void finalize_ofi(void)
 {
 	int i, j;
 
+	if (use_raw_key) {
+		for (i = 0; i < peer.num_gpus; i++) {
+			for (j = 0; j < peer.num_nics; j++) {
+				if (peer.bufs[i].rkeys[j])
+					fi_mr_unmap_key(nics[j].domain,
+							peer.bufs[i].rkeys[j]);
+			}
+		}
+	}
+
 	for (i = 0; i < num_nics; i++) {
+		if (sync_buf.mrs[i])
+			fi_close((fid_t)sync_buf.mrs[i]);
 		if (buf_location == DEVICE && use_proxy && proxy_buf.mrs[i])
 			fi_close((fid_t)proxy_buf.mrs[i]);
 		for (j = 0; j < num_gpus ; j++)
@@ -702,6 +786,56 @@ try_again:
 	return ret;
 }
 
+static int post_sync_send(int nic, size_t size)
+{
+	struct iovec iov;
+	void *desc = sync_buf.mrs[nic] ? fi_mr_desc(sync_buf.mrs[nic]) : NULL;
+	struct fi_msg msg;
+	int ret;
+
+	iov.iov_base = (char *)sync_buf.xe_buf.buf;
+	iov.iov_len = size;
+	msg.msg_iov = &iov;
+	msg.desc = sync_buf.mrs[nic] ? &desc : NULL;
+	msg.iov_count = 1;
+	msg.addr = nics[nic].peer_addr;
+	msg.context = get_context(context_pool);
+	msg.data = 0;
+
+try_again:
+	ret = fi_sendmsg(nics[nic].ep, &msg, FI_COMPLETION);
+	if (ret == -FI_EAGAIN) {
+		fi_cq_read(nics[nic].cq, NULL, 0);
+		goto try_again;
+	}
+	return ret;
+}
+
+static int post_sync_recv(int nic, size_t size)
+{
+	struct iovec iov;
+	void *desc = sync_buf.mrs[nic] ? fi_mr_desc(sync_buf.mrs[nic]) : NULL;
+	struct fi_msg msg;
+	int ret;
+
+	iov.iov_base = (char *)sync_buf.xe_buf.buf;
+	iov.iov_len = size;
+	msg.msg_iov = &iov;
+	msg.desc = sync_buf.mrs[nic] ? &desc : NULL;
+	msg.iov_count = 1;
+	msg.addr = nics[nic].peer_addr;
+	msg.context = get_context(context_pool);
+	msg.data = 0;
+
+try_again:
+	ret = fi_recvmsg(nics[nic].ep, &msg, FI_COMPLETION);
+	if (ret == -FI_EAGAIN) {
+		fi_cq_read(nics[nic].cq, NULL, 0);
+		goto try_again;
+	}
+	return ret;
+}
+
 static int post_send(int nic, int gpu, size_t size, int idx, int signaled)
 {
 	struct iovec iov;
@@ -784,8 +918,8 @@ static void sync_ofi(size_t size)
 	int i;
 
 	for (i = 0; i < num_nics; i++) {
-		EXIT_ON_ERROR(post_recv(i, 0, size, 0));
-		EXIT_ON_ERROR(post_send(i, 0, size, 0, 1));
+		EXIT_ON_ERROR(post_sync_recv(i, size));
+		EXIT_ON_ERROR(post_sync_send(i, size));
 	}
 
 	wait_completion(num_nics * 2);
@@ -797,7 +931,7 @@ static void sync_send(size_t size)
 	int i;
 
 	for (i = 0; i < num_nics; i++)
-		EXIT_ON_ERROR(post_send(i, 0, size, 0, 1));
+		EXIT_ON_ERROR(post_sync_send(i, size));
 	wait_completion(num_nics);
 	return;
 }
@@ -807,12 +941,12 @@ static void sync_recv(size_t size)
 	int i;
 
 	for (i = 0; i < num_nics; i++)
-		EXIT_ON_ERROR(post_recv(i, 0, size, 0));
+		EXIT_ON_ERROR(post_sync_recv(i, size));
 	wait_completion(num_nics);
 	return;
 }
 
-int run_test(int test_type, int size, int iters, int batch, int output_result)
+int run_test(int test_type, size_t size, int iters, int batch, int output_result)
 {
 	int i, j, k, completed, pending;
 	int n, n0;
@@ -914,13 +1048,13 @@ int run_test(int test_type, int size, int iters, int batch, int output_result)
 		return 0;
 
 	if (output_result)
-		printf("%10d (x %4d) %10.2lf us %12.2lf MB/s\n", size, iters,
+		printf("%10zd (x %4d) %10.2lf us %12.2lf MB/s\n", size, iters,
 		       (t2 - t1), (long)size * iters / (t2 - t1));
 
 	return 0;
 
 err_out:
-	printf("%10d aborted due to fail to post read request\n", size);
+	printf("%10zd aborted due to fail to post read request\n", size);
 	return -1;
 }
 
@@ -934,7 +1068,9 @@ static void usage(char *prog)
 	printf("\t-p <prov_name>   Use the OFI provider named as <prov_name>, default: the first one\n");
 	printf("\t-D <domain_names> Open OFI domain(s) specified as comma separated list of <domain_name>, default: automatic\n");
 	printf("\t-n <iters>       Set the number of iterations for each message size, default: 1000\n");
-	printf("\t-S <size>        Set the message size to test (0: all, -1: none), default: 0\n");
+	printf("\t-b <batch>       Generate completion for every <batch> iterations (default: 1)\n");
+	printf("\t-S <size>        Set the message size to test (0: all, -1: none), can use suffix K/M/G, default: 0\n");
+	printf("\t-M <size>        Set the maximum message size to test, can use suffix K/M/G, default: 4194304 (4M)\n");
 	printf("\t-t <test_type>   Type of test to perform, can be 'read', 'write', or 'send', default: read\n");
 	printf("\t-P               Proxy device buffer through host buffer (for write and send only), default: off\n");
 	printf("\t-B <block_size>  Set the block size for proxying, default: maximum message size\n");
@@ -968,11 +1104,12 @@ static inline int string_to_location(char *s, int default_loc)
 void parse_buf_location(char *string, int *loc1, int *loc2, int default_loc)
 {
 	char *s;
+	char *saveptr;
 
-	s = strtok(string, ":");
+	s = strtok_r(string, ":", &saveptr);
 	if (s) {
 		*loc1 = string_to_location(s, default_loc);
-		s = strtok(NULL, ":");
+		s = strtok_r(NULL, ":", &saveptr);
 		if (s)
 			*loc2 = string_to_location(s, default_loc);
 		else
@@ -982,6 +1119,23 @@ void parse_buf_location(char *string, int *loc1, int *loc2, int default_loc)
 	}
 }
 
+size_t parse_size(char *string)
+{
+	size_t size = MAX_SIZE;
+	char unit = '\0';
+
+	sscanf(string, "%zd%c", &size, &unit);
+
+	if (unit == 'k' || unit == 'K')
+		size *= 1024;
+	else if (unit == 'm' || unit == 'M')
+		size *= 1024 * 1024;
+	else if (unit == 'g' || unit == 'G')
+		size *= 1024 * 1024 * 1024;
+
+	return size;
+}
+
 int main(int argc, char *argv[])
 {
 	char *gpu_dev_nums = NULL;
@@ -989,26 +1143,37 @@ int main(int argc, char *argv[])
 	unsigned int port = 12345;
 	int test_type = READ;
 	int iters = 1000;
-	int batch = 1; /* was 16. however, when > 1, for large messages, rxd
-			* hangs and rxm gets i/o error (err=5, prov_errno=10)
-			*/
 	int reverse = 0;
 	int bidir = 0;
 	int sockfd;
-	int size;
+	size_t size;
 	int c;
 	int initiator;
-	int msg_size = 0;
+	ssize_t msg_size = 0;
+	size_t warm_up_size;
 	int err = 0;
 	int rank;
 	char *s;
 	int loc1 = MALLOC, loc2 = MALLOC;
 
-	while ((c = getopt(argc, argv, "2d:D:e:p:m:n:t:gPB:rRsS:x:hv")) != -1) {
+	while ((c = getopt(argc, argv, "2b:d:D:e:p:m:M:n:t:gPB:rRsS:x:hv")) != -1) {
 		switch (c) {
 		case '2':
 			bidir = 1;
 			break;
+		case 'b':
+			batch = atoi(optarg);
+			if (batch <= 0) {
+				fprintf(stderr,
+					"Batch too small, adjusted to 1\n");
+				batch = 1;
+			} else if (batch > TX_DEPTH) {
+				fprintf(stderr,
+					"Batch too large, adjusted to %d\n",
+					TX_DEPTH);
+				batch = TX_DEPTH;
+			}
+			break;
 		case 'd':
 			gpu_dev_nums = strdup(optarg);
 			break;
@@ -1060,7 +1225,11 @@ int main(int argc, char *argv[])
 			use_sync_ofi = 1;
 			break;
 		case 'S':
-			msg_size = atoi(optarg);
+			msg_size = parse_size(optarg);
+			break;
+		case 'M':
+			max_size = parse_size(optarg);
+			proxy_block = max_size;
 			break;
 		case 'x':
 			prepost = atoi(optarg);
@@ -1075,6 +1244,21 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	if (msg_size > 0 && msg_size > max_size) {
+		max_size = msg_size;
+		proxy_block = msg_size;
+		fprintf(stderr,
+			"Max_size smaller than message size, adjusted to %zd\n",
+			max_size);
+	}
+
+	if (max_size * batch / batch != max_size) {
+		fprintf(stderr,
+			"Buffer_size = Max_size (%zd) * Batch (%d) overflows\n",
+			max_size, batch);
+		exit(-1);
+	}
+
 	if (argc > optind) {
 		client = 1;
 		server_name = strdup(argv[optind]);
@@ -1111,25 +1295,26 @@ int main(int argc, char *argv[])
 	enable_multi_gpu = buf_location != MALLOC && buf_location != HOST;
 	num_gpus = xe_init(gpu_dev_nums, enable_multi_gpu);
 
-	init_buf(MAX_SIZE * batch, initiator ? 'A' : 'a');
+	init_buf(max_size * batch, initiator ? 'A' : 'a');
 	init_ofi(sockfd, server_name, port + 1000, test_type);
 
 	sync_tcp(sockfd);
 	printf("Warming up ...\n");
+	warm_up_size = msg_size > 0 ? msg_size : 1;
 	if (initiator) {
-		run_test(test_type, 1, 16, 1, 0);
+		run_test(test_type, warm_up_size, 16, 1, 0);
 		sync_send(4);
 	} else {
 		if (test_type == SEND)
-			run_test(RECV, 1, 16, 1, 0);
+			run_test(RECV, warm_up_size, 16, 1, 0);
 		else if (bidir)
-			run_test(test_type, 1, 16, 1, 0);
+			run_test(test_type, warm_up_size, 16, 1, 0);
 		sync_recv(4);
 	}
 
 	sync_tcp(sockfd);
 	printf("Start test ...\n");
-	for (size = 1; size <= MAX_SIZE && !err; size <<= 1) {
+	for (size = 1; size <= max_size && !err; size <<= 1) {
 		if (msg_size < 0)
 			break;
 		else if (msg_size > 0)
