@@ -34,7 +34,7 @@
 #include <pthread.h>
 #include <sys/stat.h>
 #include <sys/types.h>
-#include <wait.h>
+#include <sys/wait.h>
 #include <unistd.h>
 
 #include <rdma/fi_cm.h>
@@ -68,6 +68,7 @@ enum {
 struct rpc_ctrl {
 	uint32_t op;
 	uint64_t size;
+	uint64_t count;
 	union {
 		uint64_t offset;
 		uint64_t tag;
@@ -186,7 +187,7 @@ static int rpc_inject(struct rpc_hdr *hdr, fi_addr_t addr)
 
 	start = ft_gettime_ms();
 	do {
-		fi_cq_read(txcq, NULL, 0);
+		(void) fi_cq_read(txcq, NULL, 0);
 		ret = (int) fi_inject(ep, hdr, sizeof(*hdr), addr);
 	} while ((ret == -FI_EAGAIN) && (ft_gettime_ms() - start < rpc_timeout));
 
@@ -204,7 +205,7 @@ static int rpc_send(struct rpc_hdr *hdr, size_t size, fi_addr_t addr)
 
 	start = ft_gettime_ms();
 	do {
-		fi_cq_read(txcq, NULL, 0);
+		(void) fi_cq_read(txcq, NULL, 0);
 		ret = (int) fi_send(ep, hdr, size, NULL, addr, hdr);
 	} while ((ret == -FI_EAGAIN) && (ft_gettime_ms() - start < rpc_timeout));
 
@@ -235,7 +236,7 @@ static int rpc_deliver(struct rpc_hdr *hdr, size_t size, fi_addr_t addr)
 
 	start = ft_gettime_ms();
 	do {
-		fi_cq_read(txcq, NULL, 0);
+		(void) fi_cq_read(txcq, NULL, 0);
 		ret = (int) fi_sendmsg(ep, &msg, FI_DELIVERY_COMPLETE);
 	} while ((ret == -FI_EAGAIN) && (ft_gettime_ms() - start < rpc_timeout));
 
@@ -454,7 +455,7 @@ static int rpc_reg_buf(struct rpc_ctrl *ctrl, size_t size, uint64_t access)
 	return FI_SUCCESS;
 
 close:
-	fi_close(&ctrl->mr->fid);
+	FT_CLOSE_FID(ctrl->mr);
 	return ret;
 }
 
@@ -469,7 +470,9 @@ static int rpc_read_req(struct rpc_ctrl *ctrl)
 	if (!ctrl->buf)
 		return -FI_ENOMEM;
 
-	ft_fill_buf(&ctrl->buf[ctrl->offset], ctrl->size);
+	ret = ft_fill_buf(&ctrl->buf[ctrl->offset], ctrl->size);
+	if (ret)
+		goto free;
 
 	ret = rpc_reg_buf(ctrl, size, FI_REMOTE_READ);
 	if (ret)
@@ -491,7 +494,7 @@ static int rpc_read_req(struct rpc_ctrl *ctrl)
 	return 0;
 
 close:
-	fi_close(&ctrl->mr->fid);
+	FT_CLOSE_FID(ctrl->mr);
 free:
 	free(ctrl->buf);
 	return ret;
@@ -512,7 +515,7 @@ static int rpc_read_resp(struct rpc_ctrl *ctrl)
 	ret = ft_check_buf(&req->buf[req->offset], req->size);
 
 close:
-	fi_close(&req->mr->fid);
+	FT_CLOSE_FID(req->mr);
 	free(req->buf);
 	return ret;
 }
@@ -548,7 +551,7 @@ static int rpc_write_req(struct rpc_ctrl *ctrl)
 	return 0;
 
 close:
-	fi_close(&ctrl->mr->fid);
+	FT_CLOSE_FID(ctrl->mr);
 free:
 	free(ctrl->buf);
 	return ret;
@@ -569,7 +572,7 @@ static int rpc_write_resp(struct rpc_ctrl *ctrl)
 	ret = ft_check_buf(&req->buf[req->offset], req->size);
 
 close:
-	fi_close(&req->mr->fid);
+	FT_CLOSE_FID(req->mr);
 	free(req->buf);
 	return ret;
 }
@@ -610,7 +613,7 @@ int (*ctrl_op[op_last])(struct rpc_ctrl *ctrl) = {
 
 static int run_child(void)
 {
-	int i, ret;
+	int i, j, ret;
 
 	printf("(%d-?) running\n", myid);
 	ret = ft_init_fabric();
@@ -631,9 +634,11 @@ static int run_child(void)
 		goto free;
 
 	for (i = 0; i < ctrl_cnt && !ret; i++) {
-		printf("(%d-%d) rpc op %s\n", myid, id_at_server,
-		       rpc_op_str(ctrls[i].op));
-		ret = ctrl_op[ctrls[i].op](&ctrls[i]);
+		for (j = 0; j < ctrls[i].count && !ret; j++) {
+			printf("(%d-%d) rpc op %s iteration %d\n", myid, id_at_server,
+			       rpc_op_str(ctrls[i].op), j);
+			ret = ctrl_op[ctrls[i].op](&ctrls[i]);
+		}
 	}
 
 free:
@@ -710,6 +715,7 @@ static void init_rpc_ctrl(struct rpc_ctrl *ctrl)
 {
 	ctrl->op = op_last;
 	ctrl->size = 0;
+	ctrl->count = 1;
 	ctrl->offset = 0;
 	ctrl->buf = 0;
 	ctrl->mr = 0;
@@ -762,6 +768,11 @@ static bool add_ctrl(const char *js, int njts, jsmntok_t *jts,
 			t = &jts[*idx];
 			if (!get_uint64_val(js, t, &ctrl->size))
 				goto err_out;
+		} else if (FT_TOKEN_CHECK(ks, len, "count")) {
+			(*idx)++;
+			t = &jts[*idx];
+			if (!get_uint64_val(js, t, &ctrl->count))
+				goto err_out;
 		} else {
 			goto err_out;
 		}
@@ -801,7 +812,8 @@ static int init_ctrls(const char *ctrlfile)
 
 	if (stat(ctrlfile, &sb)) {
 		FT_PRINTERR("stat", -errno);
-		return -errno;
+		ret = -errno;
+		goto no_mem_out;
 	}
 
 	js = malloc(sb.st_size + 1);
@@ -978,7 +990,8 @@ static void complete_rpc(struct rpc_resp *resp)
 	}
 
 	if (resp->mr)
-		fi_close(&resp->mr->fid);
+		FT_CLOSE_FID(resp->mr);
+
 	(void) ft_check_buf(resp + 1, resp->hdr.size);
 	free(resp);
 }
@@ -1100,7 +1113,11 @@ static void start_rpc(struct rpc_hdr *req)
 		goto free;
 
 	resp->hdr = *req;
-	ft_fill_buf(resp + 1, resp->hdr.size);
+	ret = ft_fill_buf(resp + 1, resp->hdr.size);
+	if (ret) {
+		free(resp);
+		goto free;
+	}
 
 	start = ft_gettime_ms();
 	do {
