@@ -63,6 +63,7 @@ declare C_ARGS=""
 declare S_ARGS=""
 declare PIN_CORE=""
 declare PROVIDER_TESTS=0
+declare gdb_cmd=""
 
 declare cur_excludes=""
 declare file_excludes=""
@@ -100,16 +101,16 @@ functional_tests=(
 	"fi_av_xfer -e rdm"
 	"fi_av_xfer -e dgram"
 	"fi_cm_data"
-	"fi_cq_data -e msg"
-	"fi_cq_data -e rdm"
-	"fi_cq_data -e dgram"
+	"fi_cq_data -e msg -o senddata"
+	"fi_cq_data -e rdm -o senddata"
+	"fi_cq_data -e dgram -o senddata"
+	"fi_cq_data -e msg -o writedata"
+	"fi_cq_data -e rdm -o writedata"
+	"fi_cq_data -e dgram -o writedata"
 	"fi_dgram"
-	"fi_dgram_waitset"
 	"fi_msg"
 	"fi_msg_epoll"
 	"fi_msg_sockets"
-	"fi_poll -t queue"
-	"fi_poll -t counter"
 	"fi_rdm"
 	"fi_rdm -U"
 	"fi_rdm_rma_event"
@@ -128,18 +129,30 @@ functional_tests=(
 	"fi_rdm_shared_av"
 	"fi_multi_mr -e msg -V"
 	"fi_multi_mr -e rdm -V"
-	"fi_multi_ep -e msg -v"
-	"fi_multi_ep -e rdm -v"
+	"fi_multi_ep -e msg -v -A"
+	"fi_multi_ep -e rdm -v -A"
+	"fi_multi_ep -e msg -v -Q"
+	"fi_multi_ep -e rdm -v -Q"
+	"fi_multi_ep -e msg -v -A -Q"
+	"fi_multi_ep -e rdm -v -A -Q"
+	"fi_multi_ep -e msg -v --threading completion"
+	"fi_multi_ep -e rdm -v --threading completion"
+	"fi_multi_ep -e msg -v -A --threading completion"
+	"fi_multi_ep -e rdm -v -A --threading completion"
+	"fi_multi_ep -e msg -v -Q --threading completion"
+	"fi_multi_ep -e rdm -v -Q --threading completion"
+	"fi_multi_ep -e msg -v -A -Q --threading completion"
+	"fi_multi_ep -e rdm -v -A -Q --threading completion"
 	"fi_recv_cancel -e rdm -V"
-	"fi_unexpected_msg -e msg -I 10"
-	"fi_unexpected_msg -e rdm -I 10"
-	"fi_msg_inject -A inject -v"
-	"fi_msg_inject -N -A inject -v"
-	"fi_msg_inject -A inj_complete -v"
-	"fi_msg_inject -N -A inj_complete -v"
-	"fi_bw -e rdm -v -T 1"
-	"fi_bw -e rdm -v -T 1 -U"
-	"fi_bw -e msg -v -T 1"
+	"fi_unexpected_msg -e msg -I 10 -v"
+	"fi_unexpected_msg -e rdm -I 10 -v"
+	"fi_inject_test -A inject -v"
+	"fi_inject_test -N -A inject -v"
+	"fi_inject_test -A inj_complete -v"
+	"fi_inject_test -N -A inj_complete -v"
+	"fi_flood -e rdm -v -T 1"
+	"fi_flood -e rdm -v -T 1 -U"
+	"fi_flood -e msg -v -T 1"
 	"fi_rdm_multi_client -C 10 -I 5"
 	"fi_rdm_multi_client -C 10 -I 5 -U"
 )
@@ -160,6 +173,8 @@ short_tests=(
 	"fi_rma_bw -e rdm -o writedata -I 5 -U"
 	"fi_rdm_atomic -I 5 -o all"
 	"fi_rdm_atomic -I 5 -o all -U"
+	"fi_rdm_atomic -I 5 -o all -v"
+	"fi_rdm_atomic -I 5 -o all -U -v"
 	"fi_rdm_cntr_pingpong -I 5"
 	"fi_multi_recv -e rdm -I 5"
 	"fi_multi_recv -e msg -I 5"
@@ -196,6 +211,8 @@ standard_tests=(
 	"fi_rma_bw -e rdm -o writedata -U"
 	"fi_rdm_atomic -o all -I 1000"
 	"fi_rdm_atomic -o all -I 1000 -U"
+	"fi_rdm_atomic -o all -I 1000 -v"
+	"fi_rdm_atomic -o all -I 1000 -U -v"
 	"fi_rdm_cntr_pingpong"
 	"fi_multi_recv -e rdm"
 	"fi_multi_recv -e msg"
@@ -228,6 +245,7 @@ unit_tests=(
 	"fi_cq_test"
 	"fi_mr_test"
 	"fi_cntr_test"
+	"fi_setopt_test"
 )
 
 regression_tests=(
@@ -239,8 +257,8 @@ complex_tests=(
 )
 
 multinode_tests=(
-	"fi_multinode -C msg"
-	"fi_multinode -C rma"
+	"fi_multinode -x msg"
+	"fi_multinode -x rma"
 	"fi_multinode_coll"
 )
 
@@ -449,7 +467,7 @@ function unit_test {
 
 	start_time=$(date '+%s')
 
-	cmd="${BIN_PATH}${test_exe}"
+	cmd="${gdb_cmd} ${BIN_PATH}${test_exe}"
 	${SERVER_CMD} "${EXPORT_ENV} $cmd" &> $s_outp &
 	p1=$!
 
@@ -504,7 +522,7 @@ function cs_test {
 	else
 		s_arg="-s $S_INTERFACE"
 	fi
-	s_cmd="${BIN_PATH}${test_exe} ${S_ARGS} ${pin_core} $s_arg"
+	s_cmd="${gdb_cmd} ${BIN_PATH}${test_exe} ${S_ARGS} ${pin_core} $s_arg"
 	${SERVER_CMD} "${EXPORT_ENV} $s_cmd" &> $s_outp &
 	s_pid=$!
 	sleep 1
@@ -514,7 +532,7 @@ function cs_test {
 	else
 		c_arg="-s $C_INTERFACE $S_INTERFACE"
 	fi
-	c_cmd="${BIN_PATH}${test_exe} ${C_ARGS} ${pin_core} $c_arg"
+	c_cmd="${gdb_cmd} ${BIN_PATH}${test_exe} ${C_ARGS} ${pin_core} $c_arg"
 	${CLIENT_CMD} "${EXPORT_ENV} $c_cmd" &> $c_outp &
 	c_pid=$!
 
@@ -605,12 +623,12 @@ function complex_test {
 		opts+=" -E"
 	fi
 
-	s_cmd="${BIN_PATH}${test_exe} ${S_ARGS} -x $opts"
+	s_cmd="${gdb_cmd} ${BIN_PATH}${test_exe} ${S_ARGS} -x $opts"
 	FI_LOG_LEVEL=error ${SERVER_CMD} "${EXPORT_ENV} $s_cmd" &> $s_outp &
 	s_pid=$!
 	sleep 1
 
-	c_cmd="${BIN_PATH}${test_exe} ${C_ARGS} -u "${COMPLEX_CFG}" $S_INTERFACE $opts"
+	c_cmd="${gdb_cmd} ${BIN_PATH}${test_exe} ${C_ARGS} -u "${COMPLEX_CFG}" $S_INTERFACE $opts"
 	FI_LOG_LEVEL=error ${CLIENT_CMD} "${EXPORT_ENV} $c_cmd" &> $c_outp &
 	c_pid=$!
 
@@ -668,7 +686,7 @@ function multinode_test {
 
 	start_time=$(date '+%s')
 
-	s_cmd="${BIN_PATH}${test_exe} ${S_ARGS} -s ${S_INTERFACE}"
+	s_cmd="${gdb_cmd} ${BIN_PATH}${test_exe} ${S_ARGS} -s ${S_INTERFACE}"
 	${SERVER_CMD} "${EXPORT_ENV} $s_cmd" &> $s_outp &
 	s_pid=$!
 	sleep 1
@@ -677,7 +695,7 @@ function multinode_test {
 	for ((i=1; i<num_procs; i++))
 	do
 		local c_out=$(mktemp fabtests.c_outp${i}.XXXXXX)
-		c_cmd="${BIN_PATH}${test_exe} ${S_ARGS} -s ${S_INTERFACE}"
+		c_cmd="${gdb_cmd} ${BIN_PATH}${test_exe} ${S_ARGS} -s ${S_INTERFACE}"
 		${CLIENT_CMD} "${EXPORT_ENV} $c_cmd" &> $c_out &
 		c_pid_arr+=($!)
 		c_out_arr+=($c_out)
@@ -893,10 +911,11 @@ function usage {
 	errcho -e " -b\tenable out-of-band address exchange over the default port"
 	errcho -e " -P\tRun provider specific tests"
 	errcho -e " --pin-core\tSpecify cores to pin when running standard tests. Cores can specified via a comma-delimited list, e.g. 0,2-4"
+	errcho -e " -G\tRun with gdb and print backtraces"
 	exit 1
 }
 
-Options=$(getopt --options v,t:,p:,g:,e:,f:,c:,s:,u:,T:,C:,L:,N,R,S,b,k,P,E:h \
+Options=$(getopt --options v,t:,p:,g:,e:,f:,c:,s:,u:,T:,C:,L:,N,R,S,b,k,P,E:,G,h \
 		 --longoptions pin-core:,help \
 		 --quiet \
 		 -- "$@")
@@ -954,6 +973,9 @@ while true; do
 			EXPORT_ENV="$EXPORT_ENV $EXPORT_STRING ;"
 		    fi
 		    shift 2 ;;
+		-G)
+		    gdb_cmd="gdb -batch -ex run -ex bt -ex quit --args";
+		    shift ;;
 		--pin-core)
 		    PIN_CORE=$2; shift 2 ;;
 		-h | --help)
