@@ -34,6 +34,38 @@
 
 #include "fabtest.h"
 
+int ft_random_fd = -1;
+
+static inline void ft_init_random(void)
+{
+	ft_random_fd = open("/dev/urandom", O_RDONLY);
+	if (ft_random_fd < 0)
+		FT_PRINTERR("ft_open_control", ft_random_fd);
+}
+
+static inline void ft_cleanup_random(void)
+{
+	if (ft_random_fd >= 0) {
+		close(ft_random_fd);
+		ft_random_fd = -1;
+	}
+}
+
+static int get_random_data(void)
+{
+	ssize_t ret;
+	static int rdata = 1;
+
+	if (ft_random_fd < 0)
+		return rdata;
+
+	ret = read(ft_random_fd, &rdata, sizeof(rdata));
+	if (ret < 0)
+		ft_record_error(ret);
+
+	return rdata;
+}
+
 void ft_record_error(int error)
 {
 	if (!ft_ctrl.error) {
@@ -138,6 +170,8 @@ static int ft_init_control(void)
 static void ft_cleanup_xcontrol(struct ft_xcontrol *ctrl)
 {
 	ft_hmem_free(opts.iface, ctrl->buf);
+	free(ctrl->cpy_buf);
+	FT_CLOSE_FID(ctrl->mr);
 	free(ctrl->iov);
 	free(ctrl->iov_desc);
 	free(ctrl->ctx);
@@ -148,6 +182,7 @@ static void ft_cleanup_atomic_control(struct ft_atomic_control *ctrl)
 {
 	free(ctrl->res_buf);
 	free(ctrl->comp_buf);
+	FT_CLOSE_FID(ctrl->comp_mr);
 	free(ctrl->ioc);
 	free(ctrl->res_ioc);
 	free(ctrl->comp_ioc);
@@ -158,6 +193,7 @@ static void ft_cleanup_atomic_control(struct ft_atomic_control *ctrl)
 static void ft_cleanup_mr_control(struct ft_mr_control *ctrl)
 {
 	free(ctrl->buf);
+	FT_CLOSE_FID(ctrl->mr);
 	memset(ctrl, 0, sizeof *ctrl);
 }
 
@@ -243,7 +279,7 @@ static void ft_format_iov_random(struct iovec *iov, size_t cnt, char *buf,
 			 * the remaining IOV count. This is so we can reserve at
 			 * least a length of 1 for every IOV.
 			 */
-			weight = (rand() % (len - (cnt - i) + 1)) + 1;
+			weight = (get_random_data() % (len - (cnt - i) + 1)) + 1;
 		}
 
 		len -= weight;
@@ -268,7 +304,7 @@ void ft_format_iov(struct iovec *iov, size_t cnt, char *buf, size_t len)
 		ft_format_iov_random
 	};
 
-	choice = rand() % ARRAY_SIZE(options);
+	choice = get_random_data() % ARRAY_SIZE(options);
 
 	options[choice](iov, cnt, buf, len);
 }
@@ -365,7 +401,7 @@ static int ft_sync_test(int value)
 	if (ret)
 		return ret;
 
-	return ft_sock_sync(value);
+	return ft_sock_sync(sock, value);
 }
 
 static int ft_sync_manual()
@@ -408,7 +444,7 @@ static int ft_sync_progress(int value)
 {
 	if (test_info.progress == FI_PROGRESS_MANUAL)
 		return ft_sync_manual();
-	return ft_sock_sync(value);
+	return ft_sock_sync(sock, value);
 }
 
 static int ft_sync_msg_needed(void)
@@ -834,9 +870,9 @@ static int ft_unit_atomic(void)
 
 	ft_atom_ctrl.count = ft_tx_ctrl.rma_msg_size / ft_atom_ctrl.datatype_size;
 	if (ret == -FI_ENOSYS || ret == -FI_EOPNOTSUPP ||
-	    ft_atom_ctrl.count > count || ft_atom_ctrl.count == 0) {
+	    ft_atom_ctrl.count > count || ft_atom_ctrl.count == 0)
 		return 0;
-	}
+
 	if (ret)
 		return ret;
 
@@ -972,6 +1008,9 @@ static int ft_run_unit(void)
 
 void ft_cleanup(void)
 {
+	if (ft_mr_ctrl.mr && (test_info.mr_mode & FI_MR_RAW))
+		fi_mr_unmap_key(domain, ft_mr_ctrl.peer_mr_key);
+
 	FT_CLOSE_FID(ft_rx_ctrl.mr);
 	FT_CLOSE_FID(ft_tx_ctrl.mr);
 	FT_CLOSE_FID(ft_mr_ctrl.mr);
@@ -979,42 +1018,76 @@ void ft_cleanup(void)
 	FT_CLOSE_FID(ft_atom_ctrl.comp_mr);
 	ft_cleanup_xcontrol(&ft_rx_ctrl);
 	ft_cleanup_xcontrol(&ft_tx_ctrl);
+	ft_free_host_bufs();
 	ft_cleanup_mr_control(&ft_mr_ctrl);
 	ft_cleanup_atomic_control(&ft_atom_ctrl);
+	ft_cleanup_random();
 	ft_free_res();
 	memset(&ft_ctrl, 0, sizeof ft_ctrl);
 }
 
 static int ft_exchange_mr_addr_key(void)
 {
-	struct fi_rma_iov local_rma_iov = {0};
-	struct fi_rma_iov peer_rma_iov = {0};
+	char temp1[FT_MAX_CTRL_MSG];
+	char temp2[FT_MAX_CTRL_MSG];
+	struct fi_rma_iov *rma_iov = (void *) temp1;
+	struct fi_rma_iov *peer_rma_iov = (void *) temp2;
+	size_t len, key_size = 0;
+	uint64_t addr;
 	int ret;
 
 	if (!(test_info.mr_mode & (FI_MR_VIRT_ADDR | FI_MR_PROV_KEY)))
 		return 0;
 
 	if (test_info.mr_mode & FI_MR_VIRT_ADDR)
-		local_rma_iov.addr = (uint64_t) ft_mr_ctrl.buf;
+		rma_iov->addr = (uint64_t) ft_mr_ctrl.buf;
+	else
+		rma_iov->addr = 0;
 
-	if (test_info.mr_mode & FI_MR_PROV_KEY)
-		local_rma_iov.key = ft_mr_ctrl.mr_key;
+	if (ft_mr_ctrl.mr && (test_info.mr_mode & FI_MR_RAW)) {
+		ret = fi_mr_raw_attr(ft_mr_ctrl.mr, &addr, NULL, &key_size, 0);
+		if (ret != -FI_ETOOSMALL)
+			return ret;
+		len = sizeof(*rma_iov) + key_size - sizeof(rma_iov->key);
+		if (len > FT_MAX_CTRL_MSG) {
+			FT_PRINTERR("Raw key too large for ctrl message",
+				    -FI_ETOOSMALL);
+			return -FI_ETOOSMALL;
+		}
+		ret = fi_mr_raw_attr(ft_mr_ctrl.mr, &addr, (uint8_t *) &rma_iov->key,
+				     &key_size, 0);
+		if (ret)
+			return ret;
+	} else {
+		len = sizeof(*rma_iov);
+		if (test_info.mr_mode & FI_MR_PROV_KEY)
+			rma_iov->key = ft_mr_ctrl.mr_key;
+	}
 
-	ret = ft_sock_send(sock, &local_rma_iov, sizeof local_rma_iov);
+	ret = ft_sock_send(sock, rma_iov, len);
 	if (ret) {
 		FT_PRINTERR("ft_sock_send", ret);
 		return ret;
 	}
 
-	ret = ft_sock_recv(sock, &peer_rma_iov, sizeof peer_rma_iov);
+	ret = ft_sock_recv(sock, peer_rma_iov, len);
 	if (ret) {
 		FT_PRINTERR("ft_sock_recv", ret);
 		return ret;
 	}
 
-	ft_mr_ctrl.peer_mr_addr = peer_rma_iov.addr;
-	if (test_info.mr_mode & FI_MR_PROV_KEY)
-		ft_mr_ctrl.peer_mr_key = peer_rma_iov.key;
+	ft_mr_ctrl.peer_mr_addr = peer_rma_iov->addr;
+
+        if (ft_mr_ctrl.mr && (test_info.mr_mode & FI_MR_RAW)) {
+                ret = fi_mr_map_raw(domain, peer_rma_iov->addr,
+                                    (uint8_t *) &peer_rma_iov->key, key_size,
+                                    &ft_mr_ctrl.peer_mr_key, 0);
+                if (ret)
+                        return ret;
+        } else {
+		if (test_info.mr_mode & FI_MR_PROV_KEY)
+			ft_mr_ctrl.peer_mr_key = peer_rma_iov->key;
+        }
 
 	return 0;
 }
@@ -1034,6 +1107,8 @@ int ft_open_res()
 		FT_PRINTERR("ft_open_control", ret);
 		goto cleanup;
 	}
+
+	ft_init_random();
 	if (test_info.ep_type == FI_EP_MSG && listen_sock >= 0) {
 		ret = ft_open_passive();
 		if (ret) {
@@ -1059,7 +1134,7 @@ int ft_init_test()
 {
 	int ret;
 
-	ft_sock_sync(0);
+	ft_sock_sync(sock, 0);
 
 	ret = ft_enable_comm();
 	if (ret) {
