@@ -245,7 +245,7 @@ eq_wait_fd_poll()
 	}
 
 	fids[0] = &eq->fid;
-	if (fi_trywait(fabric, fids, 1) != FI_SUCCESS) {
+	if ((ret = fi_trywait(fabric, fids, 1)) != FI_SUCCESS) {
 		sprintf(err_buf, "fi_trywait ret=%d, %s", ret, fi_strerror(-ret));
 		goto fail;
 	}
@@ -610,10 +610,8 @@ int main(int argc, char **argv)
 	}
 
 	hints->mode = FI_CONTEXT | FI_CONTEXT2 | FI_MSG_PREFIX | FI_ASYNC_IOV |
-		FI_RX_CQ_DATA | FI_NOTIFY_FLAGS_ONLY | FI_RESTRICTED_COMP |
-		FI_BUFFERED_RECV;
-	hints->domain_attr->mode = FI_RESTRICTED_COMP;
-	hints->domain_attr->mr_mode = ~(FI_MR_BASIC | FI_MR_SCALABLE);
+		FI_RX_CQ_DATA;
+	hints->domain_attr->mr_mode = ~OFI_MR_DEPRECATED;
 
 	ret = fi_getinfo(FT_FIVERSION, NULL, 0, 0, hints, &fi);
 	if (ret) {
