@@ -45,7 +45,8 @@
 					    getinfo_ ## name ## _desc)
 
 typedef int (*ft_getinfo_init)(struct fi_info *);
-typedef int (*ft_getinfo_test)(char *, char *, uint64_t, struct fi_info *, struct fi_info **);
+typedef int (*ft_getinfo_test)(char *, char *, uint64_t, struct fi_info *,
+				struct fi_info **);
 typedef int (*ft_getinfo_check)(struct fi_info *);
 typedef int (*ft_getinfo_init_val)(struct fi_info *, uint64_t);
 typedef int (*ft_getinfo_check_val)(struct fi_info *, uint64_t);
@@ -142,7 +143,8 @@ static int validate_bit_combos(char *node, char *service, uint64_t flags,
 			if (check && check(fi, combinations[i])) {
 				FT_DEBUG("%s:failed check for caps [%s]\n",
 					 fi->fabric_attr->prov_name,
-					 fi_tostr(&combinations[i], FI_TYPE_CAPS));
+					 fi_tostr(&combinations[i],
+					 FI_TYPE_CAPS));
 				ret = -FI_EIO;
 			}
 		}
@@ -179,8 +181,7 @@ static int validate_tx_ordering_bits(char *node, char *service, uint64_t flags,
 		struct fi_info *hints, struct fi_info **info)
 {
 	return validate_bit_combos(node, service, flags, hints, info,
-				   FI_ORDER_STRICT | FI_ORDER_DATA,
-				   init_tx_order, check_tx_order);
+				   ~0ULL, init_tx_order, check_tx_order);
 }
 
 static int init_rx_order(struct fi_info *hints, uint64_t order)
@@ -198,8 +199,7 @@ static int validate_rx_ordering_bits(char *node, char *service, uint64_t flags,
 		struct fi_info *hints, struct fi_info **info)
 {
 	return validate_bit_combos(node, service, flags, hints, info,
-				   FI_ORDER_STRICT | FI_ORDER_DATA,
-				   init_rx_order, check_rx_order);
+				   ~0ULL, init_rx_order, check_rx_order);
 }
 
 static int init_caps(struct fi_info *hints, uint64_t bits)
@@ -212,8 +212,7 @@ static int init_caps(struct fi_info *hints, uint64_t bits)
 			 FI_MULTICAST | FI_NAMED_RX_CTX | FI_HMEM | \
 			 FI_COLLECTIVE)
 #define PRIMARY_RX_CAPS (FI_MSG | FI_RMA | FI_TAGGED | FI_ATOMIC | \
-			 FI_DIRECTED_RECV | FI_VARIABLE_MSG | \
-			 FI_HMEM | FI_COLLECTIVE)
+			 FI_DIRECTED_RECV | FI_HMEM | FI_COLLECTIVE)
 
 #define PRIMARY_CAPS (PRIMARY_TX_CAPS | PRIMARY_RX_CAPS)
 #define DOMAIN_CAPS (FI_LOCAL_COMM | FI_REMOTE_COMM | FI_SHARED_AV)
@@ -336,8 +335,10 @@ static int init_valid_rma_RAW_ordering_set_size(struct fi_info *hints)
 			-ret, fi_strerror(-ret));
 		return ret;
 	}
-	if (fi->ep_attr->max_order_raw_size > 0)
-		hints->ep_attr->max_order_raw_size = fi->ep_attr->max_order_raw_size - 1;
+	if (fi->ep_attr->max_order_raw_size > 0) {
+		hints->ep_attr->max_order_raw_size =
+				fi->ep_attr->max_order_raw_size - 1;
+	}
 
 	fi_freeinfo(fi);
 
@@ -369,8 +370,10 @@ static int init_valid_rma_WAR_ordering_set_size(struct fi_info *hints)
 			-ret, fi_strerror(-ret));
 		return ret;
 	}
-	if (fi->ep_attr->max_order_war_size > 0)
-		hints->ep_attr->max_order_war_size = fi->ep_attr->max_order_war_size - 1;
+	if (fi->ep_attr->max_order_war_size > 0) {
+		hints->ep_attr->max_order_war_size =
+				fi->ep_attr->max_order_war_size - 1;
+	}
 
 	fi_freeinfo(fi);
 
@@ -401,8 +404,10 @@ static int init_valid_rma_WAW_ordering_set_size(struct fi_info *hints)
 			-ret, fi_strerror(-ret));
 		return ret;
 	}
-	if (fi->ep_attr->max_order_waw_size > 0)
-		hints->ep_attr->max_order_waw_size = fi->ep_attr->max_order_waw_size - 1;
+	if (fi->ep_attr->max_order_waw_size > 0) {
+		hints->ep_attr->max_order_waw_size =
+				fi->ep_attr->max_order_waw_size - 1;
+	}
 
 	fi_freeinfo(fi);
 
@@ -416,7 +421,8 @@ static int check_valid_rma_ordering_sizes(struct fi_info *info)
 		if (info->ep_attr->max_order_raw_size <= 0)
 			return EXIT_FAILURE;
 		if (hints->ep_attr->max_order_raw_size) {
-			if (info->ep_attr->max_order_raw_size < hints->ep_attr->max_order_raw_size)
+			if (info->ep_attr->max_order_raw_size <
+			    hints->ep_attr->max_order_raw_size)
 				return EXIT_FAILURE;
 		}
 	}
@@ -425,7 +431,8 @@ static int check_valid_rma_ordering_sizes(struct fi_info *info)
 		if (info->ep_attr->max_order_war_size <= 0)
 			return EXIT_FAILURE;
 		if (hints->ep_attr->max_order_war_size) {
-			if (info->ep_attr->max_order_war_size < hints->ep_attr->max_order_war_size)
+			if (info->ep_attr->max_order_war_size <
+			    hints->ep_attr->max_order_war_size)
 				return EXIT_FAILURE;
 		}
 	}
@@ -434,7 +441,8 @@ static int check_valid_rma_ordering_sizes(struct fi_info *info)
 		if (info->ep_attr->max_order_waw_size <= 0)
 			return EXIT_FAILURE;
 		if (hints->ep_attr->max_order_waw_size) {
-			if (info->ep_attr->max_order_waw_size < hints->ep_attr->max_order_waw_size)
+			if (info->ep_attr->max_order_waw_size <
+			    hints->ep_attr->max_order_waw_size)
 				return EXIT_FAILURE;
 		}
 	}
@@ -460,8 +468,10 @@ static int init_invalid_rma_RAW_ordering_size(struct fi_info *hints)
 		return ret;
 	}
 
-	if (fi->ep_attr->max_order_raw_size)
-		hints->ep_attr->max_order_raw_size = fi->ep_attr->max_order_raw_size + 1;
+	if (fi->ep_attr->max_order_raw_size) {
+		hints->ep_attr->max_order_raw_size =
+				fi->ep_attr->max_order_raw_size + 1;
+	}
 
 	fi_freeinfo(fi);
 
@@ -486,8 +496,10 @@ static int init_invalid_rma_WAR_ordering_size(struct fi_info *hints)
 		return ret;
 	}
 
-	if (fi->ep_attr->max_order_war_size)
-		hints->ep_attr->max_order_war_size = fi->ep_attr->max_order_war_size + 1;
+	if (fi->ep_attr->max_order_war_size) {
+		hints->ep_attr->max_order_war_size =
+				fi->ep_attr->max_order_war_size + 1;
+	}
 
 	fi_freeinfo(fi);
 
@@ -512,8 +524,10 @@ static int init_invalid_rma_WAW_ordering_size(struct fi_info *hints)
 		return ret;
 	}
 
-	if (fi->ep_attr->max_order_waw_size)
-		hints->ep_attr->max_order_waw_size = fi->ep_attr->max_order_waw_size + 1;
+	if (fi->ep_attr->max_order_waw_size) {
+		hints->ep_attr->max_order_waw_size =
+				fi->ep_attr->max_order_waw_size + 1;
+	}
 
 	fi_freeinfo(fi);
 
@@ -524,52 +538,6 @@ static int init_invalid_rma_WAW_ordering_size(struct fi_info *hints)
 /*
  * MR mode checks
  */
-static int init_mr_basic(struct fi_info *hints)
-{
-	hints->caps |= FI_RMA;
-	hints->domain_attr->mr_mode = FI_MR_BASIC;
-	return 0;
-}
-
-static int check_mr_basic(struct fi_info *info)
-{
-	return (info->domain_attr->mr_mode != FI_MR_BASIC) ?
-		EXIT_FAILURE : 0;
-}
-
-static int init_mr_scalable(struct fi_info *hints)
-{
-	hints->caps |= FI_RMA;
-	hints->domain_attr->mr_mode = FI_MR_SCALABLE;
-	return 0;
-}
-
-static int check_mr_scalable(struct fi_info *info)
-{
-	return (info->domain_attr->mr_mode != FI_MR_SCALABLE) ?
-		EXIT_FAILURE : 0;
-}
-
-static int init_mr_unspec(struct fi_info *hints)
-{
-	hints->caps |= FI_RMA;
-	hints->domain_attr->mr_mode = FI_MR_UNSPEC;
-	return 0;
-}
-
-static int test_mr_v1_0(char *node, char *service, uint64_t flags,
-			struct fi_info *test_hints, struct fi_info **info)
-{
-	return fi_getinfo(FI_VERSION(1, 0), node, service, flags, test_hints, info);
-}
-
-static int check_mr_unspec(struct fi_info *info)
-{
-	return (info->domain_attr->mr_mode != FI_MR_BASIC &&
-		info->domain_attr->mr_mode != FI_MR_SCALABLE) ?
-		EXIT_FAILURE : 0;
-}
-
 static int init_mr_mode(struct fi_info *hints, uint64_t mode)
 {
 	hints->domain_attr->mr_mode = (uint32_t) mode;
@@ -607,18 +575,6 @@ static int init_data_auto(struct fi_info *hints)
 	return 0;
 }
 
-static int init_ctrl_manual(struct fi_info *hints)
-{
-	hints->domain_attr->control_progress = FI_PROGRESS_MANUAL;
-	return 0;
-}
-
-static int init_ctrl_auto(struct fi_info *hints)
-{
-	hints->domain_attr->control_progress = FI_PROGRESS_AUTO;
-	return 0;
-}
-
 static int check_data_manual(struct fi_info *info)
 {
 	return (info->domain_attr->data_progress != FI_PROGRESS_MANUAL) ?
@@ -631,18 +587,6 @@ static int check_data_auto(struct fi_info *info)
 		EXIT_FAILURE : 0;
 }
 
-static int check_ctrl_manual(struct fi_info *info)
-{
-	return (info->domain_attr->control_progress != FI_PROGRESS_MANUAL) ?
-		EXIT_FAILURE : 0;
-}
-
-static int check_ctrl_auto(struct fi_info *info)
-{
-	return (info->domain_attr->control_progress != FI_PROGRESS_AUTO) ?
-		EXIT_FAILURE : 0;
-}
-
 
 static int init_domain_caps(struct fi_info *hints, uint64_t caps)
 {
@@ -663,12 +607,91 @@ static int validate_domain_caps(char *node, char *service, uint64_t flags,
 				   init_domain_caps, check_domain_caps);
 }
 
+/* Some apps (MPI) request all fi_info structures, and use the output to
+ * form the hints for a second call.  This usage breaks if the provider
+ * adds a new capability bit that also requires setting a mode or mr_mode
+ * bit (new or otherwise), which the app does not set.
+ * This is really a problem with the app, but avoid a regression
+ * by verifying that providers do not add new requirements for apps that
+ * inadvertently pick up a new capability bit.
+ */
+static int test_caps_regression(char *node, char *service, uint64_t flags,
+		struct fi_info *hints, struct fi_info **info)
+{
+	struct fi_info *fi;
+	int ret;
+
+	if (!hints) {
+		printf("invalid test case: hints may not be null");
+		return -FI_EINVAL;
+	}
+
+	ret = fi_getinfo(FT_FIVERSION, node, service, flags, NULL, info);
+	if (ret)
+		return ret;
+
+	if (!hints->fabric_attr || !hints->fabric_attr->prov_name) {
+		fi = *info;
+	} else {
+		for (fi = *info; fi; fi = fi->next) {
+			if (!strcasecmp(hints->fabric_attr->prov_name,
+					(*info)->fabric_attr->prov_name))
+				break;
+		}
+	}
+
+	if (!fi)
+		return 0;
+
+	/* Limit mode bits to common, older options only */
+	hints->caps |= fi->caps;
+	hints->mode = FI_CONTEXT | FI_CONTEXT2;
+	hints->domain_attr->mr_mode = FI_MR_LOCAL | OFI_MR_BASIC_MAP;
+
+	fi_freeinfo(*info);
+	*info = NULL;
+
+	ret = fi_getinfo(FT_FIVERSION, node, service, flags, hints, info);
+	if (ret) {
+		printf("regression: new mode/mr_mode bits required...");
+		return -FI_EINVAL;
+	}
+
+	return 0;
+}
+
+/* Check if at least one of FI_LOCAL_COMM and FI_REMOTE_COMM is set */
+static int test_comm_caps(char *node, char *service, uint64_t flags,
+		struct fi_info *hints, struct fi_info **info)
+{
+	uint64_t test_caps[2] = { FI_LOCAL_COMM, FI_REMOTE_COMM };
+	int ret, i;
+
+	for (i = 0; i < 2; i++) {
+		hints->caps &= ~(FI_LOCAL_COMM | FI_REMOTE_COMM);
+		hints->domain_attr->caps &= ~(FI_LOCAL_COMM | FI_REMOTE_COMM);
+		hints->caps |= test_caps[i];
+		hints->domain_attr->caps |= test_caps[i];
+		ret = fi_getinfo(FT_FIVERSION, node, service, flags, hints, info);
+		if (ret == 0) {
+			if (((*info)->caps & test_caps[i]) == test_caps[i] &&
+			    ((*info)->domain_attr->caps & test_caps[i]) == test_caps[i])
+				return 0;
+
+			fi_freeinfo(*info);
+		}
+	}
+
+	FT_DEBUG("Neither FI_LOCAL_COMM nor FI_REMOTE_COMM is set\n");
+	return -FI_EIO;
+}
+
 /*
  * getinfo test
  */
 static int getinfo_unit_test(char *node, char *service, uint64_t flags,
-		struct fi_info *base_hints, ft_getinfo_init init, ft_getinfo_test test,
-		ft_getinfo_check check, int ret_exp)
+		struct fi_info *base_hints, ft_getinfo_init init,
+		ft_getinfo_test test, ft_getinfo_check check, int ret_exp)
 {
 	struct fi_info *info = NULL, *fi, *test_hints = NULL;
 	int ret;
@@ -685,10 +708,12 @@ static int getinfo_unit_test(char *node, char *service, uint64_t flags,
 			goto out;
 	}
 
-	if (test)
+	if (test) {
 		ret = test(node, service, flags, test_hints, &info);
-	else
-		ret = fi_getinfo(FT_FIVERSION, node, service, flags, test_hints, &info);
+	} else {
+		ret = fi_getinfo(FT_FIVERSION, node, service, flags,
+				 test_hints, &info);
+	}
 	if (ret) {
 		if (ret == ret_exp) {
 			ret = 0;
@@ -704,31 +729,40 @@ static int getinfo_unit_test(char *node, char *service, uint64_t flags,
 
 	for (fi = info; fi; fi = fi->next) {
 		FT_DEBUG("\nTesting for fabric: %s, domain: %s, endpoint type: %d",
-				fi->fabric_attr->prov_name, fi->domain_attr->name,
-				fi->ep_attr->type);
+			 fi->fabric_attr->prov_name, fi->domain_attr->name,
+			 fi->ep_attr->type);
 		ret = check(fi);
 		if (ret)
 			break;
 	}
 out:
+	/* If init was called, then there is a chance that the hints were
+	 * modified so that the application now owns some of the memory.
+	 * At the moment, only invalid_dom does this and the domain name
+	 * is the only application owned memory. Free the application owned
+	 * memory so that fi_freeinfo only frees memory that it owns. */
+	if (init && !!test_hints) {
+		free(test_hints->domain_attr->name);
+		test_hints->domain_attr->name = NULL;
+	}
 	fi_freeinfo(test_hints);
 	fi_freeinfo(info);
 	return ret;
 }
 
-#define getinfo_test(name, num, desc, node, service, flags, hints, init, test, check,	\
-		ret_exp)							\
-char *getinfo_ ## name ## num ## _desc = desc;					\
-static int getinfo_ ## name ## num(void)					\
-{										\
-	int ret, testret = FAIL;						\
-	ret = getinfo_unit_test(node, service, flags, hints, init, test, check,	\
-			ret_exp);						\
-	if (ret)								\
-		goto fail;							\
-	testret = PASS;								\
-fail:										\
-	return TEST_RET_VAL(ret, testret);					\
+#define getinfo_test(name, num, desc, node, service, flags, hints, 	\
+		     init, test, check, ret_exp)			\
+char *getinfo_ ## name ## num ## _desc = desc;				\
+static int getinfo_ ## name ## num(void)				\
+{									\
+	int ret, testret = FAIL;					\
+	ret = getinfo_unit_test(node, service, flags, hints, init,	\
+				test, check, ret_exp);			\
+	if (ret)							\
+		goto fail;						\
+	testret = PASS;							\
+fail:									\
+	return TEST_RET_VAL(ret, testret);				\
 }
 
 /*
@@ -744,98 +778,88 @@ fail:										\
 
 /* 1.1 Source address only tests */
 getinfo_test(no_hints, 1, "Test with no node, service, flags or hints",
-		NULL, NULL, 0, NULL, NULL, NULL, check_srcaddr, 0)
-getinfo_test(no_hints, 2, "Test with node, no service, FI_SOURCE flag and no hints",
-		opts.src_addr ? opts.src_addr : "localhost", NULL, FI_SOURCE,
-		NULL, NULL, NULL, check_srcaddr, 0)
-getinfo_test(no_hints, 3, "Test with service, FI_SOURCE flag and no node or hints",
+	     NULL, NULL, 0, NULL, NULL, NULL, check_srcaddr, 0)
+getinfo_test(no_hints, 2, "Test with node, no service, FI_SOURCE flag, no hints",
+	     opts.src_addr ? opts.src_addr : "localhost", NULL, FI_SOURCE,
+	     NULL, NULL, NULL, check_srcaddr, 0)
+getinfo_test(no_hints, 3, "Test with service, FI_SOURCE flag, no node, no hints",
 		 NULL, opts.src_port, FI_SOURCE, NULL, NULL,
-		 NULL, check_srcaddr, 0)	// TODO should we check for wildcard addr?
-getinfo_test(no_hints, 4, "Test with node, service, FI_SOURCE flags and no hints",
-		opts.src_addr ? opts.src_addr : "localhost", opts.src_port,
-		FI_SOURCE, NULL, NULL, NULL, check_srcaddr, 0)
+		 NULL, check_srcaddr, 0)
+getinfo_test(no_hints, 4, "Test with node, service, FI_SOURCE flag, no hints",
+	     opts.src_addr ? opts.src_addr : "localhost", opts.src_port,
+	     FI_SOURCE, NULL, NULL, NULL, check_srcaddr, 0)
 
 /* 1.2 Source and destination address tests */
-getinfo_test(no_hints, 5, "Test with node, service and no hints",
-		opts.dst_addr ? opts.dst_addr : "localhost", opts.dst_port,
-		0, NULL, NULL, NULL, check_src_dest_addr, 0)
+getinfo_test(no_hints, 5, "Test with node, service, no hints",
+	     opts.dst_addr ? opts.dst_addr : "localhost", opts.dst_port,
+	     0, NULL, NULL, NULL, check_src_dest_addr, 0)
 
-/* 2. Test with hints */
+/* 2. Tests, most with hints */
 /* 2.1 Source address only tests */
 getinfo_test(src, 1, "Test with no node, service, or flags",
-		NULL, NULL, 0, hints, NULL, NULL, check_srcaddr, 0)
+	     NULL, NULL, 0, hints, NULL, NULL, check_srcaddr, 0)
 getinfo_test(src, 2, "Test with node, no service, FI_SOURCE flag",
-		opts.src_addr ? opts.src_addr : "localhost", NULL, FI_SOURCE,
-		hints, NULL, NULL, check_srcaddr, 0)
-getinfo_test(src, 3, "Test with service, FI_SOURCE flag and no node",
-		 NULL, opts.src_port, FI_SOURCE, hints, NULL,
-		 NULL, check_srcaddr, 0)	// TODO should we check for wildcard addr?
-getinfo_test(src, 4, "Test with node, service, FI_SOURCE flags",
-		opts.src_addr ? opts.src_addr : "localhost", opts.src_port,
-		FI_SOURCE, hints, NULL, NULL, check_srcaddr, 0)
+	     opts.src_addr ? opts.src_addr : "localhost", NULL, FI_SOURCE,
+	     hints, NULL, NULL, check_srcaddr, 0)
+getinfo_test(src, 3, "Test with service, FI_SOURCE flag, no node",
+	     NULL, opts.src_port, FI_SOURCE, hints, NULL,
+	     NULL, check_srcaddr, 0)
+getinfo_test(src, 4, "Test with node, service, FI_SOURCE flag",
+	     opts.src_addr ? opts.src_addr : "localhost", opts.src_port,
+	     FI_SOURCE, hints, NULL, NULL, check_srcaddr, 0)
 
 /* 2.2 Source and destination address tests */
 getinfo_test(src_dest, 1, "Test with node, service",
-		opts.dst_addr ? opts.dst_addr : "localhost", opts.dst_port,
-		0, hints, NULL, NULL, check_src_dest_addr, 0)
+	     opts.dst_addr ? opts.dst_addr : "localhost", opts.dst_port,
+	     0, hints, NULL, NULL, check_src_dest_addr, 0)
 
 getinfo_test(src_dest, 2, "Test API version",
-		NULL, NULL, 0, hints, NULL, NULL, check_api_version , 0)
+	     NULL, NULL, 0, hints, NULL, NULL, check_api_version , 0)
 
 /* Negative tests */
 getinfo_test(neg, 1, "Test with non-existent domain name",
-		NULL, NULL, 0, hints, invalid_dom, NULL, NULL, -FI_ENODATA)
+	     NULL, NULL, 0, hints, invalid_dom, NULL, NULL, -FI_ENODATA)
 
 /* Utility provider tests */
 getinfo_test(util, 1, "Test if we get utility provider when requested",
-		NULL, NULL, 0, hints, NULL, NULL, check_util_prov, 0)
+	     NULL, NULL, 0, hints, NULL, NULL, check_util_prov, 0)
 
 /* Message Ordering Tests */
 getinfo_test(msg_ordering, 1, "Test tx ordering bits supported are set",
-		NULL, NULL, 0, hints, NULL, validate_tx_ordering_bits, NULL, 0)
+	     NULL, NULL, 0, hints, NULL, validate_tx_ordering_bits, NULL, 0)
 getinfo_test(msg_ordering, 2, "Test rx ordering bits supported are set",
-		NULL, NULL, 0, hints, NULL, validate_rx_ordering_bits, NULL, 0)
+	     NULL, NULL, 0, hints, NULL, validate_rx_ordering_bits, NULL, 0)
 
 getinfo_test(raw_ordering, 1, "Test rma RAW ordering size is set",
-		NULL, NULL, 0, hints, init_valid_rma_RAW_ordering_no_set_size,
-		NULL, check_valid_rma_ordering_sizes, 0)
+	     NULL, NULL, 0, hints, init_valid_rma_RAW_ordering_no_set_size,
+	     NULL, check_valid_rma_ordering_sizes, 0)
 getinfo_test(raw_ordering, 2, "Test rma RAW ordering size is set to hints",
-		NULL, NULL, 0, hints, init_valid_rma_RAW_ordering_set_size,
-		NULL, check_valid_rma_ordering_sizes, 0)
+	     NULL, NULL, 0, hints, init_valid_rma_RAW_ordering_set_size,
+	     NULL, check_valid_rma_ordering_sizes, 0)
 getinfo_test(war_ordering, 1, "Test rma WAR ordering size is set",
-		NULL, NULL, 0, hints, init_valid_rma_WAR_ordering_no_set_size,
-		NULL, check_valid_rma_ordering_sizes, 0)
+	     NULL, NULL, 0, hints, init_valid_rma_WAR_ordering_no_set_size,
+	     NULL, check_valid_rma_ordering_sizes, 0)
 getinfo_test(war_ordering, 2, "Test rma WAR ordering size is set to hints",
-		NULL, NULL, 0, hints, init_valid_rma_WAR_ordering_set_size,
-		NULL, check_valid_rma_ordering_sizes, 0)
+	     NULL, NULL, 0, hints, init_valid_rma_WAR_ordering_set_size,
+	     NULL, check_valid_rma_ordering_sizes, 0)
 getinfo_test(waw_ordering, 1, "Test rma WAW ordering size is set",
-		NULL, NULL, 0, hints, init_valid_rma_WAW_ordering_no_set_size,
-		NULL, check_valid_rma_ordering_sizes, 0)
+	     NULL, NULL, 0, hints, init_valid_rma_WAW_ordering_no_set_size,
+	     NULL, check_valid_rma_ordering_sizes, 0)
 getinfo_test(waw_ordering, 2, "Test rma WAW ordering size is set to hints",
-		NULL, NULL, 0, hints, init_valid_rma_WAW_ordering_set_size,
-		NULL, check_valid_rma_ordering_sizes, 0)
+	     NULL, NULL, 0, hints, init_valid_rma_WAW_ordering_set_size,
+	     NULL, check_valid_rma_ordering_sizes, 0)
 getinfo_test(bad_raw_ordering, 1, "Test invalid rma RAW ordering size",
-		NULL, NULL, 0, hints, init_invalid_rma_RAW_ordering_size,
-		NULL, NULL, -FI_ENODATA)
+	     NULL, NULL, 0, hints, init_invalid_rma_RAW_ordering_size,
+	     NULL, NULL, -FI_ENODATA)
 getinfo_test(bad_war_ordering, 1, "Test invalid rma WAR ordering size",
-		NULL, NULL, 0, hints, init_invalid_rma_WAR_ordering_size,
-		NULL, NULL, -FI_ENODATA)
+	     NULL, NULL, 0, hints, init_invalid_rma_WAR_ordering_size,
+	     NULL, NULL, -FI_ENODATA)
 getinfo_test(bad_waw_ordering, 1, "Test invalid rma WAW ordering size",
-		NULL, NULL, 0, hints, init_invalid_rma_WAW_ordering_size,
-		NULL, NULL, -FI_ENODATA)
+	     NULL, NULL, 0, hints, init_invalid_rma_WAW_ordering_size,
+	     NULL, NULL, -FI_ENODATA)
 
 /* MR mode tests */
-getinfo_test(mr_mode, 1, "Test FI_MR_BASIC", NULL, NULL, 0,
-	     hints, init_mr_basic, NULL, check_mr_basic, -FI_ENODATA)
-getinfo_test(mr_mode, 2, "Test FI_MR_SCALABLE", NULL, NULL, 0,
-	     hints, init_mr_scalable, NULL, check_mr_scalable, -FI_ENODATA)
-getinfo_test(mr_mode, 3, "Test FI_MR_UNSPEC (v1.0)", NULL, NULL, 0,
-	     hints, init_mr_unspec, test_mr_v1_0, check_mr_unspec, -FI_ENODATA)
-getinfo_test(mr_mode, 4, "Test FI_MR_BASIC (v1.0)", NULL, NULL, 0,
-	     hints, init_mr_basic, test_mr_v1_0, check_mr_basic, -FI_ENODATA)
-getinfo_test(mr_mode, 5, "Test FI_MR_SCALABLE (v1.0)", NULL, NULL, 0,
-     	     hints, init_mr_scalable, test_mr_v1_0, check_mr_scalable, -FI_ENODATA)
-getinfo_test(mr_mode, 6, "Test mr_mode bits", NULL, NULL, 0,
+getinfo_test(mr_mode, 1, "Test mr_mode bits", NULL, NULL, 0,
 	     hints, NULL, validate_mr_modes, NULL, 0)
 
 /* Progress tests */
@@ -843,24 +867,25 @@ getinfo_test(progress, 1, "Test data manual progress", NULL, NULL, 0,
 	     hints, init_data_manual, NULL, check_data_manual, 0)
 getinfo_test(progress, 2, "Test data auto progress", NULL, NULL, 0,
 	     hints, init_data_auto, NULL, check_data_auto, 0)
-getinfo_test(progress, 3, "Test ctrl manual progress", NULL, NULL, 0,
-	     hints, init_ctrl_manual, NULL, check_ctrl_manual, 0)
-getinfo_test(progress, 4, "Test ctrl auto progress", NULL, NULL, 0,
-	     hints, init_ctrl_auto, NULL, check_ctrl_auto, 0)
 
 /* Capability test */
 getinfo_test(caps, 1, "Test capability bits supported are set",
-		NULL, NULL, 0, hints, NULL, validate_primary_caps, NULL, 0)
+	     NULL, NULL, 0, hints, NULL, validate_primary_caps, NULL, 0)
 getinfo_test(caps, 2, "Test capability with no hints",
-		NULL, NULL, 0, NULL, NULL, NULL, test_null_hints_caps, 0)
+	     NULL, NULL, 0, NULL, NULL, NULL, test_null_hints_caps, 0)
 getinfo_test(caps, 3, "Test domain capabilities", NULL, NULL, 0,
 	     hints, NULL, validate_domain_caps, NULL, 0)
+getinfo_test(caps, 4, "Test for capability bit regression",
+	     NULL, NULL, 0, hints, NULL, test_caps_regression, NULL, 0)
+getinfo_test(caps, 5, "Test if either FI_LOCAL_COMM or FI_REMOTE_COMM is set",
+	     NULL, NULL, 0, hints, NULL, test_comm_caps, NULL, 0)
 
 
-static void usage(void)
+static void usage(char *name)
 {
-	ft_unit_usage("getinfo_test", "Unit tests for fi_getinfo");
-	FT_PRINT_OPTS_USAGE("-e <ep_type>", "Endpoint type: msg|rdm|dgram (default:rdm)");
+	ft_unit_usage(name, "Unit tests for fi_getinfo");
+	FT_PRINT_OPTS_USAGE("-e <ep_type>",
+			    "Endpoint type: msg|rdm|dgram (default:rdm)");
 	ft_addr_usage();
 }
 
@@ -925,18 +950,13 @@ int main(int argc, char **argv)
 		TEST_ENTRY_GETINFO(bad_waw_ordering1),
 		TEST_ENTRY_GETINFO(neg1),
 		TEST_ENTRY_GETINFO(mr_mode1),
-		TEST_ENTRY_GETINFO(mr_mode2),
-		TEST_ENTRY_GETINFO(mr_mode3),
-		TEST_ENTRY_GETINFO(mr_mode4),
-		TEST_ENTRY_GETINFO(mr_mode5),
-		TEST_ENTRY_GETINFO(mr_mode6),
 		TEST_ENTRY_GETINFO(progress1),
 		TEST_ENTRY_GETINFO(progress2),
-		TEST_ENTRY_GETINFO(progress3),
-		TEST_ENTRY_GETINFO(progress4),
 		TEST_ENTRY_GETINFO(caps1),
 		TEST_ENTRY_GETINFO(caps2),
 		TEST_ENTRY_GETINFO(caps3),
+		TEST_ENTRY_GETINFO(caps4),
+		TEST_ENTRY_GETINFO(caps5),
 		{ NULL, "" }
 	};
 
@@ -958,10 +978,8 @@ int main(int argc, char **argv)
 			ft_parseinfo(op, optarg, hints, &opts);
 			break;
 		case 'h':
-			usage();
-			return EXIT_SUCCESS;
 		case '?':
-			usage();
+			usage(argv[0]);
 			return EXIT_FAILURE;
 		}
 	}
@@ -980,10 +998,10 @@ int main(int argc, char **argv)
 		if (set_prov(hints->fabric_attr->prov_name))
 			return EXIT_FAILURE;
 	} else {
-	       FT_WARN("\nTests getinfo1 to getinfo5 may not run exclusively "
-		       "for a particular provider since we don't pass hints.\n"
-		       "So the failures in any of those tests may not be "
-		       "attributable to a single provider.\n");
+	       FT_WARN("\nSome tests do not pass in hints, and may not run "
+		       "exclusively for a particular provider.\n"
+		       "Failures in any of those tests may not be "
+		       "attributable to a specific provider.\n");
 	}
 
 	failed = run_tests(no_hint_tests, err_buf);
