@@ -528,6 +528,12 @@ static int mr_cache_test(enum alloc_type type)
 		}
 		break;
 
+	/* Add test case for this once neuron has a memory monitor cache. */
+	case FI_HMEM_NEURON:
+		ret = -FI_ENOSYS;
+		goto cleanup;
+		break;
+
 	default:
 		prime_buf = malloc(mr_buf_size);
 		if (!prime_buf) {
@@ -630,19 +636,19 @@ static int mr_cache_test(enum alloc_type type)
 
 cleanup:
 	if (realloc_mr)
-		fi_close(&realloc_mr->fid);
+		FT_CLOSE_FID(realloc_mr);
 
 	if (cached_mr)
-		fi_close(&cached_mr->fid);
+		FT_CLOSE_FID(cached_mr);
 
 	if (mr)
-		fi_close(&mr->fid);
+		FT_CLOSE_FID(mr);
 
 	if (buf)
 		mem_free(buf, type);
 
 	if (prime_mr)
-		fi_close(&prime_mr->fid);
+		FT_CLOSE_FID(prime_mr);
 
 	if (prime_buf) {
 		switch (iface) {
@@ -732,9 +738,9 @@ struct test_entry test_array[] = {
 	{ NULL, "" }
 };
 
-static void usage(void)
+static void usage(char *name)
 {
-	ft_unit_usage("fi_mr_cache_evict",
+	ft_unit_usage(name,
 		"Test a provider's ability to evict MR cache entries.\n"
 		"Evictions are verified using MMAP, BRK, SBRK, CUDA and ROCR\n"
 		"allocations. FI_HMEM support must be enabled to run CUDA and\n"
@@ -793,14 +799,14 @@ int main(int argc, char **argv)
 			break;
 		case '?':
 		case 'h':
-			usage();
+			usage(argv[0]);
 			return EXIT_FAILURE;
 		}
 	}
 
 	hints->mode = ~0;
 	hints->domain_attr->mode = ~0;
-	hints->domain_attr->mr_mode = ~(FI_MR_BASIC | FI_MR_SCALABLE);
+	hints->domain_attr->mr_mode = ~OFI_MR_DEPRECATED;
 	hints->caps |= FI_MSG | FI_RMA;
 
 	if (opts.options & FT_OPT_ENABLE_HMEM)
