@@ -149,67 +149,178 @@ typedef struct {
 } ofi_io_uring_cqe_t;
 #endif
 
+struct ofi_sockctx {
+	void *context;
+	bool uring_sqe_inuse;
+};
+
+struct ofi_sockapi_uring {
+	ofi_io_uring_t *io_uring;
+	uint64_t credits;
+};
+
 struct ofi_sockapi {
-	ofi_io_uring_t *tx_io_uring;
-	ofi_io_uring_t *rx_io_uring;
+	struct ofi_sockapi_uring tx_uring;
+	struct ofi_sockapi_uring rx_uring;
+
+	int (*connect)(struct ofi_sockapi *sockapi, SOCKET sock,
+		       const struct sockaddr *addr, socklen_t addrlen,
+		       struct ofi_sockctx *ctx);
+	int (*accept)(struct ofi_sockapi *sockapi, SOCKET sock,
+		      struct sockaddr *addr, socklen_t *addrlen,
+		      struct ofi_sockctx *ctx);
 
 	ssize_t (*send)(struct ofi_sockapi *sockapi, SOCKET sock, const void *buf,
-			size_t len, int flags, void *ctx);
+			size_t len, int flags, struct ofi_sockctx *ctx);
 	ssize_t (*sendv)(struct ofi_sockapi *sockapi, SOCKET sock,
-			 const struct iovec *iov, size_t cnt, int flags, void *ctx);
+			 const struct iovec *iov, size_t cnt, int flags,
+			 struct ofi_sockctx *ctx);
 	ssize_t (*recv)(struct ofi_sockapi *sockapi, SOCKET sock, void *buf,
-			size_t len, int flags, void *ctx);
+			size_t len, int flags, struct ofi_sockctx *ctx);
 	ssize_t (*recvv)(struct ofi_sockapi *sockapi, SOCKET sock,
-			 struct iovec *iov, size_t cnt, int flags, void *ctx);
+			 struct iovec *iov, size_t cnt, int flags,
+			 struct ofi_sockctx *ctx);
 };
 
+static inline void
+ofi_sockctx_init(struct ofi_sockctx *sockctx, void *context)
+{
+	sockctx->context = context;
+	sockctx->uring_sqe_inuse = false;
+}
+
+static inline int
+ofi_sockapi_connect_socket(struct ofi_sockapi *sockapi, SOCKET sock,
+			   const struct sockaddr *addr, socklen_t addrlen,
+			   struct ofi_sockctx *ctx)
+{
+	int ret;
+
+	OFI_UNUSED(sockapi);
+	OFI_UNUSED(ctx);
+
+	ret = connect(sock, addr, addrlen);
+	if (ret < 0)
+		return -ofi_sockerr();
+	return ret;
+}
+
+static inline int
+ofi_sockapi_accept_socket(struct ofi_sockapi *sockapi, SOCKET sock,
+			  struct sockaddr *addr, socklen_t *addrlen,
+			  struct ofi_sockctx *ctx)
+{
+	int ret;
+
+	OFI_UNUSED(sockapi);
+	OFI_UNUSED(ctx);
+
+	ret = accept(sock, addr, addrlen);
+	if (ret < 0)
+		return -ofi_sockerr();
+	return ret;
+}
+
 static inline ssize_t
 ofi_sockapi_send_socket(struct ofi_sockapi *sockapi, SOCKET sock, const void *buf,
-			size_t len, int flags, void *ctx)
+			size_t len, int flags, struct ofi_sockctx *ctx)
 {
+	ssize_t ret;
+	int err;
+
 	OFI_UNUSED(sockapi);
 	OFI_UNUSED(ctx);
-	return ofi_send_socket(sock, buf, len, flags);
+	ret = ofi_send_socket(sock, buf, len, flags);
+	if (ret < 0) {
+		err = ofi_sockerr();
+		if (err == EPIPE)
+			return -FI_ENOTCONN;
+		if (err == EWOULDBLOCK)
+			return -FI_EAGAIN;
+		return -err;
+	}
+	return ret;
 }
 
 static inline ssize_t
 ofi_sockapi_sendv_socket(struct ofi_sockapi *sockapi, SOCKET sock,
-			 const struct iovec *iov, size_t cnt, int flags, void *ctx)
+			 const struct iovec *iov, size_t cnt, int flags,
+			 struct ofi_sockctx *ctx)
 {
+	ssize_t ret;
+	int err;
+
 	OFI_UNUSED(sockapi);
 	OFI_UNUSED(ctx);
-	return ofi_sendv_socket(sock, iov, cnt, flags);
+	ret = ofi_sendv_socket(sock, iov, cnt, flags);
+	if (ret < 0) {
+		err = ofi_sockerr();
+		if (err == EPIPE)
+			return -FI_ENOTCONN;
+		if (err == EWOULDBLOCK)
+			return -FI_EAGAIN;
+		return -err;
+	}
+	return ret;
 }
 
 static inline ssize_t
 ofi_sockapi_recv_socket(struct ofi_sockapi *sockapi, SOCKET sock, void *buf,
-			size_t len, int flags, void *ctx)
+			size_t len, int flags, struct ofi_sockctx *ctx)
 {
+	ssize_t ret;
+
 	OFI_UNUSED(sockapi);
 	OFI_UNUSED(ctx);
-	return ofi_recv_socket(sock, buf, len, flags);
+	ret = ofi_recv_socket(sock, buf, len, flags);
+	if (ret <= 0)
+		return ret ? -ofi_sockerr() : -FI_ENOTCONN;
+	return ret;
 }
 
 static inline ssize_t
 ofi_sockapi_recvv_socket(struct ofi_sockapi *sockapi, SOCKET sock,
-			 struct iovec *iov, size_t cnt, int flags, void *ctx)
+			 struct iovec *iov, size_t cnt, int flags,
+			 struct ofi_sockctx *ctx)
 {
+	ssize_t ret;
+
 	OFI_UNUSED(sockapi);
 	OFI_UNUSED(ctx);
-	return ofi_recvv_socket(sock, iov, cnt, flags);
+	ret = ofi_recvv_socket(sock, iov, cnt, flags);
+	if (ret <= 0)
+		return ret ? -ofi_sockerr() : -FI_ENOTCONN;
+	return ret;
 }
 
 #ifdef HAVE_LIBURING
+int ofi_sockapi_connect_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			      const struct sockaddr *addr, socklen_t addrlen,
+			      struct ofi_sockctx *ctx);
+int ofi_sockapi_accept_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			     struct sockaddr *addr, socklen_t *addrlen,
+			     struct ofi_sockctx *ctx);
+
 ssize_t ofi_sockapi_send_uring(struct ofi_sockapi *sockapi, SOCKET sock,
-			       const void *buf, size_t len, int flags, void *ctx);
+			       const void *buf, size_t len, int flags,
+			       struct ofi_sockctx *ctx);
 ssize_t ofi_sockapi_sendv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
 				const struct iovec *iov, size_t cnt, int flags,
-				void *ctx);
+				struct ofi_sockctx *ctx);
 ssize_t ofi_sockapi_recv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
-			       void *buf, size_t len, int flags, void *ctx);
+			       void *buf, size_t len, int flags,
+			       struct ofi_sockctx *ctx);
 ssize_t ofi_sockapi_recvv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
 				struct iovec *iov, size_t cnt, int flags,
-				void *ctx);
+				struct ofi_sockctx *ctx);
+
+int ofi_sockctx_uring_cancel(struct ofi_sockapi_uring *uring,
+			     struct ofi_sockctx *canceled_ctx,
+			     struct ofi_sockctx *ctx);
+
+int ofi_sockctx_uring_poll_add(struct ofi_sockapi_uring *uring,
+			       int fd, short poll_mask, bool multishot,
+			       struct ofi_sockctx *ctx);
 
 int ofi_uring_init(ofi_io_uring_t *io_uring, size_t entries);
 int ofi_uring_destroy(ofi_io_uring_t *io_uring);
@@ -251,30 +362,67 @@ static inline void ofi_uring_cq_advance(ofi_io_uring_t *io_uring, unsigned int c
 	io_uring_cq_advance(io_uring, count);
 }
 #else
+#define IORING_CQE_F_MORE	(1U << 1)
+
+static inline int
+ofi_sockapi_connect_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			  const struct sockaddr *addr, socklen_t addrlen,
+			  struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+static inline int
+ofi_sockapi_accept_uring(struct ofi_sockapi *sockapi, SOCKET sock,
+			 struct sockaddr *addr, socklen_t *addrlen,
+			 struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
 static inline ssize_t
 ofi_sockapi_send_uring(struct ofi_sockapi *sockapi, SOCKET sock, const void *buf,
-		       size_t len, int flags, void *ctx)
+		       size_t len, int flags, struct ofi_sockctx *ctx)
 {
 	return -FI_ENOSYS;
 }
 
 static inline ssize_t
 ofi_sockapi_sendv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
-			const struct iovec *iov, size_t cnt, int flags, void *ctx)
+			const struct iovec *iov, size_t cnt, int flags,
+			struct ofi_sockctx *ctx)
 {
 	return -FI_ENOSYS;
 }
 
 static inline ssize_t
 ofi_sockapi_recv_uring(struct ofi_sockapi *sockapi, SOCKET sock, void *buf,
-		       size_t len, int flags, void *ctx)
+		       size_t len, int flags, struct ofi_sockctx *ctx)
 {
 	return -FI_ENOSYS;
 }
 
 static inline ssize_t
 ofi_sockapi_recvv_uring(struct ofi_sockapi *sockapi, SOCKET sock,
-			struct iovec *iov, size_t cnt, int flags, void *ctx)
+			struct iovec *iov, size_t cnt, int flags,
+			struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+
+static inline int
+ofi_sockctx_uring_cancel(struct ofi_sockapi_uring *uring,
+			 struct ofi_sockctx *canceled_ctx,
+			 struct ofi_sockctx *ctx)
+{
+	return -FI_ENOSYS;
+}
+
+static inline int
+ofi_sockctx_uring_poll_add(struct ofi_sockapi_uring *uring,
+			   int fd, short poll_mask, bool multishot,
+			   struct ofi_sockctx *ctx)
 {
 	return -FI_ENOSYS;
 }
@@ -385,22 +533,30 @@ size_t ofi_byteq_readv(struct ofi_byteq *byteq, struct iovec *iov,
 struct ofi_bsock {
 	SOCKET sock;
 	struct ofi_sockapi *sockapi;
+	struct ofi_sockctx tx_sockctx;
+	struct ofi_sockctx rx_sockctx;
+	struct ofi_sockctx pollin_sockctx;
 	struct ofi_byteq sq;
 	struct ofi_byteq rq;
 	size_t zerocopy_size;
 	uint32_t async_index;
 	uint32_t done_index;
+	bool async_prefetch;
 };
 
 static inline void
 ofi_bsock_init(struct ofi_bsock *bsock, struct ofi_sockapi *sockapi,
-	       ssize_t sbuf_size, ssize_t rbuf_size)
+	       ssize_t sbuf_size, ssize_t rbuf_size, void *context)
 {
 	bsock->sock = INVALID_SOCKET;
 	bsock->sockapi = sockapi;
+	ofi_sockctx_init(&bsock->tx_sockctx, context);
+	ofi_sockctx_init(&bsock->rx_sockctx, context);
+	ofi_sockctx_init(&bsock->pollin_sockctx, context);
 	ofi_byteq_init(&bsock->sq, sbuf_size);
 	ofi_byteq_init(&bsock->rq, rbuf_size);
 	bsock->zerocopy_size = SIZE_MAX;
+	bsock->async_prefetch = false;
 
 	/* first async op will wrap back to 0 as the starting index */
 	bsock->async_index = UINT32_MAX;
@@ -423,19 +579,36 @@ static inline size_t ofi_bsock_tosend(struct ofi_bsock *bsock)
 	return ofi_byteq_readable(&bsock->sq);
 }
 
-ssize_t ofi_bsock_flush(struct ofi_bsock *bsock);
-ssize_t ofi_bsock_flush_sync(struct ofi_bsock *bsock);
+static inline int ofi_bsock_connect(struct ofi_bsock *bsock,
+				    const struct sockaddr *addr,
+				    socklen_t addrlen)
+{
+	return bsock->sockapi->connect(bsock->sockapi, bsock->sock,
+				       addr, addrlen, &bsock->tx_sockctx);
+}
+
+static inline int ofi_bsock_recv_unbuffered(struct ofi_bsock *bsock, void *buf,
+					    size_t len)
+{
+	assert(!ofi_bsock_readable(bsock));
+	return bsock->sockapi->recv(bsock->sockapi, bsock->sock, buf, len,
+				    MSG_NOSIGNAL, &bsock->rx_sockctx);
+}
+
+int ofi_bsock_flush(struct ofi_bsock *bsock);
+int ofi_bsock_flush_sync(struct ofi_bsock *bsock);
 /* For sends started asynchronously, the return value will be -EINPROGRESS_ASYNC,
  * and len will be set to the number of bytes that were queued.
  */
-ssize_t ofi_bsock_send(struct ofi_bsock *bsock, const void *buf, size_t *len);
-ssize_t ofi_bsock_sendv(struct ofi_bsock *bsock, const struct iovec *iov,
-			size_t cnt, size_t *len);
-ssize_t ofi_bsock_recv(struct ofi_bsock *bsock, void *buf, size_t len);
-ssize_t ofi_bsock_recvv(struct ofi_bsock *bsock, struct iovec *iov,
-			size_t cnt);
-uint32_t ofi_bsock_async_done(const struct fi_provider *prov,
-			      struct ofi_bsock *bsock);
+int ofi_bsock_send(struct ofi_bsock *bsock, const void *buf, size_t *len);
+int ofi_bsock_sendv(struct ofi_bsock *bsock, const struct iovec *iov,
+		    size_t cnt, size_t *len);
+int ofi_bsock_recv(struct ofi_bsock *bsock, void *buf, size_t *len);
+int ofi_bsock_recvv(struct ofi_bsock *bsock, struct iovec *iov,
+		    size_t cnt, size_t *len);
+int ofi_bsock_async_done(const struct fi_provider *prov,
+			 struct ofi_bsock *bsock);
+void ofi_bsock_prefetch_done(struct ofi_bsock *bsock, size_t len);
 
 
 /*
@@ -730,14 +903,14 @@ uint32_t ofi_addr_format(const char *str);
 int ofi_str_toaddr(const char *str, uint32_t *addr_format,
 		   void **addr, size_t *len);
 
-void ofi_straddr_log_internal(const char *func, int line,
+void ofi_straddr_log_internal(const char *func, int line, uint32_t addr_format,
 			      const struct fi_provider *prov,
 			      enum fi_log_level level,
 			      enum fi_log_subsys subsys, char *log_str,
 			      const void *addr);
 
 #define ofi_straddr_log(...) \
-	ofi_straddr_log_internal(__func__, __LINE__, __VA_ARGS__)
+	ofi_straddr_log_internal(__func__, __LINE__, FI_FORMAT_UNSPEC, __VA_ARGS__)
 
 #if ENABLE_DEBUG
 #define ofi_straddr_dbg(prov, subsystem, ...) \
