@@ -48,15 +48,15 @@
  * not built: no-op call for ctor
 */
 
-#if (HAVE_GNI) && (HAVE_GNI_DL)
-#  define GNI_INI FI_EXT_INI
-#  define GNI_INIT NULL
-#elif (HAVE_GNI)
-#  define GNI_INI INI_SIG(fi_gni_ini)
-#  define GNI_INIT fi_gni_ini()
-GNI_INI ;
+#if (HAVE_CXI) && (HAVE_CXI_DL)
+#  define CXI_INI FI_EXT_INI
+#  define CXI_INIT NULL
+#elif (HAVE_CXI)
+#  define CXI_INI INI_SIG(fi_cxi_ini)
+#  define CXI_INIT fi_cxi_ini()
+CXI_INI ;
 #else
-#  define GNI_INIT NULL
+#  define CXI_INIT NULL
 #endif
 
 /* If HAVE_EFA is defined on Windows, then the VisualStudio project configures
@@ -90,17 +90,6 @@ EFA_INI ;
 #  define EFA_INIT NULL
 #endif
 
-#if (HAVE_PSM) && (HAVE_PSM_DL)
-#  define PSM_INI FI_EXT_INI
-#  define PSM_INIT NULL
-#elif (HAVE_PSM)
-#  define PSM_INI INI_SIG(fi_psm_ini)
-#  define PSM_INIT fi_psm_ini()
-PSM_INI ;
-#else
-#  define PSM_INIT NULL
-#endif
-
 #if (HAVE_PSM2) && (HAVE_PSM2_DL)
 #  define PSM2_INI FI_EXT_INI
 #  define PSM2_INIT NULL
@@ -167,17 +156,6 @@ TCP_INI ;
 #  define TCP_INIT NULL
 #endif
 
-#if (HAVE_NET) && (HAVE_NET_DL)
-#  define XNET_INI FI_EXT_INI
-#  define XNET_INIT NULL
-#elif (HAVE_NET)
-#  define XNET_INI INI_SIG(fi_xnet_ini)
-#  define XNET_INIT fi_xnet_ini()
-XNET_INI ;
-#else
-#  define XNET_INIT NULL
-#endif
-
 #if (HAVE_RXM) && (HAVE_RXM_DL)
 #  define RXM_INI FI_EXT_INI
 #  define RXM_INIT NULL
@@ -200,32 +178,6 @@ RXD_INI ;
 #  define RXD_INIT NULL
 #endif
 
-#if (HAVE_BGQ) && (HAVE_BGQ_DL)
-#  define BGQ_INI FI_EXT_INI
-#  define BGQ_INIT NULL
-#elif (HAVE_BGQ)
-#  define BGQ_INI INI_SIG(fi_bgq_ini)
-#  define BGQ_INIT fi_bgq_ini()
-BGQ_INI ;
-#else
-#  define BGQ_INIT NULL
-#endif
-
-#ifdef _WIN32
-#if (HAVE_NETDIR) && (HAVE_NETDIR_DL)
-#  define NETDIR_INI FI_EXT_INI
-#  define NETDIR_INIT NULL
-#elif (HAVE_NETDIR)
-#  define NETDIR_INI INI_SIG(fi_netdir_ini)
-#  define NETDIR_INIT fi_netdir_ini()
-NETDIR_INI ;
-#else
-#  define NETDIR_INIT NULL
-#endif
-#else /* _WIN32 */
-#  define NETDIR_INIT NULL
-#endif /* _WIN32 */
-
 #if (HAVE_SHM) && (HAVE_SHM_DL)
 #  define SHM_INI FI_EXT_INI
 #  define SHM_INIT NULL
@@ -237,6 +189,17 @@ SHM_INI ;
 #  define SHM_INIT NULL
 #endif
 
+#if (HAVE_SM2) && (HAVE_SM2_DL)
+#  define SM2_INI FI_EXT_INI
+#  define SM2_INIT NULL
+#elif (HAVE_SM2)
+#  define SM2_INI INI_SIG(fi_sm2_ini)
+#  define SM2_INIT fi_sm2_ini()
+SM2_INI ;
+#else
+#  define SM2_INIT NULL
+#endif
+
 #if (HAVE_MRAIL) && (HAVE_MRAIL_DL)
 #  define MRAIL_INI FI_EXT_INI
 #  define MRAIL_INIT NULL
@@ -248,15 +211,15 @@ MRAIL_INI ;
 #  define MRAIL_INIT NULL
 #endif
 
-#if (HAVE_RSTREAM) && (HAVE_RSTREAM_DL)
-#  define RSTREAM_INI FI_EXT_INI
-#  define RSTREAM_INIT NULL
-#elif (HAVE_RSTREAM)
-#  define RSTREAM_INI INI_SIG(fi_rstream_ini)
-#  define RSTREAM_INIT fi_rstream_ini()
-RSTREAM_INI ;
+#if (HAVE_LNX) && (HAVE_LNX_DL)
+#  define LNX_INI FI_EXT_INI
+#  define LNX_INIT NULL
+#elif (HAVE_LNX)
+#  define LNX_INI INI_SIG(fi_lnx_ini)
+#  define LNX_INIT fi_lnx_ini()
+LNX_INI ;
 #else
-#  define RSTREAM_INIT NULL
+#  define LNX_INIT NULL
 #endif
 
 #if (HAVE_PERF) && (HAVE_PERF_DL)
@@ -270,6 +233,29 @@ HOOK_PERF_INI ;
 #  define HOOK_PERF_INIT NULL
 #endif
 
+#if (HAVE_TRACE) && (HAVE_TRACE_DL)
+#  define HOOK_TRACE_INI FI_EXT_INI
+#  define HOOK_TRACE_INIT NULL
+#elif (HAVE_TRACE)
+#  define HOOK_TRACE_INI INI_SIG(fi_hook_trace_ini)
+#  define HOOK_TRACE_INIT fi_hook_trace_ini()
+HOOK_TRACE_INI ;
+#else
+#  define HOOK_TRACE_INIT NULL
+#endif
+
+#if (HAVE_PROFILE) && (HAVE_PROFILE_DL)
+#  define HOOK_PROFILE_INI FI_EXT_INI
+#  define HOOK_PROFILE_INIT NULL
+#elif (HAVE_PROFILE)
+#  define HOOK_PROFILE_INI INI_SIG(fi_hook_profile_ini)
+#  define HOOK_PROFILE_INIT fi_hook_profile_ini()
+HOOK_PROFILE_INI ;
+#else
+#  define HOOK_PROFILE_INIT NULL
+#endif
+
+
 #if (HAVE_HOOK_DEBUG) && (HAVE_HOOK_DEBUG_DL)
 #  define HOOK_DEBUG_INI FI_EXT_INI
 #  define HOOK_DEBUG_INIT NULL
@@ -318,4 +304,35 @@ OPX_INI ;
 #  define OPX_INIT NULL
 #endif
 
+#if (HAVE_UCX) && (HAVE_UCX_DL)
+#  define UCX_INI FI_EXT_INI
+#  define UCX_INIT NULL
+#elif (HAVE_UCX)
+#  define UCX_INI INI_SIG(fi_ucx_ini)
+#  define UCX_INIT fi_ucx_ini()
+UCX_INI ;
+#else
+#  define UCX_INIT NULL
+#endif
+
+#if defined(_WIN32) && (HAVE_LPP)
+#  define LPP_INIT NULL
+#else
+#  if (HAVE_LPP) && (HAVE_LPP_DL)
+#    define LPP_INI FI_EXT_INI
+#    define LPP_INIT NULL
+#  elif (HAVE_LPP)
+#    define LPP_INI INI_SIG(fi_lpp_ini)
+#    define LPP_INIT fi_lpp_ini()
+LPP_INI ;
+#  else
+#    define LPP_INIT NULL
+#  endif
+#endif
+
+/* the utility collective provider is always enabled and built-in */
+#define COLL_INI INI_SIG(fi_coll_ini)
+#define COLL_INIT fi_coll_ini()
+COLL_INI ;
+
 #endif /* _OFI_PROV_H_ */
