@@ -33,6 +33,8 @@
 #ifndef FI_EQ_H
 #define FI_EQ_H
 
+#include <poll.h>
+
 #ifndef _WIN32
 #include <pthread.h>
 #endif /* _WIN32 */
@@ -58,7 +60,9 @@ enum fi_wait_obj {
 	FI_WAIT_UNSPEC,
 	FI_WAIT_SET,
 	FI_WAIT_FD,
-	FI_WAIT_MUTEX_COND	/* pthread mutex & cond */
+	FI_WAIT_MUTEX_COND,	/* pthread mutex & cond, deprecated */
+	FI_WAIT_YIELD,
+	FI_WAIT_POLLFD,
 };
 
 struct fi_wait_attr {
@@ -66,11 +70,13 @@ struct fi_wait_attr {
 	uint64_t		flags;
 };
 
+/* deprecated */
 struct fi_ops_wait {
 	size_t	size;
 	int	(*wait)(struct fid_wait *waitset, int timeout);
 };
 
+/* deprecated */
 struct fid_wait {
 	struct fid		fid;
 	struct fi_ops_wait	*ops;
@@ -83,6 +89,11 @@ struct fi_mutex_cond {
 };
 #endif /* _WIN32 */
 
+struct fi_wait_pollfd {
+	uint64_t		change_index;
+	size_t			nfds;
+	struct pollfd		*fd;
+};
 
 /*
  * Poll Set
@@ -93,6 +104,7 @@ struct fi_poll_attr {
 	uint64_t		flags;
 };
 
+/* deprecated */
 struct fi_ops_poll {
 	size_t	size;
 	int	(*poll)(struct fid_poll *pollset, void **context, int count);
@@ -102,6 +114,7 @@ struct fi_ops_poll {
 			uint64_t flags);
 };
 
+/* deprecated */
 struct fid_poll {
 	struct fid		fid;
 	struct fi_ops_poll	*ops;
@@ -117,7 +130,7 @@ struct fi_eq_attr {
 	uint64_t		flags;
 	enum fi_wait_obj	wait_obj;
 	int			signaling_vector;
-	struct fid_wait		*wait_set;
+	struct fid_wait		*wait_set;	/* deprecated */
 };
 
 /* Standard EQ events */
@@ -230,6 +243,7 @@ struct fi_cq_err_entry {
 	/* err_data is available until the next time the CQ is read */
 	void			*err_data;
 	size_t			err_data_size;
+	fi_addr_t		src_addr;
 };
 
 enum fi_cq_wait_cond {
@@ -244,7 +258,7 @@ struct fi_cq_attr {
 	enum fi_wait_obj	wait_obj;
 	int			signaling_vector;
 	enum fi_cq_wait_cond	wait_cond;
-	struct fid_wait		*wait_set;
+	struct fid_wait		*wait_set;	/* deprecated */
 };
 
 struct fi_ops_cq {
@@ -275,13 +289,14 @@ struct fid_cq {
  */
 
 enum fi_cntr_events {
-	FI_CNTR_EVENTS_COMP
+	FI_CNTR_EVENTS_COMP,
+	FI_CNTR_EVENTS_BYTES	/* count bytes not completeion events */
 };
 
 struct fi_cntr_attr {
 	enum fi_cntr_events	events;
 	enum fi_wait_obj	wait_obj;
-	struct fid_wait		*wait_set;
+	struct fid_wait		*wait_set;	/* deprecated */
 	uint64_t		flags;
 };
 
@@ -314,25 +329,25 @@ fi_trywait(struct fid_fabric *fabric, struct fid **fids, int count)
 	return fabric->ops->trywait(fabric, fids, count);
 }
 
-static inline int
+static inline FI_DEPRECATED_FUNC int
 fi_wait(struct fid_wait *waitset, int timeout)
 {
 	return waitset->ops->wait(waitset, timeout);
 }
 
-static inline int
+static inline FI_DEPRECATED_FUNC int
 fi_poll(struct fid_poll *pollset, void **context, int count)
 {
 	return pollset->ops->poll(pollset, context, count);
 }
 
-static inline int
+static inline FI_DEPRECATED_FUNC int
 fi_poll_add(struct fid_poll *pollset, struct fid *event_fid, uint64_t flags)
 {
 	return pollset->ops->poll_add(pollset, event_fid, flags);
 }
 
-static inline int
+static inline FI_DEPRECATED_FUNC int
 fi_poll_del(struct fid_poll *pollset, struct fid *event_fid, uint64_t flags)
 {
 	return pollset->ops->poll_del(pollset, event_fid, flags);
@@ -394,6 +409,9 @@ fi_cq_readfrom(struct fid_cq *cq, void *buf, size_t count, fi_addr_t *src_addr)
 static inline ssize_t
 fi_cq_readerr(struct fid_cq *cq, struct fi_cq_err_entry *buf, uint64_t flags)
 {
+	/* For compatibility with older providers. */
+	if (buf)
+		buf->src_addr = FI_ADDR_NOTAVAIL;
 	return cq->ops->readerr(cq, buf, flags);
 }
 
