@@ -205,6 +205,9 @@ extern "C" {
 #ifndef ESTALE
 # define ESTALE			246	/* Stale NFS file handle */
 #endif
+#ifndef EREMOTEIO
+# define EREMOTEIO		247	/* Remote I/O error */
+#endif
 
 /* MSG_NOSIGNAL doesn't exist on Windows */
 #ifndef MSG_NOSIGNAL
@@ -281,6 +284,9 @@ do						\
 #define be64toh ntohll
 #define strncasecmp _strnicmp
 
+#define access(path, mode) _access(path, mode)
+#define F_OK 0
+
 typedef int pid_t;
 #define getpid (int)GetCurrentProcessId
 
@@ -684,6 +690,12 @@ static inline char* strsep(char **stringp, const char *delim)
 
 #define __attribute__(x)
 
+static inline int pthread_atfork(void (*prepare)(void), void (*parent)(void),
+				 void (*child)(void))
+{
+	return FI_ENOSYS;
+}
+
 static inline int ofi_mmap_anon_pages(void **memptr, size_t size, int flags)
 {
 	return -FI_ENOSYS;
@@ -749,6 +761,26 @@ ofi_send_socket(SOCKET fd, const void *buf, size_t count, int flags)
 	return (ssize_t) send(fd, (const char*) buf, len, flags);
 }
 
+static inline ssize_t ofi_process_vm_readv(pid_t pid,
+			const struct iovec *local_iov,
+			unsigned long liovcnt,
+			const struct iovec *remote_iov,
+			unsigned long riovcnt,
+			unsigned long flags)
+{
+	return -FI_ENOSYS;
+}
+
+static inline size_t ofi_process_vm_writev(pid_t pid,
+			 const struct iovec *local_iov,
+			 unsigned long liovcnt,
+			 const struct iovec *remote_iov,
+			 unsigned long riovcnt,
+			 unsigned long flags)
+{
+	return -FI_ENOSYS;
+}
+
 static inline ssize_t ofi_read_socket(SOCKET fd, void *buf, size_t count)
 {
 	return ofi_recv_socket(fd, buf, count, 0);
@@ -869,7 +901,34 @@ static inline char * strndup(char const *src, size_t n)
 	return dst;
 }
 
-char *strcasestr(const char *haystack, const char *needle);
+static inline char *strcasestr(const char *haystack, const char *needle)
+{
+	char *uneedle, *uhaystack, *pos = NULL;
+	int i;
+
+	uneedle = malloc(strlen(needle) + 1);
+	uhaystack = malloc(strlen(haystack) + 1);
+	if (!uneedle || !uhaystack) {
+		free(uneedle);
+		free(uhaystack);
+		return pos;
+	}
+
+	for (i = 0; i < strlen(needle); i++)
+		uneedle[i] = (char) toupper(needle[i]);
+	uneedle[i] = '\0';
+	for (i = 0; i < strlen(haystack); i++)
+		uhaystack[i] = (char) toupper(haystack[i]);
+	uhaystack[i] = '\0';
+	pos = strstr(uhaystack, uneedle);
+	if (pos)
+		pos = (char *) ((uintptr_t) haystack + (uintptr_t) pos -
+						(uintptr_t) uhaystack);
+	free(uneedle);
+	free(uhaystack);
+
+	return pos;
+}
 
 #ifndef _SC_PAGESIZE
 #define _SC_PAGESIZE	0
@@ -906,6 +965,11 @@ static inline long ofi_sysconf(int name)
 
 int ofi_shm_unmap(struct util_shm *shm);
 
+static inline ssize_t ofi_get_addr_page_size(const void *addr)
+{
+	return ofi_sysconf(_SC_PAGESIZE);
+}
+
 static inline ssize_t ofi_get_hugepage_size(void)
 {
 	return -FI_ENOSYS;
@@ -939,6 +1003,7 @@ size_t ofi_ifaddr_get_speed(struct ifaddrs *ifa);
 
 #define file2unix_time	10000000i64
 #define win2unix_epoch	116444736000000000i64
+#define CLOCK_REALTIME 0
 #define CLOCK_MONOTONIC 1
 
 /* Own implementation of clock_gettime*/
@@ -984,12 +1049,16 @@ static inline ofi_complex_## type ofi_complex_prod_## type	\
 	res.imag = a.real * b.imag + a.imag * b.real;		\
 	return res;						\
 }								\
+static bool ofi_complex_is_true_## type (ofi_complex_ ## type a)\
+{								\
+	return a.real != 0 || a.imag != 0;			\
+}								\
 static inline ofi_complex_## type ofi_complex_land_## type	\
 	(ofi_complex_## type a, ofi_complex_## type b)		\
 {								\
 	ofi_complex_## type res;				\
-	res.real = (type)(((a.real != 0) || (a.imag != 0)) &&	\
-		((b.real != 0) || (b.imag != 0)));		\
+	res.real = (type)(ofi_complex_is_true_## type (a) &&	\
+			  ofi_complex_is_true_## type (b));	\
 	res.imag = 0;						\
 	return res;						\
 }								\
@@ -997,8 +1066,8 @@ static inline ofi_complex_## type ofi_complex_lor_## type	\
 	(ofi_complex_## type a, ofi_complex_## type b)		\
 {								\
 	ofi_complex_## type res;				\
-	res.real = (type)(((a.real != 0) || (a.imag != 0)) &&	\
-		((b.real != 0) || (b.imag != 0)));		\
+	res.real = (type)(ofi_complex_is_true_## type (a) ||	\
+			  ofi_complex_is_true_## type (b));	\
 	res.imag = 0;						\
 	return res;						\
 }								\
@@ -1006,10 +1075,10 @@ static inline ofi_complex_## type ofi_complex_lxor_## type	\
 	(ofi_complex_## type a, ofi_complex_## type b)		\
 {								\
 	ofi_complex_## type res;				\
-	res.real = (type)((((a.real != 0) || (a.imag != 0)) &&	\
-		    !((b.real != 0) || (b.imag != 0))) ||	\
-		   (!((a.real != 0) || (a.imag != 0)) &&	\
-		    ((b.real != 0) || (b.imag != 0))));		\
+	res.real = (type)((ofi_complex_is_true_## type (a) &&	\
+			   !ofi_complex_is_true_## type (b))) || \
+			  (!ofi_complex_is_true_## type (a) &&	\
+			   ofi_complex_is_true_## type (b));	\
 	res.imag = 0;						\
 	return res;						\
 }
@@ -1031,6 +1100,13 @@ typedef LONGLONG ofi_atomic_int_64_t;
 #define ofi_atomic_cas_bool(radix, ptr, expected, desired)					\
 	(InterlockedCompareExchange##radix((ofi_atomic_int_##radix##_t volatile *)ptr, desired, expected) == expected)
 
+#define ofi_atomic_compare_exchange_weak(radix, ptr, expected, desired, \
+					 succ_memmodel, fail_memmodel) \
+	InterlockedCompareExchange((ofi_atomic_int_##radix##_t volatile *)ptr, desired, expected)
+#define ofi_atomic_store_explicit(radix, ptr, value, memmodel) \
+	InterlockedExchange((ofi_atomic_int_##radix##_t volatile *)ptr, value)
+#define ofi_atomic_load_explicit(radix, ptr, memmodel) \
+	InterlockedAdd((ofi_atomic_int_##radix##_t volatile *)ptr, 0)
 #endif /* HAVE_BUILTIN_ATOMICS */
 
 static inline int ofi_set_thread_affinity(const char *s)
@@ -1065,6 +1141,23 @@ ofi_cpuid(unsigned func, unsigned subfunc, unsigned cpuinfo[4])
 
 #endif /* defined(_M_X64) || defined(_M_AMD64) */
 
+#if defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 202311L)
+    // C23 and above: use thread_local directly
+    #define OFI_THREAD_LOCAL thread_local
+#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 201102L) && defined(_Thread_local)
+    // C11: use _Thread_local
+    #define OFI_THREAD_LOCAL _Thread_local
+#elif defined(__GNUC__) || defined(__INTEL_COMPILER) || defined(__SUNPRO_CC) || defined(__IBMCPP__) || defined(__clang__)
+    // GCC/Clang/Intel/SunPro/IBM compilers
+    #define OFI_THREAD_LOCAL __thread
+#elif defined(_MSC_VER)
+    // Microsoft Visual C++ compiler
+    #define OFI_THREAD_LOCAL __declspec(thread)
+#else
+    // Unsupported compiler
+    #warning "Thread-local storage is not supported on this platform"
+	#define OFI_THREAD_LOCAL
+#endif
 
 #ifdef __cplusplus
 }
