@@ -1,66 +1,29 @@
-/*
- * Copyright (c) 2022 Amazon.com, Inc. or its affiliates. All rights reserved.
- *
- * This software is available to you under a choice of one of two
- * licenses.  You may choose to be licensed under the terms of the GNU
- * General Public License (GPL) Version 2, available from the file
- * COPYING in the main directory of this source tree, or the
- * BSD license below:
- *
- *     Redistribution and use in source and binary forms, with or
- *     without modification, are permitted provided that the following
- *     conditions are met:
- *
- *      - Redistributions of source code must retain the above
- *        copyright notice, this list of conditions and the following
- *        disclaimer.
- *
- *      - Redistributions in binary form must reproduce the above
- *        copyright notice, this list of conditions and the following
- *        disclaimer in the documentation and/or other materials
- *        provided with the distribution.
- *
- * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
- * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
- * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
- * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
- * COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
- * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
- * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
- * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
- * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
- * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
- * POSSIBILITY OF SUCH DAMAGE.
- */
+/* SPDX-License-Identifier: BSD-2-Clause OR GPL-2.0-only */
+/* SPDX-FileCopyrightText: Copyright Amazon.com, Inc. or its affiliates. All rights reserved. */
 
 #include "ofi_prov.h"
 #include <ofi_util.h>
 
 #include "efa.h"
-#if HAVE_EFA_DL
-#include <ofi_shm.h>
-#endif
+#include "rdm/efa_rdm_cq.h"
+#include "rdm/efa_rdm_pkt_type.h"
 
 #define EFA_FABRIC_PREFIX "EFA-"
-#define EFA_FABRIC_NAME "efa"
 #define EFA_DOMAIN_CAPS (FI_LOCAL_COMM | FI_REMOTE_COMM)
 
 #define EFA_RDM_TX_CAPS (OFI_TX_MSG_CAPS)
 #define EFA_RDM_RX_CAPS (OFI_RX_MSG_CAPS | FI_SOURCE)
+#define EFA_RDM_CAPS (EFA_RDM_TX_CAPS | EFA_RDM_RX_CAPS | EFA_DOMAIN_CAPS)
+
 #define EFA_DGRM_TX_CAPS (OFI_TX_MSG_CAPS)
 #define EFA_DGRM_RX_CAPS (OFI_RX_MSG_CAPS | FI_SOURCE)
-#define EFA_RDM_CAPS (EFA_RDM_TX_CAPS | EFA_RDM_RX_CAPS | EFA_DOMAIN_CAPS)
 #define EFA_DGRM_CAPS (EFA_DGRM_TX_CAPS | EFA_DGRM_RX_CAPS | EFA_DOMAIN_CAPS)
 
 #define EFA_TX_OP_FLAGS (FI_TRANSMIT_COMPLETE)
+#define EFA_RDM_TX_OP_FLAGS (FI_INJECT | FI_COMPLETION | FI_TRANSMIT_COMPLETE | \
+							FI_DELIVERY_COMPLETE)
 
-#define EFA_RX_MODE (0)
-
-#define EFA_RX_RDM_OP_FLAGS (0)
-#define EFA_RX_DGRM_OP_FLAGS (0)
-
-#define EFA_MSG_ORDER (FI_ORDER_NONE)
+#define EFA_MSG_ORDER (0)
 
 #define EFA_NO_DEFAULT -1
 
@@ -74,30 +37,6 @@ const struct fi_fabric_attr efa_fabric_attr = {
 	.prov_version	= OFI_VERSION_DEF_PROV,
 };
 
-/**
- * @brief set the fabric_attr field of a prov_info
- *
- * @param 	prov_info[out]		prov_info object
- * @param	device[in]		pointer to an efa_device struct, which contains device attributes
- * @return	0 on sucessess
- * 		-FI_ENOMEM if memory allocation failed
- */
-static
-int efa_prov_info_set_fabric_attr(struct fi_info *prov_info, struct efa_device *device)
-{
-	size_t name_len = strlen(EFA_FABRIC_NAME);
-
-	*prov_info->fabric_attr	= efa_fabric_attr;
-
-	name_len = strlen(EFA_FABRIC_NAME);
-	prov_info->fabric_attr->name = calloc(1, name_len + 1);
-	if (!prov_info->fabric_attr->name)
-		return -FI_ENOMEM;
-
-	strcpy(prov_info->fabric_attr->name, EFA_FABRIC_NAME);
-	return 0;
-}
-
 /**
  * @brief default domain_attr for prov_info
  */
@@ -105,11 +44,11 @@ const struct fi_domain_attr efa_domain_attr = {
 	.caps			= EFA_DOMAIN_CAPS,
 	.threading		= FI_THREAD_DOMAIN,
 	.control_progress	= FI_PROGRESS_AUTO,
-	.data_progress		= FI_PROGRESS_AUTO,
+	.progress		= FI_PROGRESS_AUTO,
 	.resource_mgmt		= FI_RM_DISABLED,
-	.mr_mode		= OFI_MR_BASIC_MAP | FI_MR_LOCAL | FI_MR_BASIC,
+	.mr_mode		= OFI_MR_BASIC_MAP | FI_MR_LOCAL | OFI_MR_BASIC,
 	.mr_key_size		= sizeof_field(struct ibv_sge, lkey),
-	.cq_data_size		= 0,
+	.cq_data_size		= EFA_CQ_DATA_SIZE,
 	.tx_ctx_cnt		= 1024,
 	.rx_ctx_cnt		= 1024,
 	.max_ep_tx_ctx		= 1,
@@ -156,6 +95,15 @@ int efa_prov_info_set_domain_attr(struct fi_info *prov_info,
 	prov_info->domain_attr->max_ep_rx_ctx	= 1;
 	prov_info->domain_attr->resource_mgmt	= FI_RM_DISABLED;
 	prov_info->domain_attr->mr_cnt		= device->ibv_attr.max_mr;
+
+	if (ep_type == FI_EP_RDM) {
+		/* EFA direct path is thread safe */
+		prov_info->domain_attr->threading = FI_THREAD_SAFE;
+
+		/* EFA direct path retries indefinitely when Receiver Not Ready (RNR) */
+		prov_info->domain_attr->resource_mgmt = FI_RM_ENABLED;
+	}
+
 	EFA_DBG(FI_LOG_DOMAIN, "Domain attribute :\n"
 				"\t prov_info->domain_attr->cq_cnt		= %zu\n"
 				"\t prov_info->domain_attr->ep_cnt		= %zu\n"
@@ -179,7 +127,9 @@ const struct fi_ep_attr efa_ep_attr = {
 	.protocol		= FI_PROTO_EFA,
 	.protocol_version	= 1,
 	.msg_prefix_size	= 0,
+	.max_order_raw_size	= 0,
 	.max_order_war_size	= 0,
+	.max_order_waw_size	= 0,
 	.mem_tag_format		= 0,
 	.tx_ctx_cnt		= 1,
 	.rx_ctx_cnt		= 1,
@@ -199,32 +149,18 @@ void efa_prov_info_set_ep_attr(struct fi_info *prov_info,
 {
 
 	*prov_info->ep_attr = efa_ep_attr;
-	if (ep_type == FI_EP_DGRAM) {
-		prov_info->mode |= FI_MSG_PREFIX;
+	prov_info->ep_attr->protocol	= FI_PROTO_EFA;
+	prov_info->ep_attr->max_msg_size = device->ibv_port_attr.max_msg_sz;
+	prov_info->ep_attr->type = ep_type;
+
+	if (ep_type == FI_EP_RDM) {
+		/* ep_attr->max_msg_size is the maximum of both MSG and RMA operations */
+		if (prov_info->caps & FI_RMA)
+			prov_info->ep_attr->max_msg_size = MAX(device->ibv_port_attr.max_msg_sz, device->max_rdma_size);
+	} else {
+		assert(ep_type == FI_EP_DGRAM);
 		prov_info->ep_attr->msg_prefix_size = 40;
 	}
-
-	prov_info->ep_attr->protocol	= FI_PROTO_EFA;
-	prov_info->ep_attr->type	= ep_type;
-
-	if (prov_info->ep_attr->type == FI_EP_RDM) {
-		prov_info->tx_attr->inject_size = device->efa_attr.inline_buf_size;
-	} else {
-		assert(prov_info->ep_attr->type == FI_EP_DGRAM);
-                /*
-                 * Currently, there is no mechanism for EFA layer (lower layer)
-                 * to discard completions internally and FI_INJECT is not optional,
-                 * it can only be disabled by setting inject_size to 0. RXR
-                 * layer does not have this issue as completions can be read from
-                 * the EFA layer and discarded in the RXR layer. For dgram
-                 * endpoint, inject size needs to be set to 0
-                 */
-		prov_info->tx_attr->inject_size = 0;
-	}
-
-	prov_info->ep_attr->max_msg_size		= device->ibv_port_attr.max_msg_sz;
-	prov_info->ep_attr->max_order_raw_size	= device->ibv_port_attr.max_msg_sz;
-	prov_info->ep_attr->max_order_waw_size	= device->ibv_port_attr.max_msg_sz;
 }
 
 /**
@@ -235,7 +171,6 @@ const struct fi_tx_attr efa_dgrm_tx_attr = {
 	.mode			= FI_MSG_PREFIX,
 	.op_flags		= EFA_TX_OP_FLAGS,
 	.msg_order		= EFA_MSG_ORDER,
-	.comp_order		= FI_ORDER_NONE,
 	.inject_size		= 0,
 	.rma_iov_limit		= 0,
 };
@@ -245,11 +180,8 @@ const struct fi_tx_attr efa_dgrm_tx_attr = {
  */
 const struct fi_rx_attr efa_dgrm_rx_attr = {
 	.caps			= EFA_DGRM_RX_CAPS,
-	.mode			= FI_MSG_PREFIX | EFA_RX_MODE,
-	.op_flags		= EFA_RX_DGRM_OP_FLAGS,
+	.mode			= FI_MSG_PREFIX,
 	.msg_order		= EFA_MSG_ORDER,
-	.comp_order		= FI_ORDER_NONE,
-	.total_buffered_recv	= 0,
 	.iov_limit		= 1
 };
 
@@ -261,7 +193,6 @@ const struct fi_tx_attr efa_rdm_tx_attr = {
 	.mode			= 0,
 	.op_flags		= EFA_TX_OP_FLAGS,
 	.msg_order		= EFA_MSG_ORDER,
-	.comp_order		= FI_ORDER_NONE,
 	.inject_size		= 0,
 	.rma_iov_limit		= 1,
 };
@@ -271,11 +202,7 @@ const struct fi_tx_attr efa_rdm_tx_attr = {
  */
 const struct fi_rx_attr efa_rdm_rx_attr = {
 	.caps			= EFA_RDM_RX_CAPS,
-	.mode			= EFA_RX_MODE,
-	.op_flags		= EFA_RX_RDM_OP_FLAGS,
 	.msg_order		= EFA_MSG_ORDER,
-	.comp_order		= FI_ORDER_NONE,
-	.total_buffered_recv	= 0,
 	.iov_limit		= 1
 };
 
@@ -293,12 +220,26 @@ void efa_prov_info_set_tx_rx_attr(struct fi_info *prov_info,
 	if (ep_type == FI_EP_RDM) {
 		*prov_info->tx_attr	= efa_rdm_tx_attr;
 		*prov_info->rx_attr	= efa_rdm_rx_attr;
+
+		prov_info->tx_attr->op_flags |= EFA_RDM_TX_OP_FLAGS;
+		prov_info->rx_attr->op_flags |= FI_COMPLETION;
+
+		/* If the device supports FI_RMA, also set tx_attr and rx_attr */
+		if (prov_info->caps & FI_RMA) {
+			prov_info->tx_attr->caps |= OFI_TX_RMA_CAPS;
+			prov_info->rx_attr->caps |= OFI_RX_RMA_CAPS;
+		}
 	} else {
 		assert(ep_type == FI_EP_DGRAM);
 		*prov_info->tx_attr	= efa_dgrm_tx_attr;
 		*prov_info->rx_attr	= efa_dgrm_rx_attr;
 	}
 
+	/* efa-direct and DGRAM paths require FI_CONTEXT2 */
+	prov_info->tx_attr->mode |= FI_CONTEXT2;
+	prov_info->rx_attr->mode |= FI_CONTEXT2;
+
+	prov_info->tx_attr->inject_size = device->efa_attr.inline_buf_size;
 	prov_info->tx_attr->iov_limit = device->efa_attr.max_sq_sge;
 	prov_info->tx_attr->size = rounddown_power_of_two(device->efa_attr.max_sq_wr);
 	prov_info->rx_attr->iov_limit = device->efa_attr.max_rq_sge;
@@ -414,7 +355,7 @@ static int efa_prov_info_set_nic_attr(struct fi_info *prov_info, struct efa_devi
 		goto err_free;
 	}
 
-	link_attr->mtu = device->ibv_port_attr.max_msg_sz - rxr_pkt_max_hdr_size();
+	link_attr->mtu = device->ibv_port_attr.max_msg_sz;
 	link_attr->speed = ofi_vrb_speed(device->ibv_port_attr.active_speed,
 	                                 device->ibv_port_attr.active_width);
 
@@ -448,20 +389,52 @@ err_free:
 }
 
 #if HAVE_CUDA || HAVE_NEURON || HAVE_SYNAPSEAI
-void efa_prov_info_set_hmem_flags(struct fi_info *prov_info)
+void efa_prov_info_set_hmem_flags(struct fi_info *prov_info, enum fi_ep_type ep_type)
 {
-	if (prov_info->ep_attr->type == FI_EP_RDM &&
-	    (ofi_hmem_is_initialized(FI_HMEM_CUDA) ||
-	     ofi_hmem_is_initialized(FI_HMEM_NEURON) ||
-	     ofi_hmem_is_initialized(FI_HMEM_SYNAPSEAI))) {
-		prov_info->caps			|= FI_HMEM;
-		prov_info->tx_attr->caps		|= FI_HMEM;
-		prov_info->rx_attr->caps		|= FI_HMEM;
-		prov_info->domain_attr->mr_mode	|= FI_MR_HMEM;
+	int i;
+	enum fi_hmem_iface iface;
+	struct efa_hmem_info *hmem_info;
+
+	if (ep_type != FI_EP_RDM)
+		return;
+
+	/* Check if FI_HMEM_P2P_DISABLED is set */
+	if (ofi_hmem_p2p_disabled()) {
+		EFA_WARN(FI_LOG_CORE,
+			"FI_HMEM capability requires peer to peer "
+			"support, which is disabled because "
+			"FI_HMEM_P2P_DISABLED was set to 1/on/true.\n");
+		return;
 	}
+
+	/* Check if HMEM libraries are available at runtime */
+	if (!ofi_hmem_is_initialized(FI_HMEM_CUDA) &&
+	    !ofi_hmem_is_initialized(FI_HMEM_NEURON) &&
+	    !ofi_hmem_is_initialized(FI_HMEM_SYNAPSEAI)) {
+			return;
+	}
+
+	/* EFA direct only supports HMEM when p2p support is available */
+	EFA_HMEM_IFACE_FOREACH_NON_SYSTEM(i) {
+		iface = efa_hmem_ifaces[i];
+		hmem_info = &g_efa_hmem_info[iface];
+		if (hmem_info->initialized && !hmem_info->p2p_supported_by_device) {
+			EFA_INFO(FI_LOG_CORE,
+				"EFA direct provider was compiled with support for %s HMEM interface "
+				"but the interface does not support p2p transfers. "
+				"EFA direct provider does not support HMEM transfers without p2p support. "
+				"HMEM support will be disabled.\n", fi_tostr(&iface, FI_TYPE_HMEM_IFACE));
+				return;
+		}
+	}
+
+	prov_info->caps			|= FI_HMEM;
+	prov_info->tx_attr->caps	|= FI_HMEM;
+	prov_info->rx_attr->caps	|= FI_HMEM;
+	prov_info->domain_attr->mr_mode	|= FI_MR_HMEM;
 }
 #else
-void efa_prov_info_set_hmem_flags(struct fi_info *prov_info)
+void efa_prov_info_set_hmem_flags(struct fi_info *prov_info, enum fi_ep_type ep_type)
 {
 }
 #endif
@@ -494,13 +467,25 @@ int efa_prov_info_alloc(struct fi_info **prov_info_ptr,
 	if (!prov_info)
 		return -FI_ENOMEM;
 
-	if (ep_type != FI_EP_RDM && ep_type != FI_EP_DGRAM) {
-		EFA_WARN(FI_LOG_DOMAIN, "Unsupported endpoint type: %d\n",
-			 ep_type);
-		return -FI_ENODATA;
+	prov_info->mode |= FI_CONTEXT2; 	/* EFA direct path requires FI_CONTEXT2 mode */
+
+	if (ep_type == FI_EP_RDM) {
+		prov_info->caps	= EFA_RDM_CAPS;
+		/* Claim RMA support in the efa-direct path only if read, write
+		 *  and unsolicited write are all available */
+		if (efa_device_support_rdma_read() &&
+			efa_device_support_rdma_write() &&
+			efa_device_support_unsolicited_write_recv())
+			prov_info->caps |= (OFI_TX_RMA_CAPS | OFI_RX_RMA_CAPS);
+	} else {
+		if (ep_type != FI_EP_DGRAM) {
+			EFA_WARN(FI_LOG_DOMAIN, "Unsupported EFA info type: %d\n", ep_type);
+			return -FI_ENODATA;
+		}
+		prov_info->caps	= EFA_DGRM_CAPS;
+		prov_info->mode |= FI_MSG_PREFIX;
 	}
 
-	prov_info->caps	= (ep_type == FI_EP_RDM) ? EFA_RDM_CAPS : EFA_DGRM_CAPS;
 	prov_info->handle = NULL;
 	prov_info->addr_format = FI_ADDR_EFA;
 	prov_info->src_addr = calloc(1, EFA_EP_ADDR_LEN);
@@ -511,10 +496,8 @@ int efa_prov_info_alloc(struct fi_info **prov_info_ptr,
 	prov_info->src_addrlen = EFA_EP_ADDR_LEN;
 	memcpy(prov_info->src_addr, &device->ibv_gid, sizeof(device->ibv_gid));
 
-	err = efa_prov_info_set_fabric_attr(prov_info, device);
-	if (err) {
-		goto err_free;
-	}
+	/* fabric_attr->name is set in efa_util_prov_initialize */
+	*prov_info->fabric_attr	= efa_fabric_attr;
 
 	err = efa_prov_info_set_domain_attr(prov_info, device, ep_type);
 	if (err) {
@@ -530,7 +513,7 @@ int efa_prov_info_alloc(struct fi_info **prov_info_ptr,
 		goto err_free;
 	}
 
-	efa_prov_info_set_hmem_flags(prov_info);
+	efa_prov_info_set_hmem_flags(prov_info, ep_type);
 
 	*prov_info_ptr = prov_info;
 	return 0;
@@ -542,150 +525,165 @@ err_free:
 }
 
 /**
- * @brief allocate an prov_info object that matches the functionality of RxR
+ * @brief allocate an prov_info object that matches the functionality of EFA RDM endpoint
  *
- * RxR (RDM over RDM) is a software layer that implemented mulitple
- * additional functionalities for an RDM endpoint. The added functionalities
+ * EFA RDM endpoint implemented mulitple additional functionalities over rdma-core
+ * SRD QP. The added functionalities
  * include:
  *
  * tag matching, orderred send/receive, message segmentation,
  * emulated RMA/atomic emulation, multi recv.
  *
- * This function allocates a prov_info object the matches RxR's functionality.
+ * This function allocates a prov_info object the matches EFA RDM endpoint's functionality.
  *
- * @param[out]		prov_info_rxr_ptr	pointer to pointer of prov_info_rxr
+ * Note that this info object is not to be confused the with rdm_info in struct efa_device.
+ * That info object translates the rdma-core QP's capability to libfabric's info object.
+ *
+ * @param[out]		prov_info_rdm_ptr	pointer to pointer of prov_info_rdm
  * @param[in]		device			efa_device information
  * @return		0 on success
  * 			-FI_ENOMEM if memory allocation failed.
  */
-int efa_prov_info_alloc_for_rxr(struct fi_info **prov_info_rxr_ptr,
+int efa_prov_info_alloc_for_rdm(struct fi_info **prov_info_rdm_ptr,
 				struct efa_device *device)
 {
-	uint64_t rxr_added_tx_caps = FI_TAGGED | OFI_TX_RMA_CAPS | FI_ATOMIC;
+	uint64_t efa_rdm_added_tx_caps = FI_TAGGED | OFI_TX_RMA_CAPS | FI_ATOMIC;
 
-	uint64_t rxr_added_rx_caps = FI_TAGGED | OFI_RX_RMA_CAPS | FI_ATOMIC |
+	uint64_t efa_rdm_added_rx_caps = FI_TAGGED | OFI_RX_RMA_CAPS | FI_ATOMIC |
 				     FI_SOURCE | FI_MULTI_RECV | FI_DIRECTED_RECV;
 
-	uint64_t rxr_domain_caps = FI_LOCAL_COMM | FI_REMOTE_COMM;
+	uint64_t efa_domain_caps = FI_LOCAL_COMM | FI_REMOTE_COMM;
 
-	struct fi_info *prov_info_rxr;
+	struct fi_info *prov_info_rdm;
 
 	assert(device->rdm_info);
 
-	prov_info_rxr  = fi_dupinfo(device->rdm_info);
-	if (!prov_info_rxr)
+	prov_info_rdm  = fi_dupinfo(device->rdm_info);
+	if (!prov_info_rdm)
 		return -FI_ENOMEM;
 
-	prov_info_rxr->caps |= rxr_added_tx_caps | rxr_added_rx_caps | rxr_domain_caps;
+	prov_info_rdm->caps |= efa_rdm_added_tx_caps | efa_rdm_added_rx_caps | efa_domain_caps;
+
+	/* efa-direct requires FI_CONTEXT2 but RDM doesn't. So unset FI_CONTEXT2 */
+	prov_info_rdm->mode &= ~FI_CONTEXT2;
+	prov_info_rdm->tx_attr->mode &= ~FI_CONTEXT2;
+	prov_info_rdm->rx_attr->mode &= ~FI_CONTEXT2;
 
 	/* update domain_attr */
 	{
-		/* RxR ensure thread safety by pthread lock */
-		prov_info_rxr->domain_attr->threading = FI_THREAD_SAFE;
-		/* RxR handles Receiver Not Ready (RNR) events by queuing the send,
+		/* EFA RDM endpoint ensure thread safety by pthread lock */
+		prov_info_rdm->domain_attr->threading = FI_THREAD_SAFE;
+		/* EFA RDM endpoint handles Receiver Not Ready (RNR) events by queuing the send,
 		 * hence resource_mgmt is enabled.
 		 */
-		prov_info_rxr->domain_attr->resource_mgmt = FI_RM_ENABLED;
+		prov_info_rdm->domain_attr->resource_mgmt = FI_RM_ENABLED;
 		/*
 		 * The device endpoint requires a memory descriptor for any send/receive.
 		 * Therefore it set the FI_MR_LOCAL mode.
-		 * buffer. RxR does not have this requirement, hence unset the flag
+		 * buffer. EFA RDM endpoint does not have this requirement, hence unset the flag
 		 */
-		prov_info_rxr->domain_attr->mr_mode &= ~FI_MR_LOCAL;
-
-		/* RxR support writing CQ data by put it in packet header
+		prov_info_rdm->domain_attr->mr_mode &= ~FI_MR_LOCAL;
+		/*
+		 * EFA RDM path requires manual progress for connection management
+		 * and data transfers
 		 */
-		prov_info_rxr->domain_attr->cq_data_size = RXR_CQ_DATA_SIZE;
+		prov_info_rdm->domain_attr->control_progress = FI_PROGRESS_MANUAL;
+		prov_info_rdm->domain_attr->progress = FI_PROGRESS_MANUAL;
 	}
 
 	/* update ep_attr */
 	{
 		int max_atomic_size;
 
-		prov_info_rxr->ep_attr->protocol = FI_PROTO_EFA;
-		prov_info_rxr->ep_attr->mem_tag_format = FI_TAG_GENERIC;
-		prov_info_rxr->ep_attr->protocol_version = RXR_PROTOCOL_VERSION;
+		prov_info_rdm->ep_attr->protocol = FI_PROTO_EFA;
+		prov_info_rdm->ep_attr->mem_tag_format = FI_TAG_GENERIC;
+		prov_info_rdm->ep_attr->protocol_version = EFA_RDM_PROTOCOL_VERSION;
 		/*
-		 * RxR support message segmentation, hence increase the max_msg_size
+		 * EFA RDM endpoint support message segmentation, hence increase the max_msg_size
 		 */
-		prov_info_rxr->ep_attr->max_msg_size = UINT64_MAX;
+		prov_info_rdm->ep_attr->max_msg_size = UINT64_MAX;
 
 		/*
-		 * RxR implemented emulated atomic, hence set atomic size
+		 * EFA RDM endpoint implemented emulated atomic, hence set atomic size
 		 */
-		max_atomic_size = device->rdm_info->ep_attr->max_msg_size
-					- sizeof(struct rxr_rta_hdr)
+		max_atomic_size = device->ibv_port_attr.max_msg_sz
+					- sizeof(struct efa_rdm_rta_hdr)
 					- device->rdm_info->src_addrlen
-					- RXR_IOV_LIMIT * sizeof(struct fi_rma_iov);
-		prov_info_rxr->ep_attr->max_order_raw_size = max_atomic_size;
+					- EFA_RDM_IOV_LIMIT * sizeof(struct fi_rma_iov);
+		prov_info_rdm->ep_attr->max_order_raw_size = max_atomic_size;
+		prov_info_rdm->ep_attr->max_order_war_size = max_atomic_size;
+		prov_info_rdm->ep_attr->max_order_waw_size = max_atomic_size;
 	}
 
 	/* update tx_attr */
 	{
 		int min_pkt_size;
 		/*
-		 * RxR supports ordered two-sided/atomic by putting messages by
+		 * EFA RDM endpoint supports ordered two-sided/atomic by putting messages by
 		 * software reorder buffer, hence set tx_attr->message order accordingly.
 		 */
-		prov_info_rxr->tx_attr->caps |= rxr_added_tx_caps;
-		prov_info_rxr->tx_attr->msg_order = FI_ORDER_SAS | FI_ORDER_ATOMIC_RAR | FI_ORDER_ATOMIC_RAW |
+		prov_info_rdm->tx_attr->caps |= efa_rdm_added_tx_caps;
+		prov_info_rdm->tx_attr->msg_order = FI_ORDER_SAS | FI_ORDER_ATOMIC_RAR | FI_ORDER_ATOMIC_RAW |
 						    FI_ORDER_ATOMIC_WAR | FI_ORDER_ATOMIC_WAW;
 
 		/*
-		 * RxR supports injection by software emulation.
-		 * RxR supports delivery complete by using DC capable protocols.
+		 * EFA RDM endpoint supports injection by software emulation.
+		 * EFA RDM endpoint supports delivery complete by using DC capable protocols.
 		 * Therefore changing the default op_flags
 		 */
-		prov_info_rxr->tx_attr->op_flags = FI_INJECT | FI_COMPLETION | FI_TRANSMIT_COMPLETE |
-						   FI_DELIVERY_COMPLETE;
+		prov_info_rdm->tx_attr->op_flags |= EFA_RDM_TX_OP_FLAGS;
 
-		/* Here we calculate the max msg size for emulated injection of RxR.
+		/* Here we calculate the max msg size for emulated injection of EFA RDM endpoint.
 		 * The requirement for inject is: upon return, the user buffer can be reused immediately.
 		 *
 		 * For EFA, inject is implement as: construct a packet entry, copy user data to packet entry
 		 * then send the packet entry. Therefore the maximum inject size is
 		 *    pkt_entry_size - maximum_header_size.
 		 */
-		if (rxr_env.enable_shm_transfer)
-			min_pkt_size = MIN(device->rdm_info->ep_attr->max_msg_size, rxr_env.shm_max_medium_size);
+		if (efa_env.enable_shm_transfer)
+			min_pkt_size = MIN(device->ibv_port_attr.max_msg_sz - efa_rdm_pkt_type_get_max_hdr_size(),
+					   SHM_MAX_INJECT_SIZE);
 		else
-			min_pkt_size = device->rdm_info->ep_attr->max_msg_size;
+			min_pkt_size = device->ibv_port_attr.max_msg_sz - efa_rdm_pkt_type_get_max_hdr_size();
 
-		if (min_pkt_size < rxr_pkt_max_hdr_size()) {
-			prov_info_rxr->tx_attr->inject_size = 0;
+		if (min_pkt_size < efa_rdm_pkt_type_get_max_hdr_size()) {
+			prov_info_rdm->tx_attr->inject_size = 0;
 		} else {
-			prov_info_rxr->tx_attr->inject_size = min_pkt_size - rxr_pkt_max_hdr_size();
+			prov_info_rdm->tx_attr->inject_size = min_pkt_size;
 		}
 
 		/*
-		 * RxR support multiple IOV by segmentation.
+		 * EFA RDM endpoint support multiple IOV by segmentation.
 		 */
-		prov_info_rxr->tx_attr->iov_limit = RXR_IOV_LIMIT;
-		if (rxr_env.tx_iov_limit > 0)
-			prov_info_rxr->tx_attr->iov_limit = rxr_env.tx_iov_limit;
+		prov_info_rdm->tx_attr->iov_limit = EFA_RDM_IOV_LIMIT;
 
-		if (rxr_env.tx_size > 0)
-			prov_info_rxr->tx_attr->size = rxr_env.tx_size;
+		if (efa_env.tx_size > 0)
+			prov_info_rdm->tx_attr->size = efa_env.tx_size;
 
 	}
 
-	/*
-	 * Set RX attributes for RxR info
-	 */
 	{
-		prov_info_rxr->rx_attr->caps |= rxr_added_rx_caps;
-		prov_info_rxr->rx_attr->msg_order = FI_ORDER_SAS | FI_ORDER_ATOMIC_RAR | FI_ORDER_ATOMIC_RAW |
+		prov_info_rdm->rx_attr->caps |= efa_rdm_added_rx_caps;
+		prov_info_rdm->rx_attr->msg_order = FI_ORDER_SAS | FI_ORDER_ATOMIC_RAR | FI_ORDER_ATOMIC_RAW |
 						    FI_ORDER_ATOMIC_WAR | FI_ORDER_ATOMIC_WAW;
-		prov_info_rxr->rx_attr->op_flags = FI_COMPLETION | FI_MULTI_RECV;
-		prov_info_rxr->rx_attr->iov_limit = RXR_IOV_LIMIT;
-		if (rxr_env.rx_iov_limit > 0)
-			prov_info_rxr->rx_attr->iov_limit = rxr_env.rx_iov_limit;
+		prov_info_rdm->rx_attr->op_flags = FI_COMPLETION | FI_MULTI_RECV;
+		prov_info_rdm->rx_attr->iov_limit = EFA_RDM_IOV_LIMIT;
 
-		if (rxr_env.rx_size > 0)
-			prov_info_rxr->rx_attr->size = rxr_env.rx_size;
+		if (efa_env.rx_size > 0)
+			prov_info_rdm->rx_attr->size = efa_env.rx_size;
 	}
 
-	*prov_info_rxr_ptr = prov_info_rxr;
+	/* EFA RDM can support HMEM even if p2p support is not available */
+	if ((ofi_hmem_is_initialized(FI_HMEM_CUDA) ||
+	     ofi_hmem_is_initialized(FI_HMEM_NEURON) ||
+	     ofi_hmem_is_initialized(FI_HMEM_SYNAPSEAI))) {
+		prov_info_rdm->caps |= FI_HMEM;
+		prov_info_rdm->tx_attr->caps |= FI_HMEM;
+		prov_info_rdm->rx_attr->caps |= FI_HMEM;
+		prov_info_rdm->domain_attr->mr_mode |= FI_MR_HMEM;
+	}
+
+	*prov_info_rdm_ptr = prov_info_rdm;
 	return 0;
 }
 
@@ -740,3 +738,23 @@ int efa_prov_info_compare_domain_name(const struct fi_info *hints,
 
        return 0;
 }
+
+/*
+ * @brief Compare the bus id specified via hints and match it with the
+ *		  bus_id in prov_info
+ *
+ * @param[in]  info        info object
+ * @param[in]  hints       hints from user's call to fi_getinfo()
+ *
+ * return      1 - If the bus_id in hints does not match info
+ *             0 - If the bus_id in hints match info, or user did not specify bus ID in hints
+ */
+int efa_prov_info_compare_pci_bus_id(const struct fi_info *hints,
+                                     const struct fi_info *info)
+{
+	if (hints && hints->nic && hints->nic->bus_attr && hints->nic->bus_attr->bus_type == FI_BUS_PCI) {
+		return (hints->nic->bus_attr->attr.pci.bus_id == info->nic->bus_attr->attr.pci.bus_id) ? 0 : 1;
+	}
+
+	return 0;
+}
