@@ -1,8 +1,7 @@
-#include "efa_unit_tests.h"
+/* SPDX-License-Identifier: BSD-2-Clause OR GPL-2.0-only */
+/* SPDX-FileCopyrightText: Copyright Amazon.com, Inc. or its affiliates. All rights reserved. */
 
-int efa_device_construct(struct efa_device *efa_device,
-			 int device_idx,
-			 struct ibv_device *ibv_device);
+#include "efa_unit_tests.h"
 
 /*
  * test the error handling path of efa_device_construct()
