@@ -112,14 +112,8 @@ static PSMI_HAL_INLINE int psm3_hfp_sockets_context_open(int unit,
 		err = -PSM_HAL_ERROR_CANNOT_OPEN_CONTEXT;
 		goto bail;
 	}
-	cpu_set_t mycpuset;
-	if (psm3_sysfs_get_unit_cpumask(unit, &mycpuset)) {
-		_HFI_ERROR( "Failed to get %s (unit %d) cpu set\n", ep->dev_name, unit);
-		//err = -PSM_HAL_ERROR_GENERAL_ERROR;
-		goto bail;
-	}
 
-	if (psm3_context_set_affinity(ep, mycpuset))
+	if (psm3_context_set_affinity(ep, unit))
 		goto bail;
 
 // TBD - inside psm3_gen1_userinit_internal we would find CPU
@@ -140,20 +134,22 @@ static PSMI_HAL_INLINE int psm3_hfp_sockets_get_port_index2pkey(psm2_ep_t ep, in
 	return 0x8001;
 }
 
-/* Tell the driver to change the way packets can generate interrupts.
+/* Tell the driver to change the way packets can generate interrupts
+   which wakeup poll().
 
- HFI1_POLL_TYPE_URGENT: Generate interrupt only when send with
-			IPS_SEND_FLAG_INTR (HFI_KPF_INTR)
- HFI1_POLL_TYPE_ANYRCV: wakeup on any rcv packet (when polled on). [not used]
+ PSMI_HAL_POLL_TYPE_NONE: disable interrupt generation for future packets
+ PSMI_HAL_POLL_TYPE_URGENT: Generate interrupt only when receive pkt sent with
+			IPS_SEND_FLAG_INTR
+ PSMI_HAL_POLL_TYPE_ANYRCV: interrupt on any rcv packet
 
- PSM: Uses TYPE_URGENT in ips protocol
+ PSM Uses TYPE_URGENT in ips protocol and needs TYPE_ANYRCV for psm3_wait()
+ TYPE_NONE is used during job shutdown
+
+ Sockets does not implement TYPE_URGENT nor TYPE_ANYRCV
 */
 static PSMI_HAL_INLINE int psm3_hfp_sockets_poll_type(uint16_t poll_type, psm2_ep_t ep)
 {
-	// TBD - UDP does not yet implement urgent interrupt wakeup for
-	// rcvThread.  TBD if we can do that with sockets
-	//return psm3_sockets_poll_type(poll_type, ep);
-	return 0;
+	return psm3_sockets_poll_type(poll_type, ep);
 }
 
 // initialize HAL specific parts of ptl_ips
@@ -193,7 +189,7 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_ips_proto_init(
 }
 
 // Fetch current link state to update linkinfo fields in ips_proto:
-// 	ep_base_lid, ep_lmc, ep_link_rate, QoS tables, CCA tables
+// 	ep_base_lid, ep_lmc, ep_link_rate
 // These are all fields which can change during a link bounce.
 // Note "active" state is not adjusted as on link down PSM will wait for
 // the link to become usable again so it's always a viable/active device
@@ -238,13 +234,27 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_ips_ipsaddr_set_req_params(
 			(proto->ep->sockets_ep.tcp_incoming_fd &&
 			ipsaddr->sockets.tcp_fd != proto->ep->sockets_ep.tcp_incoming_fd &&
 			psm3_epid_cmp_internal(ipsaddr->epaddr.epid, proto->ep->epid) == -1)) {
-			psm3_sockets_tcp_close_fd(proto->ep, ipsaddr->sockets.tcp_fd, -1,
-				&ipsaddr->flows[EP_FLOW_GO_BACK_N_PIO]);
-			_HFI_VDBG("Replace fd=%d with %d\n", ipsaddr->sockets.tcp_fd,
-				proto->ep->sockets_ep.tcp_incoming_fd);
+			if (ipsaddr->sockets.tcp_fd > 0) {
+				psm3_sockets_tcp_close_fd(proto->ep, ipsaddr->sockets.tcp_fd, -1,
+					&ipsaddr->flows[EP_FLOW_GO_BACK_N_PIO]);
+				_HFI_VDBG("Replace fd=%d with %d\n", ipsaddr->sockets.tcp_fd,
+					proto->ep->sockets_ep.tcp_incoming_fd);
+			}
 			ipsaddr->sockets.tcp_fd = proto->ep->sockets_ep.tcp_incoming_fd;
 			ipsaddr->sockets.connected = 1;
 		}
+		struct fd_ctx *ctx = psm3_sockets_get_fd_ctx(proto->ep, ipsaddr->sockets.tcp_fd);
+		if (!ctx) {
+			// the existing ipsaddr->sockets.tcp_fd not ready yet. using the incoming fd
+			ipsaddr->sockets.tcp_fd = proto->ep->sockets_ep.tcp_incoming_fd;
+			ctx = psm3_sockets_get_fd_ctx(proto->ep, ipsaddr->sockets.tcp_fd);
+		}
+		if (ctx) {
+			ctx->ipsaddr = ipsaddr;
+			if (ctx->state == FD_STATE_NONE) {
+				ctx->state = FD_STATE_READY;
+			}
+		}
 	}
 	return PSM2_OK;
 }
@@ -258,6 +268,11 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_ips_ipsaddr_process_connect
 				ips_epaddr_t *ipsaddr,
 				const struct ips_connect_reqrep *req)
 {
+	_HFI_PRDBG("CONN ESTABLISHED fd=%d nfds=%d\n", ipsaddr->sockets.tcp_fd, proto->ep->sockets_ep.nfds);
+	struct fd_ctx *ctx = psm3_sockets_get_fd_ctx(proto->ep, ipsaddr->sockets.tcp_fd);
+	if (ctx) {
+		ctx->state = FD_STATE_ESTABLISHED;
+	}
 	return PSM2_OK;
 }
 
@@ -300,7 +315,7 @@ static PSMI_HAL_INLINE void psm3_hfp_sockets_ips_ipsaddr_init_addressing(
 {
 	psm3_epid_build_sockaddr(&ipsaddr->sockets.remote_pri_addr, epid,
 					proto->ep->sockets_ep.if_index);
-	ipsaddr->hash = ipsaddr->sockets.remote_pri_addr.sin6_port;
+	ipsaddr->hash = psm3_socket_port(&ipsaddr->sockets.remote_pri_addr);
 	psm3_epid_get_av(epid, lidp, gidp);
 	if (proto->ep->sockets_ep.sockets_mode == PSM3_SOCKETS_TCP) {
 		psm3_epid_build_aux_sockaddr(&ipsaddr->sockets.remote_aux_addr, epid,
@@ -335,37 +350,12 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_ips_ipsaddr_init_connection
 			proto->ep->sockets_ep.tcp_incoming_fd = 0;
 			PSM2_LOG_MSG("connected to %s fd=%d", psm3_epid_fmt_internal(epid, 0), ipsaddr->sockets.tcp_fd);
 		} else {
-			ipsaddr->sockets.tcp_fd = socket(AF_INET6, SOCK_STREAM | SOCK_NONBLOCK, 0);
-			// 0 is a valid fd, but we treat 0 as an uninitialized value, so make
-			// sure we didn't end up with 0 (stdin should have 0)
-			if (ipsaddr->sockets.tcp_fd == 0) {
-				// create a copy of the file descriptor using the unused one
-				ipsaddr->sockets.tcp_fd = dup(ipsaddr->sockets.tcp_fd);
-				close(0);
-			}
-			if (ipsaddr->sockets.tcp_fd == -1) {
-				_HFI_ERROR( "Unable to create TCP tx socket for %s: %s\n", proto->ep->dev_name, strerror(errno));
-				err = PSM2_INTERNAL_ERR;
-				goto fail;
-			}
-
-			if (psm3_tune_tcp_socket("socket", proto->ep, ipsaddr->sockets.tcp_fd)) {
-				_HFI_ERROR("unable to tune socket for connection to %s for %s: %s\n",
-					psm3_sockaddr_fmt((struct sockaddr *)&ipsaddr->sockets.remote_pri_addr, 0),
-					proto->ep->dev_name, strerror(errno));
-				err = PSM2_INTERNAL_ERR;
-				goto fail;
-			}
+			// will create socket and connect to remote on the first data send attempt
+			ipsaddr->sockets.tcp_fd = -1;
+			ipsaddr->sockets.connected = 0;
 		}
 	}
 	return err;
-
-fail:
-	if (ipsaddr->sockets.tcp_fd > 0) {
-		close(ipsaddr->sockets.tcp_fd);
-		ipsaddr->sockets.tcp_fd = 0;
-	}
-	return err;
 }
 
 /* handle HAL specific ipsaddr free for any HAL specific information
@@ -379,6 +369,12 @@ static PSMI_HAL_INLINE void psm3_hfp_sockets_ips_ipsaddr_free(
 			psm3_sockets_tcp_close_fd(proto->ep, ipsaddr->sockets.tcp_fd, -1, &ipsaddr->flows[EP_FLOW_GO_BACK_N_PIO]);
 			ipsaddr->sockets.tcp_fd = 0;
 		}
+		struct ips_flow *flow = &ipsaddr->flows[EP_FLOW_GO_BACK_N_PIO];
+		if (flow->partial_conn_msg) {
+			psmi_free(flow->partial_conn_msg);
+			flow->partial_conn_msg = NULL;
+			flow->partial_conn_msg_size = 0;
+		}
 	}
 }
 
@@ -397,14 +393,12 @@ static PSMI_HAL_INLINE void psm3_hfp_sockets_ips_flow_init(
 		flow->flush = psm3_ips_proto_flow_flush_pio;
 	}
 
-	flow->send_remaining = 0;
-	flow->used_snd_buff = 0;
-
 	_HFI_CONNDBG("[ipsaddr=%p] %s flow->frag_size: %u = min("
-		"proto->epinfo.ep_mtu(%u), flow->path->pr_mtu(%u))\n",
+		"proto->epinfo.ep_mtu(%u), flow->path->pr_mtu(%u)) fd=%d\n",
 		flow->ipsaddr, 
 		(proto->ep->sockets_ep.sockets_mode == PSM3_SOCKETS_TCP)?"TCP":"UDP",
-		flow->frag_size, proto->epinfo.ep_mtu, flow->path->pr_mtu);
+		flow->frag_size, proto->epinfo.ep_mtu, flow->path->pr_mtu,
+		flow->ipsaddr->sockets.tcp_fd);
 }
 
 /* handle HAL specific connection processing as part of processing an
@@ -415,7 +409,7 @@ static PSMI_HAL_INLINE void psm3_hfp_sockets_ips_ipsaddr_disconnect(
 {
 }
 
-/* Handle HAL specific initialization of ibta path record query, CCA
+/* Handle HAL specific initialization of ibta path record query
  * and dispersive routing
  */
 static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_ips_ibta_init(
@@ -449,13 +443,13 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_ips_path_rec_init(
 static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_ips_ptl_pollintr(
 		psm2_ep_t ep, struct ips_recvhdrq *recvq,
 		int fd_pipe, int next_timeout,
-		uint64_t *pollok, uint64_t *pollcyc)
+		uint64_t *pollok, uint64_t *pollcyc, uint64_t *pollintr)
 {
 	return psm3_sockets_ips_ptl_pollintr(ep, recvq, fd_pipe,
-					 next_timeout, pollok, pollcyc);
+					 next_timeout, pollok, pollcyc, pollintr);
 }
 
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+#ifdef PSM_HAVE_GPU
 static PSMI_HAL_INLINE void psm3_hfp_sockets_gdr_close(void)
 {
 }
@@ -466,16 +460,7 @@ static PSMI_HAL_INLINE void* psm3_hfp_sockets_gdr_convert_gpu_to_host_addr(unsig
 	return psm3_sockets_gdr_convert_gpu_to_host_addr(buf, size, flags,
                                 ep);
 }
-#ifdef PSM_ONEAPI
-static PSMI_HAL_INLINE void psm3_hfp_sockets_gdr_munmap_gpu_to_host_addr(unsigned long buf,
-                                size_t size, int flags,
-                                psm2_ep_t ep)
-{
-	return psm3_sockets_gdr_munmap_gpu_to_host_addr(buf, size, flags,
-                                ep);
-}
-#endif
-#endif /* PSM_CUDA || PSM_ONEAPI */
+#endif /* PSM_HAVE_GPU */
 
 #include "sockets_spio.c"
 
@@ -484,7 +469,7 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_spio_transfer_frame(struct
 					uint32_t *payload, uint32_t length,
 					uint32_t isCtrlMsg, uint32_t cksum_valid,
 					uint32_t cksum
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+#ifdef PSM_HAVE_GPU
 				, uint32_t is_gpu_payload
 #endif
 	)
@@ -505,7 +490,7 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_spio_transfer_frame(struct
 		return psm3_sockets_udp_spio_transfer_frame(proto, flow, scb,
 					payload, length, isCtrlMsg,
 					cksum_valid, cksum
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+#ifdef PSM_HAVE_GPU
 					, is_gpu_payload
 #endif
 					);
@@ -514,7 +499,7 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_spio_transfer_frame(struct
 		return psm3_sockets_tcp_spio_transfer_frame(proto, flow, scb,
 					payload, length, isCtrlMsg,
 					cksum_valid, cksum
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+#ifdef PSM_HAVE_GPU
 					, is_gpu_payload
 #endif
 					);
@@ -525,7 +510,7 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_transfer_frame(struct ips_p
 					uint32_t *payload, uint32_t length,
 					uint32_t isCtrlMsg, uint32_t cksum_valid,
 					uint32_t cksum
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+#ifdef PSM_HAVE_GPU
 				, uint32_t is_gpu_payload
 #endif
 	)
@@ -533,7 +518,7 @@ static PSMI_HAL_INLINE psm2_error_t psm3_hfp_sockets_transfer_frame(struct ips_p
 	return psm3_hfp_sockets_spio_transfer_frame(proto, flow, scb,
 					payload, length, isCtrlMsg,
 					cksum_valid, cksum
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+#ifdef PSM_HAVE_GPU
 					, is_gpu_payload
 #endif
 					);
