@@ -66,7 +66,7 @@ extern "C" {
  * @file psm2.h
  * @page psm2_main PSM2 API
  *
- * @brief PSM2 OPA Messaging Library
+ * @brief PSM2 Messaging Library
  *
  * The PSM2 OPA Messaging API, or PSM2 API, is Intel's low-level
  * user-level communications interface for the OPA family of products.
@@ -314,8 +314,8 @@ enum psm2_error {
 	/*! PSM2 is finalized */
 	PSM2_IS_FINALIZED = 13,
 
-	/*! TCP data send is successful */
-	PSM2_TCP_DATA_SENT = 14,
+	/*! data was sent reliably */
+	PSM2_RELIABLE_DATA_SENT = 14,
 
 	/*! Endpoint was closed */
 	PSM2_EP_WAS_CLOSED = 20,
@@ -666,11 +666,11 @@ typedef psm2_epid_t psm2_nid_t;
  */
 psm2_nid_t psm3_epid_nid(psm2_epid_t epid);
 
-/** @brief Get Endpoint identifier's OPA context number */
+/** @brief Get Endpoint identifier's context number */
 uint64_t psm3_epid_context(psm2_epid_t epid);
 #endif // 0
 
-/** @brief Get Endpoint identifier's OPA port (deprecated, use
+/** @brief Get Endpoint identifier's network port (deprecated, use
  * @ref psm3_epid_context instead) */
 uint64_t psm3_epid_port(psm2_epid_t epid);
 
@@ -743,10 +743,10 @@ struct psm3_ep_open_opts {
 	int imm_size;		/* Immediate data size for endpoint */
 };
 
-/** @brief OPA endpoint creation
+/** @brief PSM3 endpoint creation
  *
- * Function used to create a new local communication endpoint on an OPA
- * adapter.  The returned endpoint handle is required in all PSM2 communication
+ * Function used to create a new local communication endpoint on an adapter/NIC.
+ * The returned endpoint handle is required in all PSM2 communication
  * operations, as PSM2 can manage communication over multiple endpoints.  An
  * opened endpoint has no global context until the user connects the endpoint
  * to other global endpoints by way of @ref psm3_ep_connect.  All local endpoint
@@ -1325,10 +1325,19 @@ void *psm3_epaddr_getctxt(psm2_epaddr_t epaddr);
    * option value: Context associated with PSM2 endpoint address.
    */
 
+/* PSM2 endpoint CUDA_PERMITTED flag */
+#define PSM2_CORE_OPT_EP_CUDA_PERMITTED   0x103
+  /**< [@b uint32_t ] Set/Get the CUDA_PERMITTED flag associated with a PSM2
+   * endpoint (psm2_ep_t).
+   *
+   * component object: PSM2 endpoint (@ref psm2_ep_t).
+   * option value: Boolean flag.
+   */
+
 /* PSM2_COMPONENT_IB options */
 /* Default service level to use to communicate with remote endpoints */
 #define PSM2_IB_OPT_DF_SL 0x201
-  /**< [@b uint32_t ] Default OPA SL to use for all remote communication.
+  /**< [@b uint32_t ] Default OPA/IB SL to use for all remote communication.
    * If unset defaults to Service Level 0.
    *
    * component object: Opened PSM2 endpoint id (@ref psm2_ep_t).
@@ -1337,7 +1346,7 @@ void *psm3_epaddr_getctxt(psm2_epaddr_t epaddr);
 
 /* Set IB service level to use for communication to an endpoint */
 #define PSM2_IB_OPT_EP_SL 0x202
-  /**< [@b uint32_t ] OPA SL to use for communication to specified
+  /**< [@b uint32_t ] OPA/IB SL to use for communication to specified
    * remote endpoint.
    *
    * component object: PSM2 endpoint (@ ref psm2_epaddr_t) address.
@@ -1348,7 +1357,7 @@ void *psm3_epaddr_getctxt(psm2_epaddr_t epaddr);
 /* MQ options that can be set in psm3_mq_init and psm2_{set,get}_opt */
 #define PSM2_MQ_OPT_RNDV_IB_SZ       0x301
   /**< [@b uint32_t ] Size at which to start enabling rendezvous
-   * messaging for OPA messages (if unset, defaults to values
+   * messaging for PSM3 messages (if unset, defaults to values
    * between 56000 and 72000 depending on the system configuration)
    *
    * component object: PSM2 Matched Queue (@ref psm2_mq_t).
@@ -1376,6 +1385,16 @@ void *psm3_epaddr_getctxt(psm2_epaddr_t epaddr);
    * option value: Deprecated; this option has no effect.
    */
 
+#define PSM2_MQ_OPT_GPU_RNDV_SHM_SZ      0x304
+#define PSM2_MQ_GPU_RNDV_SHM_SZ          PSM2_MQ_OPT_GPU_RNDV_SHM_SZ
+  /**< [@b uint32_t ] Size at which to start enabling
+   * rendezvous messaging for shared memory (intra-node) GPU messages (If
+   * unset, defaults to 127 bytes for Intel GPU, 127 for NVIDIA GPU).
+   *
+   * component object: PSM2 Matched Queue (@ref psm2_mq_t).
+   * option value: Size at which to switch to rendezvous protocol for GPU send.
+   */
+
 /* PSM2_COMPONENT_AM options */
 #define PSM2_AM_OPT_FRAG_SZ          0x401
 #define PSM2_AM_MAX_FRAG_SZ          PSM2_AM_OPT_FRAG_SZ
@@ -1526,6 +1545,13 @@ psm2_error_t psm3_ep_query(int *num_of_epinfo, psm2_epinfo_t *array_of_epinfo);
  */
 psm2_error_t psm3_ep_epid_lookup(psm2_epid_t epid, psm2_epconn_t *epconn);
 
+/** @brief return psm3 process label
+ *
+ * Function to report process label for use in log messages.
+ * The process label consists of the hostname and rank.
+ */
+char *psm3_get_mylabel();
+
 /** @brief Query given PSM2 end-point for its connections.
  *
  * The need for this function comes with 'multi-ep' feature.
@@ -1598,19 +1624,11 @@ typedef enum psm2_info_query_et
                      active.  */
 	PSM2_INFO_QUERY_UNIT_PORT_STATUS,
 
-/*! Required input arguments: 1
-   1.  type: uint32_t, description: the unit for which the number of
-       free contexts is desired (use: psm2_info_query_arg_t.unit).
-   Output parameter: uint32_t, description: the number of free
-                     contexts..  */
-	PSM2_INFO_QUERY_NUM_FREE_CONTEXTS,
+/*! removed QUERY_NUM_FREE_CONTEXTS, but kept placeholder to retain values in enum */
+	PSM2_WAS_INFO_QUERY_NUM_FREE_CONTEXTS,
 
-/*! Required input arguments: 1
-   1.  type: uint32_t, description: the unit for which the number of
-       contexts is desired (use: psm2_info_query_arg_t.unit).
-   Output parameter: uint32_t, description: the number of
-                     contexts..  */
-	PSM2_INFO_QUERY_NUM_CONTEXTS,
+/*! removed QUERY_NUM_CONTEXTS, but kept placeholder to retain values in enum */
+	PSM2_WAS_INFO_QUERY_NUM_CONTEXTS,
 
 /*! removed QUERY_CONFIG, but kept placeholder to retain values in enum */
 	PSM2_WAS_INFO_QUERY_CONFIG,
@@ -1708,6 +1726,13 @@ typedef enum psm2_info_query_et
        Output parameter: char*, description: name of the device's address. */
 	PSM2_INFO_QUERY_UNIT_ADDR_NAME,
 
+/*! Required input arguments 0
+   Output parameter: uint32_t*, description: configured PSM3_GPU_THRESH_RNDV */
+	PSM2_INFO_QUERY_GPU_THRESH_RNDV,
+
+/*! Required input arguments 0
+   Output parameter: uint32_t*, description: default for PSM3_MQ_RNDV_SHM_GPU_THRESH */
+	PSM2_INFO_QUERY_MQ_RNDV_SHM_GPU_THRESH_DEFAULT,
 
 	PSM2_INFO_QUERY_LAST, /* must appear last, and the info query
 				 constants are used as an index. */
@@ -1734,6 +1759,7 @@ typedef union psm2_info_query_arg
 	size_t                         length;
 	psm2_mq_t                      mq;
 	psm2_epaddr_t                  epaddr;
+	const char *                   name;
 } psm2_info_query_arg_t;
 
 /** @brief PSM2 info query
@@ -1757,8 +1783,151 @@ typedef union psm2_info_query_arg
 psm2_error_t psm3_info_query(psm2_info_query_t, void *out,
 			     size_t nargs, psm2_info_query_arg_t []);
 
+/** @brief PSM2 wait
+ *
+ * Function that allows a client wait for PSM3 progress in rcvThread
+ * Used to support interrupt driven progress with CPU release when
+ * >1 process per core
+ *
+ * @param[in] int timeout_ms  timeout in milliseconds.  <0 is infinite timeout
+ *
+ * @returns PSM2_OK if wait completed and some progress may have been made
+ * @returns PSM2_TIMEOUT if wait timeout exceeded with no progress made
+ * @returns PSM2_INTERNAL_ERR if wait mode not allowed for given HAL
+ * @returns PSM2_PARAM_ERR if not allowed for use with current PSM settings/mode
+ */
+psm2_error_t psm3_wait(int timeout_ms);
+
+/** @brief PSM2 env initialization
+ *
+ * Function that parses /etc/psm3.conf, must be called before any
+ * psm3_getenv_* calls
+ */
+int psm3_env_initialize(void);
+
+/** @brief PSM2 env fetch
+ *
+ * Function that fetches parameters from /etc/psm3.conf or env
+ *
+ * @param[in] const char *name parameter name
+ * @retval The parameter fetched
+ * NULL The parameter was not exported
+ */
+char* psm3_env_get(const char *name);
+
+/** @brief PSM2 int parameter parsing
+ *
+ * Function that Parses int parameters
+ *
+ * @param[in] const char *str parameter value
+ * @retval 0 The string was valid, *val has value
+ * -1 The string was empty or NULL, *val not updated
+ * -2 The string had invalid syntax, *val not updated
+ */
+int psm3_parse_str_int(const char *string, int *val, int min, int max);
+
+/** @brief PSM2 unsigned int parameter parsing
+ *
+ * Function that Parses unsigned int parameters
+ *
+ * @param[in] const char *str parameter value
+ * @retval 0 The string was valid, *val has value
+ * -1 The string was empty or NULL, *val not updated
+ * -2 The string had invalid syntax, *val not updated
+ */
+int psm3_parse_str_uint(const char *string, unsigned int *val,
+							unsigned int min, unsigned int max);
+
+/** @brief PSM2 yesno parameter parsing
+ *
+ * Function that parses a string yesno parameter
+ *
+ * @param[in] const char *str parameter value
+ * @retval 0 The string was valid, *val has value
+ * -1 The string was empty or NULL, *val not updated
+ * -2 The string had invalid syntax, *val not updated
+ * @param[out] int *val
+ * 0  The string was No, False, Off or 0
+ * 1  The string was Yes, True, On or 1
+ */
+int psm3_parse_str_yesno(const char *str, int *val);
+
+// for the purposes of psmx3 accessing PSM3_DEVICES config, these
+// interfaces are defined here.  Not for general consumption
+/* We currently have 3 PTLs, 0 is reserved. */
+#define PTL_DEVID_IPS  1	// ips aka nic, network inter-node
+#define PTL_DEVID_AMSH 2	// shm, intra-node, scale-up
+#define PTL_DEVID_SELF 3	// self
+
+/* We can currently initialize up to 3 PTLs */
+#define PTL_MAX_INIT    3
+
+/** @brief PSM2 devices parameter parsing
+ *
+ * Function that gets and parses the PSM3_DEVICES string parameter
+ *
+ * @param[out] array of devices
+ * @retval PSM2_OK - devices successfully returned
+ * other (PSM2_PARAM_ERR) - error parsing devices
+ */
+psm2_error_t psm3_parse_devices(int devices[PTL_MAX_INIT]);
+
+/** @brief PSM2 devices list search
+ *
+ * Function that searches devid_enabled for a specific device
+ *
+ * @param[in] array of devices from psm3_parse_devices
+ * @param[in] devid: PTL_DEVID_IPS, PTL_DEVID_AMSH, or PTL_DEVID_SELF
+ * @retval 1 - given devid is enabled in devices[]
+ * 0  Given devid is disabled in devices[]
+ */
+int psm3_device_is_enabled(const int devices[PTL_MAX_INIT], int devid);
+
+/** @brief PSM2 env finalize
+ *
+ * Function that frees memory from psm3_env_initialize
+ * Only need to call if don't call psm3_finalize
+ */
+void psm3_env_finalize(void);
+
+/** @brief PSM2 env query
+ *
+ * Function that allows the wrapper provider to get PSM env variables
+ * including checks of the /etc/psm3.conf file.  Purposely structured
+ * similar to fi_param_get_* for easier use in psmx3 wrapper
+ *
+ * @param[in] const char *name parameter name
+ * @param[in] const char *descr parameter description (help text)
+ * @param[in] int visible  is parameter user visible (1) or hidden (0)
+ * @param[out] value  where to put value, also holds default value on input
+ *
+ * @retval 0 The value buffer has successfully been written with the
+ * value for the parameter
+ * 		   1 The default value was used, *value was unchanged
+ */
+int psm3_getenv_int(const char *name, const char *descr, int visible,
+				int *value);
+int psm3_getenv_bool(const char *name, const char *descr, int visible,
+				int *value);
+int psm3_getenv_str(const char *name, const char *descr, int visible,
+				char **value);
+
+/** @brief PSM3 memory copy
+ *
+ * Function that copies data from source to destination. Either dest or
+ * src could be GPU buffer, but not both.
+ *
+ * @param[out] void *dest parameter destination buffer pointer
+ * @param[in] const void *src parameter source buffer pointer
+ * @param[in] unint32_t parameter copy length
+ */
+void psm3_memcpy(void *dest, const void *src, uint32_t len);
+void psm3_ep_memcpy(psm2_ep_t ep, void *dest, const void *src, uint32_t len);
+
 /*! @} */
 
+void psm3_turn_off_init_cache();
+
 #ifdef __cplusplus
 } /* extern "C" */
 #endif
