@@ -59,13 +59,13 @@
 #include "ips_config.h"
 #include "psm_user.h"
 
-#include "ips_tid.h"
 #include "ips_recvhdrq.h"
 #include "ips_epstate.h"
 #include "ips_proto_am.h"
 #include "ips_tidflow.h"
 #include "ips_path_rec.h"
 
+#if defined(PSM_SOCKETS) && defined(USE_UDP)
 // when defined, this enables use of byte based flow credits in addition
 // to packet based.
 // It can help UDP to avoid overflowing the sockets kernel buffers.
@@ -73,6 +73,7 @@
 // memory at scale.
 // UD/RC, TCP and OPA HALs self configure so this has no effect
 #define PSM_BYTE_FLOW_CREDITS
+#endif
 
 typedef enum ips_path_type {
 	IPS_PATH_LOW_PRIORITY,
@@ -93,8 +94,8 @@ struct ips_epinfo {
 	uint8_t ep_lmc;
 	enum psm3_ibv_rate ep_link_rate;
 	uint16_t ep_sl;		/* PSM3_NIC_SL only when path record not used */
-	uint32_t ep_mtu;	// PSM payload after potential hdr & PSM3_MTU decrease
-				// or TCP increase beyond wire size
+	uint32_t ep_mtu;	// PSM payload after potential hdr & PSM3_MTU adjustment
+	                	// for TCP and RC it can be beyond wire size
 	uint16_t ep_pkey;	/* PSM3_PKEY only when path record not used */
 	uint64_t ep_timeout_ack;	/* PSM3_ERRCHK_TIMEOUT if no path record */
 	uint64_t ep_timeout_ack_max;
@@ -134,9 +135,11 @@ psm2_error_t psm3_ips_proto_init(psm2_ep_t ep,
 			    void *spioc,	                  /* PTL's opaque spio control */
 			    struct ips_proto *proto);	          /* output protocol */
 
-psm2_error_t psm3_ips_proto_fini(struct ips_proto *proto, int force,
+psm2_error_t psm3_ips_proto_disconnect_all(struct ips_proto *proto, int force,
 			   uint64_t timeout);
 
+psm2_error_t psm3_ips_proto_fini(struct ips_proto *proto);
+
 /*
  * Control message structures
  *
@@ -244,7 +247,8 @@ struct ips_proto_stats {
 	uint64_t stray_packets;
 	uint64_t rcv_revisit;
 #ifdef PSM_SOCKETS
-	uint64_t partial_write_cnt;
+	uint64_t partial_data_write_cnt;
+	uint64_t partial_ctr_write_cnt;
 	uint64_t partial_read_cnt;
 	uint64_t rcv_hol_blocking;
 #endif
@@ -309,19 +313,22 @@ struct ips_ibta_compliance_fn {
 
 /* please don't change the flow id order */
 typedef enum ips_epaddr_flow {
-	EP_FLOW_GO_BACK_N_PIO,
-	EP_FLOW_TIDFLOW,	/* Can either pio or dma for tidflow */
+	EP_FLOW_GO_BACK_N_PIO = 0,
 	EP_FLOW_LAST		/* Keep this the last endpoint flow */
 } ips_epaddr_flow_t;
+// number of entries in ips_epaddr.ips_flow[], use this only when
+// sizing array and testing index.  When testing flowid in other
+// places use EP_FLOW_LAST as there could later be special flow types which
+// don't need an entry in ips_flow[]
+#define EP_NUM_FLOW_ENTRIES (EP_FLOW_LAST)
 
 typedef enum psm_transfer_type {
-	PSM_TRANSFER_PIO,
+	PSM_TRANSFER_PIO = 0,
 	PSM_TRANSFER_LAST	/* Keep this the last transfer type */
 } psm_transfer_type_t;
 
 typedef enum psm_protocol_type {
-	PSM_PROTOCOL_GO_BACK_N,
-	PSM_PROTOCOL_TIDFLOW,
+	PSM_PROTOCOL_GO_BACK_N = 0,
 	PSM_PROTOCOL_LAST	/* Keep this the last protocol type */
 } psm_protocol_type_t;
 
@@ -349,7 +356,7 @@ struct ips_proto {
 	uint32_t iovec_thresh_eager_blocking;
 #endif
 #ifdef PSM_HAVE_REG_MR
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+#ifdef PSM_HAVE_GPU
 	uint32_t iovec_gpu_thresh_eager;
 	uint32_t iovec_gpu_thresh_eager_blocking;
 #endif
@@ -362,6 +369,10 @@ struct ips_proto {
 #ifdef PSM_BYTE_FLOW_CREDITS
 	uint32_t flow_credit_bytes;	// credit limit in bytes
 #endif
+	uint16_t min_credits;		// min credits
+	uint16_t max_credits;		// max credits
+	uint16_t credits_adjust;	// credit adjusting amount
+	uint16_t credits_inc_thresh;	// credit increase threshold
 	mpool_t pend_sends_pool;
 	struct ips_ibta_compliance_fn ibta;
 	struct ips_proto_stats stats;
@@ -420,21 +431,13 @@ struct ips_proto {
 	void *opp_ctxt;
 	struct opp_api opp_fn;
 
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
-	struct ips_gpu_hostbuf_mpool_cb_context cuda_hostbuf_send_cfg;
-	struct ips_gpu_hostbuf_mpool_cb_context cuda_hostbuf_small_send_cfg;
-	mpool_t cuda_hostbuf_pool_send;
-	mpool_t cuda_hostbuf_pool_small_send;
-#endif
-
-#ifdef PSM_CUDA
-	CUstream cudastream_send;
-#elif defined(PSM_ONEAPI)
-	ze_command_queue_handle_t cq_send;
-#endif
-
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
-	unsigned cuda_prefetch_limit;
+#ifdef PSM_HAVE_GPU
+	struct ips_gpu_hostbuf_mpool_cb_context gpu_hostbuf_send_cfg;
+	struct ips_gpu_hostbuf_mpool_cb_context gpu_hostbuf_small_send_cfg;
+	mpool_t gpu_hostbuf_pool_send;
+	mpool_t gpu_hostbuf_pool_small_send;
+	union ips_proto_gpu_specific gpu_specific;
+	unsigned gpu_prefetch_limit;
 #endif
 /*
  * Control message queue for pending messages.
@@ -502,8 +505,6 @@ struct ips_flow {
 	uint16_t protocol:3;	/* go-back-n or tidflow */
 	uint16_t flags:8;	/* flow state flags */
 
-	// TBD - cwin only needed for OPA for CCA
-	uint16_t cwin;		/* Size of congestion window in packets */
 	// to allow for good pipelining of send/ACK need to trigger an ack at
 	// least every ack_interval packets (roughy flow_credits/4) or every
 	// ack_inteval_bytes bytes (roughly flow_credit_bytes/4) whichever
@@ -520,18 +521,23 @@ struct ips_flow {
 #ifdef PSM_SOCKETS
 	uint32_t used_snd_buff; // number of bytes in socket send buffer
 	uint32_t send_remaining; // the length of remaining data to send
+	uint8_t* partial_conn_msg; // partial sent connection msg
+	uint16_t partial_conn_msg_size; // size of the partial_conn_msg buffer
+	uint16_t conn_msg_remainder; // the length of remaining connection msg to send
 #endif
 	// We track credits in terms of packets and bytes.
 	// For UD and OPA, packets is sufficient since per pkt buffering.
 	// For UDP, sockets has byte oriented buffering so we need to
 	// impose a credit_bytes limit to allow sufficient pkt credits
 	// but avoid sockets buffer overflow and recv side discards/flow control
-	int16_t  credits;	/* Current credits available to send on flow */
+	int16_t  credits;	 /* Current credits available to send on flow */
+	uint16_t max_credits;	 /* credits limit */
+	uint32_t credits_inc_psn; /* the reference pkt psn used for increasing credit. We increase */
+	                         /* credit if current psn - credits_inc_psn > credit_inc_thresh */
 #ifdef PSM_BYTE_FLOW_CREDITS
 	int32_t  credit_bytes;	/* Current credit bytes avail to send on flow */
 #endif
 	uint32_t ack_index;     /* Index of the last ACK message type in pending message queue */
-
 	psmi_seqnum_t xmit_seq_num;	/* next psn for xmit */
 	psmi_seqnum_t xmit_ack_num;	/* last xmited psn acked + 1 */
 	psmi_seqnum_t recv_seq_num;	/* next psn expect to recv */
@@ -559,7 +565,7 @@ struct ips_epaddr {
 
 	struct ips_epaddr *next;	/* linklist */
 
-	struct ips_flow flows[EP_FLOW_LAST - 1];	/* pio and dma */
+	struct ips_flow flows[EP_NUM_FLOW_ENTRIES];	/* pio and dma */
 	ips_path_grp_t *pathgrp;	/* pointer to slid/dlid group in hash */
 
 	uint32_t connidx_outgoing;	/* peer's connection idx */
@@ -641,6 +647,21 @@ struct ips_epaddr {
 			uint32_t use_max_inline_data;
 
 			uint8_t rc_connected;
+
+			// MR for flow->recv_seq_num
+			struct ibv_mr *recv_seq_mr;
+			// remote flow->recv_seq_num addr and rkey
+			uint64_t remote_recv_seq_addr;
+			uint32_t remote_recv_seq_rkey;
+			// psn num of remote flow->recv_seq_num
+			uint32_t remote_recv_psn;
+			// MR for remote flow->recv_seq_num storage
+			struct ibv_mr *remote_recv_psn_mr;
+			// indicare whether we have outstanding RDMA Read for
+			// remote flow->recv_seq_num
+			uint8_t remote_seq_outstanding;
+			// congestion control count
+			uint16_t cc_count;
 #endif /* USE_RC */
 		} verbs;
 #endif /* PSM_VERBS */
@@ -649,8 +670,8 @@ struct ips_epaddr {
 			// for UDP the pri_addr is a UDP socket
 			// for TCP the pri_addr is a TCP socket and the
 			// aux_addr is a UDP socket used only for disconnect
-			struct sockaddr_in6 remote_pri_addr;
-			struct sockaddr_in6 remote_aux_addr;
+			psm3_sockaddr_in_t remote_pri_addr;
+			psm3_sockaddr_in_t remote_aux_addr;
 			int tcp_fd;
 			uint8_t connected;
 		} sockets;
@@ -823,9 +844,9 @@ MOCK_DCL_EPILOGUE(psm3_ips_ibta_init);
 psm2_error_t psm3_ips_ibta_fini(struct ips_proto *proto);
 
 
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
+#ifdef PSM_HAVE_GPU
 PSMI_ALWAYS_INLINE(
-uint32_t ips_cuda_next_window(uint32_t max_window, uint32_t offset,
+uint32_t ips_gpu_next_window(uint32_t max_window, uint32_t offset,
 			      uint32_t len))
 {
 	uint32_t window_len;
