@@ -71,6 +71,7 @@ extern "C" {
 #include "ofi_list.h"
 #include "ofi_util.h"
 #include "ofi_mem.h"
+#include "ofi_hmem.h"
 #include "rbtree.h"
 #include "version.h"
 #include "psm_config.h"
@@ -83,6 +84,24 @@ extern "C" {
 #define STATIC static
 #endif
 
+/* wrapper for logging macros so we can add our process label */
+#define PSMX3_INFO(prov, subsys, format, ...) \
+		FI_INFO(prov, subsys, "%s: " format, psm3_get_mylabel(), ##__VA_ARGS__)
+#define PSMX3_DBG_TRACE(prov, subsys, format, ...) \
+		FI_DBG_TRACE(prov, subsys, "%s: " format, psm3_get_mylabel(), ##__VA_ARGS__)
+#define PSMX3_TRACE(prov, subsys, format, ...) \
+		FI_TRACE(prov, subsys, "%s: " format, psm3_get_mylabel(), ##__VA_ARGS__)
+#define PSMX3_WARN(prov, subsys, format, ...) \
+		FI_WARN(prov, subsys, "%s: " format, psm3_get_mylabel(), ##__VA_ARGS__)
+#define PSMX3_WARN_SPARSE(prov, subsys, format, ...) \
+		FI_WARN_SPARSE(prov, subsys, "%s: " format, psm3_get_mylabel(), ##__VA_ARGS__)
+#define PSMX3_WARN_ONCE(prov, subsys, format, ...) \
+		FI_WARN_ONCE(prov, subsys, "%s: " format, psm3_get_mylabel(), ##__VA_ARGS__)
+#define PSMX3_DBG(prov, subsys, func, line, format, ...) \
+		FI_DBG(prov, subsys, "%s: " format, psm3_get_mylabel(), ##__VA_ARGS__)
+#define psmx3_log(prov, level, subsys, func, line, format, ...) \
+		fi_log(prov, level, subsys, func, line, "%s: " format, psm3_get_mylabel(), ##__VA_ARGS__)
+
 extern struct fi_provider psmx3_prov;
 
 #define PSMX3_OP_FLAGS	(FI_INJECT | FI_MULTI_RECV | FI_COMPLETION | \
@@ -110,7 +129,6 @@ extern struct fi_provider psmx3_prov;
 #define PSMX3_RMA_ORDER_SIZE	(4096)
 #define PSMX3_MSG_ORDER		(FI_ORDER_SAS | OFI_ORDER_RAR_SET | OFI_ORDER_RAW_SET | \
 				 OFI_ORDER_WAR_SET | OFI_ORDER_WAW_SET)
-#define PSMX3_COMP_ORDER	FI_ORDER_NONE
 
 /*
  * Four bits are reserved from the 64-bit tag space as a flags to identify the
@@ -539,7 +557,7 @@ struct psmx3_fid_domain {
 	uint64_t		mode;
 	uint64_t		caps;
 
-	enum fi_mr_mode		mr_mode;
+	int			mr_mode;
 	ofi_spin_t		mr_lock;
 	uint64_t		mr_reserved_key;
 	RbtHandle		mr_map;
@@ -834,12 +852,14 @@ struct psmx3_env {
 #if (PSMX3_TAG_LAYOUT == PSMX3_TAG_LAYOUT_RUNTIME)
 	char	*tag_layout;
 #endif
+	int	yield_mode;
+	int	wait_enable;
 };
 
 #define PSMX3_MAX_UNITS	PSMI_MAX_RAILS /* from psm_config.h */
+#define PSMX3_MAX_EPS	64 /* no real limit, used to report max_trx_ctxt */
 struct psmx3_domain_info {
 	int max_trx_ctxt;
-	int free_trx_ctxt;
 	int num_units;	/* total HW units found by PSM3 */
 	int num_reported_units;	/* num entries in arrays below */
 	int num_active_units;	/* total active found, >= num_reported_units */
@@ -847,8 +867,6 @@ struct psmx3_domain_info {
 	int unit_is_active[PSMX3_MAX_UNITS];
 	int unit_id[PSMX3_MAX_UNITS];	/* PSM3 unit_id */
 	int addr_index[PSMX3_MAX_UNITS];/* PSM3 address index within unit_id */
-	int unit_nctxts[PSMX3_MAX_UNITS];
-	int unit_nfreectxts[PSMX3_MAX_UNITS];
 	char default_domain_name[PSMX3_MAX_UNITS * NAME_MAX]; /* autoselect:irdma0;irdma1;..... */
 	char default_fabric_name[PSMX3_MAX_UNITS * NAME_MAX]; /* RoCE 192.168.101.0/24;RoCE 192.168.102.0/24;.... */
 };
@@ -880,6 +898,16 @@ extern struct psmx3_env		psmx3_env;
 extern struct psmx3_domain_info	psmx3_domain_info;
 extern struct psmx3_fid_fabric	*psmx3_active_fabric;
 
+int psmx3_param_get_bool(struct fi_provider *provider, const char *env_var_name,
+					const char *descr, int visible, int *value);
+int psmx3_param_get_int(struct fi_provider *provider, const char *env_var_name,
+					const char *descr, int visible, int *value);
+int psmx3_param_get_str(struct fi_provider *provider, const char *env_var_name,
+					const char *descr, int visible, char **value);
+
+/* indicate if lock_level variable was set (1 if set, 0 if defaulted) */
+int psmx3_check_lock_level(void);
+
 /*
  * Lock levels:
  *     0 -- always lock
@@ -1018,7 +1046,6 @@ char	*psmx3_uuid_to_string(psm2_uuid_t uuid);
 void	*psmx3_ep_name_to_string(const struct psmx3_ep_name *name, size_t *len);
 struct	psmx3_ep_name *psmx3_string_to_ep_name(const void *s);
 int	psmx3_errno(int err);
-void	psmx3_query_mpi(void);
 
 void	psmx3_cq_enqueue_event(struct psmx3_fid_cq *cq, struct psmx3_cq_event *event);
 struct	psmx3_cq_event *psmx3_cq_create_event(struct psmx3_fid_cq *cq,
