@@ -50,7 +50,6 @@ static struct fi_tx_attr psmx3_tx_attr = {
 	.mode			= FI_CONTEXT, /* 0 */
 	.op_flags		= PSMX3_OP_FLAGS,
 	.msg_order		= PSMX3_MSG_ORDER,
-	.comp_order		= PSMX3_COMP_ORDER,
 	.inject_size		= 64, /* psmx3_env.inject_size */
 	.size			= UINT64_MAX,
 	.iov_limit		= PSMX3_IOV_MAX_COUNT,
@@ -62,8 +61,6 @@ static struct fi_rx_attr psmx3_rx_attr = {
 	.mode			= FI_CONTEXT, /* 0 */
 	.op_flags		= PSMX3_OP_FLAGS,
 	.msg_order		= PSMX3_MSG_ORDER,
-	.comp_order		= PSMX3_COMP_ORDER,
-	.total_buffered_recv	= UINT64_MAX,
 	.size			= UINT64_MAX,
 	.iov_limit		= 1,
 };
@@ -92,13 +89,13 @@ static struct fi_domain_attr psmx3_domain_attr = {
 	.data_progress		= FI_PROGRESS_AUTO,
 	.resource_mgmt		= FI_RM_ENABLED,
 	.av_type		= FI_AV_UNSPEC,
-	.mr_mode		= FI_MR_SCALABLE | FI_MR_BASIC,
+	.mr_mode		= OFI_MR_SCALABLE | OFI_MR_BASIC,
 	.mr_key_size		= sizeof(uint64_t),
 	.cq_data_size		= 0, /* 4, 8 */
 	.cq_cnt			= 65535,
 	.ep_cnt			= 65535,
-	.tx_ctx_cnt		= 1, /* psmx3_domain_info.free_trx_ctxt */
-	.rx_ctx_cnt		= 1, /* psmx3_domain_info.free_trx_ctxt */
+	.tx_ctx_cnt		= 1, /* psmx3_domain_info.max_trx_ctxt */
+	.rx_ctx_cnt		= 1, /* psmx3_domain_info.max_trx_ctxt */
 	.max_ep_tx_ctx		= 1, /* psmx3_domain_info.max_trx_ctxt */
 	.max_ep_rx_ctx		= 1, /* psmx3_domain_info.max_trx_ctxt */
 	.max_ep_stx_ctx		= 1, /* psmx3_domain_info.max_trx_ctxt */
@@ -265,39 +262,94 @@ fail:
 #define psmx3_dupinfo fi_dupinfo
 #endif /* HAVE_PSM3_DL */
 
-#if defined(PSM_CUDA) || defined(PSM_ONEAPI)
-/* mimic parsing functionality of psm3_getenv */
-static long get_psm3_env(const char *var, int default_value) {
-	char *ep;
-	long val;
-	char *e = getenv(var);
-
-	if (!e || !*e)
-		return default_value; /* no value supplied */
-
-	val = strtol(e, &ep, 10);
-	if (!ep ||  *ep) { /* parse error - didn't consume all */
-		val = strtol(e, &ep, 16); /* try hex */
-		if (!ep ||  *ep)
-			return default_value;
-	}
-	return val;
-}
-#endif // defined(PSM_CUDA) || defined(PSM_ONEAPI)
-
 static uint64_t psmx3_check_fi_hmem_cap(void) {
+#ifdef PSM_HAVE_GPU
+	/* if parses as empty or invalid use default of 0 */
+	/* psm3 below us will provide warning as needed when it parses it */
+	int gpu = 0;
+	unsigned int gpudirect = 0;
 #ifdef PSM_CUDA
-	if ((get_psm3_env("PSM3_CUDA", 0) || get_psm3_env("PSM3_GPUDIRECT", 0)) &&
-		!ofi_hmem_p2p_disabled())
-		return FI_HMEM;
+	(void)psm3_parse_str_int(psm3_env_get("PSM3_CUDA"), &gpu, INT_MIN, INT_MAX);
+#else /* PSM_ONEAPI */
+	(void)psm3_parse_str_int(psm3_env_get("PSM3_ONEAPI_ZE"), &gpu,
+							INT_MIN, INT_MAX);
 #endif
-#ifdef PSM_ONEAPI
-	if ((get_psm3_env("PSM3_ONEAPI_ZE", 0) || get_psm3_env("PSM3_GPUDIRECT", 0)) &&
-		!ofi_hmem_p2p_disabled()) {
+	(void)psm3_parse_str_uint(psm3_env_get("PSM3_GPUDIRECT"), &gpudirect,
+							0, UINT_MAX);
+	if ((gpu || gpudirect) && !ofi_hmem_p2p_disabled())
 		return FI_HMEM;
+#endif /* PSM_HAVE_GPU */
+	return 0;
+}
+
+static uint64_t get_max_inject_size(void) {
+	unsigned int thresh_rv;
+	unsigned int temp;
+	int have_shm = 1;
+	int have_nic = 1;
+	int devid_enabled[PTL_MAX_INIT];
+
+	// check PSM3_DEVICES to determine if PSM3 shm enabled
+	if ((PSM2_OK == psm3_parse_devices(devid_enabled))) {
+		have_shm = psm3_device_is_enabled(devid_enabled, PTL_DEVID_AMSH);
+		have_nic = psm3_device_is_enabled(devid_enabled, PTL_DEVID_IPS);
+	}
+
+	// figure out the smallest rendezvous threshold (GPU vs CPU ips vs shm)
+	// If middleware above is not using PSM3 for shm but leaves it in
+	// PSM3_DEVICES, this could be more restrictive than necessary,
+	// but it's safe.  Note that PSM3_DEVICES can't be set per EP open.
+	// Also not yet sure which HAL will be selected so must pick most
+	// conservative ips (NIC) config
+	thresh_rv = 65536;	// default in odd case of PSM3_DEVICES=self
+
+	if (have_nic) {
+		temp = PSM3_MQ_RNDV_NIC_THRESH;
+		psm3_parse_str_uint(psm3_env_get("PSM3_MQ_RNDV_NIC_THRESH"), &temp,
+							0, UINT_MAX);
+		if (thresh_rv > temp)
+			thresh_rv = temp;
+	}
+
+	if (have_shm) {
+		temp = PSM3_MQ_RNDV_SHM_THRESH;
+		psm3_parse_str_uint(psm3_env_get("PSM3_MQ_RNDV_SHM_THRESH"), &temp,
+							0, UINT_MAX);
+		if (thresh_rv > temp)
+			thresh_rv = temp;
+	}
+
+#ifdef PSM_HAVE_GPU
+	if (psmx3_prov_info.caps & FI_HMEM) {
+		if (have_nic) {
+			// GPU ips rendezvous threshold
+			uint32_t out;
+			if (psm3_info_query(PSM2_INFO_QUERY_GPU_THRESH_RNDV, &out, 0, NULL)) {
+				PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
+					"Unable to get PSM3_GPU_THRESH_RNDV.\n");
+			} else if (thresh_rv > out) {
+				thresh_rv = out;
+			}
+		}
+
+		if (have_shm) {
+			// GPU shm rendezvous threshold
+			// we only have default, real value may be overriden at MQ init
+			// when open PSM3 endpoint
+			uint32_t out;
+			if (psm3_info_query(PSM2_INFO_QUERY_MQ_RNDV_SHM_GPU_THRESH_DEFAULT, &out, 0, NULL)) {
+				PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
+					"Unable to get PSM3_MQ_RNDV_SHM_GPU_THRESH default.\n");
+			} else if (thresh_rv > out) {
+				thresh_rv = out;
+			}
+		}
 	}
 #endif
-	return 0;
+
+	// messages <= thresh_rv guaranteed to use eager, so thresh_rv
+	// is the max allowed inject_size.
+	return thresh_rv;
 }
 
 /*
@@ -335,6 +387,14 @@ int psmx3_init_prov_info(const struct fi_info *hints, struct fi_info **info)
 	int ep_type2 = FI_EP_DGRAM;
 	uint64_t extra_caps = 0;
 
+	/* Check if GPU is enabled */
+	extra_caps |= psmx3_check_fi_hmem_cap();
+
+	prov_info->caps |= extra_caps;
+	prov_info->tx_attr->caps |= extra_caps;
+	prov_info->rx_attr->caps |= extra_caps;
+	prov_info->domain_attr->caps |= extra_caps;
+
 	if (!hints)
 		goto alloc_info;
 
@@ -347,13 +407,13 @@ int psmx3_init_prov_info(const struct fi_info *hints, struct fi_info **info)
 			ep_type = FI_EP_DGRAM;
 			break;
 		default:
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"Unsupported endpoint type\n");
-			FI_INFO(&psmx3_prov, FI_LOG_CORE, "Supported: %s\n",
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Supported: %s\n",
 				fi_tostr(&ep_type, FI_TYPE_EP_TYPE));
-			FI_INFO(&psmx3_prov, FI_LOG_CORE, "Supported: %s\n",
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Supported: %s\n",
 				fi_tostr(&ep_type2, FI_TYPE_EP_TYPE));
-			FI_INFO(&psmx3_prov, FI_LOG_CORE, "Requested: %s\n",
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Requested: %s\n",
 				fi_tostr(&hints->ep_attr->type, FI_TYPE_EP_TYPE));
 			return -FI_ENODATA;
 		}
@@ -367,27 +427,19 @@ int psmx3_init_prov_info(const struct fi_info *hints, struct fi_info **info)
 			addr_format = FI_ADDR_STR;
 			break;
 		default:
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"Unsupported address format\n");
-			FI_INFO(&psmx3_prov, FI_LOG_CORE, "Supported: %s\n",
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Supported: %s\n",
 				fi_tostr(&addr_format, FI_TYPE_ADDR_FORMAT));
-			FI_INFO(&psmx3_prov, FI_LOG_CORE, "Supported: %s\n",
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Supported: %s\n",
 				fi_tostr(&addr_format2, FI_TYPE_ADDR_FORMAT));
-			FI_INFO(&psmx3_prov, FI_LOG_CORE, "Requested: %s\n",
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Requested: %s\n",
 				fi_tostr(&hints->addr_format, FI_TYPE_ADDR_FORMAT));
 			return -FI_ENODATA;
 	}
 
-	/* Check if CUDA is enable */
-	extra_caps |= psmx3_check_fi_hmem_cap();
-
-	prov_info->caps |= extra_caps;
-	prov_info->tx_attr->caps |= extra_caps;
-	prov_info->rx_attr->caps |= extra_caps;
-	prov_info->domain_attr->caps |= extra_caps;
-
 	if ((hints->caps & prov_info->caps) != hints->caps) {
-		FI_INFO(&psmx3_prov, FI_LOG_CORE, "caps not supported\n");
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "caps not supported\n");
 		OFI_INFO_CHECK(&psmx3_prov, prov_info, hints, caps, FI_TYPE_CAPS);
 		return -FI_ENODATA;
 	}
@@ -409,7 +461,7 @@ alloc_info:
 			info_new->rx_attr->mode = 0;
 			info_new->domain_attr->cq_data_size = 8;
 			info_out = info_new;
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"RMA only instance included\n");
 		}
 	}
@@ -423,7 +475,7 @@ alloc_info:
 		info_new->domain_attr->cq_data_size = 4;
 		info_new->next = info_out;
 		info_out = info_new;
-		FI_INFO(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 			"TAG60 instance included\n");
 	}
 
@@ -436,7 +488,7 @@ alloc_info:
 			info_new->ep_attr->type = ep_type;
 			info_new->next = info_out;
 			info_out = info_new;
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"TAG64 instance included\n");
 		}
 	}
@@ -488,7 +540,7 @@ static void psmx3_expand_default_unit(struct fi_info *info)
 					}
 					p->next = psmx3_dupinfo(p);
 					if (!p->next) {
-						FI_WARN(&psmx3_prov, FI_LOG_CORE,
+						PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 							"Failed to duplicate info for HFI unit %d\n",
 							psmx3_domain_info.active_units[i]);
 						break;
@@ -513,6 +565,8 @@ void psmx3_update_prov_info(struct fi_info *info,
 			    struct psmx3_ep_name *dest_addr)
 {
 	struct fi_info *p;
+	unsigned int max_inject_size;
+	unsigned int inject_size;
 
 	for (p = info; p; p = p->next) {
 		psmx3_dup_addr(p->addr_format, src_addr,
@@ -523,6 +577,15 @@ void psmx3_update_prov_info(struct fi_info *info,
 
 	psmx3_expand_default_unit(info);
 
+	max_inject_size = get_max_inject_size();
+	if (psmx3_env.inject_size > max_inject_size)
+		inject_size = max_inject_size;
+	else
+		inject_size = psmx3_env.inject_size;
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
+		"Using inject_size=%u based on FI_PSM3_INJECT_SIZE=%u with max %u\n",
+		inject_size, psmx3_env.inject_size, max_inject_size);
+
 	for (p = info; p; p = p->next) {
 		int unit = ((struct psmx3_ep_name *)p->src_addr)->unit;
 		int port = ((struct psmx3_ep_name *)p->src_addr)->port;
@@ -532,19 +595,11 @@ void psmx3_update_prov_info(struct fi_info *info,
 		    ! psmx3_domain_info.default_domain_name[0])
 			unit = 0;
 
-		if (unit == PSMX3_DEFAULT_UNIT || !psmx3_env.multi_ep) {
-			p->domain_attr->tx_ctx_cnt = psmx3_domain_info.free_trx_ctxt;
-			p->domain_attr->rx_ctx_cnt = psmx3_domain_info.free_trx_ctxt;
-			p->domain_attr->max_ep_tx_ctx = psmx3_domain_info.max_trx_ctxt;
-			p->domain_attr->max_ep_rx_ctx = psmx3_domain_info.max_trx_ctxt;
-			p->domain_attr->max_ep_stx_ctx = psmx3_domain_info.max_trx_ctxt;
-		} else {
-			p->domain_attr->tx_ctx_cnt = psmx3_domain_info.unit_nfreectxts[unit];
-			p->domain_attr->rx_ctx_cnt = psmx3_domain_info.unit_nfreectxts[unit];
-			p->domain_attr->max_ep_tx_ctx = psmx3_domain_info.unit_nctxts[unit];
-			p->domain_attr->max_ep_rx_ctx = psmx3_domain_info.unit_nctxts[unit];
-			p->domain_attr->max_ep_stx_ctx = psmx3_domain_info.unit_nctxts[unit];
-		}
+		p->domain_attr->tx_ctx_cnt = psmx3_domain_info.max_trx_ctxt;
+		p->domain_attr->rx_ctx_cnt = psmx3_domain_info.max_trx_ctxt;
+		p->domain_attr->max_ep_tx_ctx = psmx3_domain_info.max_trx_ctxt;
+		p->domain_attr->max_ep_rx_ctx = psmx3_domain_info.max_trx_ctxt;
+		p->domain_attr->max_ep_stx_ctx = psmx3_domain_info.max_trx_ctxt;
 
 		free(p->domain_attr->name);
 		if (unit == PSMX3_DEFAULT_UNIT)
@@ -556,22 +611,22 @@ void psmx3_update_prov_info(struct fi_info *info,
 			int addr_index = psmx3_domain_info.addr_index[unit];
 
 			args[0].unit = unit_id;
-			args[1].port = port;
+			args[1].port = port == PSMX3_DEFAULT_PORT ? 1 : port;
 			args[2].addr_index = addr_index;
 			args[3].length = sizeof(unit_name);
 
 			if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_ADDR_NAME,
 				unit_name, 4, args)) {
-				FI_WARN(&psmx3_prov, FI_LOG_CORE,
+				PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 					"Failed to read domain name for NIC unit %d (id %d, port %d, index %d)\n",
 					unit, unit_id, port, addr_index);
 				if (asprintf(&p->domain_attr->name, "UNKNOWN") < 0) {
-					FI_WARN(&psmx3_prov, FI_LOG_CORE,
+					PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 						"Failed to allocate memory for domain name for NIC unit %d\n", unit);
 				}
 			} else {
 				if (asprintf(&p->domain_attr->name, "%s", unit_name) <0) {
-					FI_WARN(&psmx3_prov, FI_LOG_CORE,
+					PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 						"Failed to allocate memory for domain name for NIC unit %d "
 						"(id %d, port %d, index %d)\n",
 						unit, unit_id, port, addr_index);
@@ -588,27 +643,27 @@ void psmx3_update_prov_info(struct fi_info *info,
 			int addr_index = psmx3_domain_info.addr_index[unit];
 
 			args[0].unit = unit_id;
-			args[1].port = port;
+			args[1].port = port == PSMX3_DEFAULT_PORT ? 1 : port;
 			args[2].addr_index = addr_index;
 			args[3].length = sizeof(fabric_name);
 
 			if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_SUBNET_NAME,
 				fabric_name, 4, args)) {
-				FI_WARN(&psmx3_prov, FI_LOG_CORE,
+				PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 					"Failed to read unit fabric name for NIC unit_id %d port %d addr %d\n", unit_id, port, addr_index);
 				if (asprintf(&p->fabric_attr->name, "UNKNOWN") < 0) {
-					FI_WARN(&psmx3_prov, FI_LOG_CORE,
+					PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 						"Failed to allocate memory for unit fabric name for NIC unit %d\n", unit);
 				}
 			} else {
 				if (asprintf(&p->fabric_attr->name, "%s", fabric_name) <0) {
-					FI_WARN(&psmx3_prov, FI_LOG_CORE,
+					PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 						"Failed to allocate memory for unit fabric name for NIC unit %d port %d addr %d\n", unit, port, addr_index);
 				}
 			}
 		}
 
-		p->tx_attr->inject_size = psmx3_env.inject_size;
+		p->tx_attr->inject_size = inject_size;
 	}
 }
 
@@ -624,13 +679,13 @@ static int psmx3_check_info(const struct util_prov *util_prov,
 	/* some revs of ofi_check_info fail to check names, so do it here */
 	if (hints && hints->domain_attr && hints->domain_attr->name &&
 	    strcasecmp(hints->domain_attr->name, info->domain_attr->name)) {
-		FI_INFO(&psmx3_prov, FI_LOG_CORE, "skipping device %s (want %s)\n",
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "skipping device %s (want %s)\n",
 			info->domain_attr->name, hints->domain_attr->name);
 		return -FI_ENODATA;
 	}
 	if (hints && hints->fabric_attr && hints->fabric_attr->name &&
 	    strcasecmp(hints->fabric_attr->name, info->fabric_attr->name)) {
-		FI_INFO(&psmx3_prov, FI_LOG_CORE, "skipping fabric %s (want %s)\n",
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "skipping fabric %s (want %s)\n",
 			info->fabric_attr->name, hints->fabric_attr->name);
 		return -FI_ENODATA;
 	}
@@ -690,8 +745,8 @@ void psmx3_alter_prov_info(uint32_t api_version,
 			info->domain_attr->data_progress =
 				FI_PROGRESS_MANUAL;
 
-		if (info->domain_attr->mr_mode == (FI_MR_BASIC | FI_MR_SCALABLE))
-			info->domain_attr->mr_mode = FI_MR_SCALABLE;
+		if (info->domain_attr->mr_mode == (OFI_MR_BASIC | OFI_MR_SCALABLE))
+			info->domain_attr->mr_mode = OFI_MR_SCALABLE;
 
 		/*
 		 * Avoid automatically adding secondary caps that may negatively
@@ -706,7 +761,7 @@ void psmx3_alter_prov_info(uint32_t api_version,
 		cnt++;
 	}
 
-	FI_INFO(&psmx3_prov, FI_LOG_CORE,
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 		"%d instances available, %d with CQ data flag set\n",
 		cnt, cq_data_cnt);
 }
