@@ -75,7 +75,7 @@ static struct psmx3_cq_event *psmx3_cq_alloc_event(struct psmx3_fid_cq *cq)
 	cq->domain->cq_unlock_fn(&cq->lock, 2);
 	event = calloc(1, sizeof(*event));
 	if (!event)
-		FI_WARN(&psmx3_prov, FI_LOG_CQ, "out of memory.\n");
+		PSMX3_WARN(&psmx3_prov, FI_LOG_CQ, "out of memory.\n");
 
 	return event;
 }
@@ -142,7 +142,7 @@ struct psmx3_cq_event *psmx3_cq_create_event(struct psmx3_fid_cq *cq,
 		break;
 
 	default:
-		FI_WARN(&psmx3_prov, FI_LOG_CQ,
+		PSMX3_WARN(&psmx3_prov, FI_LOG_CQ,
 			"unsupported CQ format %d\n", cq->format);
 		psmx3_cq_free_event(cq, event);
 		return NULL;
@@ -311,7 +311,7 @@ static inline int psmx3_cq_any_complete(struct psmx3_fid_cq *poll_cq,
 		break;
 
 	default:
-		FI_WARN(&psmx3_prov, FI_LOG_CQ,
+		PSMX3_WARN(&psmx3_prov, FI_LOG_CQ,
 			"unsupported CQ format %d\n", comp_cq->format);
 		if (event != event_in)
 			psmx3_cq_free_event(comp_cq, event);
@@ -558,7 +558,7 @@ psmx3_mq_status_copy(struct psm2_mq_req_user *req, void *status_array, int entry
 		if (OFI_UNLIKELY(am_req->op == PSMX3_AM_REQ_READV)) {
 			am_req->read.len_read += PSMX3_STATUS_RCVLEN(req);
 			if (am_req->read.len_read < am_req->read.len) {
-				FI_INFO(&psmx3_prov, FI_LOG_EP_DATA,
+				PSMX3_INFO(&psmx3_prov, FI_LOG_EP_DATA,
 					"readv: long protocol finishes early\n");
 				if (PSMX3_STATUS_ERROR(req))
 					am_req->error = psmx3_errno(PSMX3_STATUS_ERROR(req));
@@ -587,7 +587,7 @@ psmx3_mq_status_copy(struct psm2_mq_req_user *req, void *status_array, int entry
 		if (OFI_UNLIKELY(am_req->op == PSMX3_AM_REQ_READV)) {
 			am_req->read.len_read += PSMX3_STATUS_RCVLEN(req);
 			if (am_req->read.len_read < am_req->read.len) {
-				FI_INFO(&psmx3_prov, FI_LOG_EP_DATA,
+				PSMX3_INFO(&psmx3_prov, FI_LOG_EP_DATA,
 					"readv: long protocol finishes early\n");
 				if (PSMX3_STATUS_ERROR(req))
 					am_req->error = psmx3_errno(PSMX3_STATUS_ERROR(req));
@@ -622,8 +622,10 @@ psmx3_mq_status_copy(struct psm2_mq_req_user *req, void *status_array, int entry
 			data = PSMX3_GET_CQDATA(PSMX3_STATUS_TAG(req));
 			if (PSMX3_HAS_IMM(PSMX3_GET_FLAGS(PSMX3_STATUS_TAG(req))))
 				flags |= FI_REMOTE_CQ_DATA;
-			if (multi_recv_req->offset + PSMX3_STATUS_RCVLEN(req) +
-				multi_recv_req->min_buf_size > multi_recv_req->len)
+			len_remaining = multi_recv_req->len - multi_recv_req->offset -
+					PSMX3_STATUS_RCVLEN(req);
+			if (len_remaining < multi_recv_req->min_buf_size ||
+			    len_remaining == 0)
 				flags |= FI_MULTI_RECV;	/* buffer used up */
 			err = psmx3_cq_rx_complete(
 					status_data->poll_cq, ep->recv_cq, ep->av,
@@ -638,7 +640,8 @@ psmx3_mq_status_copy(struct psm2_mq_req_user *req, void *status_array, int entry
 		/* repost multi-recv buffer */
 		multi_recv_req->offset += PSMX3_STATUS_RCVLEN(req);
 		len_remaining = multi_recv_req->len - multi_recv_req->offset;
-		if (len_remaining >= multi_recv_req->min_buf_size) {
+		if (len_remaining >= multi_recv_req->min_buf_size &&
+		    len_remaining > 0) {
 			if (len_remaining > PSMX3_MAX_MSG_SIZE)
 				len_remaining = PSMX3_MAX_MSG_SIZE;
 			err = psm3_mq_irecv2(ep->rx->psm2_mq,
@@ -786,7 +789,8 @@ psmx3_mq_status_copy(struct psm2_mq_req_user *req, void *status_array, int entry
 			multi_recv_req = PSMX3_CTXT_USER(fi_context);
 			multi_recv_req->offset += PSMX3_STATUS_RCVLEN(req);
 			len_remaining = multi_recv_req->len - multi_recv_req->offset;
-			if (len_remaining >= multi_recv_req->min_buf_size) {
+			if (len_remaining >= multi_recv_req->min_buf_size &&
+			    len_remaining > 0) {
 				if (len_remaining > PSMX3_MAX_MSG_SIZE)
 					len_remaining = PSMX3_MAX_MSG_SIZE;
 				err = psm3_mq_irecv2(ep->rx->psm2_mq,
@@ -954,22 +958,16 @@ STATIC ssize_t psmx3_cq_readerr(struct fid_cq *cq, struct fi_cq_err_entry *buf,
 				uint64_t flags)
 {
 	struct psmx3_fid_cq *cq_priv;
-	uint32_t api_version;
-	size_t size;
 
 	cq_priv = container_of(cq, struct psmx3_fid_cq, cq);
 
 	cq_priv->domain->cq_lock_fn(&cq_priv->lock, 2);
 	if (cq_priv->pending_error) {
-		api_version = cq_priv->domain->fabric->util_fabric.
-			      fabric_fid.api_version;
-		size = FI_VERSION_GE(api_version, FI_VERSION(1, 5)) ?
-			sizeof(*buf) : sizeof(struct fi_cq_err_entry_1_0);
-
-		memcpy(buf, &cq_priv->pending_error->cqe, size);
+		ofi_cq_err_memcpy(cq_priv->domain->fabric->util_fabric.fabric_fid.api_version,
+				  buf, &cq_priv->pending_error->cqe.err);
 		free(cq_priv->pending_error);
 		cq_priv->pending_error = NULL;
-		psmx3_unlock(&cq_priv->lock, 2);
+		cq_priv->domain->cq_unlock_fn(&cq_priv->lock, 2);
 		return 1;
 	}
 	cq_priv->domain->cq_unlock_fn(&cq_priv->lock, 2);
@@ -1004,7 +1002,7 @@ STATIC ssize_t psmx3_cq_sreadfrom(struct fid_cq *cq, void *buf, size_t count,
 				ofi_atomic_set32(&cq_priv->signaled, 0);
 				return -FI_ECANCELED;
 			}
-			ssize_t wait_result = fi_wait((struct fid_wait *)cq_priv->wait, timeout);
+			ssize_t wait_result = ofi_wait((struct fid_wait *)cq_priv->wait, timeout);
 			if (wait_result != FI_SUCCESS) {
 				return wait_result;
 			}
@@ -1113,7 +1111,7 @@ static int psmx3_cq_close(fid_t fid)
 	ofi_spin_destroy(&cq->lock);
 
 	if (cq->wait) {
-		fi_poll_del(&cq->wait->pollset->poll_fid, &cq->cq.fid, 0);
+		ofi_poll_del(&cq->wait->pollset->poll_fid, &cq->cq.fid, 0);
 		if (cq->wait_is_local)
 			fi_close(&cq->wait->wait_fid.fid);
 	}
@@ -1204,19 +1202,24 @@ int psmx3_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 		break;
 
 	default:
-		FI_INFO(&psmx3_prov, FI_LOG_CQ,
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CQ,
 			"attr->format=%d, supported=%d...%d\n", attr->format,
 			FI_CQ_FORMAT_UNSPEC, FI_CQ_FORMAT_TAGGED);
 		return -FI_EINVAL;
 	}
 
+	if (psmx3_env.yield_mode && attr->wait_obj != FI_WAIT_NONE) {
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CQ,
+			"waitset %d not allowed when FI_PSM3_YIELD_MODE enabled\n", attr->wait_obj);
+		return -FI_EINVAL;
+	}
 	switch (attr->wait_obj) {
 	case FI_WAIT_NONE:
 		break;
 
 	case FI_WAIT_SET:
 		if (!attr->wait_set) {
-			FI_INFO(&psmx3_prov, FI_LOG_CQ,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CQ,
 				"FI_WAIT_SET is specified but attr->wait_set is NULL\n");
 			return -FI_EINVAL;
 		}
@@ -1228,15 +1231,15 @@ int psmx3_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 	case FI_WAIT_MUTEX_COND:
 		wait_attr.wait_obj = attr->wait_obj;
 		wait_attr.flags = 0;
-		err = fi_wait_open(&domain_priv->fabric->util_fabric.fabric_fid,
-				   &wait_attr, (struct fid_wait **)&wait);
+		err = ofi_wait_open(&domain_priv->fabric->util_fabric.fabric_fid,
+				    &wait_attr, (struct fid_wait **)&wait);
 		if (err)
 			return err;
 		wait_is_local = 1;
 		break;
 
 	default:
-		FI_INFO(&psmx3_prov, FI_LOG_CQ,
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CQ,
 			"attr->wait_obj=%d, supported=%d...%d\n", attr->wait_obj,
 			FI_WAIT_NONE, FI_WAIT_MUTEX_COND);
 		return -FI_EINVAL;
@@ -1249,7 +1252,7 @@ int psmx3_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 			break;
 
 		default:
-			FI_INFO(&psmx3_prov, FI_LOG_CQ,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CQ,
 				"attr->wait_cond=%d, supported=%d...%d\n",
 				attr->wait_cond, FI_CQ_COND_NONE, FI_CQ_COND_THRESHOLD);
 			return -FI_EINVAL;
@@ -1289,14 +1292,14 @@ int psmx3_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 	for (i=0; i<PSMX3_FREE_LIST_SIZE; i++) {
 		event = calloc(1, sizeof(*event));
 		if (!event) {
-			FI_WARN(&psmx3_prov, FI_LOG_CQ, "out of memory.\n");
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CQ, "out of memory.\n");
 			exit(-1);
 		}
 		slist_insert_tail(&event->list_entry, &cq_priv->free_list);
 	}
 
 	if (wait)
-		fi_poll_add(&cq_priv->wait->pollset->poll_fid, &cq_priv->cq.fid, 0);
+		ofi_poll_add(&cq_priv->wait->pollset->poll_fid, &cq_priv->cq.fid, 0);
 
 	*cq = &cq_priv->cq;
 	return 0;
