@@ -36,28 +36,28 @@
 	do {									\
 		if (!send_completion && !recv_completion) {			\
 			ep->ep.tagged = &psmx3_tagged_ops_no_event##suffix;	\
-			FI_INFO(&psmx3_prov, FI_LOG_EP_DATA,			\
+			PSMX3_INFO(&psmx3_prov, FI_LOG_EP_DATA,			\
 				"tagged ops optimized for op_flags=0 "		\
 				"and event suppression "			\
 				msg_suffix					\
 				"\n");						\
 		} else if (!send_completion) {					\
 			ep->ep.tagged = &psmx3_tagged_ops_no_send_event##suffix;\
-			FI_INFO(&psmx3_prov, FI_LOG_EP_DATA,			\
+			PSMX3_INFO(&psmx3_prov, FI_LOG_EP_DATA,			\
 				"tagged ops optimized for op_flags=0 "		\
 				"and send event suppression "			\
 				msg_suffix					\
 				"\n");						\
 		} else if (!recv_completion) {					\
 			ep->ep.tagged = &psmx3_tagged_ops_no_recv_event##suffix;\
-			FI_INFO(&psmx3_prov, FI_LOG_EP_DATA,			\
+			PSMX3_INFO(&psmx3_prov, FI_LOG_EP_DATA,			\
 				"tagged ops optimized for op_flags=0 "		\
 				"and recv event suppression "			\
 				msg_suffix					\
 				"\n");						\
 		} else {							\
 			ep->ep.tagged = &psmx3_tagged_ops_no_flag##suffix;	\
-			FI_INFO(&psmx3_prov, FI_LOG_EP_DATA,			\
+			PSMX3_INFO(&psmx3_prov, FI_LOG_EP_DATA,			\
 				"tagged ops optimized for op_flags=0 "		\
 				msg_suffix					\
 				"\n");						\
@@ -76,7 +76,7 @@ static void psmx3_ep_optimize_ops(struct psmx3_fid_ep *ep)
 	if (ep->ep.tagged) {
 		if (ep->tx_flags & mask & ~FI_COMPLETION || ep->rx_flags & mask & ~FI_COMPLETION) {
 			ep->ep.tagged = &psmx3_tagged_ops;
-			FI_INFO(&psmx3_prov, FI_LOG_EP_DATA,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_EP_DATA,
 				"generic tagged ops.\n");
 		} else {
 			send_completion = !ep->send_selective_completion || ep->tx_flags & FI_COMPLETION;
@@ -147,6 +147,46 @@ STATIC ssize_t psmx3_ep_cancel(fid_t fid, void *context)
 	return psmx3_errno(err);
 }
 
+STATIC int psmx3_ep_getopt_cuda_api_permitted(
+	struct psmx3_fid_ep *ep, bool *value)
+{
+	// invariant: if both rx and tx are set, then they are expected to be
+	// the same internal PSM endpoint
+	assert(!ep->tx || !ep->rx || ep->tx->psm2_ep == ep->rx->psm2_ep);
+
+	uint64_t size = (uint64_t)sizeof(*value);
+
+	psm2_error_t err = psm3_getopt(
+		PSM2_COMPONENT_CORE,
+		ep->tx ? ep->tx->psm2_ep : ep->rx->psm2_ep,
+		PSM2_CORE_OPT_EP_CUDA_PERMITTED,
+		value,
+		&size);
+	if (err)
+		return -FI_EINVAL;
+
+	return 0;
+}
+
+STATIC int psmx3_ep_setopt_cuda_api_permitted(
+	struct psmx3_fid_ep *ep, const bool *value)
+{
+	// invariant: if both rx and tx are set, then they are expected to be
+	// the same internal PSM endpoint
+	assert(!ep->tx || !ep->rx || ep->tx->psm2_ep == ep->rx->psm2_ep);
+
+	psm2_error_t err = psm3_setopt(
+		PSM2_COMPONENT_CORE,
+		ep->tx ? ep->tx->psm2_ep : ep->rx->psm2_ep,
+		PSM2_CORE_OPT_EP_CUDA_PERMITTED,
+		value,
+		sizeof(*value));
+	if (err)
+		return -FI_EINVAL;
+
+	return 0;
+}
+
 DIRECT_FN
 STATIC int psmx3_ep_getopt(fid_t fid, int level, int optname,
 			   void *optval, size_t *optlen)
@@ -164,6 +204,11 @@ STATIC int psmx3_ep_getopt(fid_t fid, int level, int optname,
 		*optlen = sizeof(size_t);
 		break;
 
+	case FI_OPT_CUDA_API_PERMITTED:
+		if (!optlen || *optlen != sizeof(bool))
+			return -FI_EINVAL;
+		return psmx3_ep_getopt_cuda_api_permitted(ep, (bool *)optval);
+
 	default:
 		return -FI_ENOPROTOOPT;
 	}
@@ -187,6 +232,11 @@ STATIC int psmx3_ep_setopt(fid_t fid, int level, int optname,
 		ep->min_multi_recv = *(size_t *)optval;
 		break;
 
+	case FI_OPT_CUDA_API_PERMITTED:
+		if (optlen != sizeof(bool))
+			return -FI_EINVAL;
+		return psmx3_ep_setopt_cuda_api_permitted(ep, (const bool *)optval);
+
 	default:
 		return -FI_ENOPROTOOPT;
 	}
@@ -631,7 +681,7 @@ int psmx3_ep_open(struct fid_domain *domain, struct fi_info *info,
 	}
 	if (!psmx3_override_uuid() && info && info->domain_attr && info->domain_attr->auth_key) {
 		if (info->domain_attr->auth_key_size != sizeof(psm2_uuid_t)) {
-			FI_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
 				"Invalid domain auth_key_len %"PRIu64
 				", should be %"PRIu64".\n",
 				info->domain_attr->auth_key_size,
@@ -643,7 +693,7 @@ int psmx3_ep_open(struct fid_domain *domain, struct fi_info *info,
 
 	if (!psmx3_override_uuid() && info && info->ep_attr && info->ep_attr->auth_key) {
 		if (info->ep_attr->auth_key_size != sizeof(psm2_uuid_t)) {
-			FI_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
 				"Invalid ep auth_key_len %"PRIu64
 				", should be %"PRIu64".\n",
 				info->ep_attr->auth_key_size,
@@ -664,7 +714,7 @@ int psmx3_ep_open(struct fid_domain *domain, struct fi_info *info,
 		if (!trx_ctxt)
 			goto errout;
 	} else {
-		FI_INFO(&psmx3_prov, FI_LOG_EP_CTRL,
+		PSMX3_INFO(&psmx3_prov, FI_LOG_EP_CTRL,
 			"Tx only endpoint with STX context.\n");
 	}
 
@@ -956,7 +1006,7 @@ int psmx3_sep_open(struct fid_domain *domain, struct fi_info *info,
 
 	if (!psmx3_override_uuid() && info && info->domain_attr && info->domain_attr->auth_key) {
 		if (info->domain_attr->auth_key_size != sizeof(psm2_uuid_t)) {
-			FI_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
 				"Invalid domain auth_key_len %"PRIu64
 				", should be %"PRIu64".\n",
 				info->domain_attr->auth_key_size,
@@ -969,7 +1019,7 @@ int psmx3_sep_open(struct fid_domain *domain, struct fi_info *info,
 	if (info && info->ep_attr) {
 		if (!psmx3_override_uuid() && info->ep_attr->auth_key) {
 			if (info->ep_attr->auth_key_size != sizeof(psm2_uuid_t)) {
-				FI_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
+				PSMX3_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
 					"Invalid ep auth_key_len %"PRIu64
 					", should be %"PRIu64".\n",
 					info->ep_attr->auth_key_size,
@@ -980,14 +1030,14 @@ int psmx3_sep_open(struct fid_domain *domain, struct fi_info *info,
 		}
 
 		if (info->ep_attr->tx_ctx_cnt > psmx3_domain_info.max_trx_ctxt) {
-			FI_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
 				"tx_ctx_cnt %"PRIu64" exceed limit %d.\n",
 				info->ep_attr->tx_ctx_cnt,
 				psmx3_domain_info.max_trx_ctxt);
 			goto errout;
 		}
 		if (info->ep_attr->rx_ctx_cnt > psmx3_domain_info.max_trx_ctxt) {
-			FI_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_EP_CTRL,
 				"rx_ctx_cnt %"PRIu64" exceed limit %d.\n",
 				info->ep_attr->rx_ctx_cnt,
 				psmx3_domain_info.max_trx_ctxt);
@@ -997,7 +1047,7 @@ int psmx3_sep_open(struct fid_domain *domain, struct fi_info *info,
 		if (ctxt_cnt < info->ep_attr->rx_ctx_cnt)
 			ctxt_cnt = info->ep_attr->rx_ctx_cnt;
 		if (ctxt_cnt == 0) {
-			FI_INFO(&psmx3_prov, FI_LOG_EP_CTRL,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_EP_CTRL,
 				"tx_ctx_cnt and rx_ctx_cnt are 0, use 1.\n");
 			ctxt_cnt = 1;
 		}
