@@ -41,9 +41,100 @@ static pthread_mutex_t psmx3_lib_mutex;
 
 struct psmx3_domain_info psmx3_domain_info;
 
+static const char* FI_PSM3_NAME_SERVER_HELP =
+			"Whether to turn on the name server or not (default: yes)";
+static const char* FI_PSM3_TAGGED_RMA_HELP =
+			"Whether to use tagged messages for large size RMA or not " \
+			"(default: no)";
+static const char* FI_PSM3_UUID_HELP =
+			"Unique Job ID required by the fabric";
+static const char* FI_PSM3_DELAY_HELP =
+			"Delay (seconds) before finalization (for debugging)";
+static const char* FI_PSM3_TIMEOUT_HELP =
+			"Timeout (seconds) for gracefully closing the PSM3 endpoint";
+static const char* FI_PSM3_PROG_INTERVAL_HELP =
+			"Interval (microseconds) between progress calls made in the " \
+			"progress thread (default: 1 if affinity is set, 1000 if not)";
+static const char* FI_PSM3_PROG_AFFINITY_HELP =
+			"When set, specify the set of CPU cores to set the progress " \
+			"thread affinity to. The format is " \
+			"<start>[:<end>[:<stride>]][,<start>[:<end>[:<stride>]]]*, " \
+			"where each triplet <start>:<end>:<stride> defines a block " \
+			"of core_ids. Both <start> and <end> can be either the core_id " \
+			"(when >=0) or core_id - num_cores (when <0). " \
+			"(default: affinity not set)";
+static const char* FI_PSM3_INJECT_SIZE_HELP =
+			"Maximum message size for fi_inject and fi_tinject (default: 64).";
+static const char* FI_PSM3_LOCK_LEVEL_HELP =
+			"How internal locking is used. 0 means no locking. (default: 2).";
+static const char* FI_PSM3_LAZY_CONN_HELP =
+			"Whether to force lazy connection mode. (default: no).";
+static const char* FI_PSM3_CONN_TIMEOUT_HELP =
+			"Timeout (seconds) for establishing connection between two PSM3 endpoints";
+static const char* FI_PSM3_DISCONNECT_HELP =
+			"Whether to issue disconnect request when process ends (default: no).";
+#if (PSMX3_TAG_LAYOUT == PSMX3_TAG_LAYOUT_RUNTIME)
+static const char* FI_PSM3_TAG_LAYOUT_HELP =
+			"How the 96 bit PSM3 tag is organized: " \
+			"tag60 means 32/4/60 for data/flags/tag;" \
+			"tag64 means 4/28/64 for flags/data/tag (default: tag60).";
+#endif
+static const char* FI_PSM3_YIELD_MODE_HELP =
+			"Enabled interrupt driven operation with fi_wait. (default: no).";
+static const char* FI_PSM3_WAIT_ENABLE_HELP =
+			"Enabled use of wait semantics outside of yield mode. (default: no).";
+
+#define FI_PSM3_PREFIX "FI_PSM3_"
+#define FI_PSM3_PREFIX_LEN strlen(FI_PSM3_PREFIX)
+
+// logging and behavior similar to fi_param_get_bool
+int psmx3_param_get_bool(struct fi_provider *provider, const char *env_var_name,
+					const char *descr, int visible, int *value)
+{
+	assert(0 == strncmp(env_var_name, FI_PSM3_PREFIX, FI_PSM3_PREFIX_LEN));
+	if (psm3_getenv_bool(env_var_name, descr, visible, value)) {
+		PSMX3_INFO(provider, FI_LOG_CORE, "variable %s=<not set>\n", env_var_name);
+		return -FI_ENODATA;
+	} else {
+		PSMX3_INFO(provider, FI_LOG_CORE, "read bool var %s=%d\n", env_var_name,
+				*value);
+		return FI_SUCCESS;
+	}
+}
+
+// logging and behavior similar to fi_param_get_int
+int psmx3_param_get_int(struct fi_provider *provider, const char *env_var_name,
+					const char *descr, int visible, int *value)
+{
+	assert(0 == strncmp(env_var_name, FI_PSM3_PREFIX, FI_PSM3_PREFIX_LEN));
+	if (psm3_getenv_int(env_var_name, descr, visible, value)) {
+		PSMX3_INFO(provider, FI_LOG_CORE, "variable %s=<not set>\n", env_var_name);
+		return -FI_ENODATA;
+	} else {
+		PSMX3_INFO(provider, FI_LOG_CORE, "read int var %s=%d\n", env_var_name,
+				*value);
+		return FI_SUCCESS;
+	}
+}
+
+// logging and behavior similar to fi_param_get_str
+int psmx3_param_get_str(struct fi_provider *provider, const char *env_var_name,
+					const char *descr, int visible, char **value)
+{
+	assert(0 == strncmp(env_var_name, FI_PSM3_PREFIX, FI_PSM3_PREFIX_LEN));
+	if (psm3_getenv_str(env_var_name, descr, visible, value)) {
+		PSMX3_INFO(provider, FI_LOG_CORE, "variable %s=<not set>\n", env_var_name);
+		return -FI_ENODATA;
+	} else {
+		PSMX3_INFO(provider, FI_LOG_CORE, "read string var %s=%s\n", env_var_name,
+				*value);
+		return FI_SUCCESS;
+	}
+}
+
 struct psmx3_env psmx3_env = {
 	.name_server	= 1,
-	.tagged_rma	= 1,
+	.tagged_rma	= 0,
 	.uuid		= PSMX3_DEFAULT_UUID,
 	.uuid_override  = 0,
 	.delay		= 0,
@@ -59,6 +150,8 @@ struct psmx3_env psmx3_env = {
 #if (PSMX3_TAG_LAYOUT == PSMX3_TAG_LAYOUT_RUNTIME)
 	.tag_layout	= "auto",
 #endif
+	.yield_mode	= 0,
+	.wait_enable	= 0,
 };
 
 #if (PSMX3_TAG_LAYOUT == PSMX3_TAG_LAYOUT_RUNTIME)
@@ -74,13 +167,20 @@ static void psmx3_init_env(void)
 	uint32_t uid = getuid();
 	char *uuid = NULL;
 
+	psm3_env_initialize();
+
 	if (getenv("OMPI_COMM_WORLD_RANK") || getenv("PMI_RANK") || getenv("PMIX_RANK"))
 		psmx3_env.name_server = 0;
 
-	fi_param_get_bool(&psmx3_prov, "name_server", &psmx3_env.name_server);
-	fi_param_get_bool(&psmx3_prov, "tagged_rma", &psmx3_env.tagged_rma);
-
-	if (FI_SUCCESS != fi_param_get_str(&psmx3_prov, "uuid", &psmx3_env.uuid)) {
+	//fi_param_get_bool(&psmx3_prov, "name_server", &psmx3_env.name_server);
+	psmx3_param_get_bool(&psmx3_prov, "FI_PSM3_NAME_SERVER",
+					FI_PSM3_NAME_SERVER_HELP, 0, &psmx3_env.name_server);
+	//fi_param_get_bool(&psmx3_prov, "tagged_rma", &psmx3_env.tagged_rma);
+	psmx3_param_get_bool(&psmx3_prov, "FI_PSM3_TAGGED_RMA",
+					FI_PSM3_TAGGED_RMA_HELP, 0, &psmx3_env.tagged_rma);
+	//if (FI_SUCCESS != fi_param_get_str(&psmx3_prov, "uuid", &psmx3_env.uuid)) {
+	if (FI_SUCCESS != psmx3_param_get_str(&psmx3_prov, "FI_PSM3_UUID",
+					FI_PSM3_UUID_HELP, 1, &psmx3_env.uuid)) {
 		/*
 		 * For OpenMPI 4.x only:
 		 * The job key is passed via the environment variable, but the format
@@ -91,12 +191,12 @@ static void psmx3_init_env(void)
 		unsigned long long int *u = (unsigned long long int *)ompi_uuid;
 		char *ompi_job_key = getenv("OMPI_MCA_orte_precondition_transports");
 		if (ompi_job_key) {
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"Open MPI job key: %s.\n", ompi_job_key);
 			if (sscanf(ompi_job_key, "%016llx-%016llx", &u[0], &u[1]) == 2)
 				uuid = strdup(psmx3_uuid_to_string(ompi_uuid));
 			else {
-				FI_INFO(&psmx3_prov, FI_LOG_CORE,
+				PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 					"Invalid Open MPI job key format.\n");
 			}
 		}
@@ -119,20 +219,45 @@ static void psmx3_init_env(void)
 		/* FI_PSM3_UUID has highest priority, so it can override auth_key from fi_info */
 		psmx3_env.uuid_override = 1;
 	}
-	fi_param_get_int(&psmx3_prov, "delay", &psmx3_env.delay);
-	fi_param_get_int(&psmx3_prov, "timeout", &psmx3_env.timeout);
-	fi_param_get_int(&psmx3_prov, "prog_interval", &psmx3_env.prog_interval);
-	fi_param_get_str(&psmx3_prov, "prog_affinity", &psmx3_env.prog_affinity);
-	fi_param_get_int(&psmx3_prov, "inject_size", &psmx3_env.inject_size);
-	fi_param_get_bool(&psmx3_prov, "lock_level", &psmx3_env.lock_level);
-	fi_param_get_bool(&psmx3_prov, "lazy_conn", &psmx3_env.lazy_conn);
+	//fi_param_get_int(&psmx3_prov, "delay", &psmx3_env.delay);
+	psmx3_param_get_int(&psmx3_prov, "FI_PSM3_DELAY",
+				FI_PSM3_DELAY_HELP, 0, &psmx3_env.delay);
+	//fi_param_get_int(&psmx3_prov, "timeout", &psmx3_env.timeout);
+	psmx3_param_get_int(&psmx3_prov, "FI_PSM3_TIMEOUT",
+				FI_PSM3_TIMEOUT_HELP, 0, &psmx3_env.timeout);
+	//fi_param_get_int(&psmx3_prov, "prog_interval", &psmx3_env.prog_interval);
+	psmx3_param_get_int(&psmx3_prov, "FI_PSM3_PROG_INTERVAL",
+				FI_PSM3_PROG_INTERVAL_HELP, 0, &psmx3_env.prog_interval);
+	//fi_param_get_str(&psmx3_prov, "prog_affinity", &psmx3_env.prog_affinity);
+	psmx3_param_get_str(&psmx3_prov, "FI_PSM3_PROG_AFFINITY",
+				FI_PSM3_PROG_AFFINITY_HELP, 0, &psmx3_env.prog_affinity);
+	//fi_param_get_int(&psmx3_prov, "inject_size", &psmx3_env.inject_size);
+	psmx3_param_get_int(&psmx3_prov, "FI_PSM3_INJECT_SIZE",
+				FI_PSM3_INJECT_SIZE_HELP, 1, &psmx3_env.inject_size);
+	//fi_param_get_int(&psmx3_prov, "lock_level", &psmx3_env.lock_level);
+	psmx3_param_get_int(&psmx3_prov, "FI_PSM3_LOCK_LEVEL",
+				FI_PSM3_LOCK_LEVEL_HELP, 0, &psmx3_env.lock_level);
+	//fi_param_get_bool(&psmx3_prov, "lazy_conn", &psmx3_env.lazy_conn);
+	psmx3_param_get_bool(&psmx3_prov, "FI_PSM3_LAZY_CONN",
+				FI_PSM3_LAZY_CONN_HELP, 1, &psmx3_env.lazy_conn);
 	if (psmx3_env.lazy_conn)
 		psmx3_env.conn_timeout = 30;	// more headroom since app may be busy
-	fi_param_get_int(&psmx3_prov, "conn_timeout", &psmx3_env.conn_timeout);
-	fi_param_get_bool(&psmx3_prov, "disconnect", &psmx3_env.disconnect);
+	//fi_param_get_int(&psmx3_prov, "conn_timeout", &psmx3_env.conn_timeout);
+	psmx3_param_get_int(&psmx3_prov, "FI_PSM3_CONN_TIMEOUT",
+				FI_PSM3_CONN_TIMEOUT_HELP, 0, &psmx3_env.conn_timeout);
+	//fi_param_get_bool(&psmx3_prov, "disconnect", &psmx3_env.disconnect);
+	psmx3_param_get_bool(&psmx3_prov, "FI_PSM3_DISCONNECT",
+				FI_PSM3_DISCONNECT_HELP, 0, &psmx3_env.disconnect);
 #if (PSMX3_TAG_LAYOUT == PSMX3_TAG_LAYOUT_RUNTIME)
-	fi_param_get_str(&psmx3_prov, "tag_layout", &psmx3_env.tag_layout);
+	//fi_param_get_str(&psmx3_prov, "tag_layout", &psmx3_env.tag_layout);
+	psmx3_param_get_str(&psmx3_prov, "FI_PSM3_TAG_LAYOUT",
+				FI_PSM3_TAG_LAYOUT_HELP, 0, &psmx3_env.tag_layout);
 #endif
+	//fi_param_get_bool(&psmx3_prov, "yield_mode", &psmx3_env.yield_mode);
+	psmx3_param_get_bool(&psmx3_prov, "FI_PSM3_YIELD_MODE",
+				FI_PSM3_YIELD_MODE_HELP, 0, &psmx3_env.yield_mode);
+	psmx3_param_get_bool(&psmx3_prov, "FI_PSM3_WAIT_ENABLE",
+				FI_PSM3_WAIT_ENABLE_HELP, 0, &psmx3_env.wait_enable);
 }
 
 void psmx3_init_tag_layout(struct fi_info *info)
@@ -143,7 +268,7 @@ void psmx3_init_tag_layout(struct fi_info *info)
 	use_tag64 = (psmx3_tag_mask == PSMX3_TAG_MASK_64);
 
 	if (psmx3_tag_layout_locked) {
-		FI_INFO(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 			"tag layout already set opened domain.\n");
 		goto out;
 	}
@@ -162,7 +287,7 @@ void psmx3_init_tag_layout(struct fi_info *info)
 		use_tag64 = 1;
 	} else {
 		if (strcasecmp(psmx3_env.tag_layout, "auto") != 0) {
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"Invalid tag layout '%s', using 'auto'.\n",
 				psmx3_env.tag_layout);
 			psmx3_env.tag_layout = "auto";
@@ -190,39 +315,21 @@ out:
 #else
 	use_tag64 = 0;
 #endif
-	FI_INFO(&psmx3_prov, FI_LOG_CORE,
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 		"use %s: tag_mask: %016" PRIX64 ", data_mask: %08" PRIX32 "\n",
 		use_tag64 ? "tag64" : "tag60", (uint64_t)PSMX3_TAG_MASK,
 		PSMX3_DATA_MASK);
 }
 
-static int psmx3_get_yes_no(char *s, int default_value)
-{
-	unsigned long value;
-	char *end_ptr;
-
-	if (!s || s[0] == '\0')
-		return default_value;
-
-	if (s[0] == 'Y' || s[0] == 'y')
-		return 1;
-
-	if (s[0] == 'N' || s[0] == 'n')
-		return 0;
-
-	value = strtoul(s, &end_ptr, 0);
-	if (end_ptr == s)
-		return default_value;
-
-	return value ? 1 : 0;
-}
-
 static int psmx3_check_multi_ep_cap(void)
 {
 	uint64_t caps = PSM2_MULTI_EP_CAP;
-	char *s = getenv("PSM3_MULTI_EP");
+	char *s = NULL;
+	int val = 1; /* if parses as empty (-1) or invalid (-2), use default of 1 */
 
-	if (psm3_get_capability_mask(caps) == caps && psmx3_get_yes_no(s, 1))
+	s = psm3_env_get("PSM3_MULTI_EP");
+	/* psm3 below us will provide warning as needed when it parses it again */
+	if (psm3_get_capability_mask(caps) == caps && (psm3_parse_str_yesno(s, &val) || val))
 		psmx3_env.multi_ep = 1;
 	else
 		psmx3_env.multi_ep = 0;
@@ -230,6 +337,16 @@ static int psmx3_check_multi_ep_cap(void)
 	return psmx3_env.multi_ep;
 }
 
+/* indicate if lock_level variable was set (1 if set, 0 if defaulted) */
+int psmx3_check_lock_level(void)
+{
+	int tmp = 2;
+
+	//err = fi_param_get_int(&psmx3_prov, "lock_level", &tmp);
+	return (0 == psmx3_param_get_int(&psmx3_prov, "FI_PSM3_LOCK_LEVEL",
+				FI_PSM3_LOCK_LEVEL_HELP, 0, &tmp));
+}
+
 static int psmx3_init_lib(void)
 {
 	int major, minor;
@@ -253,21 +370,21 @@ static int psmx3_init_lib(void)
 
 	err = psm3_init(&major, &minor);
 	if (err != PSM2_OK) {
-		FI_WARN(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 			"psm3_init failed: %s\n", psm3_error_get_string(err));
 		ret = err;
 		goto out;
 	}
 
-	FI_INFO(&psmx3_prov, FI_LOG_CORE,
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 		"PSM3 header version = (%d, %d)\n", PSM2_VERNO_MAJOR, PSM2_VERNO_MINOR);
-	FI_INFO(&psmx3_prov, FI_LOG_CORE,
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 		"PSM3 library version = (%d, %d)\n", major, minor);
 
 	if (psmx3_check_multi_ep_cap())
-		FI_INFO(&psmx3_prov, FI_LOG_CORE, "PSM3 multi-ep feature enabled.\n");
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "PSM3 multi-ep feature enabled.\n");
 	else
-		FI_INFO(&psmx3_prov, FI_LOG_CORE, "PSM3 multi-ep feature not available or disabled.\n");
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "PSM3 multi-ep feature not available or disabled.\n");
 
 	psmx3_lib_initialized = 1;
 
@@ -279,16 +396,12 @@ out:
 static int psmx3_update_hfi_info(void)
 {
 	unsigned short i, j, psmx3_unit;
-	int nctxts = 0;
-	int nfreectxts = 0;
 	int multirail = 0;
-	int counted_unit;
-	char *s;
+	char *s = NULL;
 	char unit_name[NAME_MAX];
 	char fabric_name[NAME_MAX];
 	uint32_t cnt = 0;
 	uint32_t addr_cnt = 0;
-	int tmp_nctxts, tmp_nfreectxts;
 	int unit_active;
 	int ret;
 	psm2_info_query_arg_t args[4];
@@ -300,32 +413,33 @@ static int psmx3_update_hfi_info(void)
 
 	if (psm3_info_query(PSM2_INFO_QUERY_NUM_UNITS, &cnt, 0, NULL)
 			|| !cnt) {
-		FI_INFO(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 			"no PSM3 device is found.\n");
 		return -FI_ENODEV;
 	} else if (cnt > PSMX3_MAX_UNITS) {
-		FI_WARN(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 			"Discovered %u devices, limited to 1st %u\n", cnt, PSMX3_MAX_UNITS);
 		cnt = PSMX3_MAX_UNITS;
 	}
 	if (psm3_info_query(PSM2_INFO_QUERY_NUM_ADDR_PER_UNIT, &addr_cnt, 0, NULL) || !addr_cnt)
 	{
-		FI_INFO(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 			"no PSM3 device is found (no addr per unit).\n");
 		return -FI_ENODEV;
 	}
 	psmx3_domain_info.num_units = cnt * addr_cnt;
 	if (psmx3_domain_info.num_units > PSMX3_MAX_UNITS) {
-		FI_WARN(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 			"Too many addresses per PSM3 device. For %d physical "
 			"devices only  %d addresses per device are supported.",
 			cnt, PSMX3_MAX_UNITS/cnt);
 		return-FI_ENODEV;
 	}
 
-	s = getenv("PSM3_MULTIRAIL");
-	if (s)
-		multirail = atoi(s);
+	// if parses as empty or invalid (-1), use default of 0 */
+	// PSM3 below us will provide warning as needed when it parses it
+	s = psm3_env_get("PSM3_MULTIRAIL");
+	(void)psm3_parse_str_int(s, &multirail, INT_MIN, INT_MAX);
 
 	psmx3_domain_info.num_reported_units = 0;
 	psmx3_domain_info.num_active_units = 0;
@@ -333,38 +447,19 @@ static int psmx3_update_hfi_info(void)
 		args[0].unit = i;
 		ret = psm3_info_query(PSM2_INFO_QUERY_UNIT_STATUS, &unit_active, 1, args);
 		if (ret != PSM2_OK) {
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 				"Failed to check active state of HFI unit_id %d\n",
 				i);
 			continue;
 		}
 
 		if (unit_active<=0) {
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 				"NIC unit_id %d STATE = INACTIVE\n",
 				i);
 			continue;
 		}
 
-		if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_NUM_FREE_CONTEXTS,
-						&tmp_nfreectxts, 1, args) || (tmp_nfreectxts < 0))
-		{
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
-				"Failed to read number of free contexts from HFI unit_id %d\n",
-				i);
-			continue;
-		}
-
-		if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_NUM_CONTEXTS,
-						&tmp_nctxts, 1, args) || (tmp_nctxts < 0))
-		{
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
-				"Failed to read number of contexts from HFI unit_id %d\n",
-				i);
-			continue;
-		}
-
-		counted_unit = 0;
 		for (j=0; j < addr_cnt; j++) {
 			psmx3_unit = i * addr_cnt + j;
 			args[1].port = 1;	// VERBS_PORT
@@ -374,7 +469,7 @@ static int psmx3_update_hfi_info(void)
 			if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_ADDR_NAME,
 							unit_name, 4, args))
 			{
-				FI_WARN(&psmx3_prov, FI_LOG_CORE,
+				PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 					"Failed to read name of HFI unit_id %d addr_index %d\n",
 					i, j);
 				continue;
@@ -387,18 +482,12 @@ static int psmx3_update_hfi_info(void)
 			if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_SUBNET_NAME,
 							fabric_name, 4, args))
 			{
-				FI_WARN(&psmx3_prov, FI_LOG_CORE,
+				PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 					"Failed to read name of HFI unit_id %d addr_index %d subnet\n",
 					i, j);
 				continue;
 			}
 
-			if (! counted_unit) {
-				nctxts += tmp_nctxts;
-				nfreectxts += tmp_nfreectxts;
-				counted_unit = 1;
-			}
-
 			psmx3_domain_info.num_active_units++;
 
 			/* for PSM3_MULTIRAIL only report 1 "autoselect" unit */
@@ -406,8 +495,6 @@ static int psmx3_update_hfi_info(void)
 				psmx3_domain_info.unit_is_active[psmx3_unit] = 1;
 				psmx3_domain_info.unit_id[psmx3_unit] = i;
 				psmx3_domain_info.addr_index[psmx3_unit] = j;
-				psmx3_domain_info.unit_nctxts[psmx3_unit] = tmp_nctxts;
-				psmx3_domain_info.unit_nfreectxts[psmx3_unit] = tmp_nfreectxts;
 				psmx3_domain_info.active_units[psmx3_domain_info.num_reported_units++] = psmx3_unit;
 			}
 			if (psmx3_domain_info.num_active_units == 1) {
@@ -440,23 +527,20 @@ static int psmx3_update_hfi_info(void)
 		strcpy(psmx3_domain_info.default_fabric_name, first_active_unit_fabric_name);
 	}
 
-	FI_INFO(&psmx3_prov, FI_LOG_CORE,
-		"hfi1 units: total %d, reported %d, active %d; "
-		"hfi1 contexts: total %d, free %d\n",
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
+		"psm3 units: total %d, reported %d, active %d\n",
 		psmx3_domain_info.num_units, psmx3_domain_info.num_reported_units,
-		psmx3_domain_info.num_active_units, nctxts, nfreectxts);
+		psmx3_domain_info.num_active_units);
 
 	if (psmx3_env.multi_ep) {
-		psmx3_domain_info.max_trx_ctxt = nctxts;
-		psmx3_domain_info.free_trx_ctxt = nfreectxts;
+		psmx3_domain_info.max_trx_ctxt = PSMX3_MAX_EPS;
 	} else {
 		psmx3_domain_info.max_trx_ctxt = 1;
-		psmx3_domain_info.free_trx_ctxt = (nfreectxts == 0) ? 0 : 1;
 	}
 
-	FI_INFO(&psmx3_prov, FI_LOG_CORE,
-		"Tx/Rx contexts: %d in total, %d available.\n",
-		psmx3_domain_info.max_trx_ctxt, psmx3_domain_info.free_trx_ctxt);
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
+		"Tx/Rx contexts: %d allowed per process.\n",
+		psmx3_domain_info.max_trx_ctxt);
 
 	return 0;
 }
@@ -471,7 +555,7 @@ static psm2_error_t psmx3_get_link_speed(int unit_id, int port, uint64_t *speed,
 	args[1].port = port == PSMX3_DEFAULT_PORT ? 1 : port;
 	ret = psm3_info_query(PSM2_INFO_QUERY_PORT_SPEED, speed, 2, args);
 	if (PSM2_OK != ret) {
-		FI_WARN(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 			"Failed to query link_attr (speed) for unit %d/%d: %s\n",
 			unit_id, port, info->nic && info->nic->device_attr?
 					info->nic->device_attr->name:"(null)");
@@ -500,7 +584,7 @@ static void psmx3_update_nic_info(struct fi_info *info)
 			if (!info->nic) {
 				info->nic = ofi_nic_dup(NULL);
 				if (!info->nic) {
-					FI_WARN(&psmx3_prov, FI_LOG_CORE,
+					PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 						"Failed to allocate nic info for unit %d\n", unit);
 					continue;
 				}
@@ -516,7 +600,7 @@ static void psmx3_update_nic_info(struct fi_info *info)
 		if (err == PSM2_EP_NO_DEVICE)
 			continue;	/* no NIC to report, probably single node job */
 		if (err != PSM2_OK) {
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 				"Failed to query bus_attr (pci bus) for unit %d\n", unit);
 			continue;
 		}
@@ -524,7 +608,7 @@ static void psmx3_update_nic_info(struct fi_info *info)
 		if (!info->nic) {
 			info->nic = ofi_nic_dup(NULL);
 			if (!info->nic) {
-				FI_WARN(&psmx3_prov, FI_LOG_CORE,
+				PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 					"Failed to allocate nic info for unit %d\n", unit);
 				continue;
 			}
@@ -540,14 +624,14 @@ static void psmx3_update_nic_info(struct fi_info *info)
 		/* Device Attr */
 		args[1].length = sizeof(buffer);
 		if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_NAME, buffer, 2, args)) {
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 				"Failed to query device_attr (name) for unit %d\n", unit);
 			continue;
 		}
 		info->nic->device_attr->name = strdup(buffer);
 
 		if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_DEVICE_ID, buffer, 2, args)) {
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 				"Failed to query device_attr (device_id) for unit %d: %s\n",
 				unit, info->nic->device_attr->name);
 			continue;
@@ -555,7 +639,7 @@ static void psmx3_update_nic_info(struct fi_info *info)
 		info->nic->device_attr->device_id = strdup(buffer);
 
 		if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_DEVICE_VERSION, buffer, 2, args)) {
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 				"Failed to query device_attr (device_version) for unit %d: %s\n",
 				unit, info->nic->device_attr->name);
 			continue;
@@ -563,7 +647,7 @@ static void psmx3_update_nic_info(struct fi_info *info)
 		info->nic->device_attr->device_version = strdup(buffer);
 
 		if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_VENDOR_ID, buffer, 2, args)) {
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 				"Failed to query device_attr (vendor_id) for unit %d: %s\n",
 				unit, info->nic->device_attr->name);
 			continue;
@@ -571,7 +655,7 @@ static void psmx3_update_nic_info(struct fi_info *info)
 		info->nic->device_attr->vendor_id = strdup(buffer);
 
 		if (PSM2_OK != psm3_info_query(PSM2_INFO_QUERY_UNIT_DRIVER, buffer, 2, args)) {
-			FI_WARN(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_WARN(&psmx3_prov, FI_LOG_CORE,
 				"Failed to query device_attr (driver) for unit %d: %s\n",
 				unit, info->nic->device_attr->name);
 			continue;
@@ -586,6 +670,7 @@ static void psmx3_update_nic_info(struct fi_info *info)
 	}
 }
 
+static int init_calls;
 static int psmx3_getinfo(uint32_t api_version, const char *node,
 			 const char *service, uint64_t flags,
 			 const struct fi_info *hints, struct fi_info **info)
@@ -598,7 +683,7 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 	void *addr;
 	uint32_t fmt;
 
-	FI_INFO(&psmx3_prov, FI_LOG_CORE,"\n");
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,"\n");
 
 	if (psmx3_init_prov_info(hints, &prov_info))
 		goto err_out;
@@ -610,11 +695,13 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 		goto err_out;
 
 	if (!psmx3_domain_info.num_reported_units) {
-		FI_INFO(&psmx3_prov, FI_LOG_CORE,
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 			"no PSM3 device is active.\n");
 		goto err_out;
 	}
 
+	init_calls += 1;
+
 	/* when available, default domain and fabric names are a superset
 	 * of all individual names, so we can do a substr search as a 1st level
 	 * filter
@@ -622,7 +709,7 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 	if (hints && hints->domain_attr && hints->domain_attr->name &&
 		psmx3_domain_info.default_domain_name[0] &&
 		NULL == strcasestr(psmx3_domain_info.default_domain_name, hints->domain_attr->name)) {
-		FI_INFO(&psmx3_prov, FI_LOG_CORE, "Unknown domain name\n");
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Unknown domain name\n");
 		OFI_INFO_STR(&psmx3_prov, psmx3_domain_info.default_domain_name,
 					   hints->domain_attr->name, "Supported", "Requested");
 		goto err_out;
@@ -631,7 +718,7 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 	if (hints && hints->fabric_attr && hints->fabric_attr->name &&
 		psmx3_domain_info.default_fabric_name[0] &&
 		NULL == strcasestr(psmx3_domain_info.default_fabric_name, hints->fabric_attr->name)) {
-		FI_INFO(&psmx3_prov, FI_LOG_CORE, "Unknown fabric name\n");
+		PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Unknown fabric name\n");
 		OFI_INFO_STR(&psmx3_prov, psmx3_domain_info.default_fabric_name,
 					   hints->fabric_attr->name, "Supported", "Requested");
 		goto err_out;
@@ -643,12 +730,12 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 	    fmt == FI_ADDR_PSMX3) {
 		if (flags & FI_SOURCE) {
 			src_addr = addr;
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"'%s' is taken as src_addr: <unit=%d, port=%d, service=%d>\n",
 				node, src_addr->unit, src_addr->port, src_addr->service);
 		} else {
 			dest_addr = addr;
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"'%s' is taken as dest_addr: <epid=%s>\n",
 				node, psm3_epid_fmt(dest_addr->epid, 0));
 		}
@@ -659,7 +746,7 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 	if (!src_addr) {
 		src_addr = calloc(1, sizeof(*src_addr));
 		if (!src_addr) {
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"failed to allocate src addr.\n");
 			goto err_out;
 		}
@@ -673,15 +760,15 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 			if (node) {
 				if (sscanf(node, "%*[^:]:%" SCNi8 ":%" SCNu8,
 				           &src_addr->unit, &src_addr->port) != 2) {
-					FI_INFO(&psmx3_prov, FI_LOG_CORE, "Invalid string representation for node.\n");
+					PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Invalid string representation for node.\n");
 				}
 			}
 			if (service) {
 				if (sscanf(service, "%" SCNu32, &src_addr->service) != 1) {
-					FI_INFO(&psmx3_prov, FI_LOG_CORE, "Invalid string representation for service.\n");
+					PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "Invalid string representation for service.\n");
 				}
 			}
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"node '%s' service '%s' converted to <unit=%d, port=%d, service=%d>\n",
 				node, service, src_addr->unit, src_addr->port, src_addr->service);
 		}
@@ -690,12 +777,12 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 	/* Check that the src address contains valid unit */
 	if (src_addr->unit != PSMX3_DEFAULT_UNIT) {
 		if (src_addr->unit < 0 || src_addr->unit >= PSMX3_MAX_UNITS) {
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"invalid source address: unit %d out of range\n", src_addr->unit);
 			goto err_out;
 		}
 		if (!psmx3_domain_info.unit_is_active[src_addr->unit]) {
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"invalid source address: unit %d is inactive\n", src_addr->unit);
 			goto err_out;
 		}
@@ -721,11 +808,11 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 		dest_addr = (struct psmx3_ep_name *)
 			ofi_ns_resolve_name(&ns, node, &svc);
 		if (dest_addr) {
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"'%s:%u' resolved to <epid=%s>:%d\n",
 				node, svc0, psm3_epid_fmt(dest_addr->epid, 0), svc);
 		} else {
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
+			PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
 				"failed to resolve '%s:%u'.\n", node, svc);
 			goto err_out;
 		}
@@ -747,6 +834,9 @@ static int psmx3_getinfo(uint32_t api_version, const char *node,
 	*info = prov_info;
 	free(src_addr);
 	free(dest_addr);
+	if (hints || init_calls >= 2) {
+		psm3_turn_off_init_cache();
+	}
 	return 0;
 
 err_out:
@@ -759,22 +849,31 @@ err_out:
 
 static void psmx3_fini(void)
 {
-	FI_INFO(&psmx3_prov, FI_LOG_CORE, "\n");
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "\n");
 
-	if (! --psmx3_init_count && psmx3_lib_initialized) {
-		/* This function is called from a library destructor, which is called
-		 * automatically when exit() is called. The call to psm3_finalize()
-		 * might cause deadlock if the applicaiton is terminated with Ctrl-C
-		 * -- the application could be inside a PSM3 call, holding a lock that
-		 * psm3_finalize() tries to acquire. This can be avoided by only
-		 * calling psm3_finalize() when PSM3 is guaranteed to be unused.
+	if (! --psmx3_init_count) {
+		/* psm3_env_finalize is called by psm3_finailize, only need to call
+		 * when we don't call psm3_finalize
 		 */
-		if (psmx3_active_fabric) {
-			FI_INFO(&psmx3_prov, FI_LOG_CORE,
-				"psmx3_active_fabric != NULL, skip psm3_finalize\n");
+		if (psmx3_lib_initialized) {
+			/* This function is called from a library destructor, which is
+			 * called automatically when exit() is called. The call to
+			 * psm3_finalize() might cause deadlock if the applicaiton is
+			 * terminated with Ctrl-C
+			 * -- the application could be inside a PSM3 call, holding a lock
+			 * that psm3_finalize() tries to acquire. This can be avoided by
+			 * only calling psm3_finalize() when PSM3 is guaranteed to be
+			 * unused.
+			 */
+			if (psmx3_active_fabric) {
+				PSMX3_INFO(&psmx3_prov, FI_LOG_CORE,
+					"psmx3_active_fabric != NULL, skip psm3_finalize\n");
+			} else {
+				psm3_finalize();
+				psmx3_lib_initialized = 0;
+			}
 		} else {
-			psm3_finalize();
-			psmx3_lib_initialized = 0;
+			psm3_env_finalize();
 		}
 	}
 }
@@ -791,7 +890,7 @@ PROVIDER_INI
 {
 	psmx3_prov.version = get_psm3_provider_version();
 
-	FI_INFO(&psmx3_prov, FI_LOG_CORE, "build options: VERSION=%u.%u=%u.%u.%u.%u, "
+	PSMX3_INFO(&psmx3_prov, FI_LOG_CORE, "build options: VERSION=%u.%u=%u.%u.%u.%u, "
 			"HAVE_PSM3_SRC=%d, PSM3_CUDA=%d\n",
 			(psmx3_prov.version >> 16), (psmx3_prov.version & 0xFFFF),
 			(psmx3_prov.version >> 16) / 100, (psmx3_prov.version >> 16) % 100,
@@ -799,56 +898,49 @@ PROVIDER_INI
 			HAVE_PSM3_SRC, PSM3_CUDA);
 
 	fi_param_define(&psmx3_prov, "name_server", FI_PARAM_BOOL,
-			"Whether to turn on the name server or not "
-			"(default: yes)");
+			FI_PSM3_NAME_SERVER_HELP);
 
 	fi_param_define(&psmx3_prov, "tagged_rma", FI_PARAM_BOOL,
-			"Whether to use tagged messages for large size "
-			"RMA or not (default: yes)");
+			FI_PSM3_TAGGED_RMA_HELP);
 
 	fi_param_define(&psmx3_prov, "uuid", FI_PARAM_STRING,
-			"Unique Job ID required by the fabric");
+			FI_PSM3_UUID_HELP);
 
 	fi_param_define(&psmx3_prov, "delay", FI_PARAM_INT,
-			"Delay (seconds) before finalization (for debugging)");
+			FI_PSM3_DELAY_HELP);
 
 	fi_param_define(&psmx3_prov, "timeout", FI_PARAM_INT,
-			"Timeout (seconds) for gracefully closing the PSM3 endpoint");
+			FI_PSM3_TIMEOUT_HELP);
 
 	fi_param_define(&psmx3_prov, "conn_timeout", FI_PARAM_INT,
-			"Timeout (seconds) for establishing connection between two PSM3 endpoints");
+			FI_PSM3_CONN_TIMEOUT_HELP);
 
 	fi_param_define(&psmx3_prov, "prog_interval", FI_PARAM_INT,
-			"Interval (microseconds) between progress calls made in the "
-			"progress thread (default: 1 if affinity is set, 1000 if not)");
+			FI_PSM3_PROG_INTERVAL_HELP);
 
 	fi_param_define(&psmx3_prov, "prog_affinity", FI_PARAM_STRING,
-			"When set, specify the set of CPU cores to set the progress "
-			"thread affinity to. The format is "
-			"<start>[:<end>[:<stride>]][,<start>[:<end>[:<stride>]]]*, "
-			"where each triplet <start>:<end>:<stride> defines a block "
-			"of core_ids. Both <start> and <end> can be either the core_id "
-			"(when >=0) or core_id - num_cores (when <0). "
-			"(default: affinity not set)");
+			FI_PSM3_PROG_AFFINITY_HELP);
 
 	fi_param_define(&psmx3_prov, "inject_size", FI_PARAM_INT,
-			"Maximum message size for fi_inject and fi_tinject (default: 64).");
+			FI_PSM3_INJECT_SIZE_HELP);
 
 	fi_param_define(&psmx3_prov, "lock_level", FI_PARAM_INT,
-			"How internal locking is used. 0 means no locking. (default: 2).");
+			FI_PSM3_LOCK_LEVEL_HELP);
 
 	fi_param_define(&psmx3_prov, "lazy_conn", FI_PARAM_BOOL,
-			"Whether to force lazy connection mode. (default: no).");
+			FI_PSM3_LAZY_CONN_HELP);
 
 	fi_param_define(&psmx3_prov, "disconnect", FI_PARAM_BOOL,
-			"Whether to issue disconnect request when process ends (default: no).");
+			FI_PSM3_DISCONNECT_HELP);
 
 #if (PSMX3_TAG_LAYOUT == PSMX3_TAG_LAYOUT_RUNTIME)
 	fi_param_define(&psmx3_prov, "tag_layout", FI_PARAM_STRING,
-			"How the 96 bit PSM3 tag is organized: "
-			"tag60 means 32/4/60 for data/flags/tag;"
-			"tag64 means 4/28/64 for flags/data/tag (default: tag60).");
+			FI_PSM3_TAG_LAYOUT_HELP);
 #endif
+	fi_param_define(&psmx3_prov, "yield_mode", FI_PARAM_BOOL,
+			FI_PSM3_YIELD_MODE_HELP);
+	fi_param_define(&psmx3_prov, "wait_enable", FI_PARAM_BOOL,
+			FI_PSM3_WAIT_ENABLE_HELP);
 
 	psmx3_init_env();
 
