@@ -69,7 +69,7 @@ static struct rxd_x_entry *rxd_tx_entry_init_rma(struct rxd_ep *ep, fi_addr_t ad
 			rxd_init_sar_hdr(&ptr, tx_entry, rma_count);
 		} else {
 			tx_entry->flags |= RXD_INLINE;
-			base_hdr->flags = tx_entry->flags;
+			base_hdr->flags = (uint16_t) tx_entry->flags;
 			tx_entry->num_segs = 1;
 		}
 		rxd_init_rma_hdr(&ptr, rma_iov, rma_count);
@@ -94,15 +94,15 @@ static ssize_t rxd_generic_write_inject(struct rxd_ep *rxd_ep,
 	ssize_t ret = -FI_EAGAIN;
 
 	assert(iov_count <= RXD_IOV_LIMIT && rma_count <= RXD_IOV_LIMIT);
-	assert(ofi_total_iov_len(iov, iov_count) <= rxd_ep_domain(rxd_ep)->max_inline_rma);
+	assert(ofi_total_iov_len(iov, iov_count) <= (size_t) rxd_ep_domain(rxd_ep)->max_inline_rma);
 
-	fastlock_acquire(&rxd_ep->util_ep.lock);
+	ofi_genlock_lock(&rxd_ep->util_ep.lock);
 
 	if (ofi_cirque_isfull(rxd_ep->util_ep.tx_cq->cirq))
 		goto out;
-	
+
 	rxd_addr = (intptr_t) ofi_idx_lookup(&(rxd_ep_av(rxd_ep)->fi_addr_idx),
-					      RXD_IDX_OFFSET(addr));
+					      RXD_IDX_OFFSET((int) addr));
 	if (!rxd_addr)
 		goto out;
 	ret = rxd_send_rts_if_needed(rxd_ep, rxd_addr);
@@ -126,11 +126,12 @@ static ssize_t rxd_generic_write_inject(struct rxd_ep *rxd_ep,
 	ret = 0;
 
 out:
-	fastlock_release(&rxd_ep->util_ep.lock);
+	ofi_genlock_unlock(&rxd_ep->util_ep.lock);
 	return ret;
 }
 
-ssize_t rxd_generic_rma(struct rxd_ep *rxd_ep, const struct iovec *iov,
+static ssize_t
+rxd_generic_rma(struct rxd_ep *rxd_ep, const struct iovec *iov,
 	size_t iov_count, const struct fi_rma_iov *rma_iov, size_t rma_count,
 	void **desc, fi_addr_t addr, void *context, uint32_t op, uint64_t data,
 	uint32_t rxd_flags)
@@ -146,12 +147,12 @@ ssize_t rxd_generic_rma(struct rxd_ep *rxd_ep, const struct iovec *iov,
 
 	assert(iov_count <= RXD_IOV_LIMIT && rma_count <= RXD_IOV_LIMIT);
 
-	fastlock_acquire(&rxd_ep->util_ep.lock);
+	ofi_genlock_lock(&rxd_ep->util_ep.lock);
 
 	if (ofi_cirque_isfull(rxd_ep->util_ep.tx_cq->cirq))
 		goto out;
 	rxd_addr = (intptr_t) ofi_idx_lookup(&(rxd_ep_av(rxd_ep)->fi_addr_idx),
-					     RXD_IDX_OFFSET(addr));
+					     RXD_IDX_OFFSET((int) addr));
 	if (!rxd_addr)
 		goto out;
 
@@ -176,11 +177,12 @@ ssize_t rxd_generic_rma(struct rxd_ep *rxd_ep, const struct iovec *iov,
 	ret = 0;
 
 out:
-	fastlock_release(&rxd_ep->util_ep.lock);
+	ofi_genlock_unlock(&rxd_ep->util_ep.lock);
 	return ret;
 }
 
-ssize_t rxd_read(struct fid_ep *ep_fid, void *buf, size_t len, void *desc,
+static ssize_t
+rxd_read(struct fid_ep *ep_fid, void *buf, size_t len, void *desc,
 	fi_addr_t src_addr, uint64_t addr, uint64_t key, void *context)
 {
 	struct rxd_ep *ep;
@@ -195,12 +197,13 @@ ssize_t rxd_read(struct fid_ep *ep_fid, void *buf, size_t len, void *desc,
 	rma_iov.len = len;
 	rma_iov.key = key;
 
-	return rxd_generic_rma(ep, &msg_iov, 1, &rma_iov, 1, &desc, 
+	return rxd_generic_rma(ep, &msg_iov, 1, &rma_iov, 1, &desc,
 			       src_addr, context, RXD_READ_REQ, 0,
 			       ep->tx_flags);
 }
 
-ssize_t rxd_readv(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
+static ssize_t
+rxd_readv(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
 	size_t count, fi_addr_t src_addr, uint64_t addr, uint64_t key,
 	void *context)
 {
@@ -218,7 +221,8 @@ ssize_t rxd_readv(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
 			       ep->tx_flags);
 }
 
-ssize_t rxd_readmsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
+static ssize_t
+rxd_readmsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
 	uint64_t flags)
 {
 	struct rxd_ep *ep;
@@ -232,7 +236,8 @@ ssize_t rxd_readmsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
 			       ep->util_ep.tx_msg_flags));
 }
 
-ssize_t rxd_write(struct fid_ep *ep_fid, const void *buf, size_t len, void *desc,
+static ssize_t
+rxd_write(struct fid_ep *ep_fid, const void *buf, size_t len, void *desc,
 	fi_addr_t dest_addr, uint64_t addr, uint64_t key, void *context)
 {
 	struct rxd_ep *ep;
@@ -247,12 +252,13 @@ ssize_t rxd_write(struct fid_ep *ep_fid, const void *buf, size_t len, void *desc
 	rma_iov.len = len;
 	rma_iov.key = key;
 
-	return rxd_generic_rma(ep, &msg_iov, 1, &rma_iov, 1, &desc, 
+	return rxd_generic_rma(ep, &msg_iov, 1, &rma_iov, 1, &desc,
 			       dest_addr, context, RXD_WRITE, 0,
 			       ep->tx_flags);
 }
 
-ssize_t rxd_writev(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
+static ssize_t
+rxd_writev(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
 		size_t count, fi_addr_t dest_addr, uint64_t addr, uint64_t key,
 		void *context)
 {
@@ -270,8 +276,8 @@ ssize_t rxd_writev(struct fid_ep *ep_fid, const struct iovec *iov, void **desc,
 			       ep->tx_flags);
 }
 
-
-ssize_t rxd_writemsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
+static ssize_t
+rxd_writemsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
 	uint64_t flags)
 {
 	struct rxd_ep *ep;
@@ -285,7 +291,8 @@ ssize_t rxd_writemsg(struct fid_ep *ep_fid, const struct fi_msg_rma *msg,
 			       ep->util_ep.tx_msg_flags));
 }
 
-ssize_t rxd_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
+static ssize_t
+rxd_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
 		      void *desc, uint64_t data, fi_addr_t dest_addr,
 		      uint64_t addr, uint64_t key, void *context)
 {
@@ -306,7 +313,8 @@ ssize_t rxd_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
 			       ep->tx_flags | RXD_REMOTE_CQ_DATA);
 }
 
-ssize_t rxd_inject_write(struct fid_ep *ep_fid, const void *buf,
+static ssize_t
+rxd_inject_write(struct fid_ep *ep_fid, const void *buf,
 	size_t len, fi_addr_t dest_addr, uint64_t addr, uint64_t key)
 {
 	struct rxd_ep *rxd_ep;
@@ -326,9 +334,10 @@ ssize_t rxd_inject_write(struct fid_ep *ep_fid, const void *buf,
 					RXD_NO_TX_COMP | RXD_INJECT);
 }
 
-ssize_t rxd_inject_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
-			     uint64_t data, fi_addr_t dest_addr, uint64_t addr,
-			     uint64_t key)
+static ssize_t
+rxd_inject_writedata(struct fid_ep *ep_fid, const void *buf, size_t len,
+		     uint64_t data, fi_addr_t dest_addr, uint64_t addr,
+		     uint64_t key)
 {
 	struct rxd_ep *rxd_ep;
 	struct iovec iov;
