@@ -127,6 +127,8 @@ extern size_t rxm_packet_size;
 
 #define RXM_IOV_LIMIT 4
 
+#define RXM_PEER_XFER_TAG_FLAG	(1ULL << 63)
+
 #define RXM_MR_MODES	(OFI_MR_BASIC_MAP | FI_MR_LOCAL)
 
 #define RXM_PASSTHRU_TX_OP_FLAGS (FI_TRANSMIT_COMPLETE)
@@ -137,10 +139,10 @@ extern size_t rxm_packet_size;
 			 FI_DELIVERY_COMPLETE | FI_COMPLETION)
 #define RXM_RX_OP_FLAGS (FI_MULTI_RECV | FI_COMPLETION)
 
-#define RXM_MR_VIRT_ADDR(info) ((info->domain_attr->mr_mode == FI_MR_BASIC) ||\
+#define RXM_MR_VIRT_ADDR(info) ((info->domain_attr->mr_mode == OFI_MR_BASIC) ||\
 				info->domain_attr->mr_mode & FI_MR_VIRT_ADDR)
 
-#define RXM_MR_PROV_KEY(info) ((info->domain_attr->mr_mode == FI_MR_BASIC) ||\
+#define RXM_MR_PROV_KEY(info) ((info->domain_attr->mr_mode == OFI_MR_BASIC) ||\
 			       info->domain_attr->mr_mode & FI_MR_PROV_KEY)
 
 #define RXM_UPDATE_STATE(subsystem, buf, new_state)			\
@@ -181,9 +183,9 @@ do {									\
 extern struct fi_provider rxm_prov;
 extern struct util_prov rxm_util_prov;
 
-extern struct fi_ops_msg rxm_msg_ops;
+extern struct fi_ops_msg rxm_msg_ops, rxm_no_recv_msg_ops;
 extern struct fi_ops_msg rxm_msg_thru_ops;
-extern struct fi_ops_tagged rxm_tagged_ops;
+extern struct fi_ops_tagged rxm_tagged_ops, rxm_no_recv_tagged_ops;
 extern struct fi_ops_tagged rxm_tagged_thru_ops;
 extern struct fi_ops_rma rxm_rma_ops;
 extern struct fi_ops_rma rxm_rma_thru_ops;
@@ -203,6 +205,7 @@ extern size_t rxm_cq_eq_fairness;
 extern int rxm_passthru;
 extern int force_auto_progress;
 extern int rxm_use_write_rndv;
+extern int rxm_detect_hmem_iface;
 extern enum fi_wait_obj def_wait_obj, def_tcp_wait_obj;
 
 struct rxm_ep;
@@ -253,21 +256,41 @@ void rxm_freeall_conns(struct rxm_ep *ep);
 struct rxm_fabric {
 	struct util_fabric util_fabric;
 	struct fid_fabric *msg_fabric;
+	struct fi_info *util_coll_info;
+	struct fi_info *offload_coll_info;
+	struct fid_fabric *util_coll_fabric;
+	struct fid_fabric *offload_coll_fabric;
 };
 
 struct rxm_domain {
 	struct util_domain util_domain;
 	struct fid_domain *msg_domain;
+	struct fid_ep rx_ep;
+	struct fid_peer_srx *srx;
 	size_t max_atomic_size;
 	size_t rx_post_size;
 	uint64_t mr_key;
-	bool dyn_rbuf;
 	bool passthru;
 	struct ofi_ops_flow_ctrl *flow_ctrl_ops;
 	struct ofi_bufpool *amo_bufpool;
 	ofi_mutex_t amo_bufpool_lock;
+	struct fid_domain *util_coll_domain;
+	struct fid_domain *offload_coll_domain;
+	uint64_t offload_coll_mask;
 };
 
+struct rxm_cq {
+	struct util_cq util_cq;
+	struct fid_peer_cq peer_cq;
+	struct fid_cq *util_coll_cq;
+	struct fid_cq *offload_coll_cq;
+};
+
+struct rxm_eq {
+	struct util_eq util_eq;
+	struct fid_eq *util_coll_eq;
+	struct fid_eq *offload_coll_eq;
+};
 
 struct rxm_cntr {
 	struct util_cntr util_cntr;
@@ -283,15 +306,28 @@ struct rxm_mr {
 	struct rxm_domain *domain;
 	enum fi_hmem_iface iface;
 	uint64_t device;
+	void *hmem_handle;
+	uint64_t hmem_flags;
 	ofi_mutex_t amo_lock;
 };
 
 static inline enum fi_hmem_iface
-rxm_mr_desc_to_hmem_iface_dev(void **desc, size_t count, uint64_t *device)
+rxm_iov_desc_to_hmem_iface_dev(const struct iovec *iov, void **desc,
+			       size_t count, uint64_t *device)
 {
-	if (!count || !desc || !desc[0]) {
+	enum fi_hmem_iface iface = FI_HMEM_SYSTEM;
+
+	if (!count) {
 		*device = 0;
-		return FI_HMEM_SYSTEM;
+		return iface;
+	}
+
+	if (!desc || !desc[0]) {
+		if (rxm_detect_hmem_iface)
+			iface = ofi_get_hmem_iface(iov[0].iov_base, device, NULL);
+		else
+			*device = 0;
+		return iface;
 	}
 
 	*device = ((struct rxm_mr *) desc[0])->device;
@@ -383,13 +419,15 @@ struct rxm_pkt {
 	char data[];
 };
 
+enum rxm_sar_seg_type {
+	RXM_SAR_SEG_FIRST	= 1,
+	RXM_SAR_SEG_MIDDLE	= 2,
+	RXM_SAR_SEG_LAST	= 3,
+};
+
 union rxm_sar_ctrl_data {
 	struct {
-		enum rxm_sar_seg_type {
-			RXM_SAR_SEG_FIRST	= 1,
-			RXM_SAR_SEG_MIDDLE	= 2,
-			RXM_SAR_SEG_LAST	= 3,
-		} seg_type : 2;
+		enum rxm_sar_seg_type seg_type : 2;
 		uint32_t offset;
 	};
 	uint64_t align;
@@ -407,24 +445,29 @@ rxm_sar_set_seg_type(struct ofi_ctrl_hdr *ctrl_hdr, enum rxm_sar_seg_type seg_ty
 	((union rxm_sar_ctrl_data *)&(ctrl_hdr->ctrl_data))->seg_type = seg_type;
 }
 
-struct rxm_recv_match_attr {
-	fi_addr_t addr;
-	uint64_t tag;
-	uint64_t ignore;
-};
-
-struct rxm_unexp_msg {
-	struct dlist_entry entry;
-	fi_addr_t addr;
-	uint64_t tag;
-};
-
 struct rxm_iov {
 	struct iovec iov[RXM_IOV_LIMIT];
 	void *desc[RXM_IOV_LIMIT];
 	uint8_t count;
 };
 
+struct rxm_proto_info {
+        /* Used for SAR protocol */
+        struct {
+                struct dlist_entry entry;
+                struct dlist_entry pkt_list;
+                struct fi_peer_rx_entry *rx_entry;
+                size_t total_recv_len;
+                struct rxm_conn *conn;
+                uint64_t msg_id;
+        } sar;
+        /* Used for Rendezvous protocol */
+        struct {
+                /* This is used to send RNDV ACK */
+                struct rxm_tx_buf *tx_buf;
+        } rndv;
+};
+
 struct rxm_buf {
 	/* Must stay at top */
 	struct fi_context fi_context;
@@ -442,10 +485,10 @@ struct rxm_rx_buf {
 	/* MSG EP / shared context to which bufs would be posted to */
 	struct fid_ep *rx_ep;
 	struct dlist_entry repost_entry;
+	struct dlist_entry unexp_entry;
 	struct rxm_conn *conn;		/* msg ep data was received on */
-	/* if recv_entry is set, then we matched dyn rbuf */
-	struct rxm_recv_entry *recv_entry;
-	struct rxm_unexp_msg unexp_msg;
+	struct fi_peer_rx_entry *peer_entry;
+	struct rxm_proto_info *proto_info;
 	uint64_t comp_flags;
 	struct fi_recv_context recv_context;
 	bool repost;
@@ -492,10 +535,23 @@ struct rxm_tx_buf {
 	struct rxm_pkt pkt;
 };
 
+struct rxm_coll_buf {
+	/* Must stay at top */
+	struct rxm_buf hdr;
+
+	struct rxm_ep *ep;
+	void *app_context;
+	uint64_t flags;
+};
+
 /* Used for application transmits, provides credit check */
 struct rxm_tx_buf *rxm_get_tx_buf(struct rxm_ep *ep);
 void rxm_free_tx_buf(struct rxm_ep *ep, struct rxm_tx_buf *buf);
 
+/* Context for collective operations */
+struct rxm_coll_buf *rxm_get_coll_buf(struct rxm_ep *ep);
+void rxm_free_coll_buf(struct rxm_ep *ep, struct rxm_coll_buf *buf);
+
 enum rxm_deferred_tx_entry_type {
 	RXM_DEFERRED_TX_RNDV_ACK,
 	RXM_DEFERRED_TX_RNDV_DONE,
@@ -560,53 +616,6 @@ struct rxm_deferred_tx_entry {
 	};
 };
 
-struct rxm_recv_entry {
-	struct dlist_entry entry;
-	struct rxm_iov rxm_iov;
-	fi_addr_t addr;
-	void *context;
-	uint64_t flags;
-	uint64_t tag;
-	uint64_t ignore;
-	uint64_t comp_flags;
-	size_t total_len;
-	struct rxm_recv_queue *recv_queue;
-
-	/* Used for SAR protocol */
-	struct {
-		struct dlist_entry entry;
-		size_t total_recv_len;
-		struct rxm_conn *conn;
-		uint64_t msg_id;
-	} sar;
-	/* Used for Rendezvous protocol */
-	struct {
-		/* This is used to send RNDV ACK */
-		struct rxm_tx_buf *tx_buf;
-	} rndv;
-};
-OFI_DECLARE_FREESTACK(struct rxm_recv_entry, rxm_recv_fs);
-
-enum rxm_recv_queue_type {
-	RXM_RECV_QUEUE_UNSPEC,
-	RXM_RECV_QUEUE_MSG,
-	RXM_RECV_QUEUE_TAGGED,
-};
-
-struct rxm_recv_queue {
-	struct rxm_ep		*rxm_ep;
-	enum rxm_recv_queue_type type;
-	struct rxm_recv_fs	*fs;
-	struct dlist_entry	recv_list;
-	struct dlist_entry	unexp_msg_list;
-	size_t			dyn_rbuf_unexp_cnt;
-	dlist_func_t		*match_recv;
-	dlist_func_t		*match_unexp;
-};
-
-ssize_t rxm_get_dyn_rbuf(struct ofi_cq_rbuf_entry *entry, struct iovec *iov,
-			 size_t *count);
-
 struct rxm_eager_ops {
 	void (*comp_tx)(struct rxm_ep *rxm_ep,
 			struct rxm_tx_buf *tx_eager_buf);
@@ -639,7 +648,14 @@ struct rxm_ep {
 	pthread_t		cm_thread;
 	struct fid_pep 		*msg_pep;
 	struct fid_eq 		*msg_eq;
-	struct fid_ep 		*srx_ctx;
+	struct fid_ep 		*msg_srx;
+	struct fid_ep		*util_coll_ep;
+	struct fid_ep		*offload_coll_ep;
+	struct fi_ops_transfer_peer *util_coll_peer_xfer_ops;
+	struct fi_ops_transfer_peer *offload_coll_peer_xfer_ops;
+	uint64_t		offload_coll_mask;
+
+	struct fid_peer_srx	*srx;
 
 	struct fid_cq 		*msg_cq;
 	uint64_t		msg_cq_last_poll;
@@ -654,7 +670,6 @@ struct rxm_ep {
 	bool			do_progress;
 	bool			enable_direct_send;
 
-	size_t			min_multi_recv_size;
 	size_t			buffered_min;
 	size_t			buffered_limit;
 	size_t			inject_limit;
@@ -662,18 +677,18 @@ struct rxm_ep {
 	size_t			eager_limit;
 	size_t			sar_limit;
 	size_t			tx_credit;
+	size_t			min_multi_recv_size;
 
 	struct ofi_bufpool	*rx_pool;
 	struct ofi_bufpool	*tx_pool;
+	struct ofi_bufpool	*coll_pool;
+	struct ofi_bufpool	*proto_info_pool;
+
 	struct rxm_pkt		*inject_pkt;
 
 	struct dlist_entry	deferred_queue;
 	struct dlist_entry	rndv_wait_list;
 
-	struct rxm_recv_queue	recv_queue;
-	struct rxm_recv_queue	trecv_queue;
-	struct ofi_bufpool	*multi_recv_pool;
-
 	struct rxm_eager_ops	*eager_ops;
 	struct rxm_rndv_ops	*rndv_ops;
 };
@@ -698,17 +713,24 @@ int rxm_info_to_rxm(uint32_t version, const struct fi_info *core_info,
 		     const struct fi_info *base_info, struct fi_info *info);
 bool rxm_passthru_info(const struct fi_info *info);
 
+int rxm_eq_open(struct fid_fabric *fabric_fid, struct fi_eq_attr *attr,
+		struct fid_eq **eq_fid, void *context);
+
 int rxm_domain_open(struct fid_fabric *fabric, struct fi_info *info,
 			     struct fid_domain **dom, void *context);
 int rxm_cq_open(struct fid_domain *domain, struct fi_cq_attr *attr,
 			 struct fid_cq **cq_fid, void *context);
 ssize_t rxm_handle_rx_buf(struct rxm_rx_buf *rx_buf);
 
+int rxm_srx_context(struct fid_domain *domain, struct fi_rx_attr *attr,
+		    struct fid_ep **rx_ep, void *context);
+
 int rxm_endpoint(struct fid_domain *domain, struct fi_info *info,
 			  struct fid_ep **ep, void *context);
-
-void rxm_cq_write_error(struct util_cq *cq, struct util_cntr *cntr,
-			void *op_context, int err);
+void rxm_cq_write_tx_error(struct rxm_ep *rxm_ep, uint8_t op, void *op_context,
+			   int err);
+void rxm_cq_write_rx_error(struct rxm_ep *rxm_ep, uint8_t op, void *op_context,
+			   int err);
 void rxm_cq_write_error_all(struct rxm_ep *rxm_ep, int err);
 void rxm_handle_comp_error(struct rxm_ep *rxm_ep);
 ssize_t rxm_handle_comp(struct rxm_ep *rxm_ep, struct fi_cq_data_entry *comp);
@@ -736,6 +758,26 @@ void rxm_rndv_hdr_init(struct rxm_ep *rxm_ep, void *buf,
 			      const struct iovec *iov, size_t count,
 			      struct fid_mr **mr);
 
+ssize_t rxm_copy_hmem(void *desc, char *host_buf, void *dev_buf, size_t size,
+		      int dir);
+ssize_t rxm_copy_hmem_iov(void **desc, char *buf, size_t buf_size,
+			  const struct iovec *hmem_iov, int iov_count,
+			  size_t iov_offset, int dir);
+static inline ssize_t rxm_copy_from_hmem_iov(void **desc, char *buf,
+					     size_t buf_size,
+					     const struct iovec *hmem_iov,
+					     int iov_count, size_t iov_offset)
+{
+	return rxm_copy_hmem_iov(desc, buf, buf_size, hmem_iov, iov_count,
+				 iov_offset, OFI_COPY_IOV_TO_BUF);
+}
+static inline ssize_t rxm_copy_to_hmem_iov(void **desc, char *buf, int buf_size,
+					   const struct iovec *hmem_iov,
+					   int iov_count, size_t iov_offset)
+{
+	return rxm_copy_hmem_iov(desc, buf, buf_size, hmem_iov, iov_count,
+				 iov_offset, OFI_COPY_BUF_TO_IOV);
+}
 
 static inline size_t rxm_ep_max_atomic_size(struct fi_info *info)
 {
@@ -807,50 +849,6 @@ int rxm_msg_mr_reg_internal(struct rxm_domain *rxm_domain, const void *buf,
 			    size_t len, uint64_t acs, uint64_t flags,
 			    struct fid_mr **mr);
 
-static inline void rxm_cntr_incerr(struct util_cntr *cntr)
-{
-	if (cntr)
-		cntr->cntr_fid.ops->adderr(&cntr->cntr_fid, 1);
-}
-
-static inline void
-rxm_cq_write(struct util_cq *cq, void *context, uint64_t flags, size_t len,
-	     void *buf, uint64_t data, uint64_t tag)
-{
-	int ret;
-
-	FI_DBG(&rxm_prov, FI_LOG_CQ, "Reporting %s completion\n",
-	       fi_tostr((void *) &flags, FI_TYPE_CQ_EVENT_FLAGS));
-
-	ret = ofi_cq_write(cq, context, flags, len, buf, data, tag);
-	if (ret) {
-		FI_WARN(&rxm_prov, FI_LOG_CQ,
-			"Unable to report completion\n");
-		assert(0);
-	}
-	if (cq->wait)
-		cq->wait->signal(cq->wait);
-}
-
-static inline void
-rxm_cq_write_src(struct util_cq *cq, void *context, uint64_t flags, size_t len,
-		 void *buf, uint64_t data, uint64_t tag, fi_addr_t addr)
-{
-	int ret;
-
-	FI_DBG(&rxm_prov, FI_LOG_CQ, "Reporting %s completion\n",
-	       fi_tostr((void *) &flags, FI_TYPE_CQ_EVENT_FLAGS));
-
-	ret = ofi_cq_write_src(cq, context, flags, len, buf, data, tag, addr);
-	if (ret) {
-		FI_WARN(&rxm_prov, FI_LOG_CQ,
-			"Unable to report completion\n");
-		assert(0);
-	}
-	if (cq->wait)
-		cq->wait->signal(cq->wait);
-}
-
 ssize_t rxm_get_conn(struct rxm_ep *rxm_ep, fi_addr_t addr,
 		     struct rxm_conn **rxm_conn);
 
@@ -885,17 +883,10 @@ ssize_t
 rxm_inject_send(struct rxm_ep *rxm_ep, struct rxm_conn *rxm_conn,
 		const void *buf, size_t len);
 
-struct rxm_recv_entry *
-rxm_recv_entry_get(struct rxm_ep *rxm_ep, const struct iovec *iov,
-		   void **desc, size_t count, fi_addr_t src_addr,
-		   uint64_t tag, uint64_t ignore, void *context,
-		   uint64_t flags, struct rxm_recv_queue *recv_queue);
-struct rxm_rx_buf *
-rxm_get_unexp_msg(struct rxm_recv_queue *recv_queue, fi_addr_t addr,
-		  uint64_t tag, uint64_t ignore);
-ssize_t rxm_handle_unexp_sar(struct rxm_recv_queue *recv_queue,
-			     struct rxm_recv_entry *recv_entry,
-			     struct rxm_rx_buf *rx_buf);
+ssize_t rxm_handle_unexp_sar(struct fi_peer_rx_entry *peer_entry);
+int rxm_srx_context(struct fid_domain *domain, struct fi_rx_attr *attr,
+		    struct fid_ep **rx_ep, void *context);
+
 int rxm_post_recv(struct rxm_rx_buf *rx_buf);
 void rxm_av_remove_handler(struct util_ep *util_ep,
 			   struct util_peer_addr *peer);
@@ -909,37 +900,13 @@ rxm_free_rx_buf(struct rxm_rx_buf *rx_buf)
 	}
 
 	/* Discard rx buffer if its msg_ep was closed */
-	if (rx_buf->repost && (rx_buf->ep->srx_ctx || rx_buf->conn->msg_ep)) {
+	if (rx_buf->repost && (rx_buf->ep->msg_srx || rx_buf->conn->msg_ep)) {
 		rxm_post_recv(rx_buf);
 	} else {
 		ofi_buf_free(rx_buf);
 	}
 }
 
-static inline void
-rxm_recv_entry_release(struct rxm_recv_entry *entry)
-{
-	if (entry->recv_queue)
-		ofi_freestack_push(entry->recv_queue->fs, entry);
-	else
-		ofi_buf_free(entry);
-}
-
-static inline void
-rxm_cq_write_recv_comp(struct rxm_rx_buf *rx_buf, void *context, uint64_t flags,
-		       size_t len, char *buf)
-{
-	if (rx_buf->ep->rxm_info->caps & FI_SOURCE)
-		rxm_cq_write_src(rx_buf->ep->util_ep.rx_cq, context,
-				 flags, len, buf, rx_buf->pkt.hdr.data,
-				 rx_buf->pkt.hdr.tag,
-				 rx_buf->conn->peer->fi_addr);
-	else
-		rxm_cq_write(rx_buf->ep->util_ep.rx_cq, context,
-			     flags, len, buf, rx_buf->pkt.hdr.data,
-			     rx_buf->pkt.hdr.tag);
-}
-
 struct rxm_mr *rxm_mr_get_map_entry(struct rxm_domain *domain, uint64_t key);
 
 struct rxm_recv_entry *
