@@ -38,7 +38,6 @@
 
 #include <ofi_prov.h>
 #include "rxm.h"
-#include "ofi_coll.h"
 
 #define RXM_ATOMIC_UNSUPPORTED_MSG_ORDER (FI_ORDER_RAW | FI_ORDER_RAR |  \
 					  FI_ORDER_WAW | FI_ORDER_WAR |  \
@@ -59,6 +58,7 @@ size_t rxm_packet_size;
 int rxm_passthru = 0; /* disable by default, need to analyze performance */
 int force_auto_progress;
 int rxm_use_write_rndv;
+int rxm_detect_hmem_iface;
 enum fi_wait_obj def_wait_obj = FI_WAIT_FD, def_tcp_wait_obj = FI_WAIT_UNSPEC;
 
 char *rxm_proto_state_str[] = {
@@ -84,12 +84,12 @@ void rxm_info_to_core_mr_modes(uint32_t version, const struct fi_info *hints,
 			       struct fi_info *core_info)
 {
 	if (hints && hints->domain_attr &&
-	    (hints->domain_attr->mr_mode & (FI_MR_SCALABLE | FI_MR_BASIC))) {
-		core_info->mode |= FI_LOCAL_MR;
+	    (hints->domain_attr->mr_mode & (OFI_MR_SCALABLE | OFI_MR_BASIC))) {
+		core_info->mode |= OFI_LOCAL_MR;
 		core_info->domain_attr->mr_mode = hints->domain_attr->mr_mode;
 	} else if (FI_VERSION_LT(version, FI_VERSION(1, 5))) {
-		core_info->mode |= FI_LOCAL_MR;
-		core_info->domain_attr->mr_mode = FI_MR_UNSPEC;
+		core_info->mode |= OFI_LOCAL_MR;
+		core_info->domain_attr->mr_mode = OFI_MR_UNSPEC;
 	} else {
 		core_info->domain_attr->mr_mode |= FI_MR_LOCAL;
 		if (!hints || !hints->domain_attr ||
@@ -233,13 +233,11 @@ int rxm_info_to_core(uint32_t version, const struct fi_info *hints,
 			core_info->tx_attr->op_flags =
 				hints->tx_attr->op_flags & RXM_PASSTHRU_TX_OP_FLAGS;
 			core_info->tx_attr->msg_order = hints->tx_attr->msg_order;
-			core_info->tx_attr->comp_order = hints->tx_attr->comp_order;
 		}
 		if (hints->rx_attr) {
 			core_info->rx_attr->op_flags =
 				hints->rx_attr->op_flags & RXM_PASSTHRU_RX_OP_FLAGS;
 			core_info->rx_attr->msg_order = hints->rx_attr->msg_order;
-			core_info->rx_attr->comp_order = hints->rx_attr->comp_order;
 		}
 		if ((hints->caps & FI_HMEM) && ofi_hmem_p2p_disabled())
 			return -FI_ENODATA;
@@ -264,6 +262,9 @@ int rxm_info_to_core(uint32_t version, const struct fi_info *hints,
 
 	core_info->rx_attr->op_flags &= ~FI_MULTI_RECV;
 
+	core_info->domain_attr->caps &= ~(FI_AV_USER_ID | FI_PEER);
+	core_info->caps &= ~(FI_AV_USER_ID | FI_PEER);
+
 	return 0;
 }
 
@@ -275,11 +276,9 @@ rxm_info_thru_rxm(uint32_t version, const struct fi_info *core_info,
 	info->mode = core_info->mode;
 
 	*info->tx_attr = *core_info->tx_attr;
-	info->tx_attr->comp_order = base_info->tx_attr->comp_order;
 	info->tx_attr->size = MIN(base_info->tx_attr->size, rxm_def_tx_size);
 
 	*info->rx_attr = *core_info->rx_attr;
-	info->rx_attr->comp_order = base_info->rx_attr->comp_order;
 	info->rx_attr->size = MIN(base_info->rx_attr->size, rxm_def_rx_size);
 
 	*info->ep_attr = *base_info->ep_attr;
@@ -321,7 +320,6 @@ int rxm_info_to_rxm(uint32_t version, const struct fi_info *core_info,
 	info->tx_attr->caps		= base_info->tx_attr->caps;
 	info->tx_attr->mode		= info->mode;
 	info->tx_attr->msg_order 	= core_info->tx_attr->msg_order;
-	info->tx_attr->comp_order 	= base_info->tx_attr->comp_order;
 
 	/* If the core provider requires registering send buffers, it's
 	 * usually faster to copy small transfer through bounce buffers
@@ -356,7 +354,6 @@ int rxm_info_to_rxm(uint32_t version, const struct fi_info *core_info,
 	info->rx_attr->caps		= base_info->rx_attr->caps;
 	info->rx_attr->mode		= info->rx_attr->mode & ~FI_RX_CQ_DATA;
 	info->rx_attr->msg_order 	= core_info->rx_attr->msg_order;
-	info->rx_attr->comp_order 	= base_info->rx_attr->comp_order;
 	info->rx_attr->iov_limit 	= MIN(base_info->rx_attr->iov_limit,
 					      core_info->rx_attr->iov_limit);
 
@@ -461,13 +458,9 @@ static void rxm_alter_info(const struct fi_info *hints, struct fi_info *info)
 				cur->caps &= ~FI_DIRECTED_RECV;
 				cur->rx_attr->caps &= ~FI_DIRECTED_RECV;
 			}
-			if (!(hints->caps & FI_SOURCE)) {
-				cur->caps &= ~FI_SOURCE;
-				cur->rx_attr->caps &= ~FI_SOURCE;
-			}
 
-			if (hints->mode & FI_BUFFERED_RECV)
-				cur->mode |= FI_BUFFERED_RECV;
+			if (hints->mode & OFI_BUFFERED_RECV)
+				cur->mode |= OFI_BUFFERED_RECV;
 
 			if (hints->caps & FI_ATOMIC) {
 				cur->tx_attr->msg_order &=
@@ -484,9 +477,9 @@ static void rxm_alter_info(const struct fi_info *hints, struct fi_info *info)
 			}
 
 			if (!ofi_mr_local(hints)) {
-				cur->mode &= ~FI_LOCAL_MR;
-				cur->tx_attr->mode &= ~FI_LOCAL_MR;
-				cur->rx_attr->mode &= ~FI_LOCAL_MR;
+				cur->mode &= ~OFI_LOCAL_MR;
+				cur->tx_attr->mode &= ~OFI_LOCAL_MR;
+				cur->rx_attr->mode &= ~OFI_LOCAL_MR;
 				cur->domain_attr->mr_mode &= ~FI_MR_LOCAL;
 			}
 
@@ -630,17 +623,6 @@ RXM_INI
 			"typically used as the eager message size. "
 			"(default %zu)", rxm_buffer_size);
 
-	fi_param_define(&rxm_prov, "eager_limit", FI_PARAM_SIZE_T,
-			"Specifies the maximum size transfer that the eager "
-			"protocol will be used.  For transfers smaller than "
-			"this limit, data may be copied into a bounce "
-			"buffer on the transmit side and received into "
-			"bounce buffer at the receiver.  The eager_limit must "
-			"be equal to the buffer_size when using rxm over "
-			"verbs, but may differ in the case of tcp."
-			"(default: %zu)", rxm_buffer_size);
-			/* rxm_buffer_size is correct here */
-
 	fi_param_define(&rxm_prov, "comp_per_progress", FI_PARAM_INT,
 			"Defines the maximum number of MSG provider CQ entries "
 			"(default: 1) that would be read per progress "
@@ -700,14 +682,6 @@ RXM_INI
 			"RMA writes rather than RMA reads during Rendezvous "
 			"transactions. (default: false/no).");
 
-	fi_param_define(&rxm_prov, "enable_dyn_rbuf", FI_PARAM_BOOL,
-			"Enable support for dynamic receive buffering, if "
-			"available by the message endpoint provider. "
-			"This allows direct placement of received messages "
-			"into application buffers, bypassing RxM bounce "
-			"buffers.  This feature targets using tcp sockets "
-			"for the message transport.  (default: true)");
-
 	fi_param_define(&rxm_prov, "enable_direct_send", FI_PARAM_BOOL,
 			"Enable support to pass application buffers directly "
 			"to the core provider when possible.  This avoids "
@@ -724,6 +698,11 @@ RXM_INI
 			"to the tcp provider, depending on the capabilities "
 			"requested by the application.");
 
+	fi_param_define(&rxm_prov, "detect_hmem_iface", FI_PARAM_BOOL,
+			"Detect iface for user buffers with NULL desc passed "
+			"in. This allows such buffers be copied or registered "
+			"internally by RxM. (default: false).");
+
 	/* passthru supported disabled - to re-enable would need to fix call to
 	 * fi_cq_read to pass in the correct data structure.  However, passthru
 	 * will not be needed at all with in-work tcp changes.
@@ -749,6 +728,8 @@ RXM_INI
 			"(FI_OFI_RXM_DATA_AUTO_PROGRESS = 1), domain threading "
 			"level would be set to FI_THREAD_SAFE\n");
 
+	fi_param_get_bool(&rxm_prov, "detect_hmem_iface", &rxm_detect_hmem_iface);
+
 #if HAVE_RXM_DL
 	ofi_mem_init();
 	ofi_hmem_init();
