@@ -51,11 +51,11 @@
 #include <rdma/fi_tagged.h>
 #include <rdma/fi_trigger.h>
 #include <rdma/providers/fi_prov.h>
-#include <rdma/fi_ext.h>
+#include <rdma/providers/fi_peer.h>
 
 #include <ofi.h>
 #include <ofi_enosys.h>
-#include <ofi_shm.h>
+#include <ofi_shm_p2p.h>
 #include <ofi_rbuf.h>
 #include <ofi_list.h>
 #include <ofi_signal.h>
@@ -65,6 +65,9 @@
 #include <ofi_iov.h>
 #include <ofi_mr.h>
 #include <ofi_lock.h>
+#include <ofi_hmem.h>
+
+#include "smr_util.h"
 
 #ifndef _SMR_H_
 #define _SMR_H_
@@ -73,6 +76,8 @@ struct smr_env {
 	size_t sar_threshold;
 	int disable_cma;
 	int use_dsa_sar;
+	size_t max_gdrcopy_size;
+	int use_xpmem;
 };
 
 extern struct smr_env smr_env;
@@ -86,7 +91,7 @@ int smr_fabric(struct fi_fabric_attr *attr, struct fid_fabric **fabric,
 
 struct smr_av {
 	struct util_av		util_av;
-	struct smr_map		*smr_map;
+	struct smr_map		smr_map;
 	size_t			used;
 };
 
@@ -109,18 +114,6 @@ int smr_query_atomic(struct fid_domain *domain, enum fi_datatype datatype,
 
 #define SMR_IOV_LIMIT		4
 
-struct smr_rx_entry {
-	struct fi_peer_rx_entry	peer_entry;
-	struct iovec		iov[SMR_IOV_LIMIT];
-	void			*desc[SMR_IOV_LIMIT];
-	int64_t			peer_id;
-	uint64_t		ignore;
-	int			multi_recv_ref;
-	uint64_t		err;
-	enum fi_hmem_iface	iface;
-	uint64_t		device;
-};
-
 struct smr_tx_entry {
 	struct smr_cmd	cmd;
 	int64_t		peer_id;
@@ -129,85 +122,34 @@ struct smr_tx_entry {
 	uint32_t	iov_count;
 	uint64_t	op_flags;
 	size_t		bytes_done;
-	int		next;
 	void		*map_ptr;
 	struct smr_ep_name *map_name;
-	enum fi_hmem_iface	iface;
-	uint64_t		device;
-	int			fd;
+	struct ofi_mr	*mr[SMR_IOV_LIMIT];
 };
 
-struct smr_sar_entry {
+struct smr_pend_entry {
 	struct dlist_entry	entry;
 	struct smr_cmd		cmd;
 	struct fi_peer_rx_entry	*rx_entry;
+	struct smr_cmd_ctx	*cmd_ctx;
 	size_t			bytes_done;
-	int			next;
 	struct iovec		iov[SMR_IOV_LIMIT];
 	size_t			iov_count;
-	enum fi_hmem_iface	iface;
-	uint64_t		device;
+	struct ofi_mr		*mr[SMR_IOV_LIMIT];
+	struct ofi_mr_entry	*ipc_entry;
+	ofi_hmem_async_event_t	async_event;
 };
 
-struct smr_cq {
-	struct util_cq util_cq;
-	struct fid_peer_cq *peer_cq;
-};
-
-typedef int (*smr_rx_comp_func)(struct smr_cq *cq, void *context,
-		uint64_t flags, size_t len, void *buf,
-		fi_addr_t fi_addr, uint64_t tag, uint64_t data);
-
-struct smr_match_attr {
-	fi_addr_t	id;
-	uint64_t	tag;
-	uint64_t	ignore;
-};
-
-static inline int smr_match_id(fi_addr_t addr, fi_addr_t match_addr)
-{
-	return (addr == FI_ADDR_UNSPEC) || (match_addr == FI_ADDR_UNSPEC) ||
-		(addr == match_addr);
-}
-
-static inline int smr_match_tag(uint64_t tag, uint64_t ignore, uint64_t match_tag)
-{
-	return ((tag | ignore) == (match_tag | ignore));
-}
-
-static inline enum fi_hmem_iface smr_get_mr_hmem_iface(struct util_domain *domain,
-				void **desc, uint64_t *device)
-{
-	if (!(domain->mr_mode & FI_MR_HMEM) || !desc || !*desc) {
-		*device = 0;
-		return FI_HMEM_SYSTEM;
-	}
-
-	*device = ((struct ofi_mr *) *desc)->device;
-	return ((struct ofi_mr *) *desc)->iface;
-}
-
-static inline uint64_t smr_get_mr_flags(void **desc)
-{
-	assert(desc && *desc);
-	return ((struct ofi_mr *) *desc)->flags;
-}
-
 struct smr_cmd_ctx {
 	struct dlist_entry entry;
+	struct fi_peer_rx_entry *rx_entry;
 	struct smr_ep *ep;
 	struct smr_cmd cmd;
+	struct smr_pend_entry *sar_entry;
+	struct slist buf_list;
 };
 
-OFI_DECLARE_FREESTACK(struct smr_rx_entry, smr_recv_fs);
-OFI_DECLARE_FREESTACK(struct smr_cmd_ctx, smr_cmd_ctx_fs);
-OFI_DECLARE_FREESTACK(struct smr_tx_entry, smr_pend_fs);
-OFI_DECLARE_FREESTACK(struct smr_sar_entry, smr_sar_fs);
-
-struct smr_queue {
-	struct dlist_entry list;
-	dlist_func_t *match_func;
-};
+OFI_DECLARE_FREESTACK(struct smr_tx_entry, smr_tx_fs);
 
 struct smr_fabric {
 	struct util_fabric	util_fabric;
@@ -218,14 +160,13 @@ struct smr_domain {
 	int			fast_rma;
 	/* cache for use with hmem ipc */
 	struct ofi_mr_cache	*ipc_cache;
+	struct fid_ep		rx_ep;
 	struct fid_peer_srx	*srx;
 };
 
 #define SMR_PREFIX	"fi_shm://"
 #define SMR_PREFIX_NS	"fi_ns://"
 
-#define SMR_ZE_SOCK_PATH	"/dev/shm/ze_"
-
 #define SMR_RMA_ORDER (OFI_ORDER_RAR_SET | OFI_ORDER_RAW_SET | FI_ORDER_RAS |	\
 		       OFI_ORDER_WAR_SET | OFI_ORDER_WAW_SET | FI_ORDER_WAS |	\
 		       FI_ORDER_SAR | FI_ORDER_SAW)
@@ -255,7 +196,7 @@ enum smr_cmap_state {
 
 struct smr_cmap_entry {
 	enum smr_cmap_state	state;
-	int			device_fds[ZE_MAX_DEVICES];
+	int			*device_fds;
 };
 
 struct smr_sock_info {
@@ -269,58 +210,36 @@ struct smr_sock_info {
 	struct smr_cmap_entry	peers[SMR_MAX_PEERS];
 };
 
-struct smr_srx_ctx {
-	struct fid_peer_srx	peer_srx;
-	struct smr_queue	recv_queue;
-	struct smr_queue	trecv_queue;
-	bool			dir_recv;
-	size_t			min_multi_recv_size;
-	uint64_t		rx_op_flags;
-	uint64_t		rx_msg_flags;
-
-	struct smr_cq		*cq;
-	struct smr_queue	unexp_msg_queue;
-	struct smr_queue	unexp_tagged_queue;
-	struct smr_recv_fs	*recv_fs;
-	ofi_spin_t		lock;
+struct smr_unexp_buf {
+	struct slist_entry entry;
+	char buf[SMR_SAR_SIZE];
 };
 
-struct smr_rx_entry *smr_alloc_rx_entry(struct smr_srx_ctx *srx);
-
 struct smr_ep {
 	struct util_ep		util_ep;
-	smr_rx_comp_func	rx_comp;
 	size_t			tx_size;
 	size_t			rx_size;
 	const char		*name;
 	uint64_t		msg_id;
 	struct smr_region	*volatile region;
-	//if double locking is needed, shm region lock must
-	//be acquired before any shm EP locks
-	ofi_spin_t		tx_lock;
-
-	struct fid_ep		*srx;
-	struct smr_cmd_ctx_fs	*cmd_ctx_fs;
-	struct smr_pend_fs	*pend_fs;
-	struct smr_sar_fs	*sar_fs;
+	struct fid_peer_srx	*srx;
+	struct ofi_bufpool	*cmd_ctx_pool;
+	struct ofi_bufpool	*unexp_buf_pool;
+	struct ofi_bufpool	*pend_buf_pool;
 
+	struct smr_tx_fs	*tx_fs;
 	struct dlist_entry	sar_list;
+	struct dlist_entry	ipc_cpy_pend_list;
+	struct dlist_entry	unexp_cmd_list;
+	size_t			min_multi_recv_size;
 
 	int			ep_idx;
+	enum ofi_shm_p2p_type	p2p_type;
 	struct smr_sock_info	*sock_info;
 	void			*dsa_context;
+	void 			(*smr_progress_ipc_list)(struct smr_ep *ep);
 };
 
-static inline struct smr_srx_ctx *smr_get_smr_srx(struct smr_ep *ep)
-{
-	return (struct smr_srx_ctx *) ep->srx->fid.context;
-}
-
-static inline struct fid_peer_srx *smr_get_peer_srx(struct smr_ep *ep)
-{
-	return container_of(ep->srx, struct fid_peer_srx, ep_fid);
-}
-
 #define smr_ep_rx_flags(smr_ep) ((smr_ep)->util_ep.rx_op_flags)
 #define smr_ep_tx_flags(smr_ep) ((smr_ep)->util_ep.tx_op_flags)
 
@@ -331,9 +250,6 @@ static inline int smr_mmap_name(char *shm_name, const char *ep_name,
 			ep_name, msg_id);
 }
 
-int smr_srx_context(struct fid_domain *domain, struct fi_rx_attr *attr,
-		struct fid_ep **rx_ep, void *context);
-
 int smr_endpoint(struct fid_domain *domain, struct fi_info *info,
 		  struct fid_ep **ep, void *context);
 void smr_ep_exchange_fds(struct smr_ep *ep, int64_t id);
@@ -346,70 +262,71 @@ int smr_cntr_open(struct fid_domain *domain, struct fi_cntr_attr *attr,
 int64_t smr_verify_peer(struct smr_ep *ep, fi_addr_t fi_addr);
 
 void smr_format_pend_resp(struct smr_tx_entry *pend, struct smr_cmd *cmd,
-			  void *context, enum fi_hmem_iface iface, uint64_t device,
+			  void *context, struct ofi_mr **mr,
 			  const struct iovec *iov, uint32_t iov_count,
 			  uint64_t op_flags, int64_t id, struct smr_resp *resp);
 void smr_generic_format(struct smr_cmd *cmd, int64_t peer_id, uint32_t op,
 			uint64_t tag, uint64_t data, uint64_t op_flags);
 size_t smr_copy_to_sar(struct smr_freestack *sar_pool, struct smr_resp *resp,
-		       struct smr_cmd *cmd, enum fi_hmem_iface, uint64_t device,
+		       struct smr_cmd *cmd, struct ofi_mr **mr,
 		       const struct iovec *iov, size_t count,
-		       size_t *bytes_done, int *next);
+		       size_t *bytes_done);
 size_t smr_copy_from_sar(struct smr_freestack *sar_pool, struct smr_resp *resp,
-			 struct smr_cmd *cmd, enum fi_hmem_iface iface,
-			 uint64_t device, const struct iovec *iov, size_t count,
-			 size_t *bytes_done, int *next);
-
-int smr_select_proto(bool use_ipc, bool cma_avail, enum fi_hmem_iface iface,
-		     uint32_t op, uint64_t total_len, uint64_t op_flags);
+			 struct smr_cmd *cmd, struct ofi_mr **mr,
+			 const struct iovec *iov, size_t count,
+			 size_t *bytes_done);
+int smr_select_proto(void **desc, size_t iov_count, bool cma_avail,
+		     bool ipc_valid, uint32_t op, uint64_t total_len,
+		     uint64_t op_flags);
 typedef ssize_t (*smr_proto_func)(struct smr_ep *ep, struct smr_region *peer_smr,
 		int64_t id, int64_t peer_id, uint32_t op, uint64_t tag,
-		uint64_t data, uint64_t op_flags, enum fi_hmem_iface iface,
-		uint64_t device, const struct iovec *iov, size_t iov_count,
-		size_t total_len, void *context);
+		uint64_t data, uint64_t op_flags, struct ofi_mr **desc,
+		const struct iovec *iov, size_t iov_count, size_t total_len,
+		void *context, struct smr_cmd *cmd);
 extern smr_proto_func smr_proto_ops[smr_src_max];
 
 int smr_write_err_comp(struct util_cq *cq, void *context,
-		       uint64_t flags, uint64_t tag, uint64_t err);
+		       uint64_t flags, uint64_t tag, int err);
 int smr_complete_tx(struct smr_ep *ep, void *context, uint32_t op,
 		    uint64_t flags);
 int smr_complete_rx(struct smr_ep *ep, void *context, uint32_t op,
 		    uint64_t flags, size_t len, void *buf, int64_t id,
 		    uint64_t tag, uint64_t data);
-int smr_rx_comp(struct smr_cq *cq, void *context, uint64_t flags, size_t len,
-		void *buf, fi_addr_t fi_addr, uint64_t tag, uint64_t data);
-int smr_rx_src_comp(struct smr_cq *cq, void *context, uint64_t flags,
-		    size_t len, void *buf, fi_addr_t fi_addr, uint64_t tag,
-		    uint64_t data);
 
-static inline uint64_t smr_rx_cq_flags(uint32_t op, uint64_t rx_flags,
-				       uint16_t op_flags)
+static inline uint64_t smr_rx_cq_flags(uint64_t rx_flags, uint16_t op_flags)
 {
-	rx_flags |= ofi_rx_cq_flags(op);
 	if (op_flags & SMR_REMOTE_CQ_DATA)
 		rx_flags |= FI_REMOTE_CQ_DATA;
 	return rx_flags;
 }
 
-
-bool smr_adjust_multi_recv(struct smr_srx_ctx *srx,
-			   struct fi_peer_rx_entry *rx_entry, size_t len);
-void smr_init_rx_entry(struct smr_rx_entry *entry, const struct iovec *iov,
-		       void **desc, size_t count, fi_addr_t addr,
-		       void *context, uint64_t tag, uint64_t flags);
-struct smr_rx_entry *smr_get_recv_entry(struct smr_srx_ctx *srx,
-		const struct iovec *iov, void **desc, size_t count, fi_addr_t addr,
-		void *context, uint64_t tag, uint64_t ignore, uint64_t flags);
-
 void smr_ep_progress(struct util_ep *util_ep);
 
-static inline bool smr_cma_enabled(struct smr_ep *ep,
+static inline bool smr_vma_enabled(struct smr_ep *ep,
 				   struct smr_region *peer_smr)
 {
 	if (ep->region == peer_smr)
-		return ep->region->cma_cap_self == SMR_CMA_CAP_ON;
+		return (ep->region->cma_cap_self == SMR_VMA_CAP_ON ||
+			ep->region->xpmem_cap_self == SMR_VMA_CAP_ON);
 	else
-		return ep->region->cma_cap_peer == SMR_CMA_CAP_ON;
+		return (ep->region->cma_cap_peer == SMR_VMA_CAP_ON ||
+			peer_smr->xpmem_cap_self == SMR_VMA_CAP_ON);
+}
+
+static inline void smr_set_ipc_valid(struct smr_region *region, uint64_t id)
+{
+	if (ofi_hmem_is_initialized(FI_HMEM_ZE) &&
+	    region->map->peers[id].pid_fd == -1)
+		smr_peer_data(region)[id].ipc_valid = 0;
+        else
+        	smr_peer_data(region)[id].ipc_valid = 1;
+}
+
+static inline bool smr_ipc_valid(struct smr_ep *ep, struct smr_region *peer_smr,
+				 int64_t id, int64_t peer_id)
+{
+	return (smr_peer_data(ep->region)[id].ipc_valid &&
+		smr_peer_data(peer_smr)[peer_id].ipc_valid);
 }
 
 static inline bool smr_ze_ipc_enabled(struct smr_region *smr,
@@ -419,34 +336,35 @@ static inline bool smr_ze_ipc_enabled(struct smr_region *smr,
 	       (peer_smr->flags & SMR_FLAG_IPC_SOCK);
 }
 
-static inline int smr_cma_loop(pid_t pid, struct iovec *local,
-			unsigned long local_cnt, struct iovec *remote,
-			unsigned long remote_cnt, unsigned long flags,
-			size_t total, bool write)
+static inline struct smr_inject_buf *
+smr_get_txbuf(struct smr_region *smr)
 {
-	ssize_t ret;
+	struct smr_inject_buf *txbuf;
 
-	while (1) {
-		if (write)
-			ret = ofi_process_vm_writev(pid, local, local_cnt, remote,
-						    remote_cnt, flags);
-		else
-			ret = ofi_process_vm_readv(pid, local, local_cnt, remote,
-						   remote_cnt, flags);
-		if (ret < 0) {
-			FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
-				"CMA error %d\n", errno);
-			return -FI_EIO;
-		}
+	pthread_spin_lock(&smr->lock);
+	if (!smr_freestack_isempty(smr_inject_pool(smr)))
+		txbuf = smr_freestack_pop(smr_inject_pool(smr));
+	else
+		txbuf = NULL;
+	pthread_spin_unlock(&smr->lock);
+	return txbuf;
+}
 
-		total -= ret;
-		if (!total)
-			return FI_SUCCESS;
-
-		ofi_consume_iov(local, &local_cnt, (size_t) ret);
-		ofi_consume_iov(remote, &remote_cnt, (size_t) ret);
-	}
+static inline void
+smr_release_txbuf(struct smr_region *smr,
+		  struct smr_inject_buf *tx_buf)
+{
+	pthread_spin_lock(&smr->lock);
+	smr_freestack_push(smr_inject_pool(smr), tx_buf);
+	pthread_spin_unlock(&smr->lock);
 }
 
 int smr_unexp_start(struct fi_peer_rx_entry *rx_entry);
+
+void smr_progress_ipc_list(struct smr_ep *ep);
+static inline void smr_progress_ipc_list_noop(struct smr_ep *ep)
+{
+	// noop
+}
+
 #endif
