@@ -35,7 +35,6 @@
 #include <sys/uio.h>
 
 #include "ofi_iov.h"
-#include "ofi_hmem.h"
 #include "smr.h"
 
 
@@ -53,41 +52,33 @@ static void smr_generic_atomic_format(struct smr_cmd *cmd, uint8_t datatype,
 	cmd->msg.hdr.atomic_op = atomic_op;
 }
 
-static void smr_format_inline_atomic(struct smr_cmd *cmd,
-				     enum fi_hmem_iface iface, uint64_t device,
+static void smr_format_inline_atomic(struct smr_cmd *cmd, struct ofi_mr **mr,
 				     const struct iovec *iov, size_t count)
 {
 	cmd->msg.hdr.op_src = smr_src_inline;
 
-	cmd->msg.hdr.size = ofi_copy_from_hmem_iov(cmd->msg.data.msg,
-					SMR_MSG_DATA_LEN, iface, device,
-					iov, count, 0);
+	cmd->msg.hdr.size = ofi_copy_from_mr_iov(cmd->msg.data.msg,
+						 SMR_MSG_DATA_LEN, mr,
+						 iov, count, 0);
 }
 
 static void smr_do_atomic_inline(struct smr_ep *ep, struct smr_region *peer_smr,
 			int64_t id, int64_t peer_id, uint32_t op,
-			uint64_t op_flags, enum fi_hmem_iface iface,
-			uint64_t device, uint8_t datatype, uint8_t atomic_op,
-			const struct iovec *iov, size_t iov_count,
-			size_t total_len)
+			uint64_t op_flags, uint8_t datatype, uint8_t atomic_op,
+			struct ofi_mr **desc, const struct iovec *iov,
+			size_t iov_count, size_t total_len, struct smr_cmd *cmd)
 {
-	struct smr_cmd *cmd;
-
-	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
 	smr_generic_format(cmd, peer_id, op, 0, 0, op_flags);
 	smr_generic_atomic_format(cmd, datatype, atomic_op);
-	smr_format_inline_atomic(cmd, iface, device, iov, iov_count);
-
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
+	smr_format_inline_atomic(cmd, desc, iov, iov_count);
 }
 
-static void smr_format_inject_atomic(struct smr_cmd *cmd,
-			enum fi_hmem_iface iface, uint64_t device,
+static void smr_format_inject_atomic(struct smr_cmd *cmd, struct ofi_mr **desc,
 			const struct iovec *iov, size_t count,
 			const struct iovec *resultv, size_t result_count,
-			const struct iovec *compv, size_t comp_count,
-			struct smr_region *smr, struct smr_inject_buf *tx_buf)
+			struct ofi_mr **comp_desc, const struct iovec *compv,
+			size_t comp_count, struct smr_region *smr,
+			struct smr_inject_buf *tx_buf)
 {
 	size_t comp_size;
 
@@ -96,72 +87,72 @@ static void smr_format_inject_atomic(struct smr_cmd *cmd,
 
 	switch (cmd->msg.hdr.op) {
 	case ofi_op_atomic:
-		cmd->msg.hdr.size = ofi_copy_from_hmem_iov(tx_buf->data,
-					SMR_INJECT_SIZE, iface, device,
-					iov, count, 0);
+		cmd->msg.hdr.size = ofi_copy_from_mr_iov(tx_buf->data,
+					SMR_INJECT_SIZE, desc, iov, count, 0);
 		break;
 	case ofi_op_atomic_fetch:
 		if (cmd->msg.hdr.atomic_op == FI_ATOMIC_READ)
 			cmd->msg.hdr.size = ofi_total_iov_len(resultv, result_count);
 		else
-			cmd->msg.hdr.size = ofi_copy_from_hmem_iov(tx_buf->data,
-						SMR_INJECT_SIZE, iface, device,
-						iov, count, 0);
+			cmd->msg.hdr.size = ofi_copy_from_mr_iov(tx_buf->data,
+						SMR_INJECT_SIZE, desc, iov,
+						count, 0);
 		break;
 	case ofi_op_atomic_compare:
-		cmd->msg.hdr.size = ofi_copy_from_hmem_iov(tx_buf->buf,
-						SMR_COMP_INJECT_SIZE, iface, device,
-						iov, count, 0);
-		comp_size = ofi_copy_from_hmem_iov(tx_buf->comp, SMR_COMP_INJECT_SIZE,
-					      iface, device, compv, comp_count, 0);
+		cmd->msg.hdr.size = ofi_copy_from_mr_iov(tx_buf->buf,
+						SMR_COMP_INJECT_SIZE,
+						desc, iov, count, 0);
+		comp_size = ofi_copy_from_mr_iov(tx_buf->comp,
+					SMR_COMP_INJECT_SIZE,
+					comp_desc, compv, comp_count, 0);
 		if (comp_size != cmd->msg.hdr.size)
 			FI_WARN(&smr_prov, FI_LOG_EP_CTRL,
 				"atomic and compare buffer size mismatch\n");
 		break;
 	default:
+		assert(0);
 		break;
 	}
 }
 
 static ssize_t smr_do_atomic_inject(struct smr_ep *ep, struct smr_region *peer_smr,
 			int64_t id, int64_t peer_id, uint32_t op,
-			uint64_t op_flags, enum fi_hmem_iface iface,
-			uint64_t device, uint8_t datatype, uint8_t atomic_op,
-			const struct iovec *iov, size_t iov_count,
+			uint64_t op_flags, uint8_t datatype, uint8_t atomic_op,
+			struct ofi_mr **desc, const struct iovec *iov,
+			size_t iov_count, struct ofi_mr **res_desc,
 			const struct iovec *resultv, size_t result_count,
-			const struct iovec *compv, size_t comp_count,
-			size_t total_len, void *context, uint16_t smr_flags)
+			struct ofi_mr **comp_desc, const struct iovec *compv,
+			size_t comp_count, size_t total_len, void *context,
+			uint16_t smr_flags, struct smr_cmd *cmd)
 {
-	struct smr_cmd *cmd;
 	struct smr_inject_buf *tx_buf;
 	struct smr_tx_entry *pend;
 	struct smr_resp *resp;
 
-	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
-	tx_buf = smr_freestack_pop(smr_inject_pool(peer_smr));
+	tx_buf = smr_get_txbuf(peer_smr);
+	if (!tx_buf)
+		return -FI_EAGAIN;
 
 	smr_generic_format(cmd, peer_id, op, 0, 0, op_flags);
 	smr_generic_atomic_format(cmd, datatype, atomic_op);
-	smr_format_inject_atomic(cmd, iface, device, iov, iov_count, 
-				 resultv, result_count, compv, comp_count,
+	smr_format_inject_atomic(cmd, desc, iov, iov_count, resultv,
+				 result_count, comp_desc, compv, comp_count,
 				 peer_smr, tx_buf);
 
 	if (smr_flags & SMR_RMA_REQ || op_flags & FI_DELIVERY_COMPLETE) {
 		if (ofi_cirque_isfull(smr_resp_queue(ep->region))) {
-			smr_freestack_push(smr_inject_pool(peer_smr), tx_buf);
+			smr_release_txbuf(peer_smr, tx_buf);
 			return -FI_EAGAIN;
 		}
 		resp = ofi_cirque_next(smr_resp_queue(ep->region));
-		pend = ofi_freestack_pop(ep->pend_fs);
-		smr_format_pend_resp(pend, cmd, context, iface, device, resultv,
+		pend = ofi_freestack_pop(ep->tx_fs);
+		smr_format_pend_resp(pend, cmd, context, res_desc, resultv,
 				     result_count, op_flags, id, resp);
 		cmd->msg.hdr.data = smr_get_offset(ep->region, resp);
 		ofi_cirque_commit(smr_resp_queue(ep->region));
 	}
 
 	cmd->msg.hdr.op_flags |= smr_flags;
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
 
 	return FI_SUCCESS;
 }
@@ -186,18 +177,17 @@ static ssize_t smr_generic_atomic(struct smr_ep *ep,
 			enum fi_op atomic_op, void *context, uint32_t op,
 			uint64_t op_flags)
 {
-	struct smr_cmd *cmd;
+	struct smr_cmd_entry *ce;
 	struct smr_region *peer_smr;
 	struct iovec iov[SMR_IOV_LIMIT];
 	struct iovec compare_iov[SMR_IOV_LIMIT];
 	struct iovec result_iov[SMR_IOV_LIMIT];
-	enum fi_hmem_iface iface;
-	uint64_t device;
 	uint16_t smr_flags = 0;
 	int64_t id, peer_id;
 	int proto;
 	ssize_t ret = 0;
 	size_t total_len;
+	int64_t pos;
 
 	assert(count <= SMR_IOV_LIMIT);
 	assert(result_count <= SMR_IOV_LIMIT);
@@ -211,13 +201,14 @@ static ssize_t smr_generic_atomic(struct smr_ep *ep,
 	peer_id = smr_peer_data(ep->region)[id].addr.id;
 	peer_smr = smr_peer_region(ep->region, id);
 
-	pthread_spin_lock(&peer_smr->lock);
-	if (peer_smr->cmd_cnt < 2 || smr_peer_data(ep->region)[id].sar_status) {
-		ret = -FI_EAGAIN;
-		goto unlock_region;
-	}
+	if (smr_peer_data(ep->region)[id].sar_status)
+		return -FI_EAGAIN;
 
-	ofi_spin_lock(&ep->tx_lock);
+	ret = smr_cmd_queue_next(smr_cmd_queue(peer_smr), &ce, &pos);
+	if (ret == -FI_ENOENT)
+		return -FI_EAGAIN;
+
+	ofi_genlock_lock(&ep->util_ep.lock);
 	total_len = ofi_datatype_size(datatype) * ofi_total_ioc_cnt(ioc, count);
 
 	switch (op) {
@@ -242,25 +233,29 @@ static ssize_t smr_generic_atomic(struct smr_ep *ep,
 		}
 		break;
 	default:
+		assert(0);
 		break;
 	}
 
-	iface = smr_get_mr_hmem_iface(ep->util_ep.domain, desc, &device);
-
 	proto = smr_select_atomic_proto(op, total_len, op_flags);
 
 	if (proto == smr_src_inline) {
 		smr_do_atomic_inline(ep, peer_smr, id, peer_id, ofi_op_atomic,
-			 	     op_flags, iface, device, datatype, atomic_op,
-				     iov, count, total_len);
+				     op_flags, datatype, atomic_op,
+				     (struct ofi_mr **) desc, iov, count,
+				     total_len, &ce->cmd);
 	} else {
 		ret = smr_do_atomic_inject(ep, peer_smr, id, peer_id, op,
-				op_flags, iface, device, datatype, atomic_op,
-				iov, count, result_iov, result_count,
+				op_flags, datatype, atomic_op,
+				(struct ofi_mr **) desc, iov, count,
+				(struct ofi_mr **) result_desc, result_iov,
+				result_count, (struct ofi_mr **) compare_desc,
 				compare_iov, compare_count, total_len, context,
-				smr_flags);
-		if (ret)
-			goto unlock_cq;
+				smr_flags, &ce->cmd);
+		if (ret) {
+			smr_cmd_queue_discard(ce, pos);
+			goto unlock;
+		}
 	}
 
 	if (!(smr_flags & SMR_RMA_REQ) && !(op_flags & FI_DELIVERY_COMPLETE)) {
@@ -271,15 +266,10 @@ static ssize_t smr_generic_atomic(struct smr_ep *ep,
 		}
 	}
 
-	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
-	smr_format_rma_ioc(cmd, rma_ioc, rma_count);
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
-	smr_signal(peer_smr);
-unlock_cq:
-	ofi_spin_unlock(&ep->tx_lock);
-unlock_region:
-	pthread_spin_unlock(&peer_smr->lock);
+	smr_format_rma_ioc(&ce->rma_cmd, rma_ioc, rma_count);
+	smr_cmd_queue_commit(ce, pos);
+unlock:
+	ofi_genlock_unlock(&ep->util_ep.lock);
 	return ret;
 }
 
@@ -343,7 +333,7 @@ static ssize_t smr_atomic_inject(struct fid_ep *ep_fid, const void *buf,
 			size_t count, fi_addr_t dest_addr, uint64_t addr,
 			uint64_t key, enum fi_datatype datatype, enum fi_op op)
 {
-	struct smr_cmd *cmd;
+	struct smr_cmd_entry *ce;
 	struct smr_ep *ep;
 	struct smr_region *peer_smr;
 	struct iovec iov;
@@ -351,6 +341,7 @@ static ssize_t smr_atomic_inject(struct fid_ep *ep_fid, const void *buf,
 	int64_t id, peer_id;
 	ssize_t ret = 0;
 	size_t total_len;
+	int64_t pos;
 
 	ep = container_of(ep_fid, struct smr_ep, util_ep.ep_fid.fid);
 
@@ -361,12 +352,15 @@ static ssize_t smr_atomic_inject(struct fid_ep *ep_fid, const void *buf,
 	peer_id = smr_peer_data(ep->region)[id].addr.id;
 	peer_smr = smr_peer_region(ep->region, id);
 
-	pthread_spin_lock(&peer_smr->lock);
-	if (peer_smr->cmd_cnt < 2 || smr_peer_data(ep->region)[id].sar_status) {
+	if (smr_peer_data(ep->region)[id].sar_status) {
 		ret = -FI_EAGAIN;
-		goto unlock_region;
+		goto out;
 	}
 
+	ret = smr_cmd_queue_next(smr_cmd_queue(peer_smr), &ce, &pos);
+	if (ret == -FI_ENOENT)
+		return -FI_EAGAIN;
+
 	total_len = count * ofi_datatype_size(datatype);
 	assert(total_len <= SMR_INJECT_SIZE);
 
@@ -379,26 +373,23 @@ static ssize_t smr_atomic_inject(struct fid_ep *ep_fid, const void *buf,
 
 	if (total_len <= SMR_MSG_DATA_LEN) {
 		smr_do_atomic_inline(ep, peer_smr, id, peer_id, ofi_op_atomic,
-			 	     0, FI_HMEM_SYSTEM, 0, datatype, op,
-				     &iov, 1, total_len);
+				     0, datatype, op, NULL, &iov, 1, total_len,
+				     &ce->cmd);
 	} else if (total_len <= SMR_INJECT_SIZE) {
 		ret = smr_do_atomic_inject(ep, peer_smr, id, peer_id,
-				ofi_op_atomic, 0, FI_HMEM_SYSTEM, 0, datatype,
-				op, &iov, 1, NULL, 0, NULL, 0, total_len,
-				NULL, 0);
-		if (ret)
-			goto unlock_region;
+				ofi_op_atomic, 0, datatype, op, NULL, &iov, 1,
+				NULL, NULL, 0, NULL, NULL, 0, total_len, NULL,
+				0, &ce->cmd);
+		if (ret) {
+			smr_cmd_queue_discard(ce, pos);
+			goto out;
+		}
 	}
 
-	cmd = ofi_cirque_next(smr_cmd_queue(peer_smr));
-	smr_format_rma_ioc(cmd, &rma_ioc, 1);
-	ofi_cirque_commit(smr_cmd_queue(peer_smr));
-	peer_smr->cmd_cnt--;
-	smr_signal(peer_smr);
-
-	ofi_ep_tx_cntr_inc_func(&ep->util_ep, ofi_op_atomic);
-unlock_region:
-	pthread_spin_unlock(&peer_smr->lock);
+	smr_format_rma_ioc(&ce->rma_cmd, &rma_ioc, 1);
+	smr_cmd_queue_commit(ce, pos);
+	ofi_ep_peer_tx_cntr_inc(&ep->util_ep, ofi_op_atomic);
+out:
 	return ret;
 }
 
