@@ -35,7 +35,7 @@
 #ifndef _SMR_SIGNAL_H_
 #define _SMR_SIGNAL_H_
 #include <signal.h>
-#include <ofi_shm.h>
+#include "smr_util.h"
 #include "smr.h"
 
 extern struct sigaction *old_action;
@@ -46,14 +46,19 @@ static void smr_handle_signal(int signum, siginfo_t *info, void *ucontext)
 	struct smr_sock_name *sock_name;
 	int ret;
 
+	pthread_mutex_lock(&ep_list_lock);
 	dlist_foreach_container(&ep_name_list, struct smr_ep_name,
 				ep_name, entry) {
 		shm_unlink(ep_name->name);
 	}
+	pthread_mutex_unlock(&ep_list_lock);
+
+	pthread_mutex_lock(&sock_list_lock);
 	dlist_foreach_container(&sock_name_list, struct smr_sock_name,
 				sock_name, entry) {
 		unlink(sock_name->name);
 	}
+	pthread_mutex_unlock(&sock_list_lock);
 
 	/* Register the original signum handler, SIG_DFL or otherwise */
 	ret = sigaction(signum, &old_action[signum], NULL);
