@@ -67,7 +67,6 @@
 /* 4k allocated for all sockets headers */
 #define SOCK_EP_MAX_MSG_SZ (OFI_MAX_SOCKET_BUF_SIZE - 4096)
 #define SOCK_EP_MAX_INJECT_SZ ((1<<8) - 1)
-#define SOCK_EP_MAX_BUFF_RECV (1<<26)
 #define SOCK_EP_MAX_ORDER_RAW_SZ SOCK_EP_MAX_MSG_SZ
 #define SOCK_EP_MAX_ORDER_WAR_SZ SOCK_EP_MAX_MSG_SZ
 #define SOCK_EP_MAX_ORDER_WAW_SZ SOCK_EP_MAX_MSG_SZ
@@ -113,8 +112,6 @@
 			   OFI_ORDER_WAR_SET | OFI_ORDER_WAW_SET | FI_ORDER_WAS | \
 			   FI_ORDER_SAR | FI_ORDER_SAW | FI_ORDER_SAS)
 
-#define SOCK_EP_COMP_ORDER (FI_ORDER_STRICT | FI_ORDER_DATA)
-
 #define SOCK_EP_CQ_FLAGS (FI_SEND | FI_TRANSMIT | FI_RECV | \
 			FI_SELECTIVE_COMPLETION)
 #define SOCK_EP_CNTR_FLAGS (FI_SEND | FI_RECV | FI_READ | \
@@ -143,7 +140,8 @@ enum {
 
 enum {
 	SOCK_OPTS_NONBLOCK  = 1<<0,
-	SOCK_OPTS_KEEPALIVE = 1<<1
+	SOCK_OPTS_KEEPALIVE = 1<<1,
+	SOCK_OPTS_BUFSIZE = 1<<2,
 };
 
 #define SOCK_WIRE_PROTO_VERSION (2)
@@ -179,7 +177,7 @@ struct sock_fabric {
 #endif
 	struct dlist_entry service_list;
 	struct dlist_entry fab_list_entry;
-	fastlock_t lock;
+	ofi_mutex_t lock;
 };
 
 struct sock_conn {
@@ -197,17 +195,17 @@ struct sock_conn {
 struct sock_conn_map {
 	struct sock_conn *table;
 	ofi_epoll_t epoll_set;
-	void **epoll_ctxs;
-	int epoll_ctxs_sz;
+	struct ofi_epollfds_event *epoll_events;
+	int epoll_size;
 	int used;
 	int size;
-	fastlock_t lock;
+	ofi_mutex_t lock;
 };
 
 struct sock_conn_listener {
 	ofi_epoll_t epollfd;
 	struct fd_signal signal;
-	fastlock_t signal_lock; /* acquire before map lock */
+	ofi_mutex_t signal_lock; /* acquire before map lock */
 	pthread_t listener_thread;
 	int do_listen;
 	bool removed_from_epollfd;
@@ -227,11 +225,10 @@ struct sock_domain {
 	struct fi_info		info;
 	struct fid_domain	dom_fid;
 	struct sock_fabric	*fab;
-	fastlock_t		lock;
+	ofi_mutex_t		lock;
 	ofi_atomic32_t		ref;
 
 	struct sock_eq		*eq;
-	struct sock_eq		*mr_eq;
 
 	enum fi_progress	progress_mode;
 	struct ofi_mr_map	mr_map;
@@ -316,9 +313,9 @@ struct sock_cntr {
 
 	struct dlist_entry	rx_list;
 	struct dlist_entry	tx_list;
-	fastlock_t		list_lock;
+	ofi_mutex_t		list_lock;
 
-	fastlock_t		trigger_lock;
+	ofi_mutex_t		trigger_lock;
 	struct dlist_entry	trigger_list;
 
 	struct fid_wait		*waitset;
@@ -362,8 +359,8 @@ struct sock_av {
 	struct util_shm shm;
 	int    shared;
 	struct dlist_entry ep_list;
-	fastlock_t list_lock;
-	fastlock_t table_lock;
+	ofi_mutex_t list_lock;
+	ofi_mutex_t table_lock;
 };
 
 struct sock_fid_list {
@@ -467,7 +464,7 @@ struct sock_eq_entry {
 	size_t len;
 	uint64_t flags;
 	struct dlist_entry entry;
-	char event[0];
+	char event[];
 };
 
 struct sock_eq_err_data_entry {
@@ -484,7 +481,7 @@ struct sock_eq {
 	struct dlistfd_head list;
 	struct dlistfd_head err_list;
 	struct dlist_entry err_data_list;
-	fastlock_t lock;
+	ofi_mutex_t lock;
 
 	struct fid_wait *waitset;
 	int signal;
@@ -525,7 +522,7 @@ struct sock_pep_cm_entry {
 
 struct sock_ep_cm_entry {
 	int sock;
-	fastlock_t lock;
+	ofi_mutex_t lock;
 	enum sock_cm_state state;
 };
 
@@ -572,7 +569,7 @@ struct sock_ep_attr {
 	int is_enabled;
 	struct sock_ep_cm_entry cm;
 	struct sock_conn_handle conn_handle;
-	fastlock_t lock;
+	ofi_mutex_t lock;
 
 	struct index_map av_idm;
 	struct sock_conn_map cmap;
@@ -654,7 +651,7 @@ struct sock_rx_ctx {
 	struct dlist_entry rx_entry_list;
 	struct dlist_entry rx_buffered_list;
 	struct dlist_entry ep_list;
-	fastlock_t lock;
+	ofi_mutex_t lock;
 
 	struct dlist_entry *progress_start;
 
@@ -671,7 +668,7 @@ struct sock_tx_ctx {
 	size_t fclass;
 
 	struct ofi_ringbuf rb;
-	fastlock_t rb_lock;
+	ofi_mutex_t rb_lock;
 
 	uint16_t tx_id;
 	uint8_t enabled;
@@ -695,7 +692,7 @@ struct sock_tx_ctx {
 	struct dlist_entry ep_list;
 
 	struct fi_tx_attr attr;
-	fastlock_t lock;
+	ofi_mutex_t lock;
 };
 
 struct sock_msg_hdr {
@@ -847,8 +844,8 @@ struct sock_pe {
 	struct sock_domain *domain;
 	int num_free_entries;
 	struct sock_pe_entry pe_table[SOCK_PE_MAX_ENTRIES];
-	fastlock_t lock;
-	fastlock_t signal_lock;
+	ofi_mutex_t lock;
+	ofi_mutex_t signal_lock;
 	pthread_mutex_t list_lock;
 	int wcnt, rcnt;
 	int signal_fds[2];
@@ -869,14 +866,14 @@ struct sock_pe {
 	ofi_epoll_t epoll_set;
 };
 
-typedef int (*sock_cq_report_fn) (struct sock_cq *cq, fi_addr_t addr,
-				  struct sock_pe_entry *pe_entry);
+typedef ssize_t (*sock_cq_report_fn) (struct sock_cq *cq, fi_addr_t addr,
+				      struct sock_pe_entry *pe_entry);
 
 struct sock_cq_overflow_entry_t {
 	size_t len;
 	fi_addr_t addr;
 	struct dlist_entry entry;
-	char cq_entry[0];
+	char cq_entry[];
 };
 
 struct sock_cq {
@@ -916,7 +913,7 @@ struct sock_conn_req {
 	struct sock_conn_hdr hdr;
 	union ofi_sock_ip src_addr;
 	uint64_t caps;
-	char cm_data[0];
+	char cm_data[];
 };
 
 enum {
@@ -1060,11 +1057,11 @@ void sock_cq_remove_rx_ctx(struct sock_cq *cq, struct sock_rx_ctx *rx_ctx);
 
 int sock_eq_open(struct fid_fabric *fabric, struct fi_eq_attr *attr,
 		struct fid_eq **eq, void *context);
-ssize_t sock_eq_report_event(struct sock_eq *sock_eq, uint32_t event,
-			     const void *buf, size_t len, uint64_t flags);
-ssize_t sock_eq_report_error(struct sock_eq *sock_eq, fid_t fid, void *context,
-			     uint64_t data, int err, int prov_errno,
-			     void *err_data, size_t err_data_size);
+int sock_eq_report_event(struct sock_eq *sock_eq, uint32_t event,
+			 const void *buf, size_t len, uint64_t flags);
+int sock_eq_report_error(struct sock_eq *sock_eq, fid_t fid, void *context,
+			 uint64_t data, int err, int prov_errno,
+			 void *err_data, size_t err_data_size);
 int sock_eq_openwait(struct sock_eq *eq, const char *service);
 
 int sock_cntr_open(struct fid_domain *domain, struct fi_cntr_attr *attr,
@@ -1199,7 +1196,7 @@ ssize_t sock_ep_tx_atomic(struct fid_ep *ep,
 			  size_t compare_count, struct fi_ioc *resultv,
 			  void **result_desc, size_t result_count, uint64_t flags);
 
-int sock_queue_work(struct sock_domain *dom, struct fi_deferred_work *work);
+ssize_t sock_queue_work(struct sock_domain *dom, struct fi_deferred_work *work);
 ssize_t sock_queue_rma_op(struct fid_ep *ep, const struct fi_msg_rma *msg,
 			  uint64_t flags, enum fi_op_type op_type);
 ssize_t sock_queue_atomic_op(struct fid_ep *ep, const struct fi_msg_atomic *msg,
