@@ -51,14 +51,16 @@ static ssize_t sock_comm_send_socket(struct sock_conn *conn,
 
 	ret = ofi_send_socket(conn->sock_fd, buf, len, MSG_NOSIGNAL);
 	if (ret < 0) {
-		if (errno == EAGAIN || errno == EWOULDBLOCK)
+		if (OFI_SOCK_TRY_SND_RCV_AGAIN(ofi_sockerr())) {
 			ret = 0;
-		else if (errno == EPIPE) {
+		} else if (ofi_sockerr() == EPIPE) {
 			conn->connected = 0;
-			SOCK_LOG_DBG("Disconnected: %s:%d\n", inet_ntoa(conn->addr.sin_addr),
-                               ntohs(conn->addr.sin_port));
-		} else
-			SOCK_LOG_DBG("write error: %s\n", strerror(errno));
+			SOCK_LOG_DBG("Disconnected port: %d\n",
+				     ofi_addr_get_port(&conn->addr.sa));
+		} else {
+			SOCK_LOG_DBG("write error: %s\n",
+				     strerror(ofi_sockerr()));
+		}
 	}
 	if (ret > 0)
 		SOCK_LOG_DBG("wrote to network: %lu\n", ret);
@@ -81,7 +83,7 @@ ssize_t sock_comm_flush(struct sock_pe_entry *pe_entry)
 	if (ret1 > 0)
 		pe_entry->comm_buf.rcnt += ret1;
 
-	if (ret1 == xfer_len && xfer_len < len) {
+	if ((size_t) ret1 == xfer_len && xfer_len < len) {
 		ret2 = sock_comm_send_socket(pe_entry->conn, (char*)pe_entry->comm_buf.buf +
 					     (pe_entry->comm_buf.rcnt & pe_entry->comm_buf.size_mask),
 					     len - xfer_len);
@@ -130,16 +132,16 @@ static ssize_t sock_comm_recv_socket(struct sock_conn *conn,
 			      void *buf, size_t len)
 {
 	ssize_t ret;
-	ret = recv(conn->sock_fd, buf, len, 0);
+	ret = ofi_recv_socket(conn->sock_fd, buf, len, 0);
 	if (ret == 0) {
 		conn->connected = 0;
-		SOCK_LOG_DBG("Disconnected: %s:%d\n", inet_ntoa(conn->addr.sin_addr),
-                               ntohs(conn->addr.sin_port));
+		SOCK_LOG_DBG("Disconnected: port %d\n",
+			     ofi_addr_get_port(&conn->addr.sa));
 		return ret;
 	}
 
 	if (ret < 0) {
-		SOCK_LOG_DBG("read %s\n", strerror(errno));
+		SOCK_LOG_DBG("read %s\n", strerror(ofi_sockerr()));
 		ret = 0;
 	}
 
@@ -150,7 +152,7 @@ static ssize_t sock_comm_recv_socket(struct sock_conn *conn,
 
 static void sock_comm_recv_buffer(struct sock_pe_entry *pe_entry)
 {
-	int ret;
+	ssize_t ret;
 	size_t max_read, avail;
 
 	avail = ofi_rbavail(&pe_entry->comm_buf);
@@ -187,7 +189,7 @@ ssize_t sock_comm_recv(struct sock_pe_entry *pe_entry, void *buf, size_t len)
 ssize_t sock_comm_peek(struct sock_conn *conn, void *buf, size_t len)
 {
 	ssize_t ret;
-	ret = recv(conn->sock_fd, buf, len, MSG_PEEK);
+	ret = ofi_recv_socket(conn->sock_fd, buf, len, MSG_PEEK);
 	if (ret == 0) {
 		conn->connected = 0;
 		SOCK_LOG_DBG("Disconnected\n");
@@ -195,7 +197,7 @@ ssize_t sock_comm_peek(struct sock_conn *conn, void *buf, size_t len)
 	}
 
 	if (ret < 0) {
-		SOCK_LOG_DBG("peek %s\n", strerror(errno));
+		SOCK_LOG_DBG("peek %s\n", strerror(ofi_sockerr()));
 		ret = 0;
 	}
 
@@ -207,7 +209,7 @@ ssize_t sock_comm_peek(struct sock_conn *conn, void *buf, size_t len)
 ssize_t sock_comm_discard(struct sock_pe_entry *pe_entry, size_t len)
 {
 	void *buf;
-	int ret;
+	ssize_t ret;
 
 	buf = malloc(len);
 	if (!buf)
