@@ -107,7 +107,7 @@ static int sock_pep_create_listener(struct sock_pep *pep)
 	sock_set_sockopts(pep->cm.sock, SOCK_OPTS_NONBLOCK);
 
 	ret = bind(pep->cm.sock, &pep->src_addr.sa,
-		   ofi_sizeofaddr(&pep->src_addr.sa));
+		   (socklen_t) ofi_sizeofaddr(&pep->src_addr.sa));
 	if (ret) {
 		SOCK_LOG_ERROR("failed to bind listener: %s\n",
 			       strerror(ofi_sockerr()));
@@ -181,18 +181,19 @@ static int sock_ep_cm_getpeer(struct fid_ep *ep, void *addr, size_t *addrlen)
 	return (len == *addrlen) ? 0 : -FI_ETOOSMALL;
 }
 
-static int sock_cm_send(int fd, const void *buf, int len)
+static int sock_cm_send(int fd, const void *buf, size_t len)
 {
-	int ret, done = 0;
+	ssize_t ret;
+	size_t done = 0;
 
-	while (done != len) {
+	while (done < len) {
 		ret = ofi_send_socket(fd, (const char*) buf + done,
 				      len - done, MSG_NOSIGNAL);
 		if (ret < 0) {
 			if (OFI_SOCK_TRY_SND_RCV_AGAIN(ofi_sockerr()))
 				continue;
 			SOCK_LOG_ERROR("failed to write to fd: %s\n",
-				       strerror(ofi_sockerr()));
+					strerror(ofi_sockerr()));
 			return -FI_EIO;
 		}
 		done += ret;
@@ -200,10 +201,12 @@ static int sock_cm_send(int fd, const void *buf, int len)
 	return 0;
 }
 
-static int sock_cm_recv(int fd, void *buf, int len)
+static int sock_cm_recv(int fd, void *buf, size_t len)
 {
-	int ret, done = 0;
-	while (done != len) {
+	ssize_t ret;
+	size_t done = 0;
+
+	while (done < len) {
 		ret = ofi_recv_socket(fd, (char*) buf + done, len - done, 0);
 		if (ret <= 0) {
 			if (ret < 0 && (errno == EAGAIN || errno == EWOULDBLOCK))
@@ -283,7 +286,7 @@ static void sock_ep_cm_shutdown_report(struct sock_ep *ep, int send_shutdown)
 	struct sock_conn_hdr msg = {0};
 	enum sock_cm_state old_state;
 
-	fastlock_acquire(&ep->attr->cm.lock);
+	ofi_mutex_lock(&ep->attr->cm.lock);
 	old_state = ep->attr->cm.state;
 	switch (ep->attr->cm.state) {
 	case SOCK_CM_STATE_REQUESTED:
@@ -298,7 +301,7 @@ static void sock_ep_cm_shutdown_report(struct sock_ep *ep, int send_shutdown)
 		assert(0);
 		break;
 	}
-	fastlock_release(&ep->attr->cm.lock);
+	ofi_mutex_unlock(&ep->attr->cm.lock);
 
 	switch (old_state) {
 	case SOCK_CM_STATE_CONNECTED:
@@ -348,12 +351,12 @@ static void sock_ep_cm_report_connect_fail(struct sock_ep *ep,
 {
 	int do_report = 0;
 
-	fastlock_acquire(&ep->attr->cm.lock);
+	ofi_mutex_lock(&ep->attr->cm.lock);
 	if (ep->attr->cm.state == SOCK_CM_STATE_REQUESTED) {
 		do_report = 1;
 		ep->attr->cm.state = SOCK_CM_STATE_DISCONNECTED;
 	}
-	fastlock_release(&ep->attr->cm.lock);
+	ofi_mutex_unlock(&ep->attr->cm.lock);
 
 	if (do_report) {
 		SOCK_LOG_DBG("reporting FI_REJECT\n");
@@ -382,10 +385,15 @@ static void sock_ep_cm_connect_handler(struct sock_ep_cm_head *cm_head,
 	struct fi_eq_cm_entry *cm_entry = NULL;
 	int cm_data_sz, response_port;
 
-	assert(hdr->type == SOCK_CONN_ACCEPT
-	       || hdr->type == SOCK_CONN_REJECT);
+	assert(hdr->type == SOCK_CONN_ACCEPT ||
+	       hdr->type == SOCK_CONN_REJECT);
 
 	cm_data_sz = ntohs(hdr->cm_data_sz);
+	if (cm_data_sz > SOCK_EP_MAX_CM_DATA_SZ) {
+		SOCK_LOG_ERROR("CM data size too large\n");
+		goto err;
+	}
+
 	response_port = ntohs(hdr->port);
 	if (cm_data_sz) {
 		param = calloc(1, cm_data_sz);
@@ -405,10 +413,11 @@ static void sock_ep_cm_connect_handler(struct sock_ep_cm_head *cm_head,
 			goto err;
 
 		cm_entry->fid = &ep->ep.fid;
-		memcpy(&cm_entry->data, param, cm_data_sz);
+		if (cm_data_sz)
+			memcpy(&cm_entry->data, param, cm_data_sz);
 		ep->attr->cm.state = SOCK_CM_STATE_CONNECTED;
 		ep->attr->cm.sock = sock_fd;
-		ep->attr->msg_dest_port = response_port;
+		ep->attr->msg_dest_port = (uint16_t) response_port;
 		SOCK_LOG_DBG("got accept - port: %d\n", response_port);
 
 		SOCK_LOG_DBG("Reporting FI_CONNECTED\n");
@@ -482,7 +491,7 @@ static int sock_ep_cm_connect(struct fid_ep *ep, const void *addr,
 
 	req->hdr.type = SOCK_CONN_REQ;
 	req->hdr.port = htons(_ep->attr->msg_src_port);
-	req->hdr.cm_data_sz = htons(paramlen);
+	req->hdr.cm_data_sz = htons((uint16_t) paramlen);
 	req->caps = _ep->attr->info.caps;
 	memcpy(&req->src_addr, _ep->attr->src_addr,
 	       ofi_sizeofaddr(&_ep->attr->src_addr->sa));
@@ -508,7 +517,7 @@ static int sock_ep_cm_connect(struct fid_ep *ep, const void *addr,
 			&handle->dest_addr);
 	sock_set_sockopts(sock_fd, SOCK_OPTS_KEEPALIVE);
 	ret = connect(sock_fd, &handle->dest_addr.sa,
-		      ofi_sizeofaddr(&handle->dest_addr.sa));
+		      (socklen_t) ofi_sizeofaddr(&handle->dest_addr.sa));
 	if (ret < 0) {
 		SOCK_LOG_ERROR("connect failed : %s\n",
 			       strerror(ofi_sockerr()));
@@ -549,7 +558,7 @@ static int sock_ep_cm_accept(struct fid_ep *ep, const void *param, size_t paraml
 	struct sock_conn_req_handle *handle;
 	struct sock_ep_attr *ep_attr;
 	struct fi_eq_cm_entry cm_entry;
-	struct sock_conn_hdr reply;
+	struct sock_conn_hdr reply = {0};
 	struct sock_ep *_ep;
 
 	_ep = container_of(ep, struct sock_ep, ep);
@@ -579,7 +588,7 @@ static int sock_ep_cm_accept(struct fid_ep *ep, const void *param, size_t paraml
 
 	reply.type = SOCK_CONN_ACCEPT;
 	reply.port = htons(ep_attr->msg_src_port);
-	reply.cm_data_sz = htons(handle->paramlen);
+	reply.cm_data_sz = htons((uint16_t) handle->paramlen);
 	ret = sock_cm_send(handle->sock_fd, &reply, sizeof(reply));
 	if (ret) {
 		SOCK_LOG_ERROR("failed to reply\n");
@@ -676,7 +685,7 @@ static int sock_pep_fi_bind(fid_t fid, struct fid *bfid, uint64_t flags)
 
 static int sock_pep_fi_close(fid_t fid)
 {
-	int ret;
+	ssize_t ret;
 	char c = 0;
 	struct sock_pep *pep;
 
@@ -740,7 +749,7 @@ static void sock_ep_cm_process_rejected(struct sock_ep_cm_head *cm_head,
 	struct sock_conn_hdr reply;
 
 	reply.type = SOCK_CONN_REJECT;
-	reply.cm_data_sz = htons(hreq->paramlen);
+	reply.cm_data_sz = htons((uint16_t) hreq->paramlen);
 
 	SOCK_LOG_DBG("sending reject message\n");
 	if (sock_cm_send(hreq->sock_fd, &reply, sizeof(reply))) {
@@ -846,6 +855,11 @@ static void sock_pep_req_handler(struct sock_ep_cm_head *cm_head,
 	}
 
 	req_cm_data_sz = ntohs(conn_req->hdr.cm_data_sz);
+	if (req_cm_data_sz > SOCK_EP_MAX_CM_DATA_SZ) {
+		SOCK_LOG_ERROR("CM data size is too large\n");
+		goto err;
+	}
+
 	if (req_cm_data_sz) {
 		ret = sock_cm_recv(handle->sock_fd, conn_req->cm_data,
 				   req_cm_data_sz);
@@ -903,7 +917,7 @@ static void *sock_pep_listener_thread(void *data)
 	struct sock_conn_req_handle *handle = NULL;
 	struct pollfd poll_fds[2];
 
-	int ret = 0, conn_fd;
+	ssize_t ret = 0, conn_fd;
 	char tmp = 0;
 
 	SOCK_LOG_DBG("Starting listener thread for PEP: %p\n", pep);
@@ -1161,13 +1175,13 @@ static void *sock_ep_cm_thread(void *arg)
 {
 	int num_fds, i;
 	struct sock_ep_cm_head *cm_head = arg;
-	void *ep_contexts[SOCK_EPOLL_WAIT_EVENTS];
+	struct ofi_epollfds_event events[SOCK_EPOLL_WAIT_EVENTS];
 	struct sock_conn_req_handle *handle;
 
 	while (cm_head->do_listen) {
 		sock_ep_cm_check_closing_rejected_list(cm_head);
 
-		num_fds = ofi_epoll_wait(cm_head->epollfd, ep_contexts,
+		num_fds = ofi_epoll_wait(cm_head->epollfd, events,
 		                        SOCK_EPOLL_WAIT_EVENTS, -1);
 		if (num_fds < 0) {
 			SOCK_LOG_ERROR("poll failed : %s\n", strerror(errno));
@@ -1185,7 +1199,7 @@ static void *sock_ep_cm_thread(void *arg)
 			goto skip;
 		}
 		for (i = 0; i < num_fds; i++) {
-			handle = ep_contexts[i];
+			handle = events[i].data.ptr;
 
 			if (handle == NULL) { /* Signal event */
 				fd_signal_reset(&cm_head->signal);
@@ -1269,7 +1283,7 @@ void sock_ep_cm_wait_handle_finalized(struct sock_ep_cm_head *cm_head,
 
 	pthread_mutex_lock(&handle->finalized_mutex);
 	while (handle->state != SOCK_CONN_HANDLE_FINALIZED)
-		fi_wait_cond(&handle->finalized_cond,
+		ofi_wait_cond(&handle->finalized_cond,
 				&handle->finalized_mutex, -1);
 	pthread_mutex_unlock(&handle->finalized_mutex);
 }
