@@ -105,9 +105,6 @@ struct sock_rx_entry *sock_rx_new_buffered_entry(struct sock_rx_ctx *rx_ctx,
 {
 	struct sock_rx_entry *rx_entry;
 
-	if (rx_ctx->buffered_len + len >= rx_ctx->attr.total_buffered_recv)
-		SOCK_LOG_ERROR("Exceeded buffered recv limit\n");
-
 	rx_entry = calloc(1, sizeof(*rx_entry) + len);
 	if (!rx_entry)
 		return NULL;
@@ -124,6 +121,7 @@ struct sock_rx_entry *sock_rx_new_buffered_entry(struct sock_rx_ctx *rx_ctx,
 
 	rx_ctx->buffered_len += len;
 	dlist_insert_tail(&rx_entry->entry, &rx_ctx->rx_buffered_list);
+	rx_ctx->progress_start = &rx_ctx->rx_buffered_list;
 
 	return rx_entry;
 }
