@@ -41,10 +41,9 @@
 
 #define USDF_DGRAM_CAPS (FI_MSG | FI_SOURCE | FI_SEND | FI_RECV)
 
-#define USDF_DGRAM_SUPP_MODE (FI_LOCAL_MR | FI_MSG_PREFIX)
+#define USDF_DGRAM_SUPP_MODE (OFI_LOCAL_MR | FI_MSG_PREFIX)
 
-#define USDF_DGRAM_MSG_ORDER (FI_ORDER_NONE)
-#define USDF_DGRAM_COMP_ORDER (FI_ORDER_NONE)
+#define USDF_DGRAM_MSG_ORDER (0)
 #define USDF_DGRAM_INJECT_SIZE                                                 \
 	(USD_SEND_MAX_COPY - sizeof(struct usd_udp_hdr))
 #define USDF_DGRAM_SUPP_SENDMSG_FLAGS                                          \
@@ -57,13 +56,13 @@
 #define USDF_DGRAM_MR_CNT (USDF_MR_CNT)
 
 
-int usdf_dgram_fill_rx_attr(uint32_t version, struct fi_info *hints,
+int usdf_dgram_fill_rx_attr(uint32_t version, const struct fi_info *hints,
 		struct fi_info *fi, struct usd_device_attrs *dap);
-int usdf_dgram_fill_tx_attr(uint32_t version, struct fi_info *hints,
+int usdf_dgram_fill_tx_attr(uint32_t version, const struct fi_info *hints,
 		struct fi_info *fi, struct usd_device_attrs *dap);
-int usdf_dgram_fill_dom_attr(uint32_t version, struct fi_info *hints,
+int usdf_dgram_fill_dom_attr(uint32_t version, const struct fi_info *hints,
 			     struct fi_info *fi, struct usd_device_attrs *dap);
-int usdf_dgram_fill_ep_attr(uint32_t version, struct fi_info *hints,
+int usdf_dgram_fill_ep_attr(uint32_t version, const struct fi_info *hints,
 		struct fi_info *fi, struct usd_device_attrs *dap);
 
 /* fi_ops_msg for DGRAM */
