@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2017, Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2014-2019, Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -53,9 +53,9 @@
 #include <rdma/fi_endpoint.h>
 #include <rdma/fi_rma.h>
 #include <rdma/fi_errno.h>
-#include "fi.h"
-#include "fi_enosys.h"
-#include "fi_util.h"
+#include "ofi.h"
+#include "ofi_enosys.h"
+#include "ofi_util.h"
 
 #include "usnic_direct.h"
 #include "usd.h"
@@ -401,7 +401,6 @@ static const struct fi_tx_attr dgram_dflt_tx_attr = {
 	.mode = USDF_DGRAM_SUPP_MODE,
 	.op_flags = 0,
 	.msg_order = USDF_DGRAM_MSG_ORDER,
-	.comp_order = USDF_DGRAM_COMP_ORDER,
 	.inject_size = USDF_DGRAM_INJECT_SIZE,
 	.iov_limit = USDF_DGRAM_IOV_LIMIT,
 	.rma_iov_limit = USDF_DGRAM_RMA_IOV_LIMIT
@@ -412,8 +411,6 @@ static const struct fi_rx_attr dgram_dflt_rx_attr = {
 	.mode = USDF_DGRAM_SUPP_MODE,
 	.op_flags = 0,
 	.msg_order = USDF_DGRAM_MSG_ORDER,
-	.comp_order = USDF_DGRAM_COMP_ORDER,
-	.total_buffered_recv = 0,
 	.iov_limit = USDF_DGRAM_IOV_LIMIT
 };
 
@@ -431,11 +428,11 @@ static const struct fi_ep_attr dgram_dflt_ep_attr = {
 
 static const struct fi_domain_attr dgram_dflt_domain_attr = {
 	.caps = USDF_DOM_CAPS,
-	.threading = FI_THREAD_ENDPOINT,
+	.threading = FI_THREAD_COMPLETION,
 	.control_progress = FI_PROGRESS_AUTO,
 	.data_progress = FI_PROGRESS_MANUAL,
 	.resource_mgmt = FI_RM_DISABLED,
-	.mr_mode = OFI_MR_BASIC_MAP | FI_MR_LOCAL,
+	.mr_mode = FI_MR_ALLOCATED | FI_MR_LOCAL | OFI_MR_BASIC,
 	.cntr_cnt = USDF_DGRAM_CNTR_CNT,
 	.mr_iov_limit = USDF_DGRAM_MR_IOV_LIMIT,
 	.mr_cnt = USDF_DGRAM_MR_CNT,
@@ -444,7 +441,7 @@ static const struct fi_domain_attr dgram_dflt_domain_attr = {
 /*******************************************************************************
  * Fill functions for attributes
  ******************************************************************************/
-int usdf_dgram_fill_ep_attr(uint32_t version, struct fi_info *hints, struct
+int usdf_dgram_fill_ep_attr(uint32_t version, const struct fi_info *hints, struct
 		fi_info *fi, struct usd_device_attrs *dap)
 {
 	struct fi_ep_attr defaults;
@@ -500,7 +497,7 @@ out:
 	return FI_SUCCESS;
 }
 
-int usdf_dgram_fill_dom_attr(uint32_t version, struct fi_info *hints,
+int usdf_dgram_fill_dom_attr(uint32_t version, const struct fi_info *hints,
 			     struct fi_info *fi, struct usd_device_attrs *dap)
 {
 	int ret;
@@ -516,9 +513,7 @@ int usdf_dgram_fill_dom_attr(uint32_t version, struct fi_info *hints,
 
 	switch (hints->domain_attr->threading) {
 	case FI_THREAD_UNSPEC:
-	case FI_THREAD_ENDPOINT:
 		break;
-	case FI_THREAD_FID:
 	case FI_THREAD_COMPLETION:
 	case FI_THREAD_DOMAIN:
 		defaults.threading = hints->domain_attr->threading;
@@ -565,7 +560,15 @@ int usdf_dgram_fill_dom_attr(uint32_t version, struct fi_info *hints,
 		return -FI_ENODATA;
 	}
 
-	if (usdf_check_mr_mode(version, hints, defaults.mr_mode))
+	switch (hints->domain_attr->av_type) {
+	case FI_AV_UNSPEC:
+	case FI_AV_MAP:
+		break;
+	default:
+		return -FI_ENODATA;
+	}
+
+	if (ofi_check_mr_mode(&usdf_ops, version, defaults.mr_mode, hints))
 		return -FI_ENODATA;
 
 	if (hints->domain_attr->mr_cnt) {
@@ -588,7 +591,7 @@ catch:
 	return FI_SUCCESS;
 }
 
-int usdf_dgram_fill_tx_attr(uint32_t version, struct fi_info *hints,
+int usdf_dgram_fill_tx_attr(uint32_t version, const struct fi_info *hints,
 			    struct fi_info *fi,
 			    struct usd_device_attrs *dap)
 {
@@ -616,9 +619,6 @@ int usdf_dgram_fill_tx_attr(uint32_t version, struct fi_info *hints,
 	if ((hints->tx_attr->msg_order | USDF_DGRAM_MSG_ORDER) !=
 			USDF_DGRAM_MSG_ORDER)
 		return -FI_ENODATA;
-	if ((hints->tx_attr->comp_order | USDF_DGRAM_COMP_ORDER) !=
-			USDF_DGRAM_COMP_ORDER)
-		return -FI_ENODATA;
 
 	if (hints->tx_attr->inject_size > defaults.inject_size)
 		return -FI_ENODATA;
@@ -667,7 +667,7 @@ out:
 	return FI_SUCCESS;
 }
 
-int usdf_dgram_fill_rx_attr(uint32_t version, struct fi_info *hints,
+int usdf_dgram_fill_rx_attr(uint32_t version, const struct fi_info *hints,
 			    struct fi_info *fi, struct usd_device_attrs *dap)
 {
 	int ret;
@@ -694,13 +694,6 @@ int usdf_dgram_fill_rx_attr(uint32_t version, struct fi_info *hints,
 	if ((hints->rx_attr->msg_order | USDF_DGRAM_MSG_ORDER) !=
 			USDF_DGRAM_MSG_ORDER)
 		return -FI_ENODATA;
-	if ((hints->rx_attr->comp_order | USDF_DGRAM_COMP_ORDER) !=
-			USDF_DGRAM_COMP_ORDER)
-		return -FI_ENODATA;
-
-	if (hints->rx_attr->total_buffered_recv >
-			defaults.total_buffered_recv)
-		return -FI_ENODATA;
 
 	if (hints->rx_attr->iov_limit > USDF_DGRAM_MAX_SGE)
 		return -FI_ENODATA;
@@ -746,6 +739,7 @@ out:
 static int usdf_ep_dgram_control(struct fid *fid, int command, void *arg)
 {
 	struct fid_ep *ep;
+	int ret;
 
 	USDF_TRACE_SYS(EP_CTRL, "\n");
 
@@ -754,15 +748,17 @@ static int usdf_ep_dgram_control(struct fid *fid, int command, void *arg)
 		ep = container_of(fid, struct fid_ep, fid);
 		switch (command) {
 		case FI_ENABLE:
-			return usdf_ep_dgram_enable(ep);
+			ret = usdf_ep_dgram_enable(ep);
 			break;
 		default:
-			return -FI_ENOSYS;
+			ret = -FI_ENOSYS;
 		}
 		break;
 	default:
-		return -FI_ENOSYS;
+		ret = -FI_ENOSYS;
 	}
+
+	return ret;
 }
 
 static struct fi_ops usdf_ep_dgram_ops = {
