@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2017, Cisco Systems, Inc. All rights reserved.
+ * Copyright (c) 2014-2019, Cisco Systems, Inc. All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -42,7 +42,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <netinet/in.h>
-#include <sys/epoll.h>
+#include <ofi_epoll.h>
 #include <sys/eventfd.h>
 #include <sys/socket.h>
 #include <stdio.h>
@@ -71,8 +71,6 @@
 #include "usdf_progress.h"
 #include "usdf_timer.h"
 #include "usdf_dgram.h"
-#include "usdf_msg.h"
-#include "usdf_rdm.h"
 #include "usdf_cm.h"
 
 struct usdf_usnic_info *__usdf_devinfo;
@@ -332,7 +330,7 @@ static int validate_modebits(uint32_t version, const struct fi_info *hints,
 
 	/* Before version 1.5, FI_LOCAL_MR is a requirement. */
 	if (FI_VERSION_LT(version, FI_VERSION(1, 5))) {
-		if ((mode & FI_LOCAL_MR) == 0)
+		if ((mode & OFI_LOCAL_MR) == 0)
 			return -FI_ENODATA;
 	}
 
@@ -341,6 +339,80 @@ static int validate_modebits(uint32_t version, const struct fi_info *hints,
 	return FI_SUCCESS;
 }
 
+static int usdf_alloc_fid_nic(struct fi_info *fi,
+			struct usd_device_attrs *dap)
+{
+	int ret;
+	struct fid_nic *nic = NULL;
+	struct fi_device_attr *da = NULL;
+	struct fi_link_attr *la = NULL;
+
+	nic = ofi_nic_dup(NULL);
+	if (!nic)
+		goto nomem;
+
+	da = nic->device_attr;
+	da->name = strdup(dap->uda_devname);
+	if (!da->name)
+		goto nomem;
+	ret = asprintf(&da->device_id, "%s (%s)",
+		usd_devid_to_pid(dap->uda_vendor_id,
+				dap->uda_device_id),
+		usd_devid_to_nicname(dap->uda_vendor_id,
+				dap->uda_device_id));
+	if (ret < 0)
+		goto nomem;
+	ret = asprintf(&da->device_version, "0x%x", dap->uda_vendor_part_id);
+	if (ret < 0)
+		goto nomem;
+	ret = asprintf(&da->vendor_id, "0x%x", dap->uda_vendor_id);
+	if (ret < 0)
+		goto nomem;
+	da->driver = strdup("usnic_verbs");
+	if (!da->driver)
+		goto nomem;
+	da->firmware = strdup(dap->uda_firmware);
+	if (!da->firmware)
+		goto nomem;
+
+	// usnic does not currently expose PCI bus information, so we
+	// set the bus type to unknown.
+	nic->bus_attr->bus_type = FI_BUS_UNKNOWN;
+
+	la = nic->link_attr;
+
+	socklen_t size = INET_ADDRSTRLEN;
+	la->address = calloc(1, size);
+	if (!la->address)
+		goto nomem;
+	inet_ntop(AF_INET, &dap->uda_ipaddr_be, la->address, size);
+	la->mtu = dap->uda_mtu;
+	la->speed = dap->uda_bandwidth;
+	switch (dap->uda_link_state) {
+	case USD_LINK_UP:
+		la->state = FI_LINK_UP;
+		break;
+	case USD_LINK_DOWN:
+		la->state = FI_LINK_DOWN;
+		break;
+	default:
+		la->state = FI_LINK_UNKNOWN;
+		break;
+	}
+	la->network_type = strdup("Ethernet");
+	if (!la->network_type)
+		goto nomem;
+
+	fi->nic = nic;
+
+	return FI_SUCCESS;
+
+nomem:
+	if (nic)
+		fi_close(&nic->fid);
+	return -FI_ENOMEM;
+}
+
 static int usdf_fill_info_dgram(
 	uint32_t version,
 	const struct fi_info *hints,
@@ -419,179 +491,7 @@ static int usdf_fill_info_dgram(
 	if (ret)
 		goto fail;
 
-	/* add to tail of list */
-	if (*fi_first == NULL) {
-		*fi_first = fi;
-	} else {
-		(*fi_last)->next = fi;
-	}
-	*fi_last = fi;
-
-	return 0;
-
-fail:
-	if (fi != NULL) {
-		fi_freeinfo(fi);
-	}
-	return ret;
-}
-
-static int usdf_fill_info_msg(
-	uint32_t version,
-	const struct fi_info *hints,
-	void *src,
-	void *dest,
-	struct usd_device_attrs *dap,
-	struct fi_info **fi_first,
-	struct fi_info **fi_last)
-{
-	struct fi_info *fi;
-	struct fi_fabric_attr *fattrp;
-	uint32_t addr_format;
-	int ret;
-
-	fi = fi_allocinfo();
-	if (fi == NULL) {
-		ret = -FI_ENOMEM;
-		goto fail;
-	}
-
-	fi->caps = USDF_MSG_CAPS;
-
-	ret = validate_modebits(version, hints,
-				  USDF_MSG_SUPP_MODE, &fi->mode);
-	if (ret)
-		goto fail;
-
-	if (hints != NULL) {
-		addr_format = hints->addr_format;
-
-		/* check that we are capable of what's requested */
-		if ((hints->caps & ~USDF_MSG_CAPS) != 0) {
-			ret = -FI_ENODATA;
-			goto fail;
-		}
-
-		fi->handle = hints->handle;
-	} else {
-		addr_format = FI_FORMAT_UNSPEC;
-	}
-
-	fi->ep_attr->type = FI_EP_MSG;
-
-	ret = usdf_fill_addr_info(fi, addr_format, src, dest, dap);
-	if (ret != 0) {
-		goto fail;
-	}
-
-	/* fabric attrs */
-	fattrp = fi->fabric_attr;
-	ret = usdf_fabric_getname(version, dap, &fattrp->name);
-	if (ret < 0 || fattrp->name == NULL) {
-		ret = -FI_ENOMEM;
-		goto fail;
-	}
-
-	ret = usdf_msg_fill_ep_attr(hints, fi, dap);
-	if (ret)
-		goto fail;
-
-	ret = usdf_msg_fill_dom_attr(version, hints, fi, dap);
-	if (ret)
-		goto fail;
-
-	ret = usdf_msg_fill_tx_attr(version, hints, fi);
-	if (ret)
-		goto fail;
-
-	ret = usdf_msg_fill_rx_attr(version, hints, fi);
-	if (ret)
-		goto fail;
-
-	/* add to tail of list */
-	if (*fi_first == NULL) {
-		*fi_first = fi;
-	} else {
-		(*fi_last)->next = fi;
-	}
-	*fi_last = fi;
-
-	return 0;
-
-fail:
-	if (fi != NULL) {
-		fi_freeinfo(fi);
-	}
-	return ret;
-}
-
-static int usdf_fill_info_rdm(
-	uint32_t version,
-	const struct fi_info *hints,
-	void *src,
-	void *dest,
-	struct usd_device_attrs *dap,
-	struct fi_info **fi_first,
-	struct fi_info **fi_last)
-{
-	struct fi_info *fi;
-	struct fi_fabric_attr *fattrp;
-	uint32_t addr_format;
-	int ret;
-
-	fi = fi_allocinfo();
-	if (fi == NULL) {
-		ret = -FI_ENOMEM;
-		goto fail;
-	}
-
-	fi->caps = USDF_RDM_CAPS;
-
-	ret = validate_modebits(version, hints,
-				  USDF_RDM_SUPP_MODE, &fi->mode);
-	if (ret)
-		goto fail;
-
-	if (hints != NULL) {
-		addr_format = hints->addr_format;
-		/* check that we are capable of what's requested */
-		if ((hints->caps & ~USDF_RDM_CAPS) != 0) {
-			ret = -FI_ENODATA;
-			goto fail;
-		}
-
-		fi->handle = hints->handle;
-	} else {
-		addr_format = FI_FORMAT_UNSPEC;
-	}
-	fi->ep_attr->type = FI_EP_RDM;
-
-	ret = usdf_fill_addr_info(fi, addr_format, src, dest, dap);
-	if (ret != 0) {
-		goto fail;
-	}
-
-	/* fabric attrs */
-	fattrp = fi->fabric_attr;
-	ret = usdf_fabric_getname(version, dap, &fattrp->name);
-	if (ret < 0 || fattrp->name == NULL) {
-		ret = -FI_ENOMEM;
-		goto fail;
-	}
-
-	ret = usdf_rdm_fill_ep_attr(hints, fi, dap);
-	if (ret)
-		goto fail;
-
-	ret = usdf_rdm_fill_dom_attr(version, hints, fi, dap);
-	if (ret)
-		goto fail;
-
-	ret = usdf_rdm_fill_tx_attr(version, hints, fi);
-	if (ret)
-		goto fail;
-
-	ret = usdf_rdm_fill_rx_attr(version, hints, fi);
+	ret = usdf_alloc_fid_nic(fi, dap);
 	if (ret)
 		goto fail;
 
@@ -929,22 +829,6 @@ usdf_getinfo(uint32_t version, const char *node, const char *service,
 				goto fail;
 			}
 		}
-
-		if (ep_type == FI_EP_MSG || ep_type == FI_EP_UNSPEC) {
-			ret = usdf_fill_info_msg(version, hints, src, dest,
-					dap, &fi_first, &fi_last);
-			if (ret != 0 && ret != -FI_ENODATA) {
-				goto fail;
-			}
-		}
-
-		if (ep_type == FI_EP_RDM || ep_type == FI_EP_UNSPEC) {
-			ret = usdf_fill_info_rdm(version, hints, src, dest,
-					dap, &fi_first, &fi_last);
-			if (ret != 0 && ret != -FI_ENODATA) {
-				goto fail;
-			}
-		}
 	}
 
 	if (fi_first != NULL) {
@@ -994,8 +878,8 @@ usdf_fabric_close(fid_t fid)
 		pthread_join(fp->fab_thread, &rv);
 	}
 	usdf_timer_deinit(fp);
-	if (fp->fab_epollfd != -1) {
-		close(fp->fab_epollfd);
+	if (fp->fab_epollfd != OFI_EPOLL_INVALID) {
+		ofi_epoll_close(fp->fab_epollfd);
 	}
 	if (fp->fab_eventfd != -1) {
 		close(fp->fab_eventfd);
@@ -1033,7 +917,6 @@ usdf_fabric_open(struct fi_fabric_attr *fattrp, struct fid_fabric **fabric,
 	struct usdf_fabric *fp;
 	struct usdf_usnic_info *dp;
 	struct usdf_dev_entry *dep;
-	struct epoll_event ev;
 	struct sockaddr_in sin;
 	int ret;
 	int d;
@@ -1060,7 +943,7 @@ usdf_fabric_open(struct fi_fabric_attr *fattrp, struct fid_fabric **fabric,
 		USDF_INFO("unable to allocate memory for fabric\n");
 		return -FI_ENOMEM;
 	}
-	fp->fab_epollfd = -1;
+	fp->fab_epollfd = OFI_EPOLL_INVALID;
 	fp->fab_arp_sockfd = -1;
 	LIST_INIT(&fp->fab_domain_list);
 
@@ -1081,9 +964,8 @@ usdf_fabric_open(struct fi_fabric_attr *fattrp, struct fid_fabric **fabric,
 
 	fp->fab_dev_attrs = &dep->ue_dattr;
 
-	fp->fab_epollfd = epoll_create(1024);
-	if (fp->fab_epollfd == -1) {
-		ret = -errno;
+	ret = ofi_epoll_create(&fp->fab_epollfd);
+	if (ret) {
 		USDF_INFO("unable to allocate epoll fd\n");
 		goto fail;
 	}
@@ -1096,11 +978,9 @@ usdf_fabric_open(struct fi_fabric_attr *fattrp, struct fid_fabric **fabric,
 	}
 	fp->fab_poll_item.pi_rtn = usdf_fabric_progression_cb;
 	fp->fab_poll_item.pi_context = fp;
-	ev.events = EPOLLIN;
-	ev.data.ptr = &fp->fab_poll_item;
-	ret = epoll_ctl(fp->fab_epollfd, EPOLL_CTL_ADD, fp->fab_eventfd, &ev);
-	if (ret == -1) {
-		ret = -errno;
+	ret = ofi_epoll_add(fp->fab_epollfd, fp->fab_eventfd, OFI_EPOLL_IN,
+			    &fp->fab_poll_item);
+	if (ret) {
 		USDF_INFO("unable to EPOLL_CTL_ADD\n");
 		goto fail;
 	}
@@ -1162,7 +1042,7 @@ static void usdf_fini(void)
 struct fi_provider usdf_ops = {
 	.name = USDF_PROV_NAME,
 	.version = USDF_PROV_VERSION,
-	.fi_version = FI_VERSION(1, 6),
+	.fi_version = OFI_VERSION_LATEST,
 	.getinfo = usdf_getinfo,
 	.fabric = usdf_fabric_open,
 	.cleanup =  usdf_fini
