@@ -505,7 +505,7 @@ usd_create_wq_pio(
     /* packet buffer */
     wq->pio_v_pkt_buf = (void *)ivaddr;
     wq->pio_p_pkt_buf = pio_paddr + ivaddr - (uint64_t)pio_vaddr;
-    ivaddr += wq->uwq_num_entries * 256;
+    ivaddr += ((uint64_t) wq->uwq_num_entries) * 256;
 
     used_size = ivaddr - (uintptr_t)pio_vaddr;
     if (used_size > pio_memsize) {
@@ -586,7 +586,6 @@ usd_vnic_rq_init(
     iowrite32(rq->urq_cq->ucq_index, &vrq->ctrl->cq_index);
     iowrite32(0, &vrq->ctrl->error_interrupt_enable);
     iowrite32(0, &vrq->ctrl->error_interrupt_offset);
-    iowrite32(0, &vrq->ctrl->dropped_packet_count);
     iowrite32(0, &vrq->ctrl->error_status);
 
     rq->urq_state |= USD_QS_VNIC_INITIALIZED;
