@@ -72,40 +72,78 @@ static int ofi_fid_match(struct dlist_entry *entry, const void *fid)
 	return (item->fid == fid);
 }
 
-int fid_list_insert(struct dlist_entry *fid_list, fastlock_t *lock,
-		    struct fid *fid)
+/* Serialization must be provided by the caller. */
+int fid_list_search(struct dlist_entry *fid_list, struct fid *fid)
 {
-	int ret = 0;
 	struct dlist_entry *entry;
 	struct fid_list_entry *item;
 
-	fastlock_acquire(lock);
 	entry = dlist_find_first_match(fid_list, ofi_fid_match, fid);
 	if (entry)
-		goto out;
+		return -FI_EALREADY;
 
 	item = calloc(1, sizeof(*item));
-	if (!item) {
-		ret = -FI_ENOMEM;
-		goto out;
-	}
+	if (!item)
+		return -FI_ENOMEM;
 
 	item->fid = fid;
 	dlist_insert_tail(&item->entry, fid_list);
-out:
-	fastlock_release(lock);
-	return ret;
+	return 0;
 }
 
-void fid_list_remove(struct dlist_entry *fid_list, fastlock_t *lock,
+int fid_list_insert(struct dlist_entry *fid_list, ofi_mutex_t *lock,
+		    struct fid *fid)
+{
+	int ret = 0;
+
+	if (lock)
+		ofi_mutex_lock(lock);
+	ret = fid_list_search(fid_list, fid);
+	if (lock)
+		ofi_mutex_unlock(lock);
+
+	return (!ret || (ret == -FI_EALREADY)) ? 0 : ret;
+}
+
+int fid_list_insert2(struct dlist_entry *fid_list, struct ofi_genlock *lock,
+		    struct fid *fid)
+{
+	int ret = 0;
+
+	ofi_genlock_lock(lock);
+	ret = fid_list_search(fid_list, fid);
+	ofi_genlock_unlock(lock);
+
+	return (!ret || (ret == -FI_EALREADY)) ? 0 : ret;
+}
+
+void fid_list_remove(struct dlist_entry *fid_list, ofi_mutex_t *lock,
 		     struct fid *fid)
 {
 	struct fid_list_entry *item;
 	struct dlist_entry *entry;
 
-	fastlock_acquire(lock);
+	if (lock)
+		ofi_mutex_lock(lock);
 	entry = dlist_remove_first_match(fid_list, ofi_fid_match, fid);
-	fastlock_release(lock);
+	if (lock)
+		ofi_mutex_unlock(lock);
+
+	if (entry) {
+		item = container_of(entry, struct fid_list_entry, entry);
+		free(item);
+	}
+}
+
+void fid_list_remove2(struct dlist_entry *fid_list, struct ofi_genlock *lock,
+		      struct fid *fid)
+{
+	struct fid_list_entry *item;
+	struct dlist_entry *entry;
+
+	ofi_genlock_lock(lock);
+	entry = dlist_remove_first_match(fid_list, ofi_fid_match, fid);
+	ofi_genlock_unlock(lock);
 
 	if (entry) {
 		item = container_of(entry, struct fid_list_entry, entry);
@@ -127,6 +165,36 @@ static int util_find_domain(struct dlist_entry *item, const void *arg)
 		 ((info->domain_attr->mr_mode & domain->mr_mode) == domain->mr_mode);
 }
 
+static int ofi_dup_auth_keys(const struct fi_info *hints, struct fi_info *info)
+{
+	if (!hints)
+		return FI_SUCCESS;
+
+	if (hints->domain_attr && hints->domain_attr->auth_key) {
+		info->domain_attr->auth_key =
+			mem_dup(hints->domain_attr->auth_key,
+				hints->domain_attr->auth_key_size);
+		if (!info->domain_attr->auth_key)
+			return -FI_ENOMEM;
+
+		info->domain_attr->auth_key_size =
+			hints->domain_attr->auth_key_size;
+	}
+
+	if (hints->ep_attr && hints->ep_attr->auth_key) {
+		info->ep_attr->auth_key =
+			mem_dup(hints->ep_attr->auth_key,
+				hints->ep_attr->auth_key_size);
+		if (!info->ep_attr->auth_key)
+			return -FI_ENOMEM;
+
+		info->ep_attr->auth_key_size =
+			hints->ep_attr->auth_key_size;
+	}
+
+	return FI_SUCCESS;
+}
+
 /*
  * Produces 1 fi_info output for each fi_info entry in the provider's base
  * list (stored with util_prov), subject to the base fi_info meeting the
@@ -173,7 +241,7 @@ int util_getinfo(const struct util_prov *util_prov, uint32_t version,
 			FI_DBG(prov, FI_LOG_CORE, "Found opened fabric\n");
 			(*info)->fabric_attr->fabric = &fabric->fabric_fid;
 
-			fastlock_acquire(&fabric->lock);
+			ofi_mutex_lock(&fabric->lock);
 			item = dlist_find_first_match(&fabric->domain_list,
 						      util_find_domain, *info);
 			if (item) {
@@ -184,7 +252,7 @@ int util_getinfo(const struct util_prov *util_prov, uint32_t version,
 				(*info)->domain_attr->domain =
 						&domain->domain_fid;
 			}
-			fastlock_release(&fabric->lock);
+			ofi_mutex_unlock(&fabric->lock);
 
 		}
 		pthread_mutex_unlock(&common_locks.util_fabric_lock);
@@ -249,6 +317,10 @@ int util_getinfo(const struct util_prov *util_prov, uint32_t version,
 					"cannot resolve source address\n");
 			}
 		}
+
+		ret = ofi_dup_auth_keys(hints, *info);
+		if (ret)
+			goto err;
 	}
 
 	*info = saved_info;
@@ -283,7 +355,9 @@ static void util_set_netif_names(struct fi_info *info,
  * given fi_info input.
  */
 #if HAVE_GETIFADDRS
-static void util_getinfo_ifs(const struct util_prov *prov, struct fi_info *src_info,
+static void util_getinfo_ifs(const struct util_prov *prov,
+			     const struct fi_info *hints,
+			     struct fi_info *src_info,
 			     struct fi_info **head, struct fi_info **tail)
 {
 	struct fi_info *cur;
@@ -302,19 +376,6 @@ static void util_getinfo_ifs(const struct util_prov *prov, struct fi_info *src_i
 	slist_foreach(&addr_list, entry, prev) {
 		addr_entry = container_of(entry, struct ofi_addr_list_entry, entry);
 
-		cur = fi_dupinfo(src_info);
-		if (!cur)
-			break;
-
-		if (!*head) {
-			*head = cur;
-			FI_INFO(prov->prov, FI_LOG_CORE, "Chosen addr for using: %s,"
-				" speed %zu\n", addr_entry->ipstr, addr_entry->speed);
-		} else {
-			(*tail)->next = cur;
-		}
-		*tail = cur;
-
 		switch (addr_entry->ipaddr.sin.sin_family) {
 		case AF_INET:
 			addrlen = sizeof(struct sockaddr_in);
@@ -328,6 +389,26 @@ static void util_getinfo_ifs(const struct util_prov *prov, struct fi_info *src_i
 			continue;
 		}
 
+		if (hints && ((hints->caps & addr_entry->comm_caps) !=
+		    (hints->caps & (FI_LOCAL_COMM | FI_REMOTE_COMM)) ||
+		     !ofi_match_addr_format(addr_format, hints->addr_format)))
+			continue;
+
+		cur = fi_dupinfo(src_info);
+		if (!cur)
+			break;
+
+		if (!*head) {
+			*head = cur;
+			FI_INFO(prov->prov, FI_LOG_CORE, "Chosen addr for using: %s,"
+				" speed %zu\n", addr_entry->ipstr, addr_entry->speed);
+		} else {
+			(*tail)->next = cur;
+		}
+		*tail = cur;
+
+		cur->caps = (cur->caps & ~(FI_LOCAL_COMM | FI_REMOTE_COMM)) |
+			    addr_entry->comm_caps;
 		cur->src_addr = mem_dup(&addr_entry->ipaddr, addrlen);
 		if (cur->src_addr) {
 			cur->src_addrlen = addrlen;
@@ -343,7 +424,9 @@ static void util_getinfo_ifs(const struct util_prov *prov, struct fi_info *src_i
 	}
 }
 #else
-static void util_getinfo_ifs(const struct util_prov *prov, struct fi_info *src_info,
+static void util_getinfo_ifs(const struct util_prov *prov,
+			     const struct fi_info *hints,
+			     struct fi_info *src_info,
 			     struct fi_info **head, struct fi_info **tail)
 {
 	*head = src_info;
@@ -377,7 +460,7 @@ int ofi_ip_getinfo(const struct util_prov *prov, uint32_t version,
 	prev = info;
 	for (cur = *info; cur; cur = cur->next) {
 		if (!cur->src_addr && !cur->dest_addr) {
-			util_getinfo_ifs(prov, cur, &head, &tail);
+			util_getinfo_ifs(prov, hints, cur, &head, &tail);
 			if (head != cur) {
 				tail->next = (*prev)->next;
 				*prev = head;
