@@ -17,6 +17,7 @@
  * Copyright (c) 2016-2020 IBM Corporation.  All rights reserved.
  * Copyright (c) 2019 Intel Corporation, Inc.  All rights reserved.
  * Copyright (c) 2020 Amazon.com, Inc. or its affiliates. All rights reserved.
+ * Copyright (c) 2023 Tactical Computing Labs, LLC. All rights reserved.
  *
  * License text from Open-MPI (www.open-mpi.org/community/license.php)
  *
@@ -68,6 +69,7 @@ struct ofi_memhooks memhooks = {
 	.monitor.cleanup = ofi_monitor_cleanup,
 	.monitor.start = ofi_memhooks_start,
 	.monitor.stop = ofi_memhooks_stop,
+	.monitor.name = "memhooks",
 };
 struct ofi_mem_monitor *memhooks_monitor = &memhooks.monitor;
 
@@ -107,12 +109,15 @@ struct ofi_mem_monitor *memhooks_monitor = &memhooks.monitor;
 
 #define OFI_INTERCEPT_MAX_PATCH 32
 
+static bool symbols_intercepted;
+
 struct ofi_intercept {
 	struct dlist_entry 		entry;
 	const char			*symbol;
 	void				*our_func;
 	void				*orig_func;
-	unsigned char			patch_data[OFI_INTERCEPT_MAX_PATCH];
+	unsigned char			patch_data[OFI_INTERCEPT_MAX_PATCH]
+		                        __attribute__ ((aligned(32)));
 	unsigned char			patch_orig_data[OFI_INTERCEPT_MAX_PATCH];
 	unsigned			patch_data_size;
 	struct dlist_entry		dl_intercept_list;
@@ -176,7 +181,7 @@ static inline void ofi_clear_instruction_cache(uintptr_t address, size_t data_si
 {
 	size_t i;
 	size_t offset_jump = 16;
-#if defined(__aarch64__)
+#if (defined(__aarch64__) || (defined(__riscv) && (__riscv_xlen == 64)))
 	offset_jump = 32;
 #endif
 	/* align the address */
@@ -186,12 +191,15 @@ static inline void ofi_clear_instruction_cache(uintptr_t address, size_t data_si
 #if (defined(__x86_64__) || defined(__amd64__))
 		__asm__ volatile("mfence;clflush %0;mfence"::
 				 "m" (*((char*) address + i)));
-#elif defined(__aarch64__)
+#elif (defined(__aarch64__))
 		__asm__ volatile ("dc cvau, %0\n\t"
 			  "dsb ish\n\t"
 			  "ic ivau, %0\n\t"
 			  "dsb ish\n\t"
 			  "isb":: "r" (address + i));
+#elif (defined(__riscv) && (__riscv_xlen == 64))
+	        __riscv_flush_icache(address, address+data_size, SYS_RISCV_FLUSH_ICACHE_LOCAL);
+		__asm__ volatile ("fence.i\n");
 #endif
 	}
 }
@@ -213,8 +221,8 @@ static inline int ofi_write_patch(unsigned char *patch_data, void *address,
 	}
 
 	base = ofi_get_page_start(address, page_size);
-	bound = ofi_get_page_end(address, page_size);
-	length = (uintptr_t) bound - (uintptr_t) base;
+	bound = ofi_get_page_end( (void *) ((uintptr_t) address + data_size - 1), page_size);
+	length = (uintptr_t) bound - (uintptr_t) base + 1;
 
 	if (mprotect(base, length, PROT_EXEC|PROT_READ|PROT_WRITE)) {
 		FI_WARN(&core_prov, FI_LOG_MR,
@@ -373,6 +381,103 @@ static bool ofi_is_function_patched(struct ofi_intercept *intercept)
         ((*(uint32_t *) (addr + 16)) & br_mask) == br(0)
 	);
 }
+#elif (defined(__riscv) && \
+       defined(__riscv_xlen) && \
+       (__riscv_xlen == 64))
+
+/* Registers numbers to use with the move immediate to register.
+ * The destination register is X31 (highest temporary).
+ * Register X28-X30 are used for block shifting and masking.
+ * Register X0 is always zero
+ */
+#define X31 31
+#define X30 30
+#define X0  0
+
+/**
+ * jalr
+ *
+ * Add 12 bit immediate to source register
+ * save to destination register
+ * jump and link from destination register
+ *
+ */
+#define jalr(_regd, _regs, _imm) \
+    (((_imm) << 20) | ((_regs) << 15) | (0b000 << 12) | ((_regd) << 7) | (0x67))
+
+/**
+ * addi
+ *
+ * Add 12 bit immediate to source register
+ * save to destination register 
+ *
+ */
+#define addi(_regd, _regs, _imm) \
+    (((_imm) << 20) | ((_regs) << 15) | (0b000 << 12) | ((_regd) << 7) | (0x13))
+
+#define add(_regd, _regs_a, _regs_b) \
+    ((_regs_b << 20) | (_regs_a << 15) | (0b000 << 12) | ((_regd) << 7) | (0x33))
+
+/**
+ * lui
+ *
+ * load upper 20 bit immediate to destination register
+ *
+ */
+#define lui(_regd, _imm) \
+    (((_imm) << 12) | ((_regd) << 7) | (0x37))
+
+/**
+ * slli
+ *
+ * left-shift immediate number of bits in source register into destination register
+ *
+ */
+#define slli(_regd, _regs, _imm) \
+    (((_imm) << 20) | ((_regs) << 15) | (0b001 << 12) | ((_regd) << 7) | (0x13))
+
+static int ofi_patch_function(struct ofi_intercept *intercept)
+{
+    /*
+     * r15 is the highest numbered temporary register. I am
+     * assuming this one is safe to use.
+     */
+    uintptr_t addr __attribute__ ((aligned(32))) = (uintptr_t) intercept->patch_data;
+    uintptr_t value __attribute__ ((aligned(32))) = (uintptr_t) intercept->our_func;
+
+    uint32_t fnaddr_hi = (uint32_t)(value >> 32) + ((uint32_t)(value >> 31) & 1);
+    uint32_t fnaddr_lo = (uint32_t) value;
+
+    *(uint32_t*) (addr +  0) = lui  (X31, ((fnaddr_hi >> 12) + ((fnaddr_hi >> 11) & 1)) & 0xFFFFF);
+    *(uint32_t*) (addr +  4) = addi (X31, X31, fnaddr_hi & 0xFFF );
+    *(uint32_t*) (addr +  8) = lui  (X30, ((fnaddr_lo >> 12) + ((fnaddr_lo >> 11) & 1)) & 0xFFFFF);
+    *(uint32_t*) (addr + 12) = slli (X31, X31, 32);
+    *(uint32_t*) (addr + 16) = add  (X31, X31, X30);
+    *(uint32_t*) (addr + 20) = jalr (X0, X31, fnaddr_lo & 0xFFF);
+
+    intercept->patch_data_size = 24;
+
+    return ofi_apply_patch(intercept);
+}
+
+/*
+ * Please see comments at other ofi_is_function_patched() function
+ */
+static bool ofi_is_function_patched(struct ofi_intercept *intercept)
+{
+    uintptr_t addr = (uintptr_t) intercept->orig_func;
+    /*
+     * focus on the instructions in the bytes.
+     */
+    return (
+        ((*(uint32_t *) (addr +  0)) & 0xFF) == 0x37 &&
+        ((*(uint32_t *) (addr +  4)) & 0xFF) == 0x13 &&
+        ((*(uint32_t *) (addr +  8)) & 0xFF) == 0x37 &&
+        ((*(uint32_t *) (addr + 12)) & 0xFF) == 0x13 &&
+        ((*(uint32_t *) (addr + 16)) & 0xFF) == 0x33 &&
+        ((*(uint32_t *) (addr + 20)) & 0xFF) == 0x67
+	);
+}
 #endif
 
 /*
@@ -610,6 +715,8 @@ static int ofi_memhooks_start(struct ofi_mem_monitor *monitor)
 		}
 	}
 
+	symbols_intercepted = true;
+
 	return 0;
 
 err_intercept_failed:
@@ -628,6 +735,12 @@ static void ofi_memhooks_stop(struct ofi_mem_monitor *monitor)
 	memhooks_monitor->unsubscribe = NULL;
 }
 
+void ofi_memhooks_atfork_handler(void)
+{
+	if (symbols_intercepted)
+		ofi_restore_intercepts();
+}
+
 #else
 
 static int ofi_memhooks_start(struct ofi_mem_monitor *monitor)
@@ -639,4 +752,8 @@ static void ofi_memhooks_stop(struct ofi_mem_monitor *monitor)
 {
 }
 
+void ofi_memhooks_atfork_handler(void)
+{
+}
+
 #endif /* memhook support checks */
