@@ -31,14 +31,15 @@
  */
 #include <config.h>
 #include <stdlib.h>
-#include <ofi_enosys.h>
-#include <ofi_util.h>
-#include <ofi_mr.h>
+#include "ofi_enosys.h"
+#include "ofi_util.h"
+#include "ofi_mr.h"
+#include "ofi_hmem.h"
 #include <assert.h>
 
 
 static struct fi_mr_attr *
-dup_mr_attr(const struct fi_mr_attr *attr)
+dup_mr_attr(const struct fi_mr_attr *attr, uint64_t flags)
 {
 	struct fi_mr_attr *dup_attr;
 
@@ -49,19 +50,29 @@ dup_mr_attr(const struct fi_mr_attr *attr)
 
 	*dup_attr = *attr;
 	dup_attr->mr_iov = (struct iovec *) (dup_attr + 1);
-	memcpy((void *) dup_attr->mr_iov, attr->mr_iov,
-		sizeof(*attr->mr_iov) * attr->iov_count);
+
+	/*
+	 * dup_mr_attr is only used insided ofi_mr_map_insert.
+	 * dmabuf must be converted to iov before the attr
+	 * is inserted to the mr_map
+	 */
+	if (flags & FI_MR_DMABUF)
+		ofi_mr_get_iov_from_dmabuf((struct iovec *)dup_attr->mr_iov,
+			attr->dmabuf, attr->iov_count);
+	else
+		memcpy((void *) dup_attr->mr_iov, attr->mr_iov,
+			sizeof(*attr->mr_iov) * attr->iov_count);
 
 	return dup_attr;
 }
 
 int ofi_mr_map_insert(struct ofi_mr_map *map, const struct fi_mr_attr *attr,
-		      uint64_t *key, void *context)
+		      uint64_t *key, void *context, uint64_t flags)
 {
 	struct fi_mr_attr *item;
 	int ret;
 
-	item = dup_mr_attr(attr);
+	item = dup_mr_attr(attr, flags);
 	if (!item)
 		return -FI_ENOMEM;
 
@@ -188,10 +199,10 @@ int ofi_mr_map_init(const struct fi_provider *prov, int mode,
 		return -FI_ENOMEM;
 
 	switch (mode) {
-	case FI_MR_BASIC:
+	case OFI_MR_BASIC:
 		map->mode = OFI_MR_BASIC_MAP;
 		break;
-	case FI_MR_SCALABLE:
+	case OFI_MR_SCALABLE:
 		map->mode = 0;
 		break;
 	default:
@@ -236,9 +247,14 @@ static struct fi_ops ofi_mr_fi_ops = {
 
 void ofi_mr_update_attr(uint32_t user_version, uint64_t caps,
 			const struct fi_mr_attr *user_attr,
-			struct fi_mr_attr *cur_abi_attr)
+			struct fi_mr_attr *cur_abi_attr,
+			uint64_t flags)
 {
-	cur_abi_attr->mr_iov = (struct iovec *) user_attr->mr_iov;
+	if (FI_VERSION_GE(user_version, FI_VERSION(1, 20))
+	    && (flags & FI_MR_DMABUF))
+		cur_abi_attr->dmabuf = user_attr->dmabuf;
+	else
+		cur_abi_attr->mr_iov = (struct iovec *) user_attr->mr_iov;
 	cur_abi_attr->iov_count = user_attr->iov_count;
 	cur_abi_attr->access = user_attr->access;
 	cur_abi_attr->offset = user_attr->offset;
@@ -256,9 +272,26 @@ void ofi_mr_update_attr(uint32_t user_version, uint64_t caps,
 	if (caps & FI_HMEM) {
 		cur_abi_attr->iface = user_attr->iface;
 		cur_abi_attr->device = user_attr->device;
+		cur_abi_attr->hmem_data = FI_VERSION_GE(user_version, FI_VERSION(1, 19))
+		                          ? user_attr->hmem_data
+		                          : NULL;
 	} else {
 		cur_abi_attr->iface = FI_HMEM_SYSTEM;
 		cur_abi_attr->device.reserved = 0;
+		cur_abi_attr->hmem_data = NULL;
+	}
+
+	if (FI_VERSION_GE(user_version, FI_VERSION(1, 22)))
+		cur_abi_attr->page_size = user_attr->page_size;
+	else
+		cur_abi_attr->page_size = 0;
+
+	if (FI_VERSION_GE(user_version, FI_VERSION(2, 0))) {
+		cur_abi_attr->base_mr = user_attr->base_mr;
+		cur_abi_attr->sub_mr_cnt = user_attr->sub_mr_cnt;
+	} else {
+		cur_abi_attr->base_mr = NULL;
+		cur_abi_attr->sub_mr_cnt = 0;
 	}
 }
 
@@ -287,7 +320,7 @@ int ofi_mr_regattr(struct fid *fid, const struct fi_mr_attr *attr,
 		return -FI_ENOMEM;
 
 	ofi_mr_update_attr(domain->fabric->fabric_fid.api_version,
-			   domain->info_domain_caps, attr, &cur_abi_attr);
+			   domain->info_domain_caps, attr, &cur_abi_attr, flags);
 
 	if ((flags & FI_HMEM_HOST_ALLOC) && (attr->iface == FI_HMEM_ZE))
 		cur_abi_attr.device.ze = -1;
@@ -308,8 +341,9 @@ int ofi_mr_regattr(struct fid *fid, const struct fi_mr_attr *attr,
 	mr->flags = flags;
 	mr->iface = cur_abi_attr.iface;
 	mr->device = cur_abi_attr.device.reserved;
+	mr->hmem_data = cur_abi_attr.hmem_data;
 
-	ret = ofi_mr_map_insert(&domain->mr_map, &cur_abi_attr, &key, mr);
+	ret = ofi_mr_map_insert(&domain->mr_map, &cur_abi_attr, &key, mr, flags);
 	if (ret) {
 		free(mr);
 		goto out;
@@ -341,6 +375,7 @@ int ofi_mr_regv(struct fid *fid, const struct iovec *iov,
 	attr.context = context;
 	attr.iface = FI_HMEM_SYSTEM;
 	attr.device.reserved = 0;
+	attr.hmem_data = NULL;
 
 	return ofi_mr_regattr(fid, &attr, flags, mr_fid);
 }
