@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2015 Intel Corporation, Inc.  All rights reserved.
+ * Copyright (c) Intel Corporation, Inc.  All rights reserved.
  *
  * This software is available to you under a choice of one of two
  * licenses.  You may choose to be licensed under the terms of the GNU
@@ -35,114 +35,103 @@
 #include <stdint.h>
 #include <ofi_mem.h>
 
-#include "fi_verbs.h"
+#include "verbs_ofi.h"
 
-static void vrb_cq_read_context_entry(struct ibv_wc *wc, void *buf)
+
+enum ibv_wc_opcode vrb_wr2wc_opcode(enum ibv_wr_opcode wr)
 {
-	struct fi_cq_entry *entry = buf;
+	static enum ibv_wc_opcode wc[] = {
+		[IBV_WR_RDMA_WRITE] = IBV_WC_RDMA_WRITE,
+		[IBV_WR_RDMA_WRITE_WITH_IMM] = IBV_WC_RDMA_WRITE,
+		[IBV_WR_SEND] = IBV_WC_SEND,
+		[IBV_WR_SEND_WITH_IMM] = IBV_WC_SEND,
+		[IBV_WR_RDMA_READ] = IBV_WC_RDMA_READ,
+		[IBV_WR_ATOMIC_CMP_AND_SWP] = IBV_WC_COMP_SWAP,
+		[IBV_WR_ATOMIC_FETCH_AND_ADD] = IBV_WC_FETCH_ADD,
+	};
 
-	entry->op_context = (void *) (uintptr_t) wc->wr_id;
+	return (wr < ARRAY_SIZE(wc)) ? wc[wr] : IBV_WC_SEND;
 }
 
-static void vrb_cq_read_msg_entry(struct ibv_wc *wc, void *buf)
+static void
+vrb_get_cq_info(struct ibv_wc *wc, uint64_t *flags, uint64_t *data,
+		size_t *len)
 {
-	struct fi_cq_msg_entry *entry = buf;
-
-	entry->op_context = (void *) (uintptr_t) wc->wr_id;
-
 	switch (wc->opcode) {
 	case IBV_WC_SEND:
-		entry->flags = (FI_SEND | FI_MSG);
+		*len = 0;
+		*flags = (FI_SEND | FI_MSG);
+		*data = 0;
 		break;
 	case IBV_WC_RDMA_WRITE:
-		entry->flags = (FI_RMA | FI_WRITE);
+		*len = 0;
+		*flags = (FI_RMA | FI_WRITE);
+		*data = 0;
 		break;
 	case IBV_WC_RDMA_READ:
-		entry->flags = (FI_RMA | FI_READ);
+		*len = 0;
+		*flags = (FI_RMA | FI_READ);
+		*data = 0;
 		break;
 	case IBV_WC_COMP_SWAP:
-		entry->flags = FI_ATOMIC;
+		*len = 0;
+		*flags = FI_ATOMIC;
+		*data = 0;
 		break;
 	case IBV_WC_FETCH_ADD:
-		entry->flags = FI_ATOMIC;
+		*len = 0;
+		*flags = FI_ATOMIC;
+		*data = 0;
 		break;
 	case IBV_WC_RECV:
-		entry->len = wc->byte_len;
-		entry->flags = (FI_RECV | FI_MSG);
+		*len = wc->byte_len;
+		if (wc->wc_flags & IBV_WC_WITH_IMM) {
+			*flags = (FI_RECV | FI_MSG | FI_REMOTE_CQ_DATA);
+			*data = ntohl(wc->imm_data);
+		} else {
+			*flags = (FI_RECV | FI_MSG);
+			*data = 0;
+		}
 		break;
 	case IBV_WC_RECV_RDMA_WITH_IMM:
-		entry->len = wc->byte_len;
-		entry->flags = (FI_RMA | FI_REMOTE_WRITE);
+		*len = wc->byte_len;
+		*flags = (FI_RMA | FI_REMOTE_WRITE | FI_REMOTE_CQ_DATA);
+		*data = ntohl(wc->imm_data);
 		break;
 	default:
+		*len = 0;
+		*flags = 0;
+		*data = 0;
 		break;
 	}
 }
 
-static void vrb_cq_read_data_entry(struct ibv_wc *wc, void *buf)
-{
-	struct fi_cq_data_entry *entry = buf;
-
-	/* fi_cq_data_entry can cast to fi_cq_msg_entry */
-	vrb_cq_read_msg_entry(wc, buf);
-	if ((wc->wc_flags & IBV_WC_WITH_IMM) &&
-	    (wc->opcode & IBV_WC_RECV)) {
-		entry->data = ntohl(wc->imm_data);
-		entry->flags |= FI_REMOTE_CQ_DATA;
-	}
-}
-
 static ssize_t
-vrb_cq_readerr(struct fid_cq *cq_fid, struct fi_cq_err_entry *entry,
-		  uint64_t flags)
+vrb_cq_readfrom(struct fid_cq *cq_fid, void *buf, size_t count,
+		fi_addr_t *src_addr)
 {
 	struct vrb_cq *cq;
-	struct vrb_wc_entry *wce;
-	struct slist_entry *slist_entry;
-	uint32_t api_version;
+	ssize_t ret;
 
 	cq = container_of(cq_fid, struct vrb_cq, util_cq.cq_fid);
+	ofi_genlock_lock(vrb_cq2_progress(cq)->active_lock);
+	ret = ofi_cq_readfrom(cq_fid, buf, count, src_addr);
+	ofi_genlock_unlock(vrb_cq2_progress(cq)->active_lock);
+	return ret;
+}
 
-	cq->util_cq.cq_fastlock_acquire(&cq->util_cq.cq_lock);
-	if (slist_empty(&cq->saved_wc_list))
-		goto err;
+static ssize_t
+vrb_cq_readerr(struct fid_cq *cq_fid, struct fi_cq_err_entry *entry,
+	       uint64_t flags)
+{
+	struct vrb_cq *cq;
+	ssize_t ret;
 
-	wce = container_of(cq->saved_wc_list.head, struct vrb_wc_entry, entry);
-	if (!wce->wc.status)
-		goto err;
-
-	api_version = cq->util_cq.domain->fabric->fabric_fid.api_version;
-
-	slist_entry = slist_remove_head(&cq->saved_wc_list);
-	cq->util_cq.cq_fastlock_release(&cq->util_cq.cq_lock);
-
-	wce = container_of(slist_entry, struct vrb_wc_entry, entry);
-
-	entry->op_context = (void *)(uintptr_t)wce->wc.wr_id;
-	entry->prov_errno = wce->wc.status;
-	if (wce->wc.status == IBV_WC_WR_FLUSH_ERR)
-		entry->err = FI_ECANCELED;
-	else
-		entry->err = EIO;
-
-	/* fi_cq_err_entry can cast to fi_cq_data_entry */
-	vrb_cq_read_data_entry(&wce->wc, (void *) entry);
-
-	if ((FI_VERSION_GE(api_version, FI_VERSION(1, 5))) &&
-		entry->err_data && entry->err_data_size) {
-		entry->err_data_size = MIN(entry->err_data_size,
-			sizeof(wce->wc.vendor_err));
-		memcpy(entry->err_data, &wce->wc.vendor_err, entry->err_data_size);
-	} else {
-		memcpy(&entry->err_data, &wce->wc.vendor_err,
-			sizeof(wce->wc.vendor_err));
-	}
-
-	ofi_buf_free(wce);
-	return 1;
-err:
-	cq->util_cq.cq_fastlock_release(&cq->util_cq.cq_lock);
-	return -FI_EAGAIN;
+	cq = container_of(cq_fid, struct vrb_cq, util_cq.cq_fid);
+	ofi_genlock_lock(vrb_cq2_progress(cq)->active_lock);
+	ret = ofi_cq_readerr(cq_fid, entry, flags);
+	ofi_genlock_unlock(vrb_cq2_progress(cq)->active_lock);
+	return ret;
 }
 
 static inline int
@@ -151,10 +140,9 @@ vrb_poll_events(struct vrb_cq *_cq, int timeout)
 	int ret, rc;
 	void *context;
 	struct pollfd fds[2];
-	char data;
 
 	fds[0].fd = _cq->channel->fd;
-	fds[1].fd = _cq->signal_fd[0];
+	fds[1].fd = fd_signal_get(&_cq->signal);
 
 	fds[0].events = fds[1].events = POLLIN;
 
@@ -173,14 +161,12 @@ vrb_poll_events(struct vrb_cq *_cq, int timeout)
 		rc--;
 	}
 	if (fds[1].revents & POLLIN) {
-		do {
-			ret = read(fds[1].fd, &data, 1);
-		} while (ret > 0);
+		fd_signal_reset(&_cq->signal);
 		ret = -FI_EAGAIN;
 		rc--;
 	}
 	if (rc) {
-		VERBS_WARN(FI_LOG_CQ, "Unknown poll error: check revents\n");
+		VRB_WARN(FI_LOG_CQ, "Unknown poll error: check revents\n");
 		return -FI_EOTHER;
 	}
 
@@ -226,12 +212,13 @@ vrb_cq_sread(struct fid_cq *cq, void *buf, size_t count, const void *cond,
 	return cur ? cur : ret;
 }
 
-/* Must be called with CQ lock held. */
 int vrb_poll_cq(struct vrb_cq *cq, struct ibv_wc *wc)
 {
 	struct vrb_context *ctx;
+	struct vrb_ep *ep;
 	int ret;
 
+	assert(ofi_genlock_held(vrb_cq2_progress(cq)->active_lock));
 	do {
 		ret = ibv_poll_cq(cq->cq, 1, wc);
 		if (ret <= 0)
@@ -239,66 +226,87 @@ int vrb_poll_cq(struct vrb_cq *cq, struct ibv_wc *wc)
 
 		ctx = (struct vrb_context *) (uintptr_t) wc->wr_id;
 		wc->wr_id = (uintptr_t) ctx->user_ctx;
-		if (ctx->flags & FI_TRANSMIT) {
-			cq->credits++;
-			ctx->ep->sq_credits++;
-		}
+		if (wc->status != IBV_WC_SUCCESS && wc->status != IBV_WC_WR_FLUSH_ERR)
+			vrb_shutdown_ep(ctx->ep);
+		if (ctx->op_queue == VRB_OP_SQ) {
+			ep = ctx->ep;
+			assert(ep);
+			assert(!slist_empty(&ep->sq_list));
+			assert(ep->sq_list.head == &ctx->entry);
+			(void) slist_remove_head(&ep->sq_list);
+			ep->sq_credits++;
 
-		if (wc->status) {
-			if (ctx->flags & FI_RECV)
-				wc->opcode |= IBV_WC_RECV;
-			else
-				wc->opcode &= ~IBV_WC_RECV;
-		}
-		if (ctx->srx) {
-			fastlock_acquire(&ctx->srx->ctx_lock);
-			ofi_buf_free(ctx);
-			fastlock_release(&ctx->srx->ctx_lock);
+			/* workaround incorrect opcode reported by verbs */
+			wc->opcode = vrb_wr2wc_opcode(ctx->sq_opcode);
+
+		} else if (ctx->op_queue == VRB_OP_RQ) {
+			ep = ctx->ep;
+			assert(ep);
+			assert(!slist_empty(&ep->rq_list));
+			assert(ep->rq_list.head == &ctx->entry);
+			(void) slist_remove_head(&ep->rq_list);
+			if (wc->status)
+				wc->opcode = IBV_WC_RECV;
 		} else {
-			ofi_buf_free(ctx);
+			assert(ctx->op_queue == VRB_OP_SRQ);
+			wc->opcode = IBV_WC_RECV;
 		}
 
+		vrb_free_ctx(vrb_cq2_progress(cq), ctx);
 	} while (wc->wr_id == VERBS_NO_COMP_FLAG);
 
 	return ret;
 }
 
-/* Must be called with CQ lock held. */
-int vrb_save_wc(struct vrb_cq *cq, struct ibv_wc *wc)
+void vrb_report_wc(struct vrb_cq *cq, struct ibv_wc *wc)
 {
-	struct vrb_wc_entry *wce;
+	struct fi_cq_err_entry err_entry;
+	uint64_t flags, data;
+	size_t len;
 
-	wce = ofi_buf_alloc(cq->wce_pool);
-	if (!wce) {
-		FI_WARN(&vrb_prov, FI_LOG_CQ,
-			"Unable to save completion, completion lost!\n");
-		return -FI_ENOMEM;
+	assert(ofi_genlock_held(vrb_cq2_progress(cq)->active_lock));
+
+	if (wc->status) {
+		vrb_get_cq_info(wc, &err_entry.flags, &err_entry.data, &len);
+
+		err_entry.op_context = (void *) (uintptr_t) wc->wr_id;
+		err_entry.len = 0;
+		err_entry.buf = NULL;
+		err_entry.tag = 0;
+		err_entry.olen = 0;
+
+		err_entry.err = (wc->status == IBV_WC_WR_FLUSH_ERR) ?
+				FI_ECANCELED : FI_EIO;
+		err_entry.prov_errno = (int) wc->status;
+		err_entry.err_data = NULL;
+		err_entry.err_data_size = 0;
+
+		(void) ofi_cq_write_error(&cq->util_cq, &err_entry);
+	} else {
+		vrb_get_cq_info(wc, &flags, &data, &len);
+		(void) ofi_cq_write(&cq->util_cq, (void *) (uintptr_t) wc->wr_id,
+				    flags, len, NULL, data, 0);
 	}
-
-	wce->wc = *wc;
-	slist_insert_tail(&wce->entry, &cq->saved_wc_list);
-	return FI_SUCCESS;
 }
 
-static void vrb_flush_cq(struct vrb_cq *cq)
+void vrb_flush_cq(struct vrb_cq *cq)
 {
 	struct ibv_wc wc;
 	ssize_t ret;
 
-	cq->util_cq.cq_fastlock_acquire(&cq->util_cq.cq_lock);
+	assert(ofi_genlock_held(vrb_cq2_progress(cq)->active_lock));
 	while (1) {
 		ret = vrb_poll_cq(cq, &wc);
 		if (ret <= 0)
 			break;
 
-		vrb_save_wc(cq, &wc);
+		vrb_report_wc(cq, &wc);
 	};
-
-	cq->util_cq.cq_fastlock_release(&cq->util_cq.cq_lock);
 }
 
 void vrb_cleanup_cq(struct vrb_ep *ep)
 {
+	assert(ofi_genlock_held(vrb_ep2_progress(ep)->active_lock));
 	if (ep->util_ep.rx_cq) {
 		vrb_flush_cq(container_of(ep->util_ep.rx_cq,
 					  struct vrb_cq, util_cq));
@@ -309,135 +317,70 @@ void vrb_cleanup_cq(struct vrb_ep *ep)
 	}
 }
 
-static ssize_t vrb_cq_read(struct fid_cq *cq_fid, void *buf, size_t count)
-{
-	struct vrb_cq *cq;
-	struct vrb_wc_entry *wce;
-	struct slist_entry *entry;
-	struct ibv_wc wc;
-	ssize_t ret = 0, i;
-
-	cq = container_of(cq_fid, struct vrb_cq, util_cq.cq_fid);
-
-	cq->util_cq.cq_fastlock_acquire(&cq->util_cq.cq_lock);
-
-	for (i = 0; i < count; i++) {
-		if (!slist_empty(&cq->saved_wc_list)) {
-			wce = container_of(cq->saved_wc_list.head,
-					   struct vrb_wc_entry, entry);
-			if (wce->wc.status) {
-				ret = -FI_EAVAIL;
-				break;
-			}
-			entry = slist_remove_head(&cq->saved_wc_list);
-			wce = container_of(entry, struct vrb_wc_entry, entry);
-			cq->read_entry(&wce->wc, (char *) buf + i * cq->entry_size);
-			ofi_buf_free(wce);
-			continue;
-		}
-
-		ret = vrb_poll_cq(cq, &wc);
-		if (ret <= 0)
-			break;
-
-		if (wc.status) {
-			wce = ofi_buf_alloc(cq->wce_pool);
-			if (!wce) {
-				cq->util_cq.cq_fastlock_release(&cq->util_cq.cq_lock);
-				return -FI_ENOMEM;
-			}
-			memset(wce, 0, sizeof(*wce));
-			memcpy(&wce->wc, &wc, sizeof wc);
-			slist_insert_tail(&wce->entry, &cq->saved_wc_list);
-			ret = -FI_EAVAIL;
-			break;
-		}
-
-		cq->read_entry(&wc, (char *)buf + i * cq->entry_size);
-	}
-
-	cq->util_cq.cq_fastlock_release(&cq->util_cq.cq_lock);
-	return i ? i : (ret < 0 ? ret : -FI_EAGAIN);
-}
-
 static const char *
 vrb_cq_strerror(struct fid_cq *eq, int prov_errno, const void *err_data,
 		   char *buf, size_t len)
 {
-	if (buf && len)
+	if (buf && len) {
 		strncpy(buf, ibv_wc_status_str(prov_errno), len);
+		return buf;
+	}
 	return ibv_wc_status_str(prov_errno);
 }
 
 int vrb_cq_signal(struct fid_cq *cq)
 {
 	struct vrb_cq *_cq;
-	char data = '0';
 
 	_cq = container_of(cq, struct vrb_cq, util_cq.cq_fid);
 
-	if (write(_cq->signal_fd[1], &data, 1) != 1) {
-		VERBS_WARN(FI_LOG_CQ, "Error signalling CQ\n");
-		return -errno;
-	}
+	fd_signal_set(&_cq->signal);
 
 	return 0;
 }
 
 int vrb_cq_trywait(struct vrb_cq *cq)
 {
-	struct ibv_wc wc;
 	void *context;
-	int ret = -FI_EAGAIN, rc;
+	int ret;
 
 	if (!cq->channel) {
-		VERBS_WARN(FI_LOG_CQ, "No wait object object associated with CQ\n");
+		VRB_WARN(FI_LOG_CQ, "No wait object object associated with CQ\n");
 		return -FI_EINVAL;
 	}
 
-	cq->util_cq.cq_fastlock_acquire(&cq->util_cq.cq_lock);
-	if (!slist_empty(&cq->saved_wc_list))
-		goto out;
-
-	rc = vrb_poll_cq(cq, &wc);
-	if (rc) {
-		if (rc > 0)
-			vrb_save_wc(cq, &wc);
+	ofi_genlock_lock(vrb_cq2_progress(cq)->active_lock);
+	if (!ofi_cirque_isempty(cq->util_cq.cirq)) {
+		ret = -FI_EAGAIN;
 		goto out;
 	}
 
 	while (!ibv_get_cq_event(cq->channel, &cq->cq, &context))
 		ofi_atomic_inc32(&cq->nevents);
 
-	rc = ibv_req_notify_cq(cq->cq, 0);
-	if (rc) {
-		VERBS_WARN(FI_LOG_CQ, "ibv_req_notify_cq error: %d\n", ret);
+	ret = ibv_req_notify_cq(cq->cq, 0);
+	if (ret) {
+		VRB_WARN(FI_LOG_CQ, "ibv_req_notify_cq error: %d\n", ret);
 		ret = -errno;
 		goto out;
 	}
 
-	/* Read again to fetch any completions that we might have missed
-	 * while rearming */
-	rc = vrb_poll_cq(cq, &wc);
-	if (rc) {
-		if (rc > 0)
-			vrb_save_wc(cq, &wc);
-		goto out;
-	}
+	/* Fetch any completions that we might have missed while rearming */
+	vrb_flush_cq(cq);
+	ret = ofi_cirque_isempty(cq->util_cq.cirq) ? FI_SUCCESS : -FI_EAGAIN;
 
-	ret = FI_SUCCESS;
 out:
-	cq->util_cq.cq_fastlock_release(&cq->util_cq.cq_lock);
+	ofi_genlock_unlock(vrb_cq2_progress(cq)->active_lock);
 	return ret;
 }
 
 static struct fi_ops_cq vrb_cq_ops = {
 	.size = sizeof(struct fi_ops_cq),
-	.read = vrb_cq_read,
-	.readfrom = fi_no_cq_readfrom,
+	.read = ofi_cq_read,
+	.readfrom = vrb_cq_readfrom,
 	.readerr = vrb_cq_readerr,
 	.sread = vrb_cq_sread,
-	.sreadfrom = fi_no_cq_sreadfrom,
+	.sreadfrom = ofi_cq_sreadfrom,
 	.signal = vrb_cq_signal,
 	.strerror = vrb_cq_strerror
 };
@@ -485,13 +428,10 @@ static int vrb_cq_control(fid_t fid, int command, void *arg)
 
 static int vrb_cq_close(fid_t fid)
 {
-	struct vrb_wc_entry *wce;
-	struct slist_entry *entry;
+	struct vrb_cq *cq = container_of(fid, struct vrb_cq, util_cq.cq_fid);
+	struct vrb_srx *srx;
+	struct dlist_entry *srx_temp;
 	int ret;
-	struct vrb_cq *cq =
-		container_of(fid, struct vrb_cq, util_cq.cq_fid);
-	struct vrb_srq_ep *srq_ep;
-	struct dlist_entry *srq_ep_temp;
 
 	if (ofi_atomic_get32(&cq->nevents))
 		ibv_ack_cq_events(cq->cq, ofi_atomic_get32(&cq->nevents));
@@ -499,27 +439,16 @@ static int vrb_cq_close(fid_t fid)
 	/* Since an RX CQ and SRX context can be destroyed in any order,
 	 * and the XRC SRQ references the RX CQ, we must destroy any
 	 * XRC SRQ using this CQ before destroying the CQ. */
-	fastlock_acquire(&cq->xrc.srq_list_lock);
-	dlist_foreach_container_safe(&cq->xrc.srq_list, struct vrb_srq_ep,
-				     srq_ep, xrc.srq_entry, srq_ep_temp) {
-		ret = vrb_xrc_close_srq(srq_ep);
+	ofi_genlock_lock(vrb_cq2_progress(cq)->active_lock);
+	dlist_foreach_container_safe(&cq->xrc.srq_list, struct vrb_srx,
+				     srx, xrc.srq_entry, srx_temp) {
+		ret = vrb_xrc_close_srq(srx);
 		if (ret) {
-			fastlock_release(&cq->xrc.srq_list_lock);
+			ofi_genlock_unlock(vrb_cq2_progress(cq)->active_lock);
 			return -ret;
 		}
 	}
-	fastlock_release(&cq->xrc.srq_list_lock);
-
-	cq->util_cq.cq_fastlock_acquire(&cq->util_cq.cq_lock);
-	while (!slist_empty(&cq->saved_wc_list)) {
-		entry = slist_remove_head(&cq->saved_wc_list);
-		wce = container_of(entry, struct vrb_wc_entry, entry);
-		ofi_buf_free(wce);
-	}
-	cq->util_cq.cq_fastlock_release(&cq->util_cq.cq_lock);
-
-	ofi_bufpool_destroy(cq->wce_pool);
-	ofi_bufpool_destroy(cq->ctx_pool);
+	ofi_genlock_unlock(vrb_cq2_progress(cq)->active_lock);
 
 	if (cq->cq) {
 		ret = ibv_destroy_cq(cq->cq);
@@ -527,19 +456,14 @@ static int vrb_cq_close(fid_t fid)
 			return -ret;
 	}
 
-	if (cq->signal_fd[0]) {
-		ofi_close_socket(cq->signal_fd[0]);
-	}
-	if (cq->signal_fd[1]) {
-		ofi_close_socket(cq->signal_fd[1]);
-	}
+	if (cq->wait_obj != FI_WAIT_NONE)
+		fd_signal_free(&cq->signal);
 
 	ofi_cq_cleanup(&cq->util_cq);
 
 	if (cq->channel)
 		ibv_destroy_comp_channel(cq->channel);
 
-	fastlock_destroy(&cq->xrc.srq_list_lock);
 	free(cq);
 	return 0;
 }
@@ -552,10 +476,12 @@ static struct fi_ops vrb_cq_fi_ops = {
 	.ops_open = fi_no_ops_open,
 };
 
-static void vrb_util_cq_progress_noop(struct util_cq *cq)
+static void vrb_cq_progress(struct util_cq *util_cq)
 {
-	/* This routine shouldn't be called */
-	assert(0);
+	struct vrb_cq *cq;
+
+	cq = container_of(util_cq, struct vrb_cq, util_cq);
+	vrb_flush_cq(cq);
 }
 
 int vrb_cq_open(struct fid_domain *domain_fid, struct fi_cq_attr *attr,
@@ -568,6 +494,7 @@ int vrb_cq_open(struct fid_domain *domain_fid, struct fi_cq_attr *attr,
 	size_t size;
 	int ret;
 	struct fi_cq_attr tmp_attr = *attr;
+	int comp_vector = 0;
 
 	cq = calloc(1, sizeof(*cq));
 	if (!cq)
@@ -576,7 +503,7 @@ int vrb_cq_open(struct fid_domain *domain_fid, struct fi_cq_attr *attr,
 	/* verbs uses its own implementation of wait objects for CQ */
 	tmp_attr.wait_obj = FI_WAIT_NONE;
 	ret = ofi_cq_init(&vrb_prov, domain_fid, &tmp_attr, &cq->util_cq,
-			  vrb_util_cq_progress_noop, context);
+			  vrb_cq_progress, context);
 	if (ret)
 		goto err1;
 
@@ -591,14 +518,27 @@ int vrb_cq_open(struct fid_domain *domain_fid, struct fi_cq_attr *attr,
 		break;
 	default:
 		ret = -FI_ENOSYS;
-		goto err4;
+		goto err2;
+	}
+
+	if (attr->flags & FI_AFFINITY) {
+		if (attr->signaling_vector < 0 ||
+		    attr->signaling_vector > domain->verbs->num_comp_vectors)  {
+
+			VRB_WARN(FI_LOG_CQ,
+				   "Invalid value for the CQ attribute signaling_vector: %d\n",
+				   attr->signaling_vector);
+			ret = -FI_EINVAL;
+			goto err2;
+		}
+		comp_vector = attr->signaling_vector;
 	}
 
 	if (cq->wait_obj != FI_WAIT_NONE) {
 		cq->channel = ibv_create_comp_channel(domain->verbs);
 		if (!cq->channel) {
 			ret = -errno;
-			VERBS_WARN(FI_LOG_CQ,
+			VRB_WARN(FI_LOG_CQ,
 				   "Unable to create completion channel\n");
 			goto err2;
 		}
@@ -607,14 +547,10 @@ int vrb_cq_open(struct fid_domain *domain_fid, struct fi_cq_attr *attr,
 		if (ret)
 			goto err3;
 
-		if (socketpair(AF_UNIX, SOCK_STREAM, 0, cq->signal_fd)) {
+		if (fd_signal_init(&cq->signal)) {
 			ret = -errno;
 			goto err3;
 		}
-
-		ret = fi_fd_nonblock(cq->signal_fd[0]);
-		if (ret)
-			goto err4;
 	}
 
 	size = attr->size ? attr->size : VERBS_DEF_CQ_SIZE;
@@ -629,77 +565,38 @@ int vrb_cq_open(struct fid_domain *domain_fid, struct fi_cq_attr *attr,
 	 * num_qp_per_cq = ibv_device_attr->max_cqe / (qp_send_wr + qp_recv_wr)
 	 */
 	cq->cq = ibv_create_cq(domain->verbs, size, cq, cq->channel,
-			       attr->signaling_vector);
+			       comp_vector);
 	if (!cq->cq) {
 		ret = -errno;
-		VERBS_WARN(FI_LOG_CQ, "Unable to create verbs CQ\n");
-		goto err4;
+		VRB_WARN(FI_LOG_CQ, "Unable to create verbs CQ\n");
+		goto err3;
 	}
 
 	if (cq->channel) {
 		ret = ibv_req_notify_cq(cq->cq, 0);
 		if (ret) {
-			VERBS_WARN(FI_LOG_CQ,
+			VRB_WARN(FI_LOG_CQ,
 				   "ibv_req_notify_cq failed\n");
-			goto err5;
+			goto err4;
 		}
 	}
 
-	ret = ofi_bufpool_create(&cq->wce_pool, sizeof(struct vrb_wc_entry),
-				16, 0, VERBS_WCE_CNT, 0);
-	if (ret) {
-		VERBS_WARN(FI_LOG_CQ, "Failed to create wce_pool\n");
-		goto err5;
-	}
-
 	cq->flags |= attr->flags;
 	cq->wait_cond = attr->wait_cond;
 	/* verbs uses its own ops for CQ */
 	cq->util_cq.cq_fid.fid.ops = &vrb_cq_fi_ops;
 	cq->util_cq.cq_fid.ops = &vrb_cq_ops;
 
-	switch (attr->format) {
-	case FI_CQ_FORMAT_UNSPEC:
-	case FI_CQ_FORMAT_CONTEXT:
-		cq->read_entry = vrb_cq_read_context_entry;
-		cq->entry_size = sizeof(struct fi_cq_entry);
-		break;
-	case FI_CQ_FORMAT_MSG:
-		cq->read_entry = vrb_cq_read_msg_entry;
-		cq->entry_size = sizeof(struct fi_cq_msg_entry);
-		break;
-	case FI_CQ_FORMAT_DATA:
-		cq->read_entry = vrb_cq_read_data_entry;
-		cq->entry_size = sizeof(struct fi_cq_data_entry);
-		break;
-	case FI_CQ_FORMAT_TAGGED:
-	default:
-		ret = -FI_ENOSYS;
-		goto err6;
-	}
-
-	ret = ofi_bufpool_create(&cq->ctx_pool, sizeof(struct fi_context),
-				 16, 0, size, OFI_BUFPOOL_NO_TRACK);
-	if (ret)
-		goto err6;
-
-	slist_init(&cq->saved_wc_list);
+	// slist_init(&cq->saved_wc_list);
 	dlist_init(&cq->xrc.srq_list);
-	fastlock_init(&cq->xrc.srq_list_lock);
 
 	ofi_atomic_initialize32(&cq->nevents, 0);
 
-	cq->credits = size;
-
 	*cq_fid = &cq->util_cq.cq_fid;
 	return 0;
-err6:
-	ofi_bufpool_destroy(cq->wce_pool);
-err5:
-	ibv_destroy_cq(cq->cq);
+
 err4:
-	ofi_close_socket(cq->signal_fd[0]);
-	ofi_close_socket(cq->signal_fd[1]);
+	ibv_destroy_cq(cq->cq);
 err3:
 	if (cq->channel)
 		ibv_destroy_comp_channel(cq->channel);
@@ -709,3 +606,43 @@ err1:
 	free(cq);
 	return ret;
 }
+
+
+int vrb_init_progress(struct vrb_progress *progress, struct fi_info *info)
+{
+	int ret;
+
+	ret = ofi_genlock_init(&progress->ep_lock, OFI_LOCK_MUTEX);
+	if (ret)
+		return ret;
+
+	/* Active lock will be needed when adding rdm ep support */
+	progress->active_lock = &progress->ep_lock;
+
+	ret = ofi_bufpool_create(&progress->ctx_pool, sizeof(struct fi_context),
+				 16, 0, 1024, OFI_BUFPOOL_NO_TRACK);
+	if (ret)
+		goto err1;
+
+	ret = ofi_bufpool_create(&progress->recv_wr_pool,
+				 sizeof(struct vrb_recv_wr) +
+				 info->rx_attr->iov_limit * sizeof(struct ibv_sge),
+				 16, 0, 1024, OFI_BUFPOOL_NO_TRACK);
+	if (ret)
+		goto err2;
+
+	return 0;
+
+err2:
+	ofi_bufpool_destroy(progress->ctx_pool);
+err1:
+	ofi_genlock_destroy(&progress->ep_lock);
+	return ret;
+}
+
+void vrb_close_progress(struct vrb_progress *progress)
+{
+	ofi_bufpool_destroy(progress->recv_wr_pool);
+	ofi_bufpool_destroy(progress->ctx_pool);
+	ofi_genlock_destroy(&progress->ep_lock);
+}
