@@ -71,7 +71,7 @@
 #include <ofi_lock.h>
 #include <ofi_osd.h>
 #include <ofi_iov.h>
-#include <shared/ofi_str.h>
+#include <ofi_str.h>
 
 struct fi_provider core_prov = {
 	.name = "core",
@@ -86,6 +86,7 @@ struct ofi_common_locks common_locks = {
 
 size_t ofi_universe_size = 1024;
 int ofi_av_remove_cleanup;
+char *ofi_offload_coll_prov_name = NULL;
 
 
 int ofi_genlock_init(struct ofi_genlock *lock,
@@ -303,6 +304,24 @@ uint32_t ofi_generate_seed(void)
 	return rand_seed;
 }
 
+uint64_t ofi_get_realtime_ns(void)
+{
+	struct timespec now;
+
+	clock_gettime(CLOCK_REALTIME, &now);
+	return now.tv_sec * 1000000000 + now.tv_nsec;
+}
+
+uint64_t ofi_get_realtime_us(void)
+{
+	return ofi_get_realtime_ns() / 1000;
+}
+
+uint64_t ofi_get_realtime_ms(void)
+{
+	return ofi_get_realtime_ns() / 1000000;
+}
+
 uint64_t ofi_gettime_ns(void)
 {
 	struct timespec now;
@@ -413,10 +432,6 @@ sa_sin6:
 			     (uint16_t)(ntohll(sib->sib_sid) >> 16) & 0xfff, /* port space */
 				 (uint8_t)ntohll(sib->sib_scope_id) & 0xff);
 		break;
-	case FI_ADDR_PSMX:
-		size = snprintf(buf, *len, "fi_addr_psmx://%" PRIx64,
-				*(uint64_t *)addr);
-		break;
 	case FI_ADDR_PSMX2:
 		size =
 		    snprintf(buf, *len, "fi_addr_psmx2://%" PRIx64 ":%" PRIx64,
@@ -428,19 +443,15 @@ sa_sin6:
 			     *(uint64_t *)addr, *((uint64_t *)addr + 1),
 			     *((uint64_t *)addr + 2), *((uint64_t *)addr + 3));
 		break;
-	case FI_ADDR_GNI:
-		size = snprintf(buf, *len, "fi_addr_gni://%" PRIx64,
-				*(uint64_t *)addr);
-		break;
-	case FI_ADDR_BGQ:
-		size = snprintf(buf, *len, "fi_addr_bgq://%p", addr);
-		break;
 	case FI_ADDR_OPX:
 		size = snprintf(buf, *len, "fi_addr_opx://%016lx", *(uint64_t *)addr);
 		break;
 	case FI_ADDR_MLX:
 		size = snprintf(buf, *len, "fi_addr_mlx://%p", addr);
 		break;
+	case FI_ADDR_UCX:
+		size = snprintf(buf, *len, "fi_addr_ucx://%p", addr);
+		break;
 	case FI_ADDR_IB_UD:
 		memset(str, 0, sizeof(str));
 		if (!inet_ntop(AF_INET6, addr, str, INET6_ADDRSTRLEN))
@@ -488,45 +499,24 @@ uint32_t ofi_addr_format(const char *str)
 		return FI_SOCKADDR_IN6;
 	else if (!strcasecmp(fmt, "fi_sockaddr_ib"))
 		return FI_SOCKADDR_IB;
-	else if (!strcasecmp(fmt, "fi_addr_psmx"))
-		return FI_ADDR_PSMX;
 	else if (!strcasecmp(fmt, "fi_addr_psmx2"))
 		return FI_ADDR_PSMX2;
 	else if (!strcasecmp(fmt, "fi_addr_psmx3"))
 		return FI_ADDR_PSMX3;
-	else if (!strcasecmp(fmt, "fi_addr_gni"))
-		return FI_ADDR_GNI;
-	else if (!strcasecmp(fmt, "fi_addr_bgq"))
-		return FI_ADDR_BGQ;
 	else if (!strcasecmp(fmt, "fi_addr_opx"))
 		return FI_ADDR_OPX;
 	else if (!strcasecmp(fmt, "fi_addr_efa"))
 		return FI_ADDR_EFA;
 	else if (!strcasecmp(fmt, "fi_addr_mlx"))
 		return FI_ADDR_MLX;
+	else if (!strcasecmp(fmt, "fi_addr_ucx"))
+		return FI_ADDR_UCX;
 	else if (!strcasecmp(fmt, "fi_addr_ib_ud"))
 		return FI_ADDR_IB_UD;
 
 	return FI_FORMAT_UNSPEC;
 }
 
-static int ofi_str_to_psmx(const char *str, void **addr, size_t *len)
-{
-	int ret;
-
-	*len = sizeof(uint64_t);
-	*addr = calloc(1, *len);
-	if (!(*addr))
-		return -FI_ENOMEM;
-
-	ret = sscanf(str, "%*[^:]://%" SCNx64, (uint64_t *) *addr);
-	if (ret == 1)
-		return 0;
-
-	free(*addr);
-	return -FI_EINVAL;
-}
-
 static int ofi_str_to_psmx2(const char *str, void **addr, size_t *len)
 {
 	int ret;
@@ -909,8 +899,6 @@ int ofi_str_toaddr(const char *str, uint32_t *addr_format,
 		return ofi_str_to_sin(str, addr, len);
 	case FI_SOCKADDR_IN6:
 		return ofi_str_to_sin6(str, addr, len);
-	case FI_ADDR_PSMX:
-		return ofi_str_to_psmx(str, addr, len);
 	case FI_ADDR_PSMX2:
 		return ofi_str_to_psmx2(str, addr, len);
 	case FI_ADDR_PSMX3:
@@ -923,9 +911,8 @@ int ofi_str_toaddr(const char *str, uint32_t *addr_format,
 		return ofi_str_to_efa(str, addr, len);
 	case FI_SOCKADDR_IB:
 		return ofi_str_to_sib(str, addr, len);
-	case FI_ADDR_GNI:
-	case FI_ADDR_BGQ:
 	case FI_ADDR_MLX:
+	case FI_ADDR_UCX:
 	default:
 		return -FI_ENOSYS;
 	}
@@ -1066,19 +1053,19 @@ size_t ofi_mask_addr(struct sockaddr *maskaddr, const struct sockaddr *srcaddr,
 	return len;
 }
 
-void ofi_straddr_log_internal(const char *func, int line,
+void ofi_straddr_log_internal(const char *func, int line, uint32_t addr_format,
 			      const struct fi_provider *prov,
 			      enum fi_log_level level,
 			      enum fi_log_subsys subsys, char *log_str,
 			      const void *addr)
 {
 	char buf[OFI_ADDRSTRLEN];
-	uint32_t addr_format;
 	size_t len = sizeof(buf);
 
 	if (fi_log_enabled(prov, level, subsys)) {
 		if (addr) {
-			addr_format = ofi_translate_addr_format(ofi_sa_family(addr));
+			if (addr_format == FI_FORMAT_UNSPEC)
+				addr_format = ofi_translate_addr_format(ofi_sa_family(addr));
 			fi_log(prov, level, subsys, func, line, "%s: %s\n", log_str,
 			       ofi_straddr(buf, &len, addr_format, addr));
 		} else {
@@ -1140,11 +1127,10 @@ void ofi_byteq_writev(struct ofi_byteq *byteq, const struct iovec *iov,
 }
 
 
-ssize_t ofi_bsock_flush(struct ofi_bsock *bsock)
+int ofi_bsock_flush(struct ofi_bsock *bsock)
 {
 	size_t avail;
 	ssize_t ret;
-	int err;
 
 	if (!ofi_bsock_tosend(bsock))
 		return 0;
@@ -1153,29 +1139,18 @@ ssize_t ofi_bsock_flush(struct ofi_bsock *bsock)
 	assert(avail);
 	ret = bsock->sockapi->send(bsock->sockapi, bsock->sock,
 				   &bsock->sq.data[bsock->sq.head],
-				   avail, MSG_NOSIGNAL, bsock);
-	if (ret < 0) {
-		if (ret == -OFI_EINPROGRESS_URING)
-		    return ret;
-
-		err = ofi_sockerr();
-		if (err == EPIPE)
-			return -FI_ENOTCONN;
-		if (err == EWOULDBLOCK)
-			return -FI_EAGAIN;
-		return -err;
-	} else {
-		ofi_byteq_consume(&bsock->sq, (size_t) ret);
-	}
+				   avail, MSG_NOSIGNAL, &bsock->tx_sockctx);
+	if (ret < 0)
+		return ret;
 
+	ofi_byteq_consume(&bsock->sq, (size_t) ret);
 	return ofi_bsock_tosend(bsock) ? -FI_EAGAIN : 0;
 }
 
-ssize_t ofi_bsock_flush_sync(struct ofi_bsock *bsock)
+int ofi_bsock_flush_sync(struct ofi_bsock *bsock)
 {
 	size_t avail;
 	ssize_t ret;
-	int err;
 
 	if (!ofi_bsock_tosend(bsock))
 		return 0;
@@ -1184,42 +1159,39 @@ ssize_t ofi_bsock_flush_sync(struct ofi_bsock *bsock)
 	assert(avail);
 	ret = ofi_send_socket(bsock->sock, &bsock->sq.data[bsock->sq.head],
 			      avail, MSG_NOSIGNAL);
-	if (ret < 0) {
-		err = ofi_sockerr();
-		if (err == EPIPE)
-			return -FI_ENOTCONN;
-		if (err == EWOULDBLOCK)
-			return -FI_EAGAIN;
-		return -err;
-	} else {
-		ofi_byteq_consume(&bsock->sq, (size_t) ret);
-	}
+	if (ret < 0)
+		return ret;
 
+	ofi_byteq_consume(&bsock->sq, (size_t) ret);
 	return ofi_bsock_tosend(bsock) ? -FI_EAGAIN : 0;
 }
 
-ssize_t ofi_bsock_send(struct ofi_bsock *bsock, const void *buf, size_t *len)
+int ofi_bsock_send(struct ofi_bsock *bsock, const void *buf, size_t *len)
 {
 	size_t avail;
 	ssize_t ret;
+	int err;
 
 	avail = ofi_bsock_tosend(bsock);
 	if (avail) {
 		if (*len < ofi_byteq_writeable(&bsock->sq)) {
 			ofi_byteq_write(&bsock->sq, buf, *len);
-			ret = ofi_bsock_flush(bsock);
-			return !ret || ret == -FI_EAGAIN ? *len : ret;
+			err = ofi_bsock_flush(bsock);
+			return !err || err == -FI_EAGAIN ? 0 : err;
 		}
 
-		ret = ofi_bsock_flush(bsock);
-		if (ret)
-			return ret;
+		err = ofi_bsock_flush(bsock);
+		if (err) {
+			*len = 0;
+			return err;
+		}
 	}
 
 	assert(!ofi_bsock_tosend(bsock));
 	if (*len > bsock->zerocopy_size) {
 		ret = bsock->sockapi->send(bsock->sockapi, bsock->sock, buf, *len,
-					   MSG_NOSIGNAL | OFI_ZEROCOPY, bsock);
+					   MSG_NOSIGNAL | OFI_ZEROCOPY,
+					   &bsock->tx_sockctx);
 		if (ret >= 0) {
 			bsock->async_index++;
 			*len = ret;
@@ -1227,28 +1199,31 @@ ssize_t ofi_bsock_send(struct ofi_bsock *bsock, const void *buf, size_t *len)
 		}
 	} else {
 		ret = bsock->sockapi->send(bsock->sockapi, bsock->sock, buf, *len,
-					   MSG_NOSIGNAL, bsock);
+					   MSG_NOSIGNAL, &bsock->tx_sockctx);
 	}
 	if (ret < 0) {
 		if (ret == -OFI_EINPROGRESS_URING)
 			return ret;
 
-		if (OFI_SOCK_TRY_SND_RCV_AGAIN(ofi_sockerr()) &&
+		if (OFI_SOCK_TRY_SND_RCV_AGAIN(-ret) &&
 		    *len < ofi_byteq_writeable(&bsock->sq)) {
 			ofi_byteq_write(&bsock->sq, buf, *len);
-			return *len;
+			return 0;
 		}
-		return ofi_sockerr() == EPIPE ? -FI_ENOTCONN : -ofi_sockerr();
+
+		*len = 0;
+		return ret;
 	}
 	*len = ret;
-	return ret;
+	return 0;
 }
 
-ssize_t ofi_bsock_sendv(struct ofi_bsock *bsock, const struct iovec *iov,
-			size_t cnt, size_t *len)
+int ofi_bsock_sendv(struct ofi_bsock *bsock, const struct iovec *iov,
+		    size_t cnt, size_t *len)
 {
 	size_t avail;
 	ssize_t ret;
+	int err;
 
 	if (cnt == 1) {
 		*len = iov[0].iov_len;
@@ -1260,20 +1235,23 @@ ssize_t ofi_bsock_sendv(struct ofi_bsock *bsock, const struct iovec *iov,
 	if (avail) {
 		if (*len < ofi_byteq_writeable(&bsock->sq)) {
 			ofi_byteq_writev(&bsock->sq, iov, cnt);
-			ret = ofi_bsock_flush(bsock);
-			return !ret || ret == -FI_EAGAIN ? *len : ret;
+			err = ofi_bsock_flush(bsock);
+			return !err || err == -FI_EAGAIN ? 0 : err;
 		}
 
-		ret = ofi_bsock_flush(bsock);
-		if (ret)
-			return ret;
+		err = ofi_bsock_flush(bsock);
+		if (err) {
+			*len = 0;
+			return err;
+		}
 	}
 
 	assert(!ofi_bsock_tosend(bsock));
 
 	if (*len > bsock->zerocopy_size) {
 		ret = bsock->sockapi->sendv(bsock->sockapi, bsock->sock, iov, cnt,
-					    MSG_NOSIGNAL | OFI_ZEROCOPY, bsock);
+					    MSG_NOSIGNAL | OFI_ZEROCOPY,
+					    &bsock->tx_sockctx);
 		if (ret >= 0) {
 			bsock->async_index++;
 			*len = ret;
@@ -1281,120 +1259,156 @@ ssize_t ofi_bsock_sendv(struct ofi_bsock *bsock, const struct iovec *iov,
 		}
 	} else {
 		ret = bsock->sockapi->sendv(bsock->sockapi, bsock->sock, iov, cnt,
-					    MSG_NOSIGNAL, bsock);
+					    MSG_NOSIGNAL, &bsock->tx_sockctx);
 	}
 	if (ret < 0) {
 		if (ret == -OFI_EINPROGRESS_URING)
 			return ret;
 
-		if (OFI_SOCK_TRY_SND_RCV_AGAIN(ofi_sockerr()) &&
+		if (OFI_SOCK_TRY_SND_RCV_AGAIN(-ret) &&
 		    *len < ofi_byteq_writeable(&bsock->sq)) {
 			ofi_byteq_writev(&bsock->sq, iov, cnt);
-			return *len;
+			return 0;
 		}
-		return ofi_sockerr() == EPIPE ? -FI_ENOTCONN : -ofi_sockerr();
+
+		*len = 0;
+		return ret;
 	}
 	*len = ret;
-	return ret;
+	return 0;
 }
 
-ssize_t ofi_bsock_recv(struct ofi_bsock *bsock, void *buf, size_t len)
+int ofi_bsock_recv(struct ofi_bsock *bsock, void *buf, size_t *len)
 {
-	size_t bytes, avail;
+	size_t bytes, avail = 0;
 	ssize_t ret;
 
-	bytes = ofi_byteq_read(&bsock->rq, buf, len);
+	bytes = ofi_byteq_read(&bsock->rq, buf, *len);
 	if (bytes) {
-		if (bytes == len)
-			return len;
+		if (bytes == *len) {
+			return 0;
+		}
 
 		buf = (char *) buf + bytes;
-		len -= bytes;
+		*len -= bytes;
 	}
 
 	assert(!ofi_bsock_readable(bsock));
-	if (len < (bsock->rq.size >> 1)) {
+	if (*len < (bsock->rq.size >> 1)) {
 		avail = ofi_byteq_writeable(&bsock->rq);
 		assert(avail);
 		ret = bsock->sockapi->recv(bsock->sockapi, bsock->sock,
 					   &bsock->rq.data[bsock->rq.tail],
-					   avail, MSG_NOSIGNAL, bsock);
+					   avail, MSG_NOSIGNAL, &bsock->rx_sockctx);
 		if (ret <= 0)
 			goto out;
 
 		ofi_byteq_add(&bsock->rq, (size_t) ret);
 		assert(ofi_bsock_readable(bsock));
-		bytes += ofi_byteq_read(&bsock->rq, buf, len);
-		return bytes;
+		bytes += ofi_byteq_read(&bsock->rq, buf, *len);
+		*len = bytes;
+		return 0;
 	}
 
-	ret = bsock->sockapi->recv(bsock->sockapi, bsock->sock, buf, len,
-				   MSG_NOSIGNAL, bsock);
-	if (ret > 0)
-		return bytes + ret;
+	ret = bsock->sockapi->recv(bsock->sockapi, bsock->sock, buf, *len,
+				   MSG_NOSIGNAL, &bsock->rx_sockctx);
+	if (ret > 0) {
+		*len = bytes + ret;
+		return 0;
+	}
 
 out:
-	assert(ret != -OFI_EINPROGRESS_URING);
-	if (bytes)
-		return bytes;
-	return ret ? -ofi_sockerr(): -FI_ENOTCONN;
+	*len = bytes;
+	if (ret == -OFI_EINPROGRESS_URING) {
+		assert(!bsock->async_prefetch);
+		bsock->async_prefetch = avail;
+		return ret;
+	}
+	if (*len)
+		return 0;
+
+	return ret;
 }
 
-ssize_t ofi_bsock_recvv(struct ofi_bsock *bsock, struct iovec *iov, size_t cnt)
+int ofi_bsock_recvv(struct ofi_bsock *bsock, struct iovec *iov, size_t cnt,
+		    size_t *len)
 {
-	size_t len, bytes, avail;
+	size_t bytes, avail = 0;
 	ssize_t ret;
 
-	if (cnt == 1)
-		return ofi_bsock_recv(bsock, iov[0].iov_base, iov[0].iov_len);
+	if (cnt == 1) {
+		*len = iov[0].iov_len;
+		return ofi_bsock_recv(bsock, iov[0].iov_base, len);
+	}
 
-	len = ofi_total_iov_len(iov, cnt);
+	*len = ofi_total_iov_len(iov, cnt);
 	if (ofi_byteq_readable(&bsock->rq)) {
 		bytes = ofi_byteq_readv(&bsock->rq, iov, cnt, 0);
-		if (bytes == len)
-			return len;
+		if (bytes == *len)
+			return 0;
 
-		len -= bytes;
+		*len -= bytes;
 	} else {
 		bytes = 0;
 	}
 
 	assert(!ofi_bsock_readable(bsock));
-	if (len < (bsock->rq.size >> 1)) {
+	if (*len < (bsock->rq.size >> 1)) {
 		avail = ofi_byteq_writeable(&bsock->rq);
 		assert(avail);
 		ret = bsock->sockapi->recv(bsock->sockapi, bsock->sock,
 					   &bsock->rq.data[bsock->rq.tail],
-					   avail, MSG_NOSIGNAL, bsock);
+					   avail, MSG_NOSIGNAL, &bsock->rx_sockctx);
 		if (ret <= 0)
 			goto out;
 
 		ofi_byteq_add(&bsock->rq, (size_t) ret);
 		assert(ofi_bsock_readable(bsock));
 		bytes += ofi_byteq_readv(&bsock->rq, iov, cnt, bytes);
-		return bytes;
+		*len = bytes;
+		return 0;
 	}
 
 	/* It's too difficult to adjust the iov without copying it, so return
 	 * what data we have.  The caller will consume the iov and retry.
 	 */
-	if (bytes)
-		return bytes;
+	if (bytes) {
+		*len = bytes;
+		return 0;
+	}
 
 	ret = bsock->sockapi->recvv(bsock->sockapi, bsock->sock, iov, cnt,
-				    MSG_NOSIGNAL, bsock);
-	if (ret > 0)
-		return ret;
+				    MSG_NOSIGNAL, &bsock->rx_sockctx);
+	if (ret > 0) {
+		*len = ret;
+		return 0;
+	}
 out:
-	assert(ret != -OFI_EINPROGRESS_URING);
-	if (bytes)
-		return bytes;
-	return ret ? -ofi_sockerr(): -FI_ENOTCONN;
+	*len = bytes;
+	if (ret == -OFI_EINPROGRESS_URING) {
+		assert(!bsock->async_prefetch);
+		bsock->async_prefetch = avail;
+		return ret;
+	}
+	if (*len)
+		return 0;
+
+	return ret;
+}
+
+void ofi_bsock_prefetch_done(struct ofi_bsock *bsock, size_t len)
+{
+	assert(ofi_byteq_writeable(&bsock->rq) >= len);
+	assert(bsock->async_prefetch);
+
+	ofi_byteq_add(&bsock->rq, len);
+	assert(ofi_bsock_readable(bsock));
+	bsock->async_prefetch = false;
 }
 
 #ifdef MSG_ZEROCOPY
-uint32_t ofi_bsock_async_done(const struct fi_provider *prov,
-			      struct ofi_bsock *bsock)
+int ofi_bsock_async_done(const struct fi_provider *prov,
+			 struct ofi_bsock *bsock)
 {
 	struct msghdr msg = {};
 	struct sock_extended_err *serr;
@@ -1407,9 +1421,12 @@ uint32_t ofi_bsock_async_done(const struct fi_provider *prov,
 	msg.msg_controllen = sizeof(ctrl);
 	ret = recvmsg(bsock->sock, &msg, MSG_ERRQUEUE);
 	if (ret < 0) {
+		if (OFI_SOCK_TRY_SND_RCV_AGAIN(errno))
+			return 0;
+
 		FI_WARN(prov, FI_LOG_EP_DATA,
 			"Error reading MSG_ERRQUEUE (%s)\n", strerror(errno));
-		goto disable;
+		return -errno;
 	}
 
 	assert(!(msg.msg_flags & MSG_CTRUNC));
@@ -1418,31 +1435,30 @@ uint32_t ofi_bsock_async_done(const struct fi_provider *prov,
 	    (cmsg->cmsg_level != SOL_IPV6 && cmsg->cmsg_type != IPV6_RECVERR)) {
 		FI_WARN(prov, FI_LOG_EP_DATA,
 			"Unexpected cmsg level (!IP) or type (!RECVERR)\n");
-		goto disable;
+		return -FI_EINVAL;
 	}
 
 	serr = (void *) CMSG_DATA(cmsg);
 	if ((serr->ee_origin != SO_EE_ORIGIN_ZEROCOPY) || serr->ee_errno) {
 		FI_WARN(prov, FI_LOG_EP_DATA,
 			"Unexpected sock err origin or errno\n");
-		goto disable;
+		return -FI_EINVAL;
 	}
 
 	bsock->done_index = serr->ee_data;
 	if (serr->ee_code & SO_EE_CODE_ZEROCOPY_COPIED) {
 		FI_WARN(prov, FI_LOG_EP_DATA,
 			"Zerocopy data was copied\n");
-disable:
 		if (bsock->zerocopy_size != SIZE_MAX) {
 			FI_WARN(prov, FI_LOG_EP_DATA, "disabling zerocopy\n");
 			bsock->zerocopy_size = SIZE_MAX;
 		}
 	}
-	return bsock->done_index;
+	return 0;
 }
 #else
-uint32_t ofi_bsock_async_done(const struct fi_provider *prov,
-			      struct ofi_bsock *bsock)
+int ofi_bsock_async_done(const struct fi_provider *prov,
+			 struct ofi_bsock *bsock)
 {
 	return 0;
 }
@@ -1477,7 +1493,6 @@ int ofi_pollfds_grow(struct ofi_pollfds *pfds, int max_size)
 
 	while (pfds->size < size) {
 		ctx[pfds->size].index = -1;
-		ctx[pfds->size].hot_index = -1;
 		fds[pfds->size++].fd = INVALID_SOCKET;
 	}
 
@@ -1830,7 +1845,7 @@ static int
 ofi_dynpoll_get_fd_epoll(struct ofi_dynpoll *dynpoll)
 {
 	assert(dynpoll->type == OFI_DYNPOLL_EPOLL);
-	return dynpoll->ep;
+	return ofi_epoll_fd(dynpoll->ep);
 }
 
 static int
@@ -2122,7 +2137,9 @@ void ofi_get_list_of_addr(const struct fi_provider *prov, const char *env_name,
 
 insert_lo:
 	/* Always add loopback address at the end */
-	ofi_insert_loopback_addr(prov, addr_list);
+	if (!iface || !strncmp(iface, "lo", strlen(iface) + 1) ||
+	    !strncmp(iface, "loopback", strlen(iface) + 1))
+		ofi_insert_loopback_addr(prov, addr_list);
 }
 
 #elif defined HAVE_MIB_IPADDRTABLE
@@ -2205,16 +2222,6 @@ void ofi_remove_comma(char *buffer)
 		buffer[sz-2] = '\0';
 }
 
-void ofi_strncatf(char *dest, size_t n, const char *fmt, ...)
-{
-	size_t len = strnlen(dest, n);
-	va_list arglist;
-
-	va_start(arglist, fmt);
-	vsnprintf(&dest[len], n - 1 - len, fmt, arglist);
-	va_end(arglist);
-}
-
 /* The provider must free any prov_attr data prior to calling this
  * routine.
  */
@@ -2260,110 +2267,6 @@ int ofi_nic_control(struct fid *fid, int command, void *arg)
 	}
 }
 
-static void ofi_tostr_device_attr(char *buf, size_t len,
-				  const struct fi_device_attr *attr)
-{
-	const char *prefix = TAB TAB;
-
-	ofi_strncatf(buf, len, "%sfi_device_attr:\n", prefix);
-
-	prefix = TAB TAB TAB;
-	ofi_strncatf(buf, len, "%sname: %s\n", prefix, attr->name);
-	ofi_strncatf(buf, len, "%sdevice_id: %s\n", prefix, attr->device_id);
-	ofi_strncatf(buf, len, "%sdevice_version: %s\n", prefix,
-		     attr->device_version);
-	ofi_strncatf(buf, len, "%svendor_id: %s\n", prefix, attr->vendor_id);
-	ofi_strncatf(buf, len, "%sdriver: %s\n", prefix, attr->driver);
-	ofi_strncatf(buf, len, "%sfirmware: %s\n", prefix, attr->firmware);
-}
-
-static void ofi_tostr_pci_attr(char *buf, size_t len,
-			       const struct fi_pci_attr *attr)
-{
-	const char *prefix = TAB TAB TAB;
-
-	ofi_strncatf(buf, len, "%sfi_pci_attr:\n", prefix);
-
-	prefix = TAB TAB TAB TAB;
-	ofi_strncatf(buf, len, "%sdomain_id: %u\n", prefix, attr->domain_id);
-	ofi_strncatf(buf, len, "%sbus_id: %u\n", prefix, attr->bus_id);
-	ofi_strncatf(buf, len, "%sdevice_id: %u\n", prefix, attr->device_id);
-	ofi_strncatf(buf, len, "%sfunction_id: %u\n", prefix, attr->function_id);
-}
-
-static void ofi_tostr_bus_type(char *buf, size_t len, int type)
-{
-	switch (type) {
-	CASEENUMSTRN(FI_BUS_UNKNOWN, len);
-	CASEENUMSTRN(FI_BUS_PCI, len);
-	default:
-		ofi_strncatf(buf, len, "Unknown");
-		break;
-	}
-}
-
-static void ofi_tostr_bus_attr(char *buf, size_t len,
-			       const struct fi_bus_attr *attr)
-{
-	const char *prefix = TAB TAB;
-
-	ofi_strncatf(buf, len, "%sfi_bus_attr:\n", prefix);
-
-	prefix = TAB TAB TAB;
-	ofi_strncatf(buf, len, "%sbus_type: ", prefix);
-	ofi_tostr_bus_type(buf, len, attr->bus_type);
-	ofi_strncatf(buf, len, "\n");
-
-	switch (attr->bus_type) {
-	case FI_BUS_PCI:
-		ofi_tostr_pci_attr(buf, len, &attr->attr.pci);
-		break;
-	default:
-		break;
-	}
-}
-
-static void ofi_tostr_link_state(char *buf, size_t len, int state)
-{
-	switch (state) {
-	CASEENUMSTRN(FI_LINK_UNKNOWN, len);
-	CASEENUMSTRN(FI_LINK_DOWN, len);
-	CASEENUMSTRN(FI_LINK_UP, len);
-	default:
-		ofi_strncatf(buf, len, "Unknown");
-		break;
-	}
-}
-
-static void ofi_tostr_link_attr(char *buf, size_t len,
-				const struct fi_link_attr *attr)
-{
-	const char *prefix = TAB TAB;
-	ofi_strncatf(buf, len, "%sfi_link_attr:\n", prefix);
-
-	prefix = TAB TAB TAB;
-	ofi_strncatf(buf, len, "%saddress: %s\n", prefix, attr->address);
-	ofi_strncatf(buf, len, "%smtu: %zu\n", prefix, attr->mtu);
-	ofi_strncatf(buf, len, "%sspeed: %zu\n", prefix, attr->speed);
-	ofi_strncatf(buf, len, "%sstate: ", prefix);
-	ofi_tostr_link_state(buf, len, attr->state);
-	ofi_strncatf(buf, len, "\n%snetwork_type: %s\n", prefix,
-		     attr->network_type);
-}
-
-int ofi_nic_tostr(const struct fid *fid_nic, char *buf, size_t len)
-{
-	const struct fid_nic *nic = (const struct fid_nic*) fid_nic;
-
-	assert(fid_nic->fclass == FI_CLASS_NIC);
-	ofi_strncatf(buf, len, "%snic:\n", TAB);
-
-	ofi_tostr_device_attr(buf, len, nic->device_attr);
-	ofi_tostr_bus_attr(buf, len, nic->bus_attr);
-	ofi_tostr_link_attr(buf, len, nic->link_attr);
-	return 0;
-}
-
 struct fi_ops default_nic_ops = {
 	.size = sizeof(struct fi_ops),
 	.close = ofi_nic_close,
@@ -2485,7 +2388,7 @@ size_t ofi_vrb_speed(uint8_t speed, uint8_t width)
 		break;
 	case 4:
 	case 8:
-		speed_val = 8 * gbit_2_bit_coef;
+		speed_val = 10 * gbit_2_bit_coef;
 		break;
 	case 16:
 		speed_val = 14 * gbit_2_bit_coef;
@@ -2493,6 +2396,12 @@ size_t ofi_vrb_speed(uint8_t speed, uint8_t width)
 	case 32:
 		speed_val = 25 * gbit_2_bit_coef;
 		break;
+	case 64:
+		speed_val = 50 * gbit_2_bit_coef;
+		break;
+	case 128:
+		speed_val = 100 * gbit_2_bit_coef;
+		break;
 	default:
 		speed_val = 0;
 		break;
@@ -2511,6 +2420,9 @@ size_t ofi_vrb_speed(uint8_t speed, uint8_t width)
 	case 8:
 		width_val = 12;
 		break;
+	case 16:
+		width_val = 2;
+		break;
 	default:
 		width_val = 0;
 		break;
@@ -2520,4 +2432,4 @@ size_t ofi_vrb_speed(uint8_t speed, uint8_t width)
 }
 
 /* log_prefix is used by fi_log and by prov/util */
-const char *log_prefix = "";
+OFI_THREAD_LOCAL const char *log_prefix = "";
