@@ -44,6 +44,8 @@
 #include <arpa/inet.h>
 
 #include "ofi.h"
+#include "ofi_str.h"
+#include "ofi_iov.h"
 #include <rdma/fabric.h>
 #include <rdma/fi_domain.h>
 #include <rdma/fi_endpoint.h>
@@ -114,11 +116,9 @@ static void ofi_tostr_addr_format(char *buf, size_t len, uint32_t addr_format)
 	CASEENUMSTRN(FI_SOCKADDR_IN, len);
 	CASEENUMSTRN(FI_SOCKADDR_IN6, len);
 	CASEENUMSTRN(FI_SOCKADDR_IB, len);
-	CASEENUMSTRN(FI_ADDR_PSMX, len);
 	CASEENUMSTRN(FI_ADDR_PSMX2, len);
-	CASEENUMSTRN(FI_ADDR_GNI, len);
-	CASEENUMSTRN(FI_ADDR_BGQ, len);
 	CASEENUMSTRN(FI_ADDR_MLX, len);
+	CASEENUMSTRN(FI_ADDR_UCX, len);
 	CASEENUMSTRN(FI_ADDR_STR, len);
 	CASEENUMSTRN(FI_ADDR_IB_UD, len);
 	CASEENUMSTRN(FI_ADDR_EFA, len);
@@ -126,10 +126,7 @@ static void ofi_tostr_addr_format(char *buf, size_t len, uint32_t addr_format)
 	CASEENUMSTRN(FI_ADDR_OPX, len);
 	CASEENUMSTRN(FI_ADDR_CXI, len);
 	default:
-		if (addr_format & FI_PROV_SPECIFIC)
-			ofi_strncatf(buf, len, "Provider specific");
-		else
-			ofi_strncatf(buf, len, "Unknown");
+		ofi_strncatf(buf, len, "Unknown");
 		break;
 	}
 }
@@ -140,6 +137,7 @@ static void ofi_tostr_progress(char *buf, size_t len, enum fi_progress progress)
 	CASEENUMSTRN(FI_PROGRESS_UNSPEC, len);
 	CASEENUMSTRN(FI_PROGRESS_AUTO, len);
 	CASEENUMSTRN(FI_PROGRESS_MANUAL, len);
+	CASEENUMSTRN(FI_PROGRESS_CONTROL_UNIFIED, len);
 	default:
 		ofi_strncatf(buf, len, "Unknown");
 		break;
@@ -152,10 +150,8 @@ ofi_tostr_threading(char *buf, size_t len, enum fi_threading threading)
 	switch (threading) {
 	CASEENUMSTRN(FI_THREAD_UNSPEC, len);
 	CASEENUMSTRN(FI_THREAD_SAFE, len);
-	CASEENUMSTRN(FI_THREAD_FID, len);
 	CASEENUMSTRN(FI_THREAD_DOMAIN, len);
 	CASEENUMSTRN(FI_THREAD_COMPLETION, len);
-	CASEENUMSTRN(FI_THREAD_ENDPOINT, len);
 	default:
 		ofi_strncatf(buf, len, "Unknown");
 		break;
@@ -185,19 +181,6 @@ static void ofi_tostr_msgorder(char *buf, size_t len, uint64_t flags)
 	ofi_remove_comma(buf);
 }
 
-static void ofi_tostr_comporder(char *buf, size_t len, uint64_t flags)
-{
-	if ((flags & FI_ORDER_STRICT) == FI_ORDER_NONE) {
-		ofi_strncatf(buf, len, "FI_ORDER_NONE, ");
-	} else if ((flags & FI_ORDER_STRICT) == FI_ORDER_STRICT) {
-		ofi_strncatf(buf, len, "FI_ORDER_STRICT, ");
-	}
-
-	IFFLAGSTRN(flags, FI_ORDER_DATA, len);
-
-	ofi_remove_comma(buf);
-}
-
 static void ofi_tostr_caps(char *buf, size_t len, uint64_t caps)
 {
 	IFFLAGSTRN(caps, FI_MSG, len);
@@ -219,7 +202,6 @@ static void ofi_tostr_caps(char *buf, size_t len, uint64_t caps)
 	IFFLAGSTRN(caps, FI_TRIGGER, len);
 	IFFLAGSTRN(caps, FI_FENCE, len);
 
-	IFFLAGSTRN(caps, FI_VARIABLE_MSG, len);
 	IFFLAGSTRN(caps, FI_RMA_PMEM, len);
 	IFFLAGSTRN(caps, FI_SOURCE_ERR, len);
 	IFFLAGSTRN(caps, FI_LOCAL_COMM, len);
@@ -229,6 +211,8 @@ static void ofi_tostr_caps(char *buf, size_t len, uint64_t caps)
 	IFFLAGSTRN(caps, FI_SOURCE, len);
 	IFFLAGSTRN(caps, FI_NAMED_RX_CTX, len);
 	IFFLAGSTRN(caps, FI_DIRECTED_RECV, len);
+	IFFLAGSTRN(caps, FI_AV_USER_ID, len);
+	IFFLAGSTRN(caps, FI_PEER, len);
 	IFFLAGSTRN(caps, FI_HMEM, len);
 
 	ofi_remove_comma(buf);
@@ -241,8 +225,6 @@ static void ofi_tostr_ep_type(char *buf, size_t len, enum fi_ep_type ep_type)
 	CASEENUMSTRN(FI_EP_MSG, len);
 	CASEENUMSTRN(FI_EP_DGRAM, len);
 	CASEENUMSTRN(FI_EP_RDM, len);
-	CASEENUMSTRN(FI_EP_SOCK_STREAM, len);
-	CASEENUMSTRN(FI_EP_SOCK_DGRAM, len);
 	default:
 		ofi_strncatf(buf, len, "Unknown");
 		break;
@@ -256,19 +238,17 @@ static void ofi_tostr_protocol(char *buf, size_t len, uint32_t protocol)
 	CASEENUMSTRN(FI_PROTO_RDMA_CM_IB_RC, len);
 	CASEENUMSTRN(FI_PROTO_IWARP, len);
 	CASEENUMSTRN(FI_PROTO_IB_UD, len);
-	CASEENUMSTRN(FI_PROTO_PSMX, len);
 	CASEENUMSTRN(FI_PROTO_PSMX2, len);
 	CASEENUMSTRN(FI_PROTO_UDP, len);
 	CASEENUMSTRN(FI_PROTO_SOCK_TCP, len);
 	CASEENUMSTRN(FI_PROTO_IB_RDM, len);
 	CASEENUMSTRN(FI_PROTO_IWARP_RDM, len);
-	CASEENUMSTRN(FI_PROTO_GNI, len);
 	CASEENUMSTRN(FI_PROTO_RXM, len);
 	CASEENUMSTRN(FI_PROTO_RXD, len);
 	CASEENUMSTRN(FI_PROTO_MLX, len);
+	CASEENUMSTRN(FI_PROTO_UCX, len);
 	CASEENUMSTRN(FI_PROTO_NETWORKDIRECT, len);
 	CASEENUMSTRN(FI_PROTO_SHM, len);
-	CASEENUMSTRN(FI_PROTO_RSTREAM, len);
 	CASEENUMSTRN(FI_PROTO_RDMA_CM_IB_XRC, len);
 	CASEENUMSTRN(FI_PROTO_EFA, len);
 	CASEENUMSTRN(FI_PROTO_PSMX3, len);
@@ -276,11 +256,12 @@ static void ofi_tostr_protocol(char *buf, size_t len, uint32_t protocol)
 	CASEENUMSTRN(FI_PROTO_OPX, len);
 	CASEENUMSTRN(FI_PROTO_CXI, len);
 	CASEENUMSTRN(FI_PROTO_XNET, len);
+	CASEENUMSTRN(FI_PROTO_SM2, len);
+	CASEENUMSTRN(FI_PROTO_CXI_RNR, len);
+	CASEENUMSTRN(FI_PROTO_LPP, len);
+	CASEENUMSTRN(FI_PROTO_LNX, len);
 	default:
-		if (protocol & FI_PROV_SPECIFIC)
-			ofi_strncatf(buf, len, "Provider specific");
-		else
-			ofi_strncatf(buf, len, "Unknown");
+		ofi_strncatf(buf, len, "Unknown");
 		break;
 	}
 }
@@ -291,11 +272,8 @@ static void ofi_tostr_mode(char *buf, size_t len, uint64_t mode)
 	IFFLAGSTRN(mode, FI_MSG_PREFIX, len);
 	IFFLAGSTRN(mode, FI_ASYNC_IOV, len);
 	IFFLAGSTRN(mode, FI_RX_CQ_DATA, len);
-	IFFLAGSTRN(mode, FI_LOCAL_MR, len);
-	IFFLAGSTRN(mode, FI_NOTIFY_FLAGS_ONLY, len);
-	IFFLAGSTRN(mode, FI_RESTRICTED_COMP, len);
+	IFFLAGSTRN2(mode, OFI_LOCAL_MR, FI_LOCAL_MR, len);
 	IFFLAGSTRN(mode, FI_CONTEXT2, len);
-	IFFLAGSTRN(mode, FI_BUFFERED_RECV, len);
 
 	ofi_remove_comma(buf);
 }
@@ -343,10 +321,6 @@ ofi_tostr_tx_attr(char *buf, size_t len, const struct fi_tx_attr *attr,
 	ofi_tostr_msgorder(buf, len, attr->msg_order);
 	ofi_strncatf(buf, len, " ]\n");
 
-	ofi_strncatf(buf, len, "%s%scomp_order: [ ", prefix, TAB);
-	ofi_tostr_comporder(buf, len, attr->comp_order);
-	ofi_strncatf(buf, len, " ]\n");
-
 	ofi_strncatf(buf, len, "%s%sinject_size: %zu\n", prefix, TAB,
 		     attr->inject_size);
 	ofi_strncatf(buf, len, "%s%ssize: %zu\n", prefix, TAB, attr->size);
@@ -383,12 +357,6 @@ ofi_tostr_rx_attr(char *buf, size_t len, const struct fi_rx_attr *attr,
 	ofi_tostr_msgorder(buf, len, attr->msg_order);
 	ofi_strncatf(buf, len, " ]\n");
 
-	ofi_strncatf(buf, len, "%s%scomp_order: [ ", prefix, TAB);
-	ofi_tostr_comporder(buf, len, attr->comp_order);
-	ofi_strncatf(buf, len, " ]\n");
-
-	ofi_strncatf(buf, len, "%s%stotal_buffered_recv: %zu\n", prefix, TAB,
-		     attr->total_buffered_recv);
 	ofi_strncatf(buf, len, "%s%ssize: %zu\n", prefix, TAB, attr->size);
 	ofi_strncatf(buf, len, "%s%siov_limit: %zu\n", prefix, TAB,
 		     attr->iov_limit);
@@ -467,8 +435,8 @@ static void ofi_tostr_av_type(char *buf, size_t len, enum fi_av_type type)
 
 static void ofi_tostr_mr_mode(char *buf, size_t len, int mr_mode)
 {
-	IFFLAGSTRN(mr_mode, FI_MR_BASIC, len);
-	IFFLAGSTRN(mr_mode, FI_MR_SCALABLE, len);
+	IFFLAGSTRN2(mr_mode, OFI_MR_BASIC, FI_MR_BASIC, len);
+	IFFLAGSTRN2(mr_mode, OFI_MR_SCALABLE, FI_MR_SCALABLE, len);
 	IFFLAGSTRN(mr_mode, FI_MR_LOCAL, len);
 	IFFLAGSTRN(mr_mode, FI_MR_RAW, len);
 	IFFLAGSTRN(mr_mode, FI_MR_VIRT_ADDR, len);
@@ -521,11 +489,8 @@ ofi_tostr_domain_attr(char *buf, size_t len, const struct fi_domain_attr *attr,
 	ofi_tostr_threading(buf, len, attr->threading);
 	ofi_strncatf(buf, len, "\n");
 
-	ofi_strncatf(buf, len, "%s%scontrol_progress: ", prefix,TAB);
-	ofi_tostr_progress(buf, len, attr->control_progress);
-	ofi_strncatf(buf, len, "\n");
-	ofi_strncatf(buf, len, "%s%sdata_progress: ", prefix, TAB);
-	ofi_tostr_progress(buf, len, attr->data_progress);
+	ofi_strncatf(buf, len, "%s%sprogress: ", prefix, TAB);
+	ofi_tostr_progress(buf, len, attr->progress);
 	ofi_strncatf(buf, len, "\n");
 	ofi_strncatf(buf, len, "%s%sresource_mgmt: ", prefix, TAB);
 	ofi_tostr_resource_mgmt(buf, len, attr->resource_mgmt);
@@ -774,6 +739,184 @@ ofi_tostr_cq_format(char *buf, size_t len, enum fi_cq_format cq_format)
 	}
 }
 
+static void
+ofi_tostr_av_flags(char *buf, size_t len, uint64_t flags)
+{
+	IFFLAGSTRN(flags, FI_EVENT, len);
+	IFFLAGSTRN(flags, FI_READ, len);
+	IFFLAGSTRN(flags, FI_SYMMETRIC, len);
+	ofi_remove_comma(buf);
+}
+
+static void
+ofi_tostr_av_attr(char *buf, size_t len, const struct fi_av_attr *attr)
+{
+	if (!attr) {
+		ofi_strncatf(buf, len, "fi_av_attr: (null)\n");
+		return;
+	}
+
+	ofi_strncatf(buf, len, "fi_av_attr:\n");
+	ofi_strncatf(buf, len, "%stype: ", TAB);
+	ofi_tostr_av_type(buf, len, attr->type);
+	ofi_strncatf(buf, len, "\n");
+	ofi_strncatf(buf, len, "%scount: %zu\n", TAB, attr->count);
+	ofi_strncatf(buf, len, "%sep_per_node: %zu\n", TAB,
+	             attr->ep_per_node);
+	ofi_strncatf(buf, len, "%sname: %s\n", TAB, attr->name);
+	ofi_strncatf(buf, len, "%sflags: [ ", TAB);
+	ofi_tostr_av_flags(buf, len, attr->flags);
+	ofi_strncatf(buf, len, " ]\n");
+}
+
+static void
+ofi_tostr_wait_obj(char *buf, size_t len, enum fi_wait_obj obj)
+{
+	switch (obj) {
+	CASEENUMSTRN(FI_WAIT_NONE, len);
+	CASEENUMSTRN(FI_WAIT_UNSPEC, len);
+	CASEENUMSTRN(FI_WAIT_SET, len);
+	CASEENUMSTRN(FI_WAIT_FD, len);
+	CASEENUMSTRN(FI_WAIT_MUTEX_COND, len);
+	CASEENUMSTRN(FI_WAIT_YIELD, len);
+	CASEENUMSTRN(FI_WAIT_POLLFD, len);
+	default:
+		ofi_strncatf(buf, len, "Unknown");
+		break;
+	}
+}
+
+static void
+ofi_tostr_cq_wait_cond(char *buf, size_t len, enum fi_cq_wait_cond cond)
+{
+	switch (cond) {
+	CASEENUMSTRN(FI_CQ_COND_NONE, len);
+	CASEENUMSTRN(FI_CQ_COND_THRESHOLD, len);
+	default:
+		ofi_strncatf(buf, len, "Unknown");
+		break;
+	}
+}
+
+static void
+ofi_tostr_cq_flags(char *buf, size_t len, uint64_t flags)
+{
+	IFFLAGSTRN(flags, FI_AFFINITY, len);
+	ofi_remove_comma(buf);
+}
+
+static void
+ofi_tostr_cq_attr(char *buf, size_t len, const struct fi_cq_attr *attr)
+{
+	if (!attr) {
+		ofi_strncatf(buf, len, "fi_cq_attr: (null)\n");
+		return;
+	}
+	ofi_strncatf(buf, len, "fi_cq_attr:\n");
+	ofi_strncatf(buf, len, "%ssize: %zu\n", TAB, attr->size);
+	ofi_strncatf(buf, len, "%sflags: [ ", TAB);
+	ofi_tostr_cq_flags(buf, len, attr->flags);
+	ofi_strncatf(buf, len, " ]\n");
+	ofi_strncatf(buf, len, "%sformat: ", TAB);
+	ofi_tostr_cq_format(buf, len, attr->format);
+	ofi_strncatf(buf, len, "\n");
+	ofi_strncatf(buf, len, "%swait_obj: ", TAB);
+	ofi_tostr_wait_obj(buf, len, attr->wait_obj);
+	ofi_strncatf(buf, len, "\n");
+	ofi_strncatf(buf, len, "%ssignaling_vector: %d\n", TAB,
+	             attr->signaling_vector);
+	ofi_strncatf(buf, len, "%swait_cond: ", TAB);
+	ofi_tostr_cq_wait_cond(buf, len, attr->wait_cond);
+	ofi_strncatf(buf, len, "\n");
+}
+
+static void ofi_tostr_mr_access(char *buf, size_t len, uint64_t access)
+{
+	IFFLAGSTRN(access, FI_RECV, len);
+	IFFLAGSTRN(access, FI_SEND, len);
+	IFFLAGSTRN(access, FI_READ, len);
+	IFFLAGSTRN(access, FI_WRITE, len);
+	IFFLAGSTRN(access, FI_REMOTE_READ, len);
+	IFFLAGSTRN(access, FI_REMOTE_WRITE, len);
+	IFFLAGSTRN(access, FI_COLLECTIVE, len);
+	ofi_remove_comma(buf);
+}
+
+static void
+ofi_tostr_mr_attr(char *buf, size_t len, const struct fi_mr_attr *attr)
+{
+	if (!attr) {
+		ofi_strncatf(buf, len, "fi_mr_attr: (null)\n");
+		return;
+	}
+	ofi_strncatf(buf, len, "fi_mr_attr:\n");
+	ofi_strncatf(buf, len, "%smr_iov: %p\n", TAB, attr->mr_iov);
+	ofi_strncatf(buf, len, "%siov_count: %zu\n", TAB, attr->iov_count);
+	ofi_strncatf(buf, len, "%saccess: [ ", TAB);
+	ofi_tostr_mr_access(buf, len, attr->access);
+	ofi_strncatf(buf, len, " ]\n");
+	ofi_strncatf(buf, len, "%soffset: 0x%lx\n", TAB, attr->offset);
+	ofi_strncatf(buf, len, "%srequested_key: 0x%lx\n",
+	             TAB, attr->requested_key);
+	ofi_strncatf(buf, len, "%scontext: %p\n", TAB, attr->context);
+	ofi_strncatf(buf, len, "%sauth_key_size: %zu\n", TAB,
+	             attr->auth_key_size);
+	ofi_strncatf(buf, len, "%siface: ", TAB);
+	ofi_tostr_hmem_iface(buf, len, attr->iface);
+	ofi_strncatf(buf, len, "\n");
+}
+
+static void
+ofi_tostr_cntr_events(char *buf, size_t len, enum fi_cntr_events events)
+{
+	switch (events) {
+	CASEENUMSTRN(FI_CNTR_EVENTS_COMP, len);
+	CASEENUMSTRN(FI_CNTR_EVENTS_BYTES, len);
+	default:
+		ofi_strncatf(buf, len, "Unknown");
+		break;
+	}
+}
+
+static void
+ofi_tostr_cntr_attr(char *buf, size_t len, const struct fi_cntr_attr *attr)
+{
+	if (!attr) {
+		ofi_strncatf(buf, len, "fi_cntr_attr: (null)\n");
+		return;
+	}
+	ofi_strncatf(buf, len, "fi_cntr_attr:\n");
+	ofi_strncatf(buf, len, "%sevents: ", TAB);
+	ofi_tostr_cntr_events(buf, len, attr->events);
+	ofi_strncatf(buf, len, "\n");
+	ofi_strncatf(buf, len, "%swait_obj: ", TAB);
+	ofi_tostr_wait_obj(buf, len, attr->wait_obj);
+	ofi_strncatf(buf, len, "\n");
+	ofi_strncatf(buf, len, "%sflags: 0x%lx\n", TAB, attr->flags);
+}
+
+static void
+ofi_tostr_cq_err_entry(char *buf, size_t len,
+                       const struct fi_cq_err_entry *entry)
+{
+	if (!entry) {
+		ofi_strncatf(buf, len, "fi_cq_err_entry: (null)\n");
+		return;
+	}
+	ofi_strncatf(buf, len, "fi_cq_err_entry:\n");
+	ofi_strncatf(buf, len, "%sop_context: %p\n", TAB, entry->op_context);
+	ofi_strncatf(buf, len, "%sflags: 0x%lx\n", TAB, entry->flags);
+	ofi_strncatf(buf, len, "%slen: %zu\n", TAB, entry->len);
+	ofi_strncatf(buf, len, "%sbuf: %p\n", TAB, entry->buf);
+	ofi_strncatf(buf, len, "%sdata: %lu\n", TAB, entry->data);
+	ofi_strncatf(buf, len, "%stag: 0x%lx\n", TAB, entry->tag);
+	ofi_strncatf(buf, len, "%solen: %zu\n", TAB, entry->olen);
+	ofi_strncatf(buf, len, "%serr: %d\n", TAB, entry->err);
+	ofi_strncatf(buf, len, "%sprov_errno: %d\n", TAB, entry->err);
+	ofi_strncatf(buf, len, "%serr_data: %p\n", TAB, entry->err_data);
+	ofi_strncatf(buf, len, "%serr_data_size: %zu\n", TAB, entry->err_data_size);
+}
+
 __attribute__((visibility ("default"),EXTERNALLY_VISIBLE))
 char *DEFAULT_SYMVER_PRE(fi_tostr_r)(char *buf, size_t len,
 				     const void *data, enum fi_type datatype)
@@ -880,6 +1023,21 @@ char *DEFAULT_SYMVER_PRE(fi_tostr_r)(char *buf, size_t len,
 	case FI_TYPE_LOG_SUBSYS:
 		ofi_tostr_log_subsys(buf, len, *enumval);
 		break;
+	case FI_TYPE_AV_ATTR:
+		ofi_tostr_av_attr(buf, len, data);
+		break;
+	case FI_TYPE_CQ_ATTR:
+		ofi_tostr_cq_attr(buf, len, data);
+		break;
+	case FI_TYPE_MR_ATTR:
+		ofi_tostr_mr_attr(buf, len, data);
+		break;
+	case FI_TYPE_CNTR_ATTR:
+		ofi_tostr_cntr_attr(buf, len, data);
+		break;
+	case FI_TYPE_CQ_ERR_ENTRY:
+		ofi_tostr_cq_err_entry(buf, len, data);
+		break;
 	default:
 		ofi_strncatf(buf, len, "Unknown type");
 		break;
